(***********************************************************
*                                                          *
*      FUNGI - Fast UNsupervised Grammar Induction         *
*                                                          *
*   Originator: Prachya Boonkwan [1], [2]                  *
*               [1] School of Informatics,                 *
*                   University of Edinburgh, UK            *
*               [2] National Electronics and Computer      *
*                   Technology Center (NECTEC), Thailand   *
*                                                          *
*   Copyright (C) 2011 Prachya Boonkwan                    *
*   <p.boonkwan@sms.ed.ac.uk, kaamanita@hotmail.com>       *
*                                                          *
*   This library is free software; you can redistribute    *
*   it and/or modify it under the terms of the GNU         *
*   Lesser General Public License as published by the      *
*   Free Software Foundation; either version 2.1 of the    *
*   License, or (at your option) any later version.        *
*                                                          *
*   This library is distributed in the hope that it will   *
*   be useful, but WITHOUT ANY WARRANTY; without even      *
*   the implied warranty of MERCHANTABILITY or FITNESS     *
*   FOR A PARTICULAR PURPOSE.  See the GNU Lesser          *
*   General Public License for more details.               *
*                                                          *
*   You should have received a copy of the GNU Lesser      *
*   General Public License along with this library; if     *
*   not, write to the Free Software Foundation, Inc.,      *
*   59 Temple Place, Suite 330, Boston, MA  02111-1307     *
*   USA.                                                   *
*                                                          *
***********************************************************)

open Cg
open Chart
open Prob
open Deriv
open Models
open Inout
open Em
open Utils
open Printf

module type PARAM_TYPE = sig
    type t
    val title : string
    val to_string : t -> string
    val scale_prob : t -> float -> Deriv.t -> float
    val localscore : t -> int -> int -> float
    val inc : t -> t
    val is_stop_iter : t -> bool
end

module Annealer(Param : PARAM_TYPE) = struct

    type syncat_t = Syncat.t

    type word_t = string

    type chart_t = Chart.t

    type genmodel_t = DerivGenModel.t

    type deriv_t = Deriv.t

    type fn_get_chart_t = int -> chart_t

    type fn_scale_prob_t = float -> deriv_t -> float

    type fn_localscore_t = int -> int -> float
    
    type fn_compute_inout_t =
        ?scale_prob:fn_scale_prob_t ->
        ?localscore:fn_localscore_t ->
        ?run_as_server:bool -> ?run_as_client:bool -> 
        int -> float -> int -> float ->
        genmodel_t -> genmodel_t -> 
        syncat_t -> fn_get_chart_t -> int -> Hmm.fwdbwd_t -> unit

    type param_t = Param.t

    type t = {
        noentries     : int;
        diffthr       : float;
        rndthr        : int;
        epsilon       : float;
        ruleprob      : genmodel_t;
        ruleprior     : genmodel_t;
        start         : syncat_t;
        get_chart     : fn_get_chart_t;
        nocharts      : int;
        fwdbwd        : Hmm.fwdbwd_t;
        compute_inout : fn_compute_inout_t;
        run_as_server : bool;
        run_as_client : bool;
        mutable param : param_t;
        mutable rnd   : int;
    }

    let create 
        n diffthr rndthr 
        epsilon ruleprob ruleprior start get_chart nocharts fwdbwd
        compute_inout initparam run_as_server run_as_client =
    {
        noentries     = n;
        diffthr       = diffthr;
        rndthr        = rndthr;
        epsilon       = epsilon;
        ruleprob      = ruleprob;
        ruleprior     = ruleprior;
        start         = start;
        get_chart     = get_chart;
        nocharts      = nocharts;
        fwdbwd        = fwdbwd;
        compute_inout = compute_inout;
        param         = initparam;
        run_as_server = run_as_server;
        run_as_client = run_as_client;
        rnd           = 0;
    }

    let get_round ann =
        ann.rnd

    let reset_round ann =
        ann.rnd <- 0

    let compute_annealing ann =
        if ann.run_as_server then begin
            Scale.add Scale.NumberNoMax ">> Annealing " " "
        end else ();
        while not (Param.is_stop_iter ann.param) do
            if ann.run_as_server then begin
                Scale.clearline ();
                Scale.print (Printf.sprintf ">> Annealing parameters: %s\n" (Param.to_string ann.param));
                Scale.redisplay ();
                Scale.inc_and_display ();
                Scale.add Scale.StaticInfix (Param.to_string ann.param) " > ";
                Scale.display ()
            end else ();
            ann.compute_inout
                ~scale_prob:(Param.scale_prob ann.param)
                ~localscore:(Param.localscore ann.param)
                ann.noentries ann.diffthr ann.rndthr ann.epsilon
                ann.ruleprob ann.ruleprior 
                ann.start ann.get_chart ann.nocharts ann.fwdbwd
                ~run_as_server:ann.run_as_server ~run_as_client:ann.run_as_client;
            ann.param <- Param.inc ann.param;
            if ann.run_as_server then
                Scale.discard ()
            else ();
        done;
        if ann.run_as_server then
            Scale.discard ()
        else ()

    let compute
            ?(run_as_server=false) ?(run_as_client=false)
            n diffthr rndthr 
            epsilon ruleprob ruleprior start get_chart nocharts fwdbwd
            compute_inout initparam =
        let ann = create
            n diffthr rndthr 
            epsilon ruleprob ruleprior start get_chart nocharts fwdbwd
            compute_inout initparam 
            run_as_server run_as_client in
        compute_annealing ann

end

module EmParam = struct

    let title = "None"

    type t = {
        mutable cnt : int;
        iterno      : int;
        mutable s   : bool;
    }

    let create iterno = 
    {
        cnt    = 0;
        iterno = iterno;
        s      = false;
    }

    let to_string param =
        sprintf "[%d iters]" param.iterno

    let scale_prob param prob deriv = prob

    let localscore param i j = 1.0

    let inc param =
    {
        cnt    = param.cnt + 1;
        iterno = param.iterno;
        s      = param.cnt = 0;
    }

    let is_stop_iter param = param.s

end

module DetParam = struct

    let title = "Deterministic Annealing"

    type t = {
        temp      : float;
        epoch     : float;
        mutable s : bool;
    }

    let create inittemp epoch =
        assert (inittemp > 0.0 && epoch > 1.0);
        {
            temp      = inittemp;
            epoch     = epoch;
            s         = false;
        }

    let to_string param =
        sprintf "[temperature = %.2e]"
            param.temp

    let scale_prob param prob deriv =
        prob ** param.temp

    let localscore param i j = 1.0

    let inc param =
        let newtemp = param.temp *. param.epoch in
        {
            temp  = if newtemp < 1.0 then newtemp else 1.0;
            epoch = param.epoch;
            s     = param.temp >= 1.0;
        }

    let is_stop_iter param = param.s

end

module SkewedDetParam = struct

    let title = "Skewed Deterministic Annealing"

    type genmodel_t = DerivGenModel.t

    type t = {
        temp      : float;
        epoch     : float;
        rulebias  : genmodel_t;
        mutable s : bool;
    }

    let create inittemp epoch rulebias =
        assert (inittemp > 0.0 && epoch > 1.0);
        {
            temp     = inittemp;
            epoch    = epoch;
            rulebias = rulebias;
            s        = false;
        }

    let to_string param =
        sprintf "[temperature = %.2e]"
            param.temp

    let scale_prob param prob deriv =
        let bias = DerivGenModel.get_smooth_prob param.rulebias deriv in
        (prob ** param.temp) *. (bias ** (1.0 -. param.temp))

    let localscore param i j = 1.0

    let inc param =
        let newtemp = param.temp *. param.epoch in
        {
            temp     = if newtemp < 1.0 then newtemp else 1.0;
            epoch    = param.epoch;
            rulebias = param.rulebias;
            s        = param.temp >= 1.0;
        }

    let is_stop_iter param = param.s

end

module StructParam = struct

    let title = "Structural Annealing"

    type t = {
        delta      : float;
        epoch      : float;
        finaldelta : float;
    }

    let create initdelta epoch finaldelta =
        assert (epoch <> 0.0);
        {
            delta = initdelta;
            epoch = epoch;
            finaldelta = finaldelta;
        }

    let to_string param =
        sprintf "[delta = %.2e]" param.delta

    let scale_prob param prob deriv = prob

    let localscore param i j =
        exp (param.delta *. float_of_int (abs (i - j)))

    let inc param =
    {
        delta      = param.delta +. param.epoch;
        epoch      = param.epoch;
        finaldelta = param.finaldelta;
    }

    let is_stop_iter param =
        if param.epoch > 0.0 then
            param.delta > param.finaldelta
        else
            param.delta < param.finaldelta

end

module EmAnnealer = Annealer(EmParam)

module DetAnnealer = Annealer(DetParam)

module SkewedDetAnnealer = Annealer(SkewedDetParam)

module StructAnnealer = Annealer(StructParam)
