(***********************************************************
*                                                          *
*      FUNGI - Fast UNsupervised Grammar Induction         *
*                                                          *
*   Originator: Prachya Boonkwan [1], [2]                  *
*               [1] School of Informatics,                 *
*                   University of Edinburgh, UK            *
*               [2] National Electronics and Computer      *
*                   Technology Center (NECTEC), Thailand   *
*                                                          *
*   Copyright (C) 2011 Prachya Boonkwan                    *
*   <p.boonkwan@sms.ed.ac.uk, kaamanita@hotmail.com>       *
*                                                          *
*   This library is free software; you can redistribute    *
*   it and/or modify it under the terms of the GNU         *
*   Lesser General Public License as published by the      *
*   Free Software Foundation; either version 2.1 of the    *
*   License, or (at your option) any later version.        *
*                                                          *
*   This library is distributed in the hope that it will   *
*   be useful, but WITHOUT ANY WARRANTY; without even      *
*   the implied warranty of MERCHANTABILITY or FITNESS     *
*   FOR A PARTICULAR PURPOSE.  See the GNU Lesser          *
*   General Public License for more details.               *
*                                                          *
*   You should have received a copy of the GNU Lesser      *
*   General Public License along with this library; if     *
*   not, write to the Free Software Foundation, Inc.,      *
*   59 Temple Place, Suite 330, Boston, MA  02111-1307     *
*   USA.                                                   *
*                                                          *
***********************************************************)

open Cg
open Chart
open Dict
open Cky
open Chartdb
open Prob
open Deriv
open Probinit
open Models
open Inout
open Em
open Anneal
open Printf
open Utils

type inoutmode_t = StandardInOut | VBInOut

type ann_t = NoAnn | DetAnn | SkewedDetAnn | StructAnn

type annmode_t = 
    NoAnnealing
  | DetAnnealing of float * float
  | SkewedDetAnnealing of float * float
  | StructAnnealing of float * float * float

module AnnealRunner = struct

    type syncat_t = Syncat.t

    let is_bo_consq consq =
        if not (Deriv.is_consq_dtrs consq) then false
        else
            let cwlist = Deriv.get_dtrs consq in
            let rec check cwlist' =
                match cwlist' with
                    [] -> false
                  | (c, w) :: rest ->
                        if Syncat.is_bo_cat c then true
                        else check rest in
            check cwlist

    let score_prior 
            deriv dict start defprior knownderivprior startderivprior =
        let prems = Deriv.get_prems deriv
        and consq = Deriv.get_consq deriv in
        if Syncat.is_bo_cat (Deriv.get_syncat prems) || is_bo_consq consq then
            defprior
        else if Deriv.is_consq_lex consq 
                && not (Dict.is_uncertain dict (Deriv.get_word consq)) then
            knownderivprior
        else if Syncat.equal_ign_attrs (Deriv.get_syncat prems) start then
            startderivprior
        else defprior
    
    let init_ruleprior 
            n epsilon dict start defprior knownderivprior startderivprior 
            ruleprob =
        let ruleprior = Models.create_pcfg n epsilon in
        DerivGenModel.iter
            ruleprob
            (fun deriv prob ->
                let prior = score_prior deriv dict start 
                    defprior knownderivprior startderivprior in
                DerivGenModel.set_prob ruleprior deriv prior);
        ruleprior
    
    let init_probs 
            wrapper n epsilon probflags dict start 
            defprior knownderivprior startderivprior 
            run_as_server run_as_client =
        let ruleprob = Models.create_model n epsilon probflags
        and rulebias = Models.create_model n epsilon probflags 
        and fwdbwd = Hmm.create epsilon in
        ProbInit.learn ruleprob rulebias 
            (ChartDbWrapper.find wrapper) 
            (ChartDbWrapper.get_noitems wrapper)
            fwdbwd
            ~run_as_server:run_as_server
            ~run_as_client:run_as_client;
        let ruleprior = init_ruleprior 
            n epsilon dict start defprior knownderivprior startderivprior 
            ruleprob in
        (ruleprob, rulebias, ruleprior, fwdbwd)
    
    let run_anneal
            inoutmode babystep annmode rulebias 
            run_as_server run_as_client
            n diffthr rndthr epsilon ruleprob ruleprior start
            get_chart nocharts fwdbwd =
        let compute_inout = match inoutmode with
            StandardInOut -> 
                if not babystep then 
                    let svr =
                        if run_as_server then 
                            Some (Em.StandardInOutAlgo.server (Utils.server_conf_estep ()) ())
                        else None in
                    StandardInOutAlgo.compute ~svr:svr
                else 
                    let svr =
                        if run_as_server then 
                            Some (Em.StandardBabystepInOutAlgo.server (Utils.server_conf_estep ()) ())
                        else None in
                    StandardBabystepInOutAlgo.compute ~svr:svr
          | VBInOut -> 
                if not babystep then 
                    let svr =
                        if run_as_server then 
                            Some (Em.VBInOutAlgo.server (Utils.server_conf_estep ()) ())
                        else None in
                    VBInOutAlgo.compute ~svr:svr
                else 
                    let svr =
                        if run_as_server then 
                            Some (Em.VBBabystepInOutAlgo.server (Utils.server_conf_estep ()) ())
                        else None in
                    VBBabystepInOutAlgo.compute ~svr:svr
        in
        match annmode with
            NoAnnealing -> 
                let initparam = EmParam.create rndthr in
                EmAnnealer.compute
                    ~run_as_server:run_as_server ~run_as_client:run_as_client
                    n diffthr rndthr epsilon ruleprob ruleprior 
                    start get_chart nocharts fwdbwd compute_inout initparam
          | DetAnnealing (inittemp, epoch) -> 
                (* finaltemp = 1.0 always *)
                let initparam = DetParam.create inittemp epoch in
                DetAnnealer.compute 
                    ~run_as_server:run_as_server ~run_as_client:run_as_client
                    n diffthr rndthr epsilon ruleprob ruleprior 
                    start get_chart nocharts fwdbwd compute_inout initparam
          | SkewedDetAnnealing (inittemp, epoch) -> 
                (* finaltemp = 1.0 always *)
                let initparam = SkewedDetParam.create 
                    inittemp epoch rulebias in
                SkewedDetAnnealer.compute
                    ~run_as_server:run_as_server ~run_as_client:run_as_client
                    n diffthr rndthr epsilon ruleprob ruleprior 
                    start get_chart nocharts fwdbwd compute_inout initparam
          | StructAnnealing (initdelta, epoch, finaldelta) -> 
                let initparam = StructParam.create 
                    initdelta epoch finaldelta in
                StructAnnealer.compute
                    ~run_as_server:run_as_server ~run_as_client:run_as_client
                    n diffthr rndthr epsilon ruleprob ruleprior 
                    start get_chart nocharts fwdbwd compute_inout initparam
    
    let train 
            compute n diffthr rndthr epsilon ruleprob ruleprior 
            start wrapper fwdbwd =
        let get_chart = ChartDbWrapper.find wrapper 
        and nocharts = ChartDbWrapper.length wrapper in
        compute 
            n diffthr rndthr epsilon 
            ruleprob ruleprior start get_chart nocharts fwdbwd (* ;
        ChartDbWrapper.close wrapper *)
    
    let save_prob ruleprob ofname =
        DerivGenModel.to_file ruleprob ofname

    let save_hmm fwdbwd ohmm =
        Hmm.to_file fwdbwd ohmm
    
    let compute 
            probflags inoutmode babystep annmode diffthr rndthr epsilon
            start dict defprior knownderivprior startderivprior leapfrogthr
            fnames ofname ohmm run_as_server run_as_client_probinit run_as_client_estep =
        let n = 1000 in
        let wrapper = ChartDbWrapper.from_files fnames in
        let (ruleprob, rulebias, ruleprior, fwdbwd) = 
            if not run_as_client_estep then
                init_probs 
                    wrapper n epsilon probflags dict start 
                    defprior knownderivprior startderivprior 
                    run_as_server run_as_client_probinit 
            else (
                DerivGenModel.create 100 0.0 [],
                DerivGenModel.create 100 0.0 [],
                DerivGenModel.create 100 0.0 [],
                Hmm.create epsilon;
            ) in
        if babystep then Em.leapfrogthr := leapfrogthr
        else ();
        if not run_as_client_probinit then begin
            train (run_anneal inoutmode babystep annmode rulebias run_as_server run_as_client_estep) 
                n diffthr rndthr epsilon ruleprob ruleprior start wrapper fwdbwd;
            save_prob ruleprob ofname;
            if ohmm <> "" then
                save_hmm fwdbwd ohmm
            else ();
            flush stdout
        end else ()

end

let set_probflags (probflags, flagstrs) str =
    let probflag = match str with
        "roleemis" -> RoleEmis
      | "depemis" -> JuliaEmis
      | "lexemis" -> LexEmis
      | "heademis" -> HeadEmis 
      | _ -> failwith "set_probflags" in
    probflags := probflag :: !probflags;
    match str with
        "roleemis" -> flagstrs := "role emission" :: !flagstrs
      | "depemis" -> flagstrs := "dependent word emission" :: !flagstrs
      | "lexemis" -> flagstrs := "lexicon emission" :: !flagstrs
      | "heademis" -> flagstrs := "head-word emission" :: !flagstrs
      | _ -> failwith "set_probflags"

let set_inoutmode (inoutmode, inoutstr) str =
    let m = match str with
        "standard" -> StandardInOut
      | "vb" -> VBInOut
      | _ -> failwith "set_inoutmode" in
    inoutmode := m;
    match str with
        "standard" -> inoutstr := "Standard EM"
      | "vb" -> inoutstr := "Variational Bayesian EM"
      | _ -> failwith "set_inoutmode"

let set_annmode (annmode, annstr) str =
    let m = match str with
        "none" -> NoAnn
      | "da" -> DetAnn
      | "sda" -> SkewedDetAnn
      | "sa" -> StructAnn
      | _ -> failwith "set_annmode" in
    annmode := m;
    match str with
        "none" -> annstr := "None"
      | "da" -> annstr := "Deterministic Annealing"
      | "sda" -> annstr := "Skewed Deterministic Annealing"
      | "sa" -> annstr := "Structural Annealing"
      | _ -> failwith "set_annmode"

let set_start start str =
    let lexbuf = Lexing.from_string str in
    let syncat = Ruleparser.syncat Rulelexer.token lexbuf in
    start := syncat

let conv_annparam annmode initparam epoch finalparam =
    match annmode with
        NoAnn -> NoAnnealing
      | DetAnn -> DetAnnealing (initparam, epoch)
      | SkewedDetAnn -> SkewedDetAnnealing (initparam, epoch)
      | StructAnn -> StructAnnealing (initparam, epoch, finalparam)
