(***********************************************************
*                                                          *
*      FUNGI - Fast UNsupervised Grammar Induction         *
*                                                          *
*   Originator: Prachya Boonkwan [1], [2]                  *
*               [1] School of Informatics,                 *
*                   University of Edinburgh, UK            *
*               [2] National Electronics and Computer      *
*                   Technology Center (NECTEC), Thailand   *
*                                                          *
*   Copyright (C) 2011 Prachya Boonkwan                    *
*   <p.boonkwan@sms.ed.ac.uk, kaamanita@hotmail.com>       *
*                                                          *
*   This library is free software; you can redistribute    *
*   it and/or modify it under the terms of the GNU         *
*   Lesser General Public License as published by the      *
*   Free Software Foundation; either version 2.1 of the    *
*   License, or (at your option) any later version.        *
*                                                          *
*   This library is distributed in the hope that it will   *
*   be useful, but WITHOUT ANY WARRANTY; without even      *
*   the implied warranty of MERCHANTABILITY or FITNESS     *
*   FOR A PARTICULAR PURPOSE.  See the GNU Lesser          *
*   General Public License for more details.               *
*                                                          *
*   You should have received a copy of the GNU Lesser      *
*   General Public License along with this library; if     *
*   not, write to the Free Software Foundation, Inc.,      *
*   59 Temple Place, Suite 330, Boston, MA  02111-1307     *
*   USA.                                                   *
*                                                          *
***********************************************************)

open Cg
open Dict
open Printf
open Utils

module HashedInt = struct
    type t = int
    let hash = Hashtbl.hash
    let equal = Pervasives.(=)
end

module ItemTbl = Hashtbl.Make(HashedInt)

module HashedSpan = struct
    type t = int * int
    let hash = Hashtbl.hash
    let equal = Pervasives.(=)
end

module SpanTbl = Hashtbl.Make(HashedSpan)

module Item = struct

    type id_t = int

    type syncat_t = CG.syncat_t

    type depdir_t = CG.depdir_t

    type span_t = int * int

    type word_t = string

    type proddtr_t =
        Lex of word_t
      | Ids of id_t list

    let create_proddtr_lex word = Lex word

    let create_proddtr_ids ids = Ids ids

    let to_string_proddtr proddtr =
        match proddtr with
            Lex word -> 
                "'" ^ word ^ "'"
          | Ids ids -> 
                string_of_list string_of_int ids

    let is_proddtr_lex proddtr =
        match proddtr with
            Lex _ -> true
          | Ids _ -> false

    let is_proddtr_ids proddtr =
        match proddtr with
            Lex _ -> false
          | Ids _ -> true

    let get_proddtr_lex proddtr =
        match proddtr with
            Lex w -> w
          | Ids _ -> failwith "get_proddtr_lex"

    let get_proddtr_ids proddtr =
        match proddtr with
            Lex _ -> failwith "get_proddtr_lex"
          | Ids ids -> ids

    type prod_t = {
        proddtr : proddtr_t;
        headpos : int;
        depdir  : depdir_t;
    }

    let create_prod_ids dtrids headpos depdir = {
        proddtr = Ids dtrids;
        headpos = headpos;
        depdir  = depdir;
    }

    let create_prod_lex word pos = {
        proddtr = Lex word;
        headpos = pos;
        depdir  = Syncat.Unary;
    }

    let to_string_prod prod =
        sprintf "(%s, %d, %s)" 
            (to_string_proddtr prod.proddtr) prod.headpos
            (Syncat.depdir_to_string prod.depdir)

    let get_proddtr prod = prod.proddtr

    let get_headpos prod = prod.headpos

    let get_depdir prod = prod.depdir

    let is_terminal prod =
        is_proddtr_lex prod.proddtr

    let get_word prod =
        match prod.proddtr with
            Lex word -> word
          | Ids _ -> failwith "get_word"

    let get_ids prod =
        match prod.proddtr with
            Lex word -> failwith "get_ids"
          | Ids ids -> ids

    type enumlevel_t = Dict.enumlevel_t

    type t = {
        id            : id_t;
        syncat        : syncat_t;
        span          : span_t;
        headpos       : int;
        mutable prods : prod_t list;
    }

    let create id syncat span headpos prods = {
        id      = id;
        syncat  = syncat;
        span    = span;
        headpos = headpos;
        prods   = prods;
    }

    let to_string item =
        let (begpos, endpos) = item.span in
        sprintf "Item %d (%d, %d, %s, %d): %s"
            item.id begpos endpos (Syncat.to_string item.syncat) 
            item.headpos
            (string_of_list to_string_prod item.prods ~bracket:false)

    let get_id item = item.id

    let get_syncat item = item.syncat

    let get_span item = item.span

    let get_headpos_of_item item = item.headpos

    let get_prods item = item.prods

    let is_term item =
        item.prods = []

    let add_prod item prod =
        if not (List.mem prod item.prods) then
            item.prods <- prod :: item.prods
        else ()

    let set_prods item prods =
        item.prods <- prods

    let equal item1 item2 =
        item1.id = item2.id
            && Syncat.equal item1.syncat item2.syncat
            && item1.span = item2.span
            && item1.headpos = item2.headpos

    let hash item =
        Hashtbl.hash item.id + Syncat.hash item.syncat
            + Hashtbl.hash item.span + Hashtbl.hash item.headpos

end

module Chart = struct

    type item_t = Item.t

    type id_t = Item.id_t

    type span_t = int * int

    type prod_t = Item.prod_t

    type syncat_t = Item.syncat_t

    type word_t = string

    type depdir_t = Item.depdir_t

    type enumlevel_t = Dict.enumlevel_t

    type dtrargs_t = (syncat_t * span_t * int * word_t) list
   
    type t = {
        itemtbl           : item_t ItemTbl.t;
        spantbl           : id_t list SpanTbl.t;
        mutable idcnt     : id_t;
        sentence          : word_t list;
        mutable enumlevel : enumlevel_t;
        mutable derivcnt  : int;
        mutable treecnt   : int;
    }

    type fn_word_t = syncat_t -> span_t -> int -> word_t -> unit

    type fn_dtrs_t = 
        syncat_t -> span_t -> int -> word_t -> depdir_t -> 
        (syncat_t * span_t * int * word_t) list -> unit

    type srchstr_t = BottomUp | TopDown

    let treecnttbl = ItemTbl.create 100

    let create sentence = 
        let n = List.length sentence in
        {
            itemtbl    = ItemTbl.create (n * n);
            spantbl    = SpanTbl.create (n * (n - 1) / 2);
            idcnt      = 0;
            sentence   = sentence;
            enumlevel  = Dict.NoEnum;
            derivcnt   = 0;
            treecnt    = 0;
        }

    let copy chart =
    {
        itemtbl    = ItemTbl.copy chart.itemtbl;
        spantbl    = SpanTbl.copy chart.spantbl;
        idcnt      = chart.idcnt;
        sentence   = chart.sentence;
        enumlevel  = chart.enumlevel;
        derivcnt   = chart.derivcnt;
        treecnt    = chart.treecnt;
    }

    let length chart =
        ItemTbl.length chart.itemtbl

    let to_string_itemtbl itemtbl idcnt =
        let result = ref [] in
        for id = 1 to idcnt do
            if ItemTbl.mem itemtbl id then
                let item = ItemTbl.find itemtbl id in
                result := !result @ [(Item.to_string item)]
            else ()
        done;
        !result

    let to_string_spanentry spantbl begpos endpos =
        let ids = List.sort compare 
            (SpanTbl.find spantbl (begpos, endpos)) in
        sprintf "Span (%d, %d): %s" 
            begpos endpos
            (string_of_list string_of_int ids ~bracket:false)

    let to_string_spantbl spantbl sentlen =
        let result = ref [] in
        for edgelen = 0 to sentlen - 1 do
            for begpos = 0 to sentlen - edgelen - 1 do
                let endpos = begpos + edgelen in
                let span = (begpos, endpos) in
                if SpanTbl.mem spantbl span then
                    result := !result @ [
                        to_string_spanentry spantbl begpos endpos
                    ]
                else ()
            done
        done;
        !result

    let to_string chart =
        let sentlen = List.length chart.sentence in
        sprintf "%s\n%s\nItem count = %d\nSentence = [%s]\nEnumeration = %s\nDerivation count = %d\nTree count = %d"
            (String.concat "\n" 
                (to_string_itemtbl chart.itemtbl chart.idcnt))
            (String.concat "\n" (to_string_spantbl chart.spantbl sentlen))
            chart.idcnt
            (String.concat " " chart.sentence)
            (Dict.enumlevel_to_string chart.enumlevel)
            chart.derivcnt chart.treecnt

    let set_enumlevel chart enumlevel =
        chart.enumlevel <- enumlevel

    let get_enumlevel chart =
        chart.enumlevel

    let get_derivcnt chart =
        chart.derivcnt

    let get_treecnt chart =
        chart.treecnt

    let get_sentence chart = 
        chart.sentence

    let get_nonterms chart =
        let result = ref [] in
        ItemTbl.iter
            (fun _ item ->
                let syncat = Item.get_syncat item in
                if not (List.mem syncat !result) then
                    result := syncat :: !result
                else ())
            chart.itemtbl;
        !result

    let get_item chart id =
        ItemTbl.find chart.itemtbl id

    let get_all_ids chart =
        ItemTbl.fold 
            (fun id _ result -> id :: result) 
            chart.itemtbl []

    let mem chart span =
        SpanTbl.mem chart.spantbl span 

    let add_new_item_to_chart chart item =
        let span = Item.get_span item in
        let oldset =
            if not (mem chart span) then []
            else SpanTbl.find chart.spantbl span in
        let newset = (Item.get_id item) :: oldset in
        SpanTbl.replace chart.spantbl span newset

    let create_new_item chart syncat span headpos =
        chart.idcnt <- chart.idcnt + 1;
        let id = chart.idcnt in
        let prods = [] in
        let item = Item.create id syncat span headpos prods in
        ItemTbl.replace chart.itemtbl id item;
        add_new_item_to_chart chart item;
        item

    let find_syncat_in_span chart span syncat headpos =
        if not (mem chart span) then
            raise Not_found
        else
            (* Note that List.find can also raise Not_found. *)
            let id = List.find 
                (fun id -> 
                    let item = get_item chart id in
                    Item.get_syncat item = syncat
                      && Item.get_headpos_of_item item = headpos)
                (SpanTbl.find chart.spantbl span) in
            ItemTbl.find chart.itemtbl id

    let add_deriv chart span syncat headpos prod =
        let item =
            try
                find_syncat_in_span chart span syncat headpos
            with Not_found ->
                create_new_item chart syncat span headpos in
        Item.add_prod item prod

    let add_deriv_ret_item chart span syncat headpos prod =
        let item =
            try
                find_syncat_in_span chart span syncat headpos
            with Not_found ->
                create_new_item chart syncat span headpos in
        Item.add_prod item prod;
        item
   
    let get_ids_from_span chart span =
        if not (mem chart span) then []
        else SpanTbl.find chart.spantbl span

    let get_items chart span =
        List.map 
            (get_item chart) 
            (get_ids_from_span chart span)

    let rec find_dtrids_from_prods prods result =
        match prods with
            [] -> elim_dups result
          | prod :: rest ->
            begin
                let ids' = match Item.get_proddtr prod with
                    Item.Lex _ -> []
                  | Item.Ids idlist -> idlist in
                find_dtrids_from_prods rest (result @ ids')
            end

    let find_dtrids_from_id chart id result =
        let item = get_item chart id in
        let id = Item.get_id item
        and prods = Item.get_prods item in
        if not (List.mem id result) then
            find_dtrids_from_prods prods result
        else result

    let rec find_rchids_by_id ?(depstr=None) chart rchids id =
        let find_headdep ids depdir =
            let len = List.length ids in
            let leftmost = List.nth ids 0
            and rightmost = List.nth ids (len - 1) in
            let headid, depid = begin match depdir with
                Syncat.Left -> leftmost, rightmost
              | Syncat.Right -> rightmost, leftmost
              | Syncat.Coor -> leftmost, rightmost
              | Syncat.Unary -> leftmost, rightmost
            end in
            (* Printf.printf "***** now get the headid %d and the depid %d\n" headid depid; *)
            let headitem = get_item chart headid
            and depitem = get_item chart depid in
            (* Printf.printf "***** about to get the headpos and the deppos\n"; *)
            let headpos = Item.get_headpos_of_item headitem
            and deppos = Item.get_headpos_of_item depitem in
            (* Printf.printf "***** now get the headpos %d and the deppos %d\n" headpos deppos; *)
            (headpos, deppos) in
        let assoc depstr' headpos deppos =
            Dep.DepStruct.mem depstr' headpos deppos in
        let is_any_dtrs_survive item = begin
            match depstr with
                None -> true
              | Some depstr' -> begin
                    let flag = ref false in
                    List.iter
                        (fun prod -> 
                            let proddtr = Item.get_proddtr prod in
                            match proddtr with
                                Item.Lex _ -> flag := true
                              | Item.Ids ids -> begin
                                    let depdir = Item.get_depdir prod in
                                    let (headpos, deppos) = find_headdep ids depdir in
                                    (* Printf.printf "***** the headpos %d and the deppos %d are %s\n" headpos deppos (if assoc depstr' headpos deppos then "associated" else "not associated"); *)
                                    if headpos = deppos then ()
                                    else if assoc depstr' headpos deppos then 
                                        flag := true
                                    else ()
                                end)
                        (Item.get_prods item);
                    !flag
                end 
            end in
        let run rchids ids = 
            List.iter 
                (find_rchids_by_id ~depstr:depstr chart rchids) 
                ids in
        if not (List.mem id !rchids) then begin
            rchids := id :: !rchids;
            (* Printf.printf "***** start find_rchids_by_id for item %d\n" id; *)
            let item = get_item chart id in
            (* Printf.printf "***** check if any dtrs survive\n"; *)
            let survive = is_any_dtrs_survive item in
            (* Printf.printf "***** survive = %s\n" (if survive then "yes" else "no"); *)
            List.iter
                (fun prod ->
                    let depdir = Item.get_depdir prod in
                    let proddtr = Item.get_proddtr prod in
                    (* Printf.printf "***** check the proddtr\n"; *)
                    match proddtr with
                        Item.Lex _ -> ()
                      | Item.Ids ids -> begin
                            let (headpos, deppos) = find_headdep ids depdir in
                            match depstr with
                                None -> run rchids ids
                              | Some depstr' ->
                                    if headpos = deppos then
                                        run rchids ids
                                    else if not survive then 
                                        run rchids ids
                                    else if assoc depstr' headpos deppos then 
                                        run rchids ids
                                    else ()
                        end (* ;
                    Printf.printf "***** end of checking\n"; *)
                )
                (Item.get_prods item) (* ;
            Printf.printf "***** end find_rchids_by_id for item %d\n" id;
            Printf.printf "***** rchids = [%s]\n" (String.concat ", " (List.map string_of_int !rchids)); *)
        end else ()

    let find_rchids ?(depstr=None) chart span syncat =
        if SpanTbl.mem chart.spantbl span then begin
            (* let (x, y) = span in
            Printf.printf "***** start find_rchids for span (%d, %d)\n" x y; *)
            let rchids = ref [] in
            List.iter
                (fun id ->
                    let item = get_item chart id in
                    let syncat' = Item.get_syncat item in
                    if Syncat.equal_ign_attrs syncat' syncat then
                        find_rchids_by_id ~depstr:depstr chart rchids id
                    else ())
                (SpanTbl.find chart.spantbl span);
            (* Printf.printf "***** end find_rchids for span (%d, %d)\n" x y; *)
            !rchids
        end else []

    let clean_item chart id unrchids =
        let item = get_item chart id in
        if Item.is_term item then ()
        else begin
            let newprods = ref [] in
            List.iter
                (fun prod ->
                    let proddtr = Item.get_proddtr prod in
                    match proddtr with
                        Item.Lex _ -> 
                            newprods := prod :: !newprods
                      | Item.Ids ids -> begin
                            if (intersect ids unrchids) = [] then
                                newprods := prod :: !newprods
                            else ()
                        end
                )
                (Item.get_prods item);
            Item.set_prods item !newprods
        end

    let elim_excess ?(depstr=None) chart syncat =
        (* Printf.printf "***** start elim_excess\n"; *)
        let span = (0, List.length chart.sentence - 1) in
        let rchids = find_rchids chart span syncat ~depstr:depstr in
        let unrchids = diff (get_all_ids chart) rchids in
        let emptyspans = ref [] in
        SpanTbl.iter
            (fun span ids ->
                let newids = diff ids unrchids in
                SpanTbl.replace chart.spantbl span newids;
                if List.length newids = 0 then
                    emptyspans := span :: !emptyspans
                else ())
            chart.spantbl;
        List.iter
            (fun span ->
                SpanTbl.remove chart.spantbl span)
            !emptyspans;
        List.iter
            (fun id ->
                ItemTbl.remove chart.itemtbl id)
            unrchids;
        List.iter
            (fun id ->
                clean_item chart id unrchids)
            rchids
        (* ;
        Printf.printf "***** end elim_excess\n" *)

    let rec iter_dtrargs ?(backoff=false) chart fn_dtrs 
            syncat span headpos word depdir ids dtrargs =
        match ids with
            [] ->
            begin
                if not backoff then
                    fn_dtrs syncat span headpos word depdir dtrargs
                else
                    let deptag = Syncat.get_deptag syncat in
                    fn_dtrs 
                        (Syncat.bo_cat deptag) span headpos word 
                        depdir dtrargs
            end
          | id :: rest ->
            begin
                let item' = get_item chart id in
                let syncat' = Item.get_syncat item'
                and span' = Item.get_span item' in
                List.iter
                    (fun prod' ->
                        let headpos' = Item.get_headpos prod' in
                        let word' = List.nth 
                            (get_sentence chart) headpos' in
                        let dtr' = 
                            if not backoff 
                                    || syncat' = Syncat.conn_cat then
                                (syncat', span', headpos', word') 
                            else
                                let deptag' = Syncat.get_deptag syncat' in
                                (Syncat.bo_cat deptag', 
                                    span', headpos', word') in
                        iter_dtrargs ~backoff:backoff chart fn_dtrs 
                            syncat span headpos word depdir rest 
                            (dtrargs @ [dtr']))
                    (Item.get_prods item')
            end

    let iter_deriv_by_dtrs_onelevel ?(backoff=false) chart 
            fn_word fn_dtrs syncat span headpos word depdir ids =
        iter_dtrargs ~backoff:backoff chart 
            fn_dtrs syncat span headpos word depdir ids []

    let iter_derivs_by_prod_onelevel 
            ?(backoff=false) chart fn_word fn_dtrs item prod =
        let syncat = Item.get_syncat item
        and span = Item.get_span item 
        and headpos = Item.get_headpos prod 
        and depdir = Item.get_depdir prod in
        match (Item.get_proddtr prod) with
            Item.Lex word -> 
                fn_word syncat span headpos word
          | Item.Ids ids ->
                let word = List.nth (get_sentence chart) headpos in
                iter_deriv_by_dtrs_onelevel ~backoff:backoff chart 
                    fn_word fn_dtrs syncat span headpos word depdir ids

    let iter_derivs_by_id_onelevel 
            ?(backoff=false) chart fn_word fn_dtrs id =
        let item = get_item chart id in
        List.iter
            (iter_derivs_by_prod_onelevel 
                ~backoff:backoff chart fn_word fn_dtrs item)
            (Item.get_prods item)

    let iter_derivs_onelevel ?(backoff=false) chart fn_word fn_dtrs span =
        List.iter
            (iter_derivs_by_id_onelevel 
                ~backoff:backoff chart fn_word fn_dtrs)
            (get_ids_from_span chart span)

    let iter_derivs_topdown ?(backoff=false) chart fn_word fn_dtrs span =
        let (begpos, endpos) = span in
        for edgelen = (endpos - begpos) downto 0 do
            for b = begpos to endpos - edgelen do
                let e = b + edgelen in
                let span' = (b, e) in
                if mem chart span' then
                    let ids' = get_ids_from_span chart span' in
                    List.iter 
                        (iter_derivs_by_id_onelevel 
                            ~backoff:backoff chart fn_word fn_dtrs) 
                        ids'
                else ()
            done
        done

    let iter_derivs_bottomup ?(backoff=false) chart fn_word fn_dtrs span =
        let (begpos, endpos) = span in
        for edgelen = 0 to (endpos - begpos) do
            for b = begpos to endpos - edgelen do
                let e = b + edgelen in
                let span' = (b, e) in
                if mem chart span' then
                    let ids' = get_ids_from_span chart span' in
                    List.iter 
                        (iter_derivs_by_id_onelevel 
                            ~backoff:backoff chart fn_word fn_dtrs) 
                        ids'
                else ()
            done
        done

    let iter_derivs chart srchstr fn_word fn_dtrs span =
        match srchstr with
            BottomUp -> 
            begin
                iter_derivs_bottomup 
                    ~backoff:false chart fn_word fn_dtrs span
            end
          | TopDown ->
            begin
                iter_derivs_topdown 
                    ~backoff:false chart fn_word fn_dtrs span
            end

    let is_success chart =
        mem chart (0, List.length chart.sentence - 1)

    let count_derivs chart =
        (* Printf.printf "***** start count derivs\n"; *)
        let n = List.length chart.sentence in
        let (begpos, endpos) = (0, n - 1) in
        let cnt = ref 0 in
        for edgelen = 0 to (endpos - begpos) do
            for b = begpos to endpos - edgelen do
                let e = b + edgelen in
                let span' = (b, e) in
                if mem chart span' then
                    let ids' = get_ids_from_span chart span' in
                    List.iter 
                        (iter_derivs_by_id_onelevel 
                            ~backoff:false chart 
                            (fun _ _ _ _ -> incr cnt)
                            (fun _ _ _ _ _ _ -> incr cnt))
                        ids'
                else ()
            done
        done;
        (* Printf.printf "***** end count derivs\n"; *)
        chart.derivcnt <- !cnt

    let rec count_trees_item chart item =
        let id = Item.get_id item in
        if not (ItemTbl.mem treecnttbl id) then begin
            ItemTbl.replace treecnttbl id 0;
            List.iter
                (fun prod ->
                    let proddtr = Item.get_proddtr prod in
                    if Item.is_proddtr_lex proddtr then begin
                        let oldcnt = ItemTbl.find treecnttbl id in
                        ItemTbl.replace treecnttbl id (oldcnt + 1)
                    end else begin
                        let oldcnt = ItemTbl.find treecnttbl id in
                        let ids = Item.get_proddtr_ids proddtr in
                        let cnt = List.fold_left 
                            (fun x id' -> 
                                let item' = get_item chart id' in
                                x * count_trees_item chart item') 
                            1 ids in
                        ItemTbl.replace treecnttbl id (oldcnt + cnt)
                    end)
                (Item.get_prods item)
        end else ();
        ItemTbl.find treecnttbl id

    let count_trees chart =
        let n = List.length chart.sentence in
        let span = (0, n - 1) in
        let ids = get_ids_from_span chart span in
        chart.treecnt <- List.fold_left
            (fun x id ->
                let item = get_item chart id in
                x + count_trees_item chart item)
            0 ids;
        ItemTbl.clear treecnttbl

end
