(** Chart Representation. *)

(***********************************************************
*                                                          *
*      FUNGI - Fast UNsupervised Grammar Induction         *
*                                                          *
*   Originator: Prachya Boonkwan [1], [2]                  *
*               [1] School of Informatics,                 *
*                   University of Edinburgh, UK            *
*               [2] National Electronics and Computer      *
*                   Technology Center (NECTEC), Thailand   *
*                                                          *
*   Copyright (C) 2011 Prachya Boonkwan                    *
*   <p.boonkwan@sms.ed.ac.uk, kaamanita@hotmail.com>       *
*                                                          *
*   This library is free software; you can redistribute    *
*   it and/or modify it under the terms of the GNU         *
*   Lesser General Public License as published by the      *
*   Free Software Foundation; either version 2.1 of the    *
*   License, or (at your option) any later version.        *
*                                                          *
*   This library is distributed in the hope that it will   *
*   be useful, but WITHOUT ANY WARRANTY; without even      *
*   the implied warranty of MERCHANTABILITY or FITNESS     *
*   FOR A PARTICULAR PURPOSE.  See the GNU Lesser          *
*   General Public License for more details.               *
*                                                          *
*   You should have received a copy of the GNU Lesser      *
*   General Public License along with this library; if     *
*   not, write to the Free Software Foundation, Inc.,      *
*   59 Temple Place, Suite 330, Boston, MA  02111-1307     *
*   USA.                                                   *
*                                                          *
***********************************************************)

(** Chart Item. *)
module Item :
sig

    (** Item ID. *)
    type id_t = int
    
    (** Syntactic category. *)
    type syncat_t = Cg.CG.syncat_t
    
    (** Dependency direction. *)
    type depdir_t = Cg.CG.depdir_t
    
    (** Position span. *)
    type span_t = int * int
    
    (** Lexical item. *)
    type word_t = string
    
    (** Daughter of production. *)
    type proddtr_t
    
    (** [create_proddtr_lex w] creates a lexical-item daughter from a
        word [w]. *)
    val create_proddtr_lex : word_t -> proddtr_t
    
    (** [create_proddtr_ids dtrids] create an item-list daughter from a
        list of daughter item IDs. *)
    val create_proddtr_ids : id_t list -> proddtr_t

    (** [to_string_proddtr p] represents a daughter of production [p]
        with a string. *)
    val to_string_proddtr : proddtr_t -> string
    
    (** [is_proddtr_lex p] checks if [p] is a lexical-item daughter. *)
    val is_proddtr_lex : proddtr_t -> bool
    
    (** [is_proddtrs_ids p] checks if [p] is an item-list daughter. *)
    val is_proddtr_ids : proddtr_t -> bool

    val get_proddtr_lex : proddtr_t -> string

    val get_proddtr_ids : proddtr_t -> id_t list
    
    (** Item production. *)
    type prod_t
    
    (** [create_prod_ids ids h d] creates an item production with a 
        list of daughter item IDs [ids], a position of the syntactic 
        head [h], and a dependency direction [d]. *)
    val create_prod_ids : id_t list -> int -> depdir_t -> prod_t
    
    (** [create_prod_lex w h] creates an item production with a lexical 
        item [w] and its position [h]. *)
    val create_prod_lex : word_t -> int -> prod_t

    (** [to_string_prod p] represents a production [p] with a string. *)
    val to_string_prod : prod_t -> string
    
    (** [get_proddtr p] gets the daughter production from a production
        [p]. *)
    val get_proddtr : prod_t -> proddtr_t
    
    (** [get_headpos p] gets the position of the syntactic head from 
        a production [p]. *)
    val get_headpos : prod_t -> int
    
    (** [get_depdir p] gets the dependency direction from a production 
        [p]. *)
    val get_depdir : prod_t -> depdir_t
    
    (** [is_terminal p] checks if a production [p] is a terminal node. *)
    val is_terminal : prod_t -> bool
    
    (** [get_word p] gets the word from a production [p] if it is a 
        terminal node. If [p] is not a terminal node, a failure
        exception will be thrown. *)
    val get_word : prod_t -> word_t
    
    (** [get_ids p] get the daughter item list from the production if 
        it is not a terminal node. If [p] is not a nonterminal node, a
        failure exception will be thrown. *)
    val get_ids : prod_t -> id_t list

    (** Chart item. *)
    type t
    
    (** [create id c s h ps] creates a chart item from an ID [id], 
        a syntactic category [c], a span [s], a head word's position [h],
        and a list of productions [ps]. *)
    val create : id_t -> syncat_t -> span_t -> int -> prod_t list -> t

    (** [to_string item] represents a chart item [item] with a string. *)
    val to_string : t -> string
    
    (** [get_id item] gets the item ID of an item [item]. *)
    val get_id : t -> id_t
    
    (** [get_syncat item] gets the syntactic category of an item 
        [item]. *)
    val get_syncat : t -> syncat_t
    
    (** [get_span item] gets the position span of an item [item]. *)
    val get_span : t -> span_t

    val get_headpos_of_item : t -> int
    
    (** [get_prods item] get the production list of an item [item]. *)
    val get_prods : t -> prod_t list
    
    (** [is_term item] checks if an item [item] is a terminal node. *)
    val is_term : t -> bool
    
    (** [add_prod item p] adds a new production [p] to an item [item]. *)
    val add_prod : t -> prod_t -> unit

    (** [set_prods item ps] sets the production list of an item [item]
        to be [ps]. *)
    val set_prods : t -> prod_t list -> unit
    
    (** [equal i1 i2] checks if two items [i1] and [i2] are equal. 
        Two items are said to equal if they have the same IDs, 
        syntactic categories, and spans. *)
    val equal : t -> t -> bool
    
    (** [hash item] calculates the hash value of an item [item]. *)
    val hash : t -> int

end

(** Packed Chart. *)
module Chart :
sig

    (** Item. *)
    type item_t = Item.t
    
    (** Item ID. *)
    type id_t = Item.id_t
    
    (** Span. *)
    type span_t = int * int
    
    (** Production. *)
    type prod_t = Item.prod_t
    
    (** Syntactic category. *)
    type syncat_t = Item.syncat_t

    (** Lexical item. *)
    type word_t = string

    (** Dependency direction. *)
    type depdir_t = Item.depdir_t

    type dtrargs_t = (syncat_t * span_t * int * word_t) list
    
    (** Chart. *)
    type t
    
    (** Function for dealing with lexicon items: 
        syncat -> span -> headpos -> word. *)
    type fn_word_t = syncat_t -> span_t -> int -> word_t -> unit

    (** Function for dealing with all other derivations:
        syncat -> span -> headpos -> head -> depdir -> 
        (syncat, span, headpos, head) list. *)
    type fn_dtrs_t = 
        syncat_t -> span_t -> int -> word_t -> depdir_t ->
        dtrargs_t -> unit
    
    (** Search strategy. *)
    type srchstr_t = 
        BottomUp
      | TopDown

    (** Enumeration level. *)
    type enumlevel_t = Dict.Dict.enumlevel_t

    (** [create sent] creates a packed chart with a sentence [sent]. *)
    val create : word_t list -> t

    val copy : t -> t

    (** [length chart] finds the length of the sentence in a packed chart
        [chart]. *)
    val length : t -> int

    (** [to_string chart] represents a packed chart [chart] with a 
        string. *)
    val to_string : t -> string

    (** [set_enumlevel chart e] sets the enumeration level for a 
        packed chart [chart]. *)
    val set_enumlevel : t -> enumlevel_t -> unit

    (** [get_enumlevel chart] gets the enumeration level for a packed
        chart [chart]. *)
    val get_enumlevel : t -> enumlevel_t

    (** [get_derivcnt chart] gets the total number of derivations in
        a packed chart [chart]. *)
    val get_derivcnt : t -> int

    val get_treecnt : t -> int

    (** [get_sentence chart] gets the sentence from a packed chart 
        [chart]. *)
    val get_sentence : t -> word_t list

    val get_nonterms : t -> syncat_t list
    
    (** [get_item chart id] finds a chart item in a packed chart 
        [chart] by an ID [id]. *)
    val get_item : t -> id_t -> item_t
    
    (** [get_all_ids chart] finds all item IDs in a packed chart 
        [chart]. *)
    val get_all_ids : t -> id_t list
    
    (** [mem chart span] checks if a span [span] is in a packed chart 
        [chart]. *)
    val mem : t -> span_t -> bool
    
    (** [find_syncat_in_span chart sp c h] finds an item which has a 
        specified syntactic category [c], span [sp], and head word's 
        position [h] in a packed chart [chart]. *)
    val find_syncat_in_span : t -> span_t -> syncat_t -> int -> item_t
    
    (** [add_deriv chart sp c h prod] adds a derivation containing
        a span [sp], a syntactic category [c], and a head word's position
        [h] into a packed chart [chart]. *)
    val add_deriv : t -> span_t -> syncat_t -> int -> prod_t -> unit
    
    val add_deriv_ret_item : t -> span_t -> syncat_t -> int -> prod_t -> item_t
    
    (** [get_ids_from_span chart sp] finds all item IDs in a span [sp]
        of a packed chart [chart]. *)
    val get_ids_from_span : t -> span_t -> id_t list
    
    (** [get_items chart sp] finds all items in a span [sp] of a packed
        chart [chart]. *)
    val get_items : t -> span_t -> item_t list
    
    (** [elim_excess chart] eliminates all edges that do not compose into
        a complete parse. *)
    val elim_excess : ?depstr:Dep.DepStruct.t option -> t -> syncat_t -> unit

    (** [iter_derivs chart str fn_word fn_dtrs sp] iterates all 
        derivations in a packed chart [chart] with the specified search 
        strategy (bottom-up or top-down) [str] beginning from the 
        specified span with two functions: [fn_word] dealing with 
        preterminal derivations and [fn_dtrs] dealing with all other 
        derivations. *)
    val iter_derivs : 
        t -> srchstr_t -> fn_word_t -> fn_dtrs_t -> span_t -> unit

    (** [iter_derivs_onelevel chart fn_word fn_dtrs sp ~backoff:b] 
        iterates all derivations in a packed chart [chart] only in a
        specified span [sp] with two functions: [fn_word] dealing 
        with preterminal derivations and [fn_dtrs] dealing with all 
        other derivations. If the backoff flag [b] is set to true,
        all iterated derivations are backed-off by replacing syntactic
        categories with a backoff category. The backoff flag is set to
        false by default. *)
    val iter_derivs_onelevel : 
        ?backoff:bool -> t -> fn_word_t -> fn_dtrs_t -> span_t -> unit

    (** [is_success chart] checks if at least one successful parse is 
        obtained in a packed chart [chart]. *)
    val is_success : t -> bool

    (** [count_derivs chart] counts the total number of derivations in
        a packed chart [chart]. *)
    val count_derivs : t -> unit

    val count_trees : t -> unit

end
