(***********************************************************
*                                                          *
*      FUNGI - Fast UNsupervised Grammar Induction         *
*                                                          *
*   Originator: Prachya Boonkwan [1], [2]                  *
*               [1] School of Informatics,                 *
*                   University of Edinburgh, UK            *
*               [2] National Electronics and Computer      *
*                   Technology Center (NECTEC), Thailand   *
*                                                          *
*   Copyright (C) 2011 Prachya Boonkwan                    *
*   <p.boonkwan@sms.ed.ac.uk, kaamanita@hotmail.com>       *
*                                                          *
*   This library is free software; you can redistribute    *
*   it and/or modify it under the terms of the GNU         *
*   Lesser General Public License as published by the      *
*   Free Software Foundation; either version 2.1 of the    *
*   License, or (at your option) any later version.        *
*                                                          *
*   This library is distributed in the hope that it will   *
*   be useful, but WITHOUT ANY WARRANTY; without even      *
*   the implied warranty of MERCHANTABILITY or FITNESS     *
*   FOR A PARTICULAR PURPOSE.  See the GNU Lesser          *
*   General Public License for more details.               *
*                                                          *
*   You should have received a copy of the GNU Lesser      *
*   General Public License along with this library; if     *
*   not, write to the Free Software Foundation, Inc.,      *
*   59 Temple Place, Suite 330, Boston, MA  02111-1307     *
*   USA.                                                   *
*                                                          *
***********************************************************)

open Cg
open Dict
open Dictloader
open Models
open Annealrunner
open Printf

let main =
    Utils.print_version ();
    let clientpath = ref "/usr/local/bin"
    and hostname = ref "localhost"
    and port_probinit = ref 12345
    and port_estep = ref 12346
    and noclients = ref 4
    and recurrence = ref 0
    and fnames = ref [] in
    let args = [ 
        ("--recur",
            Arg.Set_int recurrence,
            "[int]   Set the recurrence time for the client side to launch (0 and negative mean no termination; default = 0)");
        ("--host",
            Arg.Set_string hostname,
            "[hostname]   Set the Map/Reduce server's hostname");
        ("--port-probinit",
            Arg.Set_int port_probinit,
            "[int]   Set the Map/Reduce server's port");
        ("--port-estep",
            Arg.Set_int port_estep,
            "Set the Map/Reduce server's port");
        ("--noclients",
            Arg.Set_int noclients,
            "[int]   Set the number of clients to automatically fork (must be more than 0; default = 4)");
        ("--clientpath",
            Arg.Set_string clientpath,
            "[path/to/bin]   Set the path of client programs (default = /usr/local/bin)")
    ]
    and fn_anon = (fun f -> fnames := !fnames @ [f])
    and usagemsg =
        "Client program for probability table initialization.\n\n"
        ^ (sprintf "Usage: %s" Sys.argv.(0)) in
    Arg.parse args fn_anon usagemsg;
    Utils.set_poolsize !noclients;
    Utils.set_clientpath !clientpath;
    if !noclients <= 0 then Arg.usage args usagemsg else ();
    fprintf stderr ">> Server\n";
    fprintf stderr "     Hostname          : %s (%s)\n" !hostname (Utils.get_inet_addr !hostname);
    fprintf stderr "     Port for probinit : %d\n" !port_probinit;
    fprintf stderr "     Port for E-Step   : %d\n" !port_estep;
    fprintf stderr "     Clients           : %d\n" !noclients;
    if !recurrence > 0 then
        fprintf stderr "     Recurrence        : %d\n\n" !recurrence
    else
        fprintf stderr "     Recurrence        : no termination\n\n";
    flush stderr;
    Utils.set_hostname !hostname;
    Utils.set_port_probinit !port_probinit;
    Utils.set_port_estep !port_estep;
    Utils.set_poolsize !noclients;
    let arglist_probinit = [
        "--host"; !hostname;
        "--port"; string_of_int !port_probinit;
        "--clientpath"; !clientpath ]
    and arglist_estep = [
        "--host"; !hostname;
        "--port"; string_of_int !port_estep;
        "--clientpath"; !clientpath ] in
    let args_probinit = Array.of_list arglist_probinit in
    let args_estep = Array.of_list arglist_estep in
    let forked_probinit = !clientpath ^ "/fungi_probinit" 
    and forked_estep = !clientpath ^ "/fungi_estep" in
    let run i = begin
        fprintf stderr ">> Iteration %d\n" i;
        fprintf stderr ">> Press Ctrl-C to terminate manually.\n";
        flush stderr;
        let pids_probinit =
            JoinHelper.do_forks 
                forked_probinit 
                (JoinHelper.Same_arguments args_probinit) 
                !noclients in
        let pids_estep =
            JoinHelper.do_forks 
                forked_estep 
                (JoinHelper.Same_arguments args_estep) 
                !noclients in
        fprintf stderr ">> These processes will automatically terminate when the trainer finishes its task.\n";
        fprintf stderr ">> However, if necessary, you can manually kill them by copying and pasting the following command:\n";
        fprintf stderr "\n     kill %s %s\n\n" 
            (String.concat " " (List.map string_of_int pids_probinit))
            (String.concat " " (List.map string_of_int pids_estep));
        flush stderr;
        for i = 1 to 2 * !noclients do
            ignore (Unix.wait ())
        done 
    end in
    let i = ref 1 in
    while (!recurrence > 0 && !i <= !recurrence) || !recurrence <= 0 do
        run !i;
        incr i
    done
    (* JoinHelper.wait_forever () *)

