(***********************************************************
*                                                          *
*      FUNGI - Fast UNsupervised Grammar Induction         *
*                                                          *
*   Originator: Prachya Boonkwan [1], [2]                  *
*               [1] School of Informatics,                 *
*                   University of Edinburgh, UK            *
*               [2] National Electronics and Computer      *
*                   Technology Center (NECTEC), Thailand   *
*                                                          *
*   Copyright (C) 2011 Prachya Boonkwan                    *
*   <p.boonkwan@sms.ed.ac.uk, kaamanita@hotmail.com>       *
*                                                          *
*   This library is free software; you can redistribute    *
*   it and/or modify it under the terms of the GNU         *
*   Lesser General Public License as published by the      *
*   Free Software Foundation; either version 2.1 of the    *
*   License, or (at your option) any later version.        *
*                                                          *
*   This library is distributed in the hope that it will   *
*   be useful, but WITHOUT ANY WARRANTY; without even      *
*   the implied warranty of MERCHANTABILITY or FITNESS     *
*   FOR A PARTICULAR PURPOSE.  See the GNU Lesser          *
*   General Public License for more details.               *
*                                                          *
*   You should have received a copy of the GNU Lesser      *
*   General Public License along with this library; if     *
*   not, write to the Free Software Foundation, Inc.,      *
*   59 Temple Place, Suite 330, Boston, MA  02111-1307     *
*   USA.                                                   *
*                                                          *
***********************************************************)

open Printf
open Utils
open Dep
open Tree

module ConllDep = struct

    type entry_t = {
        id            : int;
        form          : string;
        lemma         : string;
        cpostag       : string;
        postag        : string;
        feats         : string list;
        mutable head  : int;
        deprel        : string;
        mutable phead : int option;
        pdeprel       : string option;
    }

    let repr_entry entry =
        sprintf "Head (%d) -> Dep (%d, %s)" entry.head entry.id entry.form

    let compare_entry entry1 entry2 =
        Pervasives.compare entry1.id entry2.id

    let entry_to_conll entry =
        sprintf "%d\t%s\t%s\t%s\t%s\t%s\t%d\t%s\t%s\t%s"
            (entry.id + 1) entry.form entry.lemma entry.cpostag entry.postag
            (String.concat "|" entry.feats) (entry.head + 1) entry.deprel
            (match entry.phead with
                None -> "_"
              | Some phead -> string_of_int (phead + 1))
            (match entry.pdeprel with
                None -> "_"
              | Some pdeprel -> pdeprel)

    type entries_t = entry_t list * string list * int list * int

    let entries_to_conll (entries : entries_t) : string =
        let (entrylist, _, _, _) = entries in
        String.concat "\n"
            (List.map entry_to_conll entrylist)

    type content_t = entries_t list

    type depstr_t = DepStruct.t

    exception End_of_entries

    exception Comment_line

    let prohforms = [
        "!"; "\""; "#"; "$"; "%"; "&";
        "\'"; "("; ")"; "*"; "+"; ",";
        "-"; "."; "/"; ":"; ";"; "<";
        "="; ">"; "?"; "@"; "["; "\\";
        "]"; "^"; "_"; "`"; "{"; "|"; 
        "}"; "~"; "`"
    ]

    let prohtags = [
        "punct"; "punc"; "pnct"; "pu";
        "PUNCT"; "PUNC"; "PNCT"; "PU";
        "."; ","; ":"; "``"; "\'\'"; "$";
        ":pound:"; "-LRB-"; "-RRB-"; "-NONE-"]
    
    let create_entry 
            id form lemma cpostag postag feats 
            head deprel phead pdeprel =
    {
        id      = id;
        form    = form;
        lemma   = lemma;
        cpostag = cpostag;
        postag  = postag;
        feats   = feats;
        head    = head;
        deprel  = deprel;
        phead   = phead;
        pdeprel = pdeprel;
    }

    let conv_id_option id =
        match id with
            "_" -> None
          | _ -> Some (int_of_string id - 1)

    let conv_deprel_option deprel =
        match deprel with
            "_" -> None
          | _ -> Some deprel

    let read_entry line =
        let tokens = split line in
        if List.length tokens = 0 || String.length line = 0 then
            raise End_of_entries
        else if List.nth tokens 0 = "#" then
            raise Comment_line
        else if List.length tokens = 10 then
            try
                let id = int_of_string (List.nth tokens 0) - 1
                and form = List.nth tokens 1
                and lemma = List.nth tokens 2
                and cpostag = List.nth tokens 3
                and postag = List.nth tokens 4
                and feats = split_delim "|" (List.nth tokens 5)
                and head = int_of_string (List.nth tokens 6) - 1
                and deprel = List.nth tokens 7
                and phead = conv_id_option (List.nth tokens 8)
                and pdeprel = conv_deprel_option (List.nth tokens 9) in
                create_entry
                    id form lemma cpostag postag feats 
                    head deprel phead pdeprel
            with (Failure _) ->
                failwith (sprintf "Unrecognizable entry 1: %s" line)
        else
            failwith (sprintf "Unrecognizable entry 2: %s" line)

    let is_proh entry =
        (* List.mem (String.sub entry.form 0 1) prohforms *)
        (* || List.mem (String.sub entry.postag 0 1) prohforms *)
        (* || *) List.mem entry.cpostag prohtags

    let sort_words entries =
        entries := List.sort compare_entry !entries

    let correct_links (entries : entry_t list ref) =
        let no_words = List.length !entries in
        (* printf "no_words = %d\n" no_words; *)
        List.iter
            (fun entry ->
                (* printf "entry.head = %d\n" entry.head; *)
                if entry.head >= no_words then begin
                    (* printf "correction!\n"; *)
                    entry.head <- -1
                end else ();
                match entry.phead with
                    None -> ()
                  | Some phead ->
                        if phead >= no_words then
                            entry.phead <- Some (-1)
                        else ())
            !entries

    let elim_punc entries sent =
        let n = List.length !entries in
        let i = ref 0 in
        let new_entries = ref [] in
        let new_sent = ref [] in
        List.iter
            (fun entry ->
                if (!i < n - 1) then begin
                    new_entries := !new_entries @ [entry];
                    new_sent := !new_sent @ [List.nth !sent !i]
                end else if not (is_proh entry) then begin
                    new_entries := !new_entries @ [entry];
                    new_sent := !new_sent @ [List.nth !sent !i]
                end else ();
                incr i)
            !entries;
        entries := !new_entries;
        sent := !new_sent

    let rec elim_longsents_inner is_elim sentlen content : content_t =
        match content with
            [] -> []
          | depstr :: rest ->
            begin
                let (entries, sent, delpos, rootpos) = depstr in
                let e' = ref entries
                and s' = ref sent in
                if is_elim then elim_punc e' s'
                else ();
                correct_links e';
                let delposno = 
                    if is_elim then List.length delpos 
                    else 0 in
                let len = List.length entries - delposno in
                if len <= sentlen then
                    let depstr' = (!e', !s', delpos, rootpos) in
                    depstr' :: elim_longsents_inner is_elim sentlen rest
                else
                    elim_longsents_inner is_elim sentlen rest
            end

    let elim_longsents is_elim sentlen content =
        match sentlen with
            None -> content
          | Some sentlen' -> 
                elim_longsents_inner is_elim sentlen' content

    let compile is_elim fname : content_t =
        Scale.add (Scale.WindMill 3000) "Loading content ... " "";
        let ifhdl = open_in fname in
        let content = ref [] in
        let entries = ref []
        and sent = ref []
        and delpos = ref []
        and rootpos = ref (-1) in begin
            try
                while true do
                    try begin
                        let line = input_line ifhdl in
                        let entry = read_entry line in
                        begin
                            Scale.inc ();
                            Scale.display ();
                            entries := !entries @ [entry];
                            if is_proh entry then
                                delpos := !delpos @ [entry.id]
                            else ();
                            if entry.head = -1 then
                                rootpos := entry.id
                            else ()
                        end
                    end with 
                        End_of_entries -> begin
                            sort_words entries;
                            List.iter
                                (fun entry ->
                                    sent := !sent @ [entry.cpostag])
                                !entries;
                            if is_elim then elim_punc entries sent
                            else ();
                            correct_links entries;
                            if List.length !entries > 0 then begin
                                content := (!entries, !sent, !delpos, !rootpos) :: !content;
                                entries := [];
                                sent := [];
                                delpos := [];
                                rootpos := 0
                            end else ()
                        end
                      | Comment_line -> ()
                done
            with End_of_file -> begin
                if List.length !entries > 0 then
                    content := (!entries, !sent, !delpos, !rootpos) :: !content
                else ();
                close_in ifhdl
            end
        end;
        Scale.discard ();
        let result = List.rev !content in
        result

    let compile_many is_elim fnames =
        let result = ref [] in
        List.iter
            (fun fname ->
                let entries = compile is_elim fname in
                result := !result @ entries)
            fnames;
        !result

    let repl_id id idmap =
        List.assoc id idmap

    let repl_id_option id idmap =
        match id with
            None -> None
          | Some id' -> Some (repl_id id' idmap)

    let conv_entries_to_depstr entries sent delpos rootpos : depstr_t =
        let depstr = DepStruct.create (List.length entries) sent rootpos in
        List.iter
            (fun entry ->
                if entry.head >= 0 then
                    DepStruct.add depstr entry.head entry.id
                else ())
            entries;
        depstr

    let rec conv_to_depstrs content =
        match content with
            [] -> []
          | (entries, sent, delpos, rootpos) :: rest ->
                conv_entries_to_depstr entries sent delpos rootpos
                :: conv_to_depstrs rest

    let read_postags_deprels content =
        let postags = ref []
        and deprels = ref [] in
        List.iter
            (fun (entries, _, _, _) ->
                List.iter
                    (fun entry ->
                        if not (List.mem entry.cpostag !postags) then
                            postags := entry.cpostag :: !postags
                        else ();
                        if not (List.mem entry.deprel !deprels) then
                            deprels := entry.deprel :: !deprels
                        else ())
                    entries)
            content;
        (!postags, !deprels)

    let from_file fname =
        let content = compile false fname in
        (* let sentid = ref 0 in *)
        let result = List.map
            (fun (entries, sent, delpos, rootpos) ->
                (* incr sentid;
                printf "Sent ID = %d\n" !sentid;
                printf "Sentence = %s\n" (String.concat " " sent); *)
                let depstr = conv_entries_to_depstr entries sent delpos rootpos in
                (* printf "%s" (DepStruct.to_string depstr); *)
                depstr)
            content in
        result

    let from_files fnames =
        let result = ref [] in
        List.iter
            (fun fname ->
                let depstrs = from_file fname in
                result := !result @ depstrs)
            fnames;
        !result

end
