(***********************************************************
*                                                          *
*      FUNGI - Fast UNsupervised Grammar Induction         *
*                                                          *
*   Originator: Prachya Boonkwan [1], [2]                  *
*               [1] School of Informatics,                 *
*                   University of Edinburgh, UK            *
*               [2] National Electronics and Computer      *
*                   Technology Center (NECTEC), Thailand   *
*                                                          *
*   Copyright (C) 2011 Prachya Boonkwan                    *
*   <p.boonkwan@sms.ed.ac.uk, kaamanita@hotmail.com>       *
*                                                          *
*   This library is free software; you can redistribute    *
*   it and/or modify it under the terms of the GNU         *
*   Lesser General Public License as published by the      *
*   Free Software Foundation; either version 2.1 of the    *
*   License, or (at your option) any later version.        *
*                                                          *
*   This library is distributed in the hope that it will   *
*   be useful, but WITHOUT ANY WARRANTY; without even      *
*   the implied warranty of MERCHANTABILITY or FITNESS     *
*   FOR A PARTICULAR PURPOSE.  See the GNU Lesser          *
*   General Public License for more details.               *
*                                                          *
*   You should have received a copy of the GNU Lesser      *
*   General Public License along with this library; if     *
*   not, write to the Free Software Foundation, Inc.,      *
*   59 Temple Place, Suite 330, Boston, MA  02111-1307     *
*   USA.                                                   *
*                                                          *
***********************************************************)

open Cg
open Models
open Decoderrunner
open Printf

let main =
    Utils.print_version ();
    let output = ref None 
    and decoding = ref ViterbiDecoding
    and probtbl = ref None
    and fwdbwd = ref None
    and stepsize = ref 0.05
    and rndthr = ref 10
    and start = ref (Syncat.create_atom "ROOT")
    and epsilon = ref 1e-25
    and catconst = ref 1.0
    and fnames = ref [] in
    let args = [
        ("--probtbl",
            Arg.String (fun p -> probtbl := Some p),
            "[ruleprob.probtbl]   Specify the rule probabilities (required)");
        ("--hmm",
            Arg.String (set_fwdbwd fwdbwd),
            "[ruleprob.hmm]   Specify the HMM. The dual decomposition will be automatically selected if this option is specified. (optional)");
        ("--stepsize",
            Arg.Set_float stepsize,
            "[float]   Set the step size of the dual decomposition (default = 0.05)");
        ("--rndthr",
            Arg.Set_int rndthr,
            "[int]   Set the threshold number of rounds of the dual decomposition (default = 10)");
        ("--decoding",
            Arg.Symbol (
                ["viterbi"; "mbr"],
                set_decoding decoding),
            "  Select the decoding method (Standard Viterbi or Minimum Bayes Risk)");
        ("--epsilon",
            Arg.Set_float epsilon,
            "[float]   Set the value of epsilon (a very small value) (default = 1e-25)");
        ("--catconst",
            Arg.Set_float catconst,
            "[float]   Set the value of category preference constant (default = 1.0)");
        ("--output",
            Arg.String (fun o -> output := Some o),
            "[output.treedb]   Specify the output tree database (required)")
    ]
    and fn_anon = (fun f -> fnames := !fnames @ [f])
    and usagemsg =
        "Run the Viterbi parser.\n\n" 
        ^ (sprintf "Usage: %s [options] chart1 chart2 ..." Sys.argv.(0)) in
    Arg.parse args fn_anon usagemsg;
    begin
        match !fwdbwd with
            Some fwdbwd ->
                decoding := DualDecomp (fwdbwd, !stepsize, !rndthr)
          | None -> ()
    end;
    match !probtbl, !output with
        Some probtblfname, Some treedbfname ->
        begin
            fprintf stderr "[1mDecoding the best parses from the charts [%s] with the model %s.[0m\n"
                (String.concat ", " !fnames) probtblfname;
            fprintf stderr ">> Progress: ";
            flush stderr;
            let vtb = 
                DecoderRunner.create !decoding !start
                    !epsilon probtblfname !catconst 
                    !fnames treedbfname in
            DecoderRunner.run vtb;
            fprintf stderr "Done\n";
            flush stderr
        end
      | _ -> Arg.usage args usagemsg
