(***********************************************************
*                                                          *
*      FUNGI - Fast UNsupervised Grammar Induction         *
*                                                          *
*   Originator: Prachya Boonkwan [1], [2]                  *
*               [1] School of Informatics,                 *
*                   University of Edinburgh, UK            *
*               [2] National Electronics and Computer      *
*                   Technology Center (NECTEC), Thailand   *
*                                                          *
*   Copyright (C) 2011 Prachya Boonkwan                    *
*   <p.boonkwan@sms.ed.ac.uk, kaamanita@hotmail.com>       *
*                                                          *
*   This library is free software; you can redistribute    *
*   it and/or modify it under the terms of the GNU         *
*   Lesser General Public License as published by the      *
*   Free Software Foundation; either version 2.1 of the    *
*   License, or (at your option) any later version.        *
*                                                          *
*   This library is distributed in the hope that it will   *
*   be useful, but WITHOUT ANY WARRANTY; without even      *
*   the implied warranty of MERCHANTABILITY or FITNESS     *
*   FOR A PARTICULAR PURPOSE.  See the GNU Lesser          *
*   General Public License for more details.               *
*                                                          *
*   You should have received a copy of the GNU Lesser      *
*   General Public License along with this library; if     *
*   not, write to the Free Software Foundation, Inc.,      *
*   59 Temple Place, Suite 330, Boston, MA  02111-1307     *
*   USA.                                                   *
*                                                          *
***********************************************************)

open Cg
open Chart
open Prob
open Deriv
open Tree
open Utils

module TreeInst = struct

    type span_t = int * int
    
    type syncat_t = Syncat.t
    
    type word_t = string
    
    type t = span_t * syncat_t * word_t

    let create span syncat head = (span, syncat, head)

    let get_span inst =
        let (span, _, _) = inst in
        span

    let get_syncat inst =
        let (_, syncat, _) = inst in
        syncat

    let get_head inst =
        let (_, _, head) = inst in
        head
    
    let hash inst = 
        let (span, cat, head) = inst in
        Hashtbl.hash span + Syncat.hash cat + Hashtbl.hash head
    
    let equal inst1 inst2 =
        let (span1, cat1, head1) = inst1
        and (span2, cat2, head2) = inst2 in
        span1 = span2 && Syncat.equal cat1 cat2 && head1 = head2

    let to_string inst =
        let ((begpos, endpos), syncat, word) = inst in
        Printf.sprintf "(%d, %d, %s, %s)" begpos endpos (Syncat.to_string syncat) word

end

module HmmInst = struct

    type idx_t = int
    
    type syncat_t = Syncat.t
    
    type t = idx_t * syncat_t

    let create idx syncat = (idx, syncat)

    let get_idx inst =
        let (idx,_) = inst in
        idx

    let get_syncat inst =
        let (_, syncat) = inst in
        syncat
    
    let hash inst = 
        let (idx, cat) = inst in
        Hashtbl.hash idx + Syncat.hash cat
    
    let equal inst1 inst2 =
        let (idx1, cat1) = inst1
        and (idx2, cat2) = inst2 in
        idx1 = idx2 && Syncat.equal cat1 cat2

    let to_string inst =
        let (idx, syncat) = inst in
        Printf.sprintf "(%d, %s)" idx (Syncat.to_string syncat)

end

module TreeTbl = Hashtbl.Make(TreeInst)

module HmmTbl = Hashtbl.Make(HmmInst)

module TreeViterbi = struct

    type syncat_t = Syncat.t

    type word_t = string

    type genmodel_t = DerivGenModel.t

    type chart_t = Chart.t

    type tree_t = Tree.t

    type span_t = int * int

    type inst_t = TreeInst.t

    type insttbl_t = (tree_t * float) TreeTbl.t

    type t = {
        insttbl         : insttbl_t;
        ruleprob        : genmodel_t;
        chart           : chart_t;
        start           : syncat_t;
        endpos          : int;
        catconst        : float;
        mutable maxroot : (inst_t * float) option;
    }

    let create ruleprob chart start catconst =
        (* Printf.printf "(viterbi.create           ) Sentence = [%s]\n" (String.concat " " (Chart.get_sentence chart)); *)
        let sentlen = List.length (Chart.get_sentence chart) in
        {   
            insttbl  = TreeTbl.create (sentlen * sentlen);
            ruleprob = ruleprob;
            chart    = chart;
            start    = start;
            endpos   = sentlen - 1;
            catconst = catconst;
            maxroot  = None;
        }

    let mem_inst viterbi inst =
        TreeTbl.mem viterbi.insttbl inst

    let get_inst viterbi inst =
        if mem_inst viterbi inst then
            TreeTbl.find viterbi.insttbl inst
        else raise Not_found

    let get_tree viterbi inst =
        let (tree, _) = get_inst viterbi inst in
        tree

    let get_prob viterbi inst =
        let (_, prob) = get_inst viterbi inst in
        prob

    let set_inst viterbi inst tree prob =
        if not (mem_inst viterbi inst) 
                || get_prob viterbi inst < prob then
            TreeTbl.replace viterbi.insttbl inst (tree, prob)
        else ()

    let set_maxroot viterbi inst prob =
        if TreeInst.get_span inst = (0, viterbi.endpos) 
                && Syncat.equal_ign_attrs 
                    (TreeInst.get_syncat inst) viterbi.start then
            match viterbi.maxroot with
                None ->
                    viterbi.maxroot <- Some (inst, prob)
              | Some (inst', prob') ->
                begin
                    if prob > prob' then
                        viterbi.maxroot <- Some (inst, prob)
                    else ()
                end
        else () 

    let build_lex_tree viterbi syncat span headpos word =
        Scale.inc ();
        Scale.display ();
        let inst = TreeInst.create span syncat word in
        let deriv = conv_word_to_deriv syncat word in
        let lexnode = Tree.create_term word span in
        let catscore = Pervasives.log (Syncat.score viterbi.catconst syncat) in
        let tree = Tree.create_nonterm 
            syncat word [lexnode] span headpos Syncat.Unary in
        let prob = 
            catscore +. Pervasives.log (DerivGenModel.get_smooth_prob viterbi.ruleprob deriv) in
        set_inst viterbi inst tree prob;
        set_maxroot viterbi inst prob

    let build_dtrs_tree viterbi syncat span headpos word depdir dtrargs =
        Scale.inc ();
        Scale.display ();
        let inst = TreeInst.create span syncat word in
        let deriv = conv_dtrargs_to_deriv syncat word depdir dtrargs in
        let dtrs = ref [] in
        let catscore = Pervasives.log (Syncat.score viterbi.catconst syncat) in
        let prob = 
            ref (catscore +. Pervasives.log (DerivGenModel.get_smooth_prob viterbi.ruleprob deriv)) in
        List.iter
            (fun (syncat', span', headpos', word') ->
                let inst' = TreeInst.create span' syncat' word' in
                let (tree', prob') = get_inst viterbi inst' in
                dtrs := !dtrs @ [tree'];
                prob := !prob +. prob')
            dtrargs;
        let tree = Tree.create_nonterm
            syncat word !dtrs span headpos depdir in
        set_inst viterbi inst tree !prob;
        set_maxroot viterbi inst !prob

    let compute_viterbi viterbi =
        Scale.add (Scale.WindMill 5000) " " "";
        Chart.iter_derivs viterbi.chart
            Chart.BottomUp
            (build_lex_tree viterbi) (build_dtrs_tree viterbi)
            (0, viterbi.endpos);
        Scale.discard ()

    let decapsulate_tree start tree =
        if Syncat.equal_ign_attrs (Tree.get_syncat tree) start then begin
            if Tree.is_nonterm tree then begin
                let dtrs = Tree.get_dtrs tree in
                let depdir = Tree.get_depdir tree in
                match depdir with
                    Syncat.Left -> List.nth dtrs 0
                  | Syncat.Right -> List.nth dtrs 1
                  | Syncat.Coor -> List.nth dtrs 0
                  | Syncat.Unary -> List.nth dtrs 0
            end else tree
        end else tree

    let find_max viterbi =
        let (tree, prob) = match viterbi.maxroot with
            None -> raise Not_found
          | Some (inst, prob) -> begin
                (* Printf.printf "TreeInst = %s\n" (TreeInst.to_string inst); *)
                get_inst viterbi inst
            end in
        (* Printf.printf "tree      = %s\n" (Tree.to_string tree); *)
        let decaptree = decapsulate_tree viterbi.start tree in
        (* Printf.printf "decaptree = %s\n" (Tree.to_string decaptree); *)
        (decaptree, prob)

    let compute ruleprob chart start catconst =
        (* Printf.printf "(viterbi.compute          ) Sentence = [%s]\n" (String.concat " " (Chart.get_sentence chart)); *)
        let viterbi = create ruleprob chart start catconst in
        compute_viterbi viterbi;
        find_max viterbi

end

module TreeDualDecomp = struct

    type syncat_t = Syncat.t

    type word_t = string

    type genmodel_t = DerivGenModel.t

    type chart_t = Chart.t

    type tree_t = Tree.t

    type span_t = int * int

    type treeinst_t = TreeInst.t

    type hmminst_t = HmmInst.t

    type treetbl_t = (tree_t * float) TreeTbl.t

    type hmmtbl_t = (syncat_t list * float) HmmTbl.t

    type pentbl_t = float HmmTbl.t

    type t = {
        treetbl             : treetbl_t;
        hmmtbl              : hmmtbl_t;
        pentbl              : pentbl_t;
        stepsize            : float;
        ruleprob            : genmodel_t;
        initprob            : Hmm.InitProb.t;
        transprob           : Hmm.TransProb.t;
        emisprob            : Hmm.EmisProb.t;
        chart               : chart_t;
        start               : syncat_t;
        endpos              : int;
        catconst            : float;
        rndthr              : int;
        mutable treemaxroot : (treeinst_t * float) option;
        mutable hmmmaxroot  : (hmminst_t * float) option;
    }

    let create ruleprob fwdbwd chart start catconst stepsize rndthr =
        (* Printf.printf "(viterbi.create           ) Sentence = [%s]\n" (String.concat " " (Chart.get_sentence chart)); *)
        let sentlen = List.length (Chart.get_sentence chart) in
        {   
            treetbl     = TreeTbl.create (sentlen * sentlen);
            hmmtbl      = HmmTbl.create (sentlen * sentlen);
            pentbl      = HmmTbl.create sentlen;
            stepsize    = stepsize;
            ruleprob    = ruleprob;
            initprob    = Hmm.get_initprob fwdbwd;
            transprob   = Hmm.get_transprob fwdbwd;
            emisprob    = Hmm.get_emisprob fwdbwd;
            chart       = chart;
            start       = start;
            endpos      = sentlen - 1;
            catconst    = catconst;
            rndthr      = rndthr;
            treemaxroot = None;
            hmmmaxroot  = None;
        }

    let mem_treeinst viterbi inst =
        TreeTbl.mem viterbi.treetbl inst

    let mem_hmminst viterbi inst =
        HmmTbl.mem viterbi.hmmtbl inst

    let mem_peninst viterbi inst =
        HmmTbl.mem viterbi.pentbl inst

    let get_treeinst viterbi inst =
        if mem_treeinst viterbi inst then
            TreeTbl.find viterbi.treetbl inst
        else raise Not_found

    let get_hmminst viterbi inst =
        if mem_hmminst viterbi inst then
            HmmTbl.find viterbi.hmmtbl inst
        else raise Not_found

    let get_penalty viterbi idx syncat =
        let inst = HmmInst.create idx syncat in
        if mem_peninst viterbi inst then
            HmmTbl.find viterbi.pentbl inst
        else 0.0

    let set_penalty viterbi idx syncat value =
        let inst = HmmInst.create idx syncat in
        HmmTbl.replace viterbi.pentbl inst value

    let get_tree viterbi inst =
        let (tree, _) = get_treeinst viterbi inst in
        tree

    let get_treeprob viterbi inst =
        let (_, prob) = get_treeinst viterbi inst in
        prob

    let get_path viterbi inst =
        let (path, _) = get_hmminst viterbi inst in
        path

    let get_pathprob viterbi inst =
        let (_, prob) = get_hmminst viterbi inst in
        prob

    let set_treeinst viterbi inst tree prob =
        if not (mem_treeinst viterbi inst) 
                || get_treeprob viterbi inst < prob then
            TreeTbl.replace viterbi.treetbl inst (tree, prob)
        else ()

    let set_hmminst viterbi inst path prob =
        if not (mem_hmminst viterbi inst)
                || get_pathprob viterbi inst < prob then
            HmmTbl.replace viterbi.hmmtbl inst (path, prob)
        else ()

    let set_treemaxroot viterbi inst prob =
        if TreeInst.get_span inst = (0, viterbi.endpos) 
                && Syncat.equal_ign_attrs 
                    (TreeInst.get_syncat inst) viterbi.start then
            match viterbi.treemaxroot with
                None ->
                    viterbi.treemaxroot <- Some (inst, prob)
              | Some (inst', prob') ->
                begin
                    if prob > prob' then
                        viterbi.treemaxroot <- Some (inst, prob)
                    else ()
                end
        else () 

    let set_hmmmaxroot viterbi inst prob =
        if HmmInst.get_idx inst = viterbi.endpos then
            match viterbi.hmmmaxroot with
                None ->
                    viterbi.hmmmaxroot <- Some (inst, prob)
              | Some (inst', prob') ->
                begin
                    if prob > prob' then
                        viterbi.hmmmaxroot <- Some (inst, prob)
                    else ()
                end
        else ()

    let build_lex_tree viterbi syncat span headpos word =
        Scale.inc ();
        Scale.display ();
        let inst = TreeInst.create span syncat word in
        let deriv = conv_word_to_deriv syncat word in
        let lexnode = Tree.create_term word span in
        let catscore = Pervasives.log (Syncat.score viterbi.catconst syncat) in
        let tree = Tree.create_nonterm 
            syncat word [lexnode] span headpos Syncat.Unary in
        let prob = catscore 
            +. Pervasives.log (DerivGenModel.get_smooth_prob viterbi.ruleprob deriv)
            +. get_penalty viterbi headpos syncat
        in
        set_treeinst viterbi inst tree prob;
        set_treemaxroot viterbi inst prob

    let build_dtrs_tree viterbi syncat span headpos word depdir dtrargs =
        Scale.inc ();
        Scale.display ();
        let inst = TreeInst.create span syncat word in
        let deriv = conv_dtrargs_to_deriv syncat word depdir dtrargs in
        let dtrs = ref [] in
        let catscore = Pervasives.log (Syncat.score viterbi.catconst syncat) in
        let prob = 
            ref (catscore +. Pervasives.log (DerivGenModel.get_smooth_prob viterbi.ruleprob deriv)) in
        List.iter
            (fun (syncat', span', headpos', word') ->
                let inst' = TreeInst.create span' syncat' word' in
                let (tree', prob') = get_treeinst viterbi inst' in
                dtrs := !dtrs @ [tree'];
                prob := !prob +. prob')
            dtrargs;
        let tree = Tree.create_nonterm
            syncat word !dtrs span headpos depdir in
        set_treeinst viterbi inst tree !prob;
        set_treemaxroot viterbi inst !prob

    let compute_tree viterbi =
        Scale.add (Scale.WindMill 500) " PCFG " "";
        Scale.display ();
        Chart.iter_derivs viterbi.chart
            Chart.BottomUp
            (build_lex_tree viterbi) (build_dtrs_tree viterbi)
            (0, viterbi.endpos);
        Scale.discard ()

    let compute_hmm viterbi =
        Scale.add (Scale.WindMill 500) " HMM  " "";
        Scale.display ();
        Chart.iter_derivs_onelevel viterbi.chart
            (fun syncat _ _ word ->
                Scale.inc ();
                Scale.display ();
                let inst = HmmInst.create 0 syncat in
                let p = Pervasives.log 
                        (Hmm.InitProb.get_prob viterbi.initprob 
                            (Hmm.InitInst.create syncat))
                    +. Pervasives.log 
                        (Hmm.EmisProb.get_prob viterbi.emisprob 
                            (Hmm.EmisInst.create syncat word))
                    -. get_penalty viterbi 0 syncat
                in
                set_hmminst viterbi inst [syncat] p;
                set_hmmmaxroot viterbi inst p
            )
            (fun _ _ _ _ _ _ -> ())
            (0, 0);
        let s1 = viterbi.endpos - 1 in
        for i = 0 to s1 do
            let span = (i, i) in
            Chart.iter_derivs_onelevel viterbi.chart
                (fun syncat _ _ word ->
                    let inst = HmmInst.create i syncat in
                    let (path, p) = get_hmminst viterbi inst in
                    let j = i + 1 in
                    let span' = (j, j) in
                    Chart.iter_derivs_onelevel viterbi.chart
                        (fun syncat' _ _ word' ->
                            Scale.inc ();
                            Scale.display ();
                            let inst' = HmmInst.create j syncat' in
                            let p' = p 
                                +. Pervasives.log 
                                    (Hmm.TransProb.get_prob viterbi.transprob 
                                        (Hmm.TransInst.create syncat syncat'))
                                +. Pervasives.log
                                    (Hmm.EmisProb.get_prob viterbi.emisprob
                                        (Hmm.EmisInst.create syncat' word'))
                                -. get_penalty viterbi j syncat'
                            in
                            set_hmminst viterbi inst' (path @ [syncat']) p';
                            set_hmmmaxroot viterbi inst' p'
                        )
                        (fun _ _ _ _ _ _ -> ())
                        span'
                )
                (fun _ _ _ _ _ _ -> ())
                span
        done;
        Scale.discard ()

    let compare_tree_and_path viterbi =
        let treepath = match viterbi.treemaxroot with
            Some (treeinst, _) ->
                let (tree, _) = get_treeinst viterbi treeinst in
                Tree.to_catwordlist tree
          | None -> raise Not_found in
        let hmmpath = match viterbi.hmmmaxroot with
            Some (hmminst, _) ->
                let (hmmpath, _) = get_hmminst viterbi hmminst in
                hmmpath
          | None -> raise Not_found in
        let flag = ref true in
        for i = 0 to List.length treepath - 1 do
            let (c1, _) = List.nth treepath i
            and c2 = List.nth hmmpath i in
            if not (Syncat.equal c1 c2) then begin
                let p1 = get_penalty viterbi i c1
                and p2 = get_penalty viterbi i c2 in
                let p1' = p1 -. viterbi.stepsize
                and p2' = p2 +. viterbi.stepsize in
                set_penalty viterbi i c1 p1';
                set_penalty viterbi i c2 p2';
                flag := false
            end else ()
        done;
        !flag

    let decapsulate_tree start tree =
        if Syncat.equal_ign_attrs (Tree.get_syncat tree) start then begin
            if Tree.is_nonterm tree then begin
                let dtrs = Tree.get_dtrs tree in
                let depdir = Tree.get_depdir tree in
                match depdir with
                    Syncat.Left -> List.nth dtrs 0
                  | Syncat.Right -> List.nth dtrs 1
                  | Syncat.Coor -> List.nth dtrs 0
                  | Syncat.Unary -> List.nth dtrs 0
            end else tree
        end else tree

    let run viterbi =
        let i = ref 0 
        and stopflag = ref false in
        Scale.add (Scale.Number viterbi.rndthr) " Round  " " > ";
        while !i < viterbi.rndthr && not (!stopflag) do
            Scale.inc ();
            Scale.display ();
            compute_tree viterbi;
            compute_hmm viterbi;
            stopflag := compare_tree_and_path viterbi;
            i := !i + 1;
            if not (!stopflag) && !i < viterbi.rndthr then begin
                TreeTbl.clear viterbi.treetbl;
                HmmTbl.clear viterbi.hmmtbl
            end;
        done;
        Scale.discard ()

    let find_treemax viterbi =
        let (tree, prob) = 
            match viterbi.treemaxroot with
                None -> raise Not_found
              | Some (inst, prob) -> begin
                    (* Printf.printf "TreeInst = %s\n" (TreeInst.to_string inst); *)
                    flush stderr;
                    get_treeinst viterbi inst
                end 
        in
        (* Printf.printf "tree      = %s\n" (Tree.to_string tree); *)
        let decaptree = decapsulate_tree viterbi.start tree in
        (* Printf.printf "decaptree = %s\n" (Tree.to_string decaptree); *)
        (decaptree, prob)

    let compute ruleprob fwdbwd chart start catconst stepsize rndthr =
        (* Printf.printf "(viterbi.compute          ) Sentence = [%s]\n" (String.concat " " (Chart.get_sentence chart)); *)
        let viterbi = create ruleprob fwdbwd chart start catconst stepsize rndthr in
        run viterbi;
        find_treemax viterbi

end

module HeadInst = struct

    type syncat_t = Syncat.t

    type t = int * int * syncat_t * string

    let create begpos endpos headcat headword =
        (begpos, endpos, headcat, headword)

    let hash = Hashtbl.hash

    let equal = Pervasives.(=)

end

module HeadCountTbl = Hashtbl.Make (HeadInst)

module SpanInst = struct

    type t = int * int

    let create begpos endpos = (begpos, endpos)

    let hash = Hashtbl.hash

    let equal = Pervasives.(=)

end

module SpanCountTbl = Hashtbl.Make (SpanInst)

module TreeMinBayesRisk = struct

    type syncat_t = Syncat.t

    type word_t = string

    type genmodel_t = DerivGenModel.t

    type chart_t = Chart.t

    type tree_t = Tree.t

    type span_t = int * int

    type inst_t = TreeInst.t

    type insttbl_t = (tree_t * float) TreeTbl.t

    type headinst_t = HeadInst.t

    type headcounttbl_t = float HeadCountTbl.t

    type spaninst_t = SpanInst.t

    type spancounttbl_t = float SpanCountTbl.t

    type t = {
        insttbl         : insttbl_t;
        headcounttbl    : headcounttbl_t;
        spancounttbl    : spancounttbl_t;
        ruleprob        : genmodel_t;
        chart           : chart_t;
        start           : syncat_t;
        endpos          : int;
        catconst        : float;
        mutable minroot : (inst_t * float) option;
    }

    let create ruleprob chart start catconst =
        (* Printf.printf "(mbr.create           ) Sentence = [%s]\n" (String.concat " " (Chart.get_sentence chart)); *)
        let sentlen = List.length (Chart.get_sentence chart) in
        {   
            insttbl      = TreeTbl.create (sentlen * sentlen);
            headcounttbl = HeadCountTbl.create (sentlen * sentlen * 100);
            spancounttbl = SpanCountTbl.create (sentlen * sentlen * 10);
            ruleprob     = ruleprob;
            chart        = chart;
            start        = start;
            endpos       = sentlen - 1;
            catconst     = catconst;
            minroot      = None;
        }

    let mem_inst mbr inst =
        TreeTbl.mem mbr.insttbl inst

    let get_inst mbr inst =
        if mem_inst mbr inst then
            TreeTbl.find mbr.insttbl inst
        else raise Not_found

    let get_tree mbr inst =
        let (tree, _) = get_inst mbr inst in
        tree

    let get_prob mbr inst =
        let (_, prob) = get_inst mbr inst in
        prob

    let set_inst mbr inst tree prob =
        if not (mem_inst mbr inst) 
                || get_prob mbr inst < prob then
            TreeTbl.replace mbr.insttbl inst (tree, prob)
        else ()

    let set_minroot mbr inst prob =
        if TreeInst.get_span inst = (0, mbr.endpos) 
                && Syncat.equal_ign_attrs 
                    (TreeInst.get_syncat inst) mbr.start then
            match mbr.minroot with
                None ->
                    mbr.minroot <- Some (inst, prob)
              | Some (inst', prob') ->
                begin
                    if prob < prob' then
                        mbr.minroot <- Some (inst, prob)
                    else ()
                end
        else () 

    let inc_headscore mbr begpos endpos syncat word value =
        let headinst = HeadInst.create begpos endpos syncat word in
        if not (HeadCountTbl.mem mbr.headcounttbl headinst) then
            HeadCountTbl.replace mbr.headcounttbl headinst value
        else
            let score = HeadCountTbl.find mbr.headcounttbl headinst in
            HeadCountTbl.replace mbr.headcounttbl headinst (value +. score)
    
    let inc_spanscore mbr begpos endpos value =
        let spaninst = SpanInst.create begpos endpos in
        if not (SpanCountTbl.mem mbr.spancounttbl spaninst) then
            SpanCountTbl.replace mbr.spancounttbl spaninst value
        else
            let score = SpanCountTbl.find mbr.spancounttbl spaninst in
            SpanCountTbl.replace mbr.spancounttbl spaninst (value +. score)

    let build_risktbl_lex mbr syncat span headpos word =
        Scale.inc ();
        Scale.display ();
        let (begpos, endpos) = span in
        inc_headscore mbr begpos endpos syncat word 1.0;
        inc_spanscore mbr begpos endpos 1.0

    let build_risktbl_dtrs mbr syncat span headpos word depdir dtrargs =
        Scale.inc ();
        Scale.display ();
        let (begpos, endpos) = span in
        List.iter
            (fun (syncat', span', headpos', word') ->
                let (begpos', endpos') = span' in
                let headinst' = HeadInst.create begpos' endpos' syncat' word' in
                let spaninst' = SpanInst.create begpos' endpos' in
                inc_headscore mbr begpos endpos syncat word
                    (HeadCountTbl.find mbr.headcounttbl headinst');
                inc_spanscore mbr begpos endpos
                    (SpanCountTbl.find mbr.spancounttbl spaninst'))
            dtrargs;
        inc_headscore mbr begpos endpos syncat word 1.0;
        inc_spanscore mbr begpos endpos 1.0

    let compute_risktbl mbr =
        Scale.add (Scale.WindMill 5000) " MBR " "";
        Chart.iter_derivs mbr.chart
            Chart.BottomUp
            (build_risktbl_lex mbr) (build_risktbl_dtrs mbr)
            (0, mbr.endpos);
        Scale.discard ()

    let riskscore mbr begpos endpos syncat word =
        let headinst = HeadInst.create begpos endpos syncat word
        and spaninst = SpanInst.create begpos endpos in
        let spanscore = SpanCountTbl.find mbr.spancounttbl spaninst
        and headscore = HeadCountTbl.find mbr.headcounttbl headinst in
        spanscore -. headscore

    let build_lex_tree mbr syncat span headpos word =
        Scale.inc ();
        Scale.display ();
        let inst = TreeInst.create span syncat word in
        let deriv = conv_word_to_deriv syncat word in
        let lexnode = Tree.create_term word span in
        let catscore = Pervasives.log (Syncat.score mbr.catconst syncat) in
        let tree = Tree.create_nonterm 
            syncat word [lexnode] span headpos Syncat.Unary in
        let prob = 
            catscore *. (DerivGenModel.get_smooth_prob mbr.ruleprob deriv) in
        let (begpos, endpos) = span in
        let risk = riskscore mbr begpos endpos syncat word in
        set_inst mbr inst tree (prob *. risk);
        set_minroot mbr inst prob

    let build_dtrs_tree mbr syncat span headpos word depdir dtrargs =
        Scale.inc ();
        Scale.display ();
        let inst = TreeInst.create span syncat word in
        let deriv = conv_dtrargs_to_deriv syncat word depdir dtrargs in
        let dtrs = ref [] in
        let catscore = Syncat.score mbr.catconst syncat in
        let prob = 
            ref (catscore *. DerivGenModel.get_smooth_prob mbr.ruleprob deriv) in
        let (begpos, endpos) = span in
        let risk = riskscore mbr begpos endpos syncat word in
        prob := !prob *. risk;
        List.iter
            (fun (syncat', span', headpos', word') ->
                let inst' = TreeInst.create span' syncat' word' in
                let (tree', prob') = get_inst mbr inst' in
                dtrs := !dtrs @ [tree'];
                prob := !prob +. prob')
            dtrargs;
        let tree = Tree.create_nonterm
            syncat word !dtrs span headpos depdir in
        set_inst mbr inst tree !prob;
        set_minroot mbr inst !prob

    let compute_viterbi mbr =
        Scale.add (Scale.WindMill 5000) " DP  " "";
        Chart.iter_derivs mbr.chart
            Chart.BottomUp
            (build_lex_tree mbr) (build_dtrs_tree mbr)
            (0, mbr.endpos);
        Scale.discard ()

    let decapsulate_tree start tree =
        if Syncat.equal_ign_attrs (Tree.get_syncat tree) start then begin
            if Tree.is_nonterm tree then begin
                let dtrs = Tree.get_dtrs tree in
                let depdir = Tree.get_depdir tree in
                match depdir with
                    Syncat.Left -> List.nth dtrs 0
                  | Syncat.Right -> List.nth dtrs 1
                  | Syncat.Coor -> List.nth dtrs 0
                  | Syncat.Unary -> List.nth dtrs 0
            end else tree
        end else tree

    let find_min mbr =
        let (tree, prob) = match mbr.minroot with
            None -> raise Not_found
          | Some (inst, prob) -> begin
                (* Printf.printf "TreeInst = %s\n" (TreeInst.to_string inst); *)
                get_inst mbr inst
            end in
        (* Printf.printf "tree      = %s\n" (Tree.to_string tree); *)
        let decaptree = decapsulate_tree mbr.start tree in
        (* Printf.printf "decaptree = %s\n" (Tree.to_string decaptree); *)
        (decaptree, prob)

    let compute ruleprob chart start catconst =
        (* Printf.printf "(mbr.compute          ) Sentence = [%s]\n" (String.concat " " (Chart.get_sentence chart)); *)
        let mbr = create ruleprob chart start catconst in
        compute_risktbl mbr;
        compute_viterbi mbr;
        find_min mbr

end
