(***********************************************************
*                                                          *
*      FUNGI - Fast UNsupervised Grammar Induction         *
*                                                          *
*   Originator: Prachya Boonkwan [1], [2]                  *
*               [1] School of Informatics,                 *
*                   University of Edinburgh, UK            *
*               [2] National Electronics and Computer      *
*                   Technology Center (NECTEC), Thailand   *
*                                                          *
*   Copyright (C) 2011 Prachya Boonkwan                    *
*   <p.boonkwan@sms.ed.ac.uk, kaamanita@hotmail.com>       *
*                                                          *
*   This library is free software; you can redistribute    *
*   it and/or modify it under the terms of the GNU         *
*   Lesser General Public License as published by the      *
*   Free Software Foundation; either version 2.1 of the    *
*   License, or (at your option) any later version.        *
*                                                          *
*   This library is distributed in the hope that it will   *
*   be useful, but WITHOUT ANY WARRANTY; without even      *
*   the implied warranty of MERCHANTABILITY or FITNESS     *
*   FOR A PARTICULAR PURPOSE.  See the GNU Lesser          *
*   General Public License for more details.               *
*                                                          *
*   You should have received a copy of the GNU Lesser      *
*   General Public License along with this library; if     *
*   not, write to the Free Software Foundation, Inc.,      *
*   59 Temple Place, Suite 330, Boston, MA  02111-1307     *
*   USA.                                                   *
*                                                          *
***********************************************************)

open Cg
open Chart
open Chartdb
open Prob
open Deriv
open Models
open Decoderalgo
open Tree
open Treedb
open Utils
open Printf

type decoding_t = 
    ViterbiDecoding 
  | MBRDecoding 
  | DualDecomp of Hmm.fwdbwd_t * float * int

module DecoderRunner = struct

    type genmodel_t = DerivGenModel.t

    type t = {
        decoding : decoding_t;
        start    : Syncat.t;
        ruleprob : genmodel_t;
        wrapper  : ChartDbWrapper.t;
        catconst : float;
        treedb   : TreeDb.t;
        tdbfname : string;
    }

    let create decoding start epsilon probtblfname catconst fnames treedbfname =
        {
            decoding = decoding;
            start    = start;
            ruleprob = DerivGenModel.from_file probtblfname epsilon;
            wrapper  = ChartDbWrapper.from_files fnames;
            catconst = catconst;
            treedb   = TreeDb.create treedbfname;
            tdbfname = treedbfname;
        }

    let parse_chart decoding ruleprob chart start catconst =
        (* Printf.printf "(viterbirunner.parse_chart) Sentence = [%s]\n" (String.concat " " (Chart.get_sentence chart)); *)
        let (tree, prob) = match decoding with
            ViterbiDecoding -> 
                TreeViterbi.compute ruleprob chart start catconst 
          | MBRDecoding -> 
                TreeMinBayesRisk.compute ruleprob chart start catconst
          | DualDecomp (fwdbwd, stepsize, rndthr) ->
                TreeDualDecomp.compute ruleprob fwdbwd chart start catconst stepsize rndthr
        in
        let depstr = Tree.to_depstr tree in
        (tree, prob, depstr)
    
    let parse_wrapper decoding ruleprob start wrapper treedb catconst =
        let nocharts = ChartDbWrapper.length wrapper in
        Scale.add (Scale.GuageWithNumber nocharts) "parsing charts " "";
        for idx = 0 to nocharts - 1 do
            Scale.inc ();
            Scale.display ();
            let chart = ChartDbWrapper.find wrapper idx in
            try 
                let (tree, prob, depstr) = parse_chart decoding ruleprob chart start catconst in
                TreeDb.add treedb (MarshTree.create tree prob depstr)
            with Not_found ->
                TreeDb.add treedb None
        done;
        Scale.discard ()

    let run vtb =
        parse_wrapper vtb.decoding vtb.ruleprob vtb.start vtb.wrapper vtb.treedb vtb.catconst;
        TreeDb.to_file vtb.treedb vtb.tdbfname

    let compute decoding start epsilon probtblfname catconst fnames treedbfname =
        let vtb = create decoding start epsilon probtblfname catconst fnames treedbfname in
        run vtb
        
end

let set_decoding decoding str =
    decoding := match str with
        "viterbi" -> ViterbiDecoding
      | "mbr" -> MBRDecoding
      | _ -> failwith "set_decoding"

let set_fwdbwd fwdbwd str =
    let model = Hmm.from_file str 1e-25 in
    fwdbwd := Some model

let set_start start str =
    let lexbuf = Lexing.from_string str in
    let syncat = Ruleparser.syncat Rulelexer.token lexbuf in
    start := syncat
