(***********************************************************
*                                                          *
*      FUNGI - Fast UNsupervised Grammar Induction         *
*                                                          *
*   Originator: Prachya Boonkwan [1], [2]                  *
*               [1] School of Informatics,                 *
*                   University of Edinburgh, UK            *
*               [2] National Electronics and Computer      *
*                   Technology Center (NECTEC), Thailand   *
*                                                          *
*   Copyright (C) 2011 Prachya Boonkwan                    *
*   <p.boonkwan@sms.ed.ac.uk, kaamanita@hotmail.com>       *
*                                                          *
*   This library is free software; you can redistribute    *
*   it and/or modify it under the terms of the GNU         *
*   Lesser General Public License as published by the      *
*   Free Software Foundation; either version 2.1 of the    *
*   License, or (at your option) any later version.        *
*                                                          *
*   This library is distributed in the hope that it will   *
*   be useful, but WITHOUT ANY WARRANTY; without even      *
*   the implied warranty of MERCHANTABILITY or FITNESS     *
*   FOR A PARTICULAR PURPOSE.  See the GNU Lesser          *
*   General Public License for more details.               *
*                                                          *
*   You should have received a copy of the GNU Lesser      *
*   General Public License along with this library; if     *
*   not, write to the Free Software Foundation, Inc.,      *
*   59 Temple Place, Suite 330, Boston, MA  02111-1307     *
*   USA.                                                   *
*                                                          *
***********************************************************)

open Cg
open Utils
open Printf

module HashedInt = struct
    type t = int
    let hash = Hashtbl.hash
    let equal = Pervasives.(=)
end

type depdir_t = Syncat.depdir_t

module DepPair = struct

    type word_t = string

    type t = word_t * word_t * depdir_t option

    let create head dep depdir =
        (head, dep, depdir)

    let get_head deppair =
        let (head, _, _) = deppair in
        head

    let get_dep deppair =
        let (_, dep, _) = deppair in
        dep

    let get_depdir deppair =
        let (_, _, depdir) = deppair in
        match depdir with
            Some dir -> dir
          | None -> failwith "get_depdir"

    let to_string deppair =
        let (head, dep, depdir) = deppair in
        match depdir with
            Some dir ->
                sprintf "'%s' %s '%s'" head 
                    (Syncat.depdir_to_string dir) dep
          | None ->
                sprintf "'%s' <-> '%s'" head dep

    let is_dirdep deppair =
        let (_, _, depdir) = deppair in
        match depdir with
            Some _ -> true
          | None -> false

    let hash = Hashtbl.hash

    let equal = Pervasives.(=)
    
    let compare = Pervasives.compare

end

module DepStruct = struct

    type word_t = string

    module DepTbl = Hashtbl.Make(HashedInt)

    type deptbl_t = int list DepTbl.t
    
    type headtbl_t = int DepTbl.t

    type deppair_t = DepPair.t

    type t = {
        sentence : word_t list;
        rootpos  : int;
        deptbl   : deptbl_t;
        headtbl  : headtbl_t;
    }

    let create n sentence rootpos =
    {
        sentence = sentence;
        rootpos  = rootpos;
        deptbl   = DepTbl.create n;
        headtbl  = DepTbl.create n;
    }

    let get_sentence depstr =
        depstr.sentence

    let get_rootpos depstr =
        depstr.rootpos

    let mem_head depstr headpos =
        DepTbl.mem depstr.deptbl headpos

    let get depstr headpos =
        DepTbl.find depstr.deptbl headpos

    let find_head depstr deppos =
        if DepTbl.mem depstr.headtbl deppos then
            DepTbl.find depstr.headtbl deppos
        else raise Not_found

    let mem depstr headpos deppos =
        if mem_head depstr headpos then
            List.mem deppos (get depstr headpos)
        else false

    let remove depstr headpos =
        if mem_head depstr headpos then
            DepTbl.remove depstr.deptbl headpos
        else ()

    let entry_to_string headpos deppos sentence =
        let headword = List.nth sentence headpos
        and depword = List.nth sentence deppos in
        sprintf "Head '%s'/%d -> Dep '%s'/%d"
            headword headpos depword deppos

    let deptbl_to_string depstr =
        let lines = ref [] in
        let sentlen = List.length depstr.sentence in
        let deppos_to_string headpos deppos =
            entry_to_string headpos deppos depstr.sentence in
        for headpos = sentlen - 1 downto 0 do
            if DepTbl.mem depstr.deptbl headpos then
                let depposes = DepTbl.find depstr.deptbl headpos in
                lines := (List.map (deppos_to_string headpos) depposes) @ !lines
            else ()
        done;
        if List.length !lines <> 0 then String.concat "\n" !lines
        else "Empty dependency"

    let to_string depstr =
        sprintf "<Dependency>\nSentence = [%s]\nRoot = '%s'/%d\n%s"
            (String.concat " " depstr.sentence) 
            (List.nth depstr.sentence depstr.rootpos) depstr.rootpos
            (deptbl_to_string depstr)

    let add depstr headpos deppos =
        if headpos = deppos then ()
        else if mem_head depstr headpos then begin
            let oldmembers = get depstr headpos in
            if not (List.mem deppos oldmembers) then begin
                DepTbl.replace depstr.deptbl headpos (deppos :: oldmembers);
                DepTbl.replace depstr.headtbl deppos headpos
            end else ()
        end else begin
            DepTbl.replace depstr.deptbl headpos [deppos];
            DepTbl.replace depstr.headtbl deppos headpos
        end

    let swap_pos headpos headword deppos depword =
        if headpos < deppos then 
            (headword, depword, headpos, deppos, Syncat.Left)
        else 
            (depword, headword, deppos, headpos, Syncat.Right)

    let depset depstr : deppair_t list =
        let result = ref [] in
        DepTbl.iter
            (fun headpos depposes ->
                let headword = List.nth depstr.sentence headpos in
                List.iter 
                    (fun deppos ->
                        let depword = List.nth depstr.sentence deppos in
                        let (left, right, leftpos, rightpos, depdir) = 
                            swap_pos headpos headword deppos depword in
                        let left = sprintf "%d" leftpos
                        and right = sprintf "%d" rightpos in
                        let dep = DepPair.create left right (Some depdir) in
                        result := !result @ [dep]) 
                    depposes)
            depstr.deptbl;
        !result

    let depset_noorder depstr : deppair_t list =
        let result = ref [] in
        DepTbl.iter
            (fun headpos depposes ->
                let headword = sprintf "%d" headpos in
                List.iter 
                    (fun deppos ->
                        let depword = sprintf "%d" deppos in
                        let dep =
                            if headpos < deppos then
                                DepPair.create headword depword None 
                            else 
                                DepPair.create depword headword None 
                        in
                        result := !result @ [dep]) 
                    depposes)
            depstr.deptbl;
        !result

    let depset_words depstr : deppair_t list =
        let result = ref [] in
        DepTbl.iter
            (fun headpos depposes ->
                let headword = List.nth depstr.sentence headpos in
                List.iter 
                    (fun deppos ->
                        let depword = List.nth depstr.sentence deppos in
                        let (left, right, leftpos, rightpos, depdir) = 
                            swap_pos headpos headword deppos depword in
                        let dep = DepPair.create left right (Some depdir) in
                        result := !result @ [dep]) 
                    depposes)
            depstr.deptbl;
        !result

    let depset_words_noorder depstr : deppair_t list =
        let result = ref [] in
        DepTbl.iter
            (fun headpos depposes ->
                let headword = List.nth depstr.sentence headpos in
                List.iter 
                    (fun deppos ->
                        let depword = List.nth depstr.sentence deppos in
                        let dep =
                            if headpos < deppos then
                                DepPair.create headword depword None 
                            else 
                                DepPair.create depword headword None 
                        in
                        result := !result @ [dep]) 
                    depposes)
            depstr.deptbl;
        !result

end
