(***********************************************************
*                                                          *
*      FUNGI - Fast UNsupervised Grammar Induction         *
*                                                          *
*   Originator: Prachya Boonkwan [1], [2]                  *
*               [1] School of Informatics,                 *
*                   University of Edinburgh, UK            *
*               [2] National Electronics and Computer      *
*                   Technology Center (NECTEC), Thailand   *
*                                                          *
*   Copyright (C) 2011 Prachya Boonkwan                    *
*   <p.boonkwan@sms.ed.ac.uk, kaamanita@hotmail.com>       *
*                                                          *
*   This library is free software; you can redistribute    *
*   it and/or modify it under the terms of the GNU         *
*   Lesser General Public License as published by the      *
*   Free Software Foundation; either version 2.1 of the    *
*   License, or (at your option) any later version.        *
*                                                          *
*   This library is distributed in the hope that it will   *
*   be useful, but WITHOUT ANY WARRANTY; without even      *
*   the implied warranty of MERCHANTABILITY or FITNESS     *
*   FOR A PARTICULAR PURPOSE.  See the GNU Lesser          *
*   General Public License for more details.               *
*                                                          *
*   You should have received a copy of the GNU Lesser      *
*   General Public License along with this library; if     *
*   not, write to the Free Software Foundation, Inc.,      *
*   59 Temple Place, Suite 330, Boston, MA  02111-1307     *
*   USA.                                                   *
*                                                          *
***********************************************************)

open Cg
open Chart
open Prob
open Printf

module Deriv = struct

    type syncat_t = CG.syncat_t

    type depdir_t = CG.depdir_t

    type word_t = string

    type prems_t = {
        syncat : syncat_t;
        head   : word_t;
    }

    let create_prems syncat head = {
        syncat = syncat;
        head   = head;
    }

    let prems_to_string prems =
        sprintf "%s{%s}" (Syncat.to_string prems.syncat) prems.head

    let get_syncat prems =
        prems.syncat

    let get_head prems =
        prems.head

    type dtrs_t = {
        dtrs   : (syncat_t * word_t) list;
        depdir : depdir_t;
    }

    type consq_t =
        Lex of word_t
      | Dtrs of dtrs_t

    let create_consq_dtrs dtrs depdir = 
        Dtrs {
            dtrs   = dtrs;
            depdir = depdir;
        }

    let create_consq_lex word =
        Lex word

    let consq_to_string consq =
        match consq with
            Lex head -> sprintf "'%s'" head
          | Dtrs dtrs -> 
                sprintf "%s : %s"
                    (String.concat " " 
                        (List.map 
                            (fun (c, w) -> 
                                sprintf "%s{%s}" 
                                    (Syncat.to_string c) w) 
                            dtrs.dtrs))
                    (Syncat.depdir_to_string dtrs.depdir)

    let get_dtrs consq =
        match consq with
            Lex _ -> failwith "get_dtrs"
          | Dtrs content -> content.dtrs

    let get_depdir consq =
        match consq with
            Lex _ -> failwith "get_depdir"
          | Dtrs content -> content.depdir

    let get_word consq =
        match consq with
            Lex word -> word
          | Dtrs _ -> failwith "get_word"

    let hash_consq consq =
        match consq with
            Lex word ->
                Hashtbl.hash word
          | Dtrs content ->
                Hashtbl.hash content.depdir
                + List.fold_left
                    (fun v (dtr, word) -> 
                        v + Syncat.hash dtr + Hashtbl.hash word)
                    0 content.dtrs

    let is_consq_dtrs consq =
        match consq with
            Lex _ -> false
          | Dtrs _ -> true

    let is_consq_lex consq =
        match consq with
            Lex _ -> true
          | Dtrs _ -> false

    type t = {
        prems  : prems_t;
        consq  : consq_t;
    }

    let create prems consq = {
        prems  = prems;
        consq  = consq;
    }

    let create_deriv_dtrs syncat head dtrcats depdir = {
        prems = create_prems syncat head;
        consq = create_consq_dtrs dtrcats depdir;
    }

    let create_deriv_lex syncat word = {
        prems = create_prems syncat word;
        consq = create_consq_lex word;
    }

    let to_string deriv =
        sprintf "%s -> %s" 
            (prems_to_string deriv.prems) 
            (consq_to_string deriv.consq)

    let get_prems deriv =
        deriv.prems

    let get_consq deriv =
        deriv.consq

    let equal_prems prems1 prems2 = 
        Syncat.equal prems1.syncat prems2.syncat
        && prems1.head = prems2.head

    let hash_prems prems = 
        Syncat.hash prems.syncat + Hashtbl.hash prems.head

    let equal deriv1 deriv2 =
        deriv1.prems = deriv2.prems && deriv1.consq = deriv2.consq

    let hash deriv =
        hash_prems deriv.prems + hash_consq deriv.consq

    let backoffword = "<*>"

    let is_backoff deriv =
        if Syncat.is_bo_cat deriv.prems.syncat then true
        else
            match deriv.consq with
                Lex word -> word = backoffword
              | Dtrs content -> begin
                    let flag = ref false in
                    List.iter
                        (fun (c, w) ->
                            if Syncat.is_bo_cat c || w = backoffword then
                                flag := true)
                        content.dtrs;
                    !flag
                end

    let backoff_word_dtrcats dtrs depdir =
        match depdir, dtrs with
            Syncat.Left, [(_, w1); (_, w2)] ->
                [(Syncat.bo_cat Syncat.NoDep, w1); (Syncat.bo_cat Syncat.NoDep, w2)]
          | Syncat.Right, [(_, w1); (_, w2)] ->
                [(Syncat.bo_cat Syncat.NoDep, w1); (Syncat.bo_cat Syncat.NoDep, w2)]
          | Syncat.Coor, [(_, w1); (_, w2); (_, w3)] ->
                [(Syncat.bo_cat Syncat.NoDep, w1); (Syncat.bo_cat Syncat.NoDep, w2); (Syncat.bo_cat Syncat.NoDep, w3)]
          | Syncat.Unary, [(_, w1)] ->
                [(Syncat.bo_cat Syncat.NoDep, w1)]
          | _, _ -> failwith "make_argbackoff_dtrcats"

    let backoff_word ?(headbackoff=false) deriv =
        let prems_word = deriv.prems.head in
        let syncat = 
            if headbackoff then Syncat.bo_cat Syncat.NoDep
            else deriv.prems.syncat in
        match deriv.consq with
            Lex _ ->
                create_deriv_lex syncat prems_word
          | Dtrs content -> 
                let dtrcats = 
                    backoff_word_dtrcats content.dtrs content.depdir in
                create_deriv_dtrs 
                    syncat prems_word dtrcats content.depdir

    let backoff_head_dtrcats dtrs depdir =
        match depdir, dtrs with
            Syncat.Left, [(s1, w1); (_, _)] ->
                [(s1, w1); (Syncat.bo_cat Syncat.NoDep, backoffword)]
          | Syncat.Right, [(_, _); (s2, w2)] ->
                [(Syncat.bo_cat Syncat.NoDep, backoffword); (s2, w2)]
          | Syncat.Coor, [(s1, w1); (_, _); (_, _)] ->
                [(s1, w1); (Syncat.bo_cat Syncat.NoDep, backoffword); (Syncat.bo_cat Syncat.NoDep, backoffword)]
          | Syncat.Unary, [(s1, w1)] ->
                [(s1, w1)]
          | _, _ -> failwith "make_argbackoff_dtrcats"

    let backoff_head ?(headbackoff=false) deriv =
        let prems_word = deriv.prems.head in
        let syncat = 
            if headbackoff then Syncat.bo_cat Syncat.NoDep
            else deriv.prems.syncat in
        match deriv.consq with
            Lex _ ->
                create_deriv_lex syncat prems_word
          | Dtrs content -> 
                let dtrcats = 
                    backoff_head_dtrcats content.dtrs content.depdir in
                create_deriv_dtrs 
                    syncat prems_word dtrcats content.depdir

    let backoff_dep_dtrcats dtrs depdir =
        match depdir, dtrs with
            Syncat.Left, [(_, _); (s2, w2)] ->
                [(Syncat.bo_cat Syncat.NoDep, backoffword); (s2, w2)]
          | Syncat.Right, [(s1, w1); (_, _)] ->
                [(s1, w1); (Syncat.bo_cat Syncat.NoDep, backoffword)]
          | Syncat.Coor, [(_, _); (_, _); (s3, w3)] ->
                [(Syncat.bo_cat Syncat.NoDep, backoffword); (Syncat.bo_cat Syncat.NoDep, backoffword); (s3, w3)]
          | Syncat.Unary, [(_, _)] ->
                [(Syncat.bo_cat Syncat.NoDep, backoffword)]
          | _, _ -> failwith "make_argbackoff_dtrcats"

    let backoff_dep ?(headbackoff=false) deriv =
        let prems_word = deriv.prems.head in
        let syncat = 
            if headbackoff then Syncat.bo_cat Syncat.NoDep
            else deriv.prems.syncat in
        match deriv.consq with
            Lex _ ->
                create_deriv_lex syncat prems_word
          | Dtrs content -> 
                let dtrcats = 
                    backoff_dep_dtrcats content.dtrs content.depdir in
                create_deriv_dtrs 
                    syncat prems_word dtrcats content.depdir

    let backoffs deriv = 
        (*
        let deriv1' = backoff_word deriv in
        let deriv2' = backoff_head deriv in
        let deriv3' = backoff_dep deriv in
        let deriv4' = backoff_word deriv ~headbackoff:true in
        let deriv5' = backoff_head deriv ~headbackoff:true in
        let deriv6' = backoff_dep deriv ~headbackoff:true in
        *)
        let deriv' = backoff_word deriv ~headbackoff:true in
        let result = ref [] in
        List.iter
            (fun deriv' ->
                if not (equal deriv' deriv) then
                    result := !result @ [Some deriv']
                else
                    result := !result @ [None])
            (* [deriv1'; deriv2'; deriv3'; deriv4'; deriv5'; deriv6'] *)
            [deriv']
            ;
        !result

    let get_nobackoffs () = 1

end

type deriv_t = Deriv.t

module DerivProbTbl = ProbTbl(Deriv)

(* module DerivGenModel = GenModel(Deriv) *)
module DerivGenModel = GenModelBackoff(Deriv)

let conv_word_to_deriv syncat word =
    let prems = Deriv.create_prems syncat word
    and consq = Deriv.create_consq_lex word in
    Deriv.create prems consq

let conv_dtrargs_to_deriv syncat word depdir dtrargs =
    let dtrs = List.map
        (fun (syncat', _, _, word') -> (syncat', word'))
        dtrargs in
    let prems = Deriv.create_prems syncat word
    and consq = Deriv.create_consq_dtrs dtrs depdir in
    Deriv.create prems consq

module type FTR_TYPE = sig
    type prems_t
    type t
    val equal_prems : prems_t -> prems_t -> bool
    val hash_prems : prems_t -> int
    val equal : t -> t -> bool
    val hash : t -> int
    val extract_mtrftrs : deriv_t -> (prems_t * t) list
    val extract_dtrftrs : deriv_t -> (prems_t * t) list
    val to_string : t -> string
    val prems_to_string : prems_t -> string
end

module ProbInst (Ftr : FTR_TYPE) = struct

    type prems_t = Ftr.prems_t

    type consq_t = Ftr.t

    type t = {
        prems : prems_t;
        consq : consq_t;
    }

    let create prems consq = {
        prems = prems;
        consq = consq;
    }

    let get_prems probinst =
        probinst.prems

    let get_consq probinst =
        probinst.consq

    let equal_prems = Ftr.equal_prems

    let hash_prems = Ftr.hash_prems

    let equal probinst1 probinst2 =
        equal_prems probinst1.prems probinst2.prems
        && Ftr.equal probinst1.consq probinst2.consq

    let hash probinst =
        hash_prems probinst.prems + Ftr.hash probinst.consq

    let extract_mtrftrs deriv =
        List.map 
            (fun (prems, ftr) -> create prems ftr)
            (Ftr.extract_mtrftrs deriv)

    let extract_dtrftrs deriv =
        List.map
            (fun (prems, ftr) -> create prems ftr)
            (Ftr.extract_dtrftrs deriv)

    let extract_ftrs deriv =
        (extract_mtrftrs deriv) @ (extract_dtrftrs deriv)

    let to_string probinst = 
        sprintf "%s : %s" 
            (Ftr.prems_to_string probinst.prems) 
            (Ftr.to_string probinst.consq)

    let prems_to_string = Ftr.prems_to_string

    let get_nobackoffs () = 0

end

module DerivFtr (Ftr : FTR_TYPE) = struct
  
    module Inst = ProbInst(Ftr)

    type inst_t = Inst.t

    module CondProbTbl = CondProbTbl(Inst)

    type t = CondProbTbl.t

    type probftr_t = DerivGenModel.probftr_t

    let create_cprobtbl n epsilon =
        CondProbTbl.create n epsilon

    let get_prob probmodel deriv =
        List.fold_left
            (fun prob symbol ->
                prob *. CondProbTbl.get_prob probmodel symbol)
            1.0 (Inst.extract_ftrs deriv)

    let get_smooth_prob probmodel deriv =
        List.fold_left
            (fun prob symbol ->
                prob *. CondProbTbl.get_smooth_prob probmodel symbol)
            1.0 (Inst.extract_ftrs deriv)

    let set_prob probmodel deriv prob =
        List.iter
            (fun symbol ->
                CondProbTbl.set_prob probmodel symbol prob)
            (Inst.extract_ftrs deriv)

    let acc_prob probmodel deriv prob =
        List.iter
            (fun symbol ->
                CondProbTbl.acc_prob probmodel symbol prob)
            (Inst.extract_ftrs deriv)

    let create_probftr probmodel =
        let fn_get_prob = get_prob probmodel
        and fn_get_smooth_prob = get_smooth_prob probmodel
        and fn_set_prob = set_prob probmodel
        and fn_acc_prob = acc_prob probmodel in
        DerivGenModel.create_probftr 
            fn_get_prob fn_get_smooth_prob fn_set_prob fn_acc_prob

    let create n epsilon : probftr_t =
        let probmodel = create_cprobtbl n epsilon in
        create_probftr probmodel

end
