(***********************************************************
*                                                          *
*      FUNGI - Fast UNsupervised Grammar Induction         *
*                                                          *
*   Originator: Prachya Boonkwan [1], [2]                  *
*               [1] School of Informatics,                 *
*                   University of Edinburgh, UK            *
*               [2] National Electronics and Computer      *
*                   Technology Center (NECTEC), Thailand   *
*                                                          *
*   Copyright (C) 2011 Prachya Boonkwan                    *
*   <p.boonkwan@sms.ed.ac.uk, kaamanita@hotmail.com>       *
*                                                          *
*   This library is free software; you can redistribute    *
*   it and/or modify it under the terms of the GNU         *
*   Lesser General Public License as published by the      *
*   Free Software Foundation; either version 2.1 of the    *
*   License, or (at your option) any later version.        *
*                                                          *
*   This library is distributed in the hope that it will   *
*   be useful, but WITHOUT ANY WARRANTY; without even      *
*   the implied warranty of MERCHANTABILITY or FITNESS     *
*   FOR A PARTICULAR PURPOSE.  See the GNU Lesser          *
*   General Public License for more details.               *
*                                                          *
*   You should have received a copy of the GNU Lesser      *
*   General Public License along with this library; if     *
*   not, write to the Free Software Foundation, Inc.,      *
*   59 Temple Place, Suite 330, Boston, MA  02111-1307     *
*   USA.                                                   *
*                                                          *
***********************************************************)

open Cg
open Printf
open Utils

module HashedString = struct
    type t = string
    let equal = Pervasives.(=)
    let hash = Hashtbl.hash
end

module StrDict = Hashtbl.Make(HashedString)

module Dict = struct

    type word_t = StrDict.key

    type syncat_t = CG.syncat_t

    type enumlevel_t = NoEnum | EnumOnUncertain | EnumOnAll

    let enumlevel_to_string enumlevel =
        match enumlevel with
            NoEnum -> "NoEnum"
          | EnumOnUncertain -> "EnumOnUncertain"
          | EnumOnAll -> "EnumOnAll"

    type t = {
        lexdict           : syncat_t list StrDict.t;  (** Content *)
        mutable unclexs   : word_t list;          (** Uncertainty list *)
        mutable allcats   : syncat_t list;        (** All categories *)
        mutable enumlevel : enumlevel_t;          (** Enumeration level *)
        mutable accepts   : syncat_t list;        (** Accepted categories *)
        mutable unarydrv  : CG.transf_t;          (** Unary transformation rules *)
    }

    let create n enumlevel = 
    {
        lexdict   = StrDict.create n;
        unclexs   = [];
        allcats   = [];
        enumlevel = enumlevel;
        accepts   = [];
        unarydrv  = [];
    }

    let rec to_string dict =
        sprintf "Lexicon:\n%s\nUncertainty: %s\nAll cats: %s\nEnumeration: %s\nTransformation rules:\n%s\nAccepted Categories: %s"
            (to_string_lexdict dict.lexdict)
            (to_string_unclexs dict.unclexs)
            (to_string_allcats dict.allcats)
            (enumlevel_to_string dict.enumlevel)
            (to_string_unarydrv dict.unarydrv)
            (to_string_accepts dict.accepts)

    and to_string_lexdict lexdict =
        let lexs = List.sort Pervasives.compare 
            (StrDict.fold (fun lex _ v -> lex :: v) lexdict []) in
        String.concat "\n"
            (List.fold_right
                (fun lex v ->
                    let cats = StrDict.find lexdict lex in
                    let s = sprintf "    '%s' :- %s" lex
                        (String.concat ", " 
                            (List.map Syncat.to_string cats)) in
                    s :: v)
                lexs [])

    and to_string_unclexs unclexs =
        String.concat ", " 
            (List.map (fun s -> sprintf "'%s'" s) unclexs)

    and to_string_allcats allcats =
        String.concat ", "
            (List.map (Syncat.to_string) allcats)
        ^ (sprintf " (%d categories)" (List.length allcats))

    and to_string_unarydrv unarydrv =
        String.concat "\n"
            (List.map
                (fun (syncat, newsyncats) ->
                    Printf.sprintf "    %s => {%s}"
                        (Syncat.to_string syncat)
                        (String.concat ", " (List.map Syncat.to_string newsyncats)))
                unarydrv)

    and to_string_accepts accepts =
        String.concat ", " 
            (List.map (Syncat.to_string) accepts)

    let get_enumlevel dict =
        dict.enumlevel

    let set_enumlevel dict enumlevel =
        dict.enumlevel <- enumlevel

    let mem dict word =
        StrDict.mem dict.lexdict word

    let find dict word =
        StrDict.find dict.lexdict word

    let is_uncertain dict word =
        List.mem word dict.unclexs

    let add_cat dict cat =
        if not (List.mem cat dict.allcats) then
            dict.allcats <- cat :: dict.allcats
        else ()

    let add_entry dict word cat =
        let oldset =
            if mem dict word then find dict word
            else [] in
        if not (List.mem cat oldset) then begin
            let newset = cat :: oldset in
            StrDict.replace dict.lexdict word newset;
            add_cat dict cat
        end else ()

    let add_accept dict cat =
        if not (List.mem cat dict.accepts) then begin
            dict.accepts <- cat :: dict.accepts;
            let acccat = 
                Syncat.create_bwd Syncat.root_cat cat Syncat.Left in
            add_entry dict ending_mark acccat
        end else ()

    let add_uncertain dict word =
        if not (List.mem word dict.unclexs) then
            dict.unclexs <- word :: dict.unclexs
        else ()

    let add_unarydrv dict syncat newcat =
        if not (List.mem_assoc syncat dict.unarydrv) then
            dict.unarydrv <- (syncat, [newcat]) :: dict.unarydrv
        else begin
            let entry = List.assoc syncat dict.unarydrv in
            if not (List.mem newcat entry) then
                dict.unarydrv <- (syncat, newcat :: entry) :: (List.remove_assoc syncat dict.unarydrv)
            else ()
        end

    let get_cats_noenum dict word =
        if mem dict word then
            StrDict.find dict.lexdict word
        else []

    let get_cats dict word =
        match dict.enumlevel with
            NoEnum | EnumOnUncertain -> 
                get_cats_noenum dict word
          | EnumOnAll -> 
                if word = ending_mark then get_cats_noenum dict word
                else dict.allcats

    let get_allcats dict =
        match dict.enumlevel with
            NoEnum -> []
          | EnumOnUncertain | EnumOnAll -> dict.allcats

    let get_accepts dict =
        dict.accepts

    let get_unarydrv dict =
        dict.unarydrv

end
