(***********************************************************
*                                                          *
*      FUNGI - Fast UNsupervised Grammar Induction         *
*                                                          *
*   Originator: Prachya Boonkwan [1], [2]                  *
*               [1] School of Informatics,                 *
*                   University of Edinburgh, UK            *
*               [2] National Electronics and Computer      *
*                   Technology Center (NECTEC), Thailand   *
*                                                          *
*   Copyright (C) 2011 Prachya Boonkwan                    *
*   <p.boonkwan@sms.ed.ac.uk, kaamanita@hotmail.com>       *
*                                                          *
*   This library is free software; you can redistribute    *
*   it and/or modify it under the terms of the GNU         *
*   Lesser General Public License as published by the      *
*   Free Software Foundation; either version 2.1 of the    *
*   License, or (at your option) any later version.        *
*                                                          *
*   This library is distributed in the hope that it will   *
*   be useful, but WITHOUT ANY WARRANTY; without even      *
*   the implied warranty of MERCHANTABILITY or FITNESS     *
*   FOR A PARTICULAR PURPOSE.  See the GNU Lesser          *
*   General Public License for more details.               *
*                                                          *
*   You should have received a copy of the GNU Lesser      *
*   General Public License along with this library; if     *
*   not, write to the Free Software Foundation, Inc.,      *
*   59 Temple Place, Suite 330, Boston, MA  02111-1307     *
*   USA.                                                   *
*                                                          *
***********************************************************)

open Cg
open Chart
open Prob
open Deriv
open Models
open Inout
open Utils
open Printf

module ScoreInst = struct

    type syncat_t = Syncat.t
    type word_t = string

    type t = {
        syncat : syncat_t;
        head   : word_t;
    }

    let create syncat head = {
        syncat = syncat;
        head   = head;
    }

    let get_syncat inst =
        inst.syncat

    let get_head inst =
        inst.head

    let equal inst1 inst2 =
        Syncat.equal inst1.syncat inst2.syncat
        && inst1.head = inst2.head

    let hash inst =
        Syncat.hash inst.syncat + Hashtbl.hash inst.head

    let to_string inst =
        sprintf "%s{%s}" (Syncat.to_string inst.syncat) inst.head

end

module ScoreTbl = ProbTbl(ScoreInst)

module type EXPCALC_TYPE = sig
    val calc_prob : float -> float -> float
end

let uniform_scale_prob prob _ = prob

let uniform_localscore _ _ = 1.0

let goldmodel = ref None

let set_goldmodel genmodel =
    goldmodel := Some genmodel

module EStep(ExpCalc : EXPCALC_TYPE) = struct

    type syncat_t = Syncat.t

    type word_t = string

    type chart_t = Chart.t

    type scoreinst_t = ScoreInst.t

    type scoretbl_t = ScoreTbl.t

    module IOScoreInst = Inout.ScoreInst

    module IOScoreTbl = Inout.ScoreTbl

    type ioscoretbl_t = IOScoreTbl.t

    type genmodel_t = DerivGenModel.t

    type deriv_t = Deriv.t

    type fn_scale_prob_t = float -> deriv_t -> float

    type fn_localscore_t = int -> int -> float

    type t = {
        epsilon         : float;
        chart           : chart_t;
        ruleprob        : genmodel_t;
        ruleprior       : genmodel_t;
        insctbl         : ioscoretbl_t;
        outsctbl        : ioscoretbl_t;
        ruleexp         : genmodel_t;
        catexp          : scoretbl_t;
        start           : syncat_t;
        endpos          : int;
        scale_prob      : fn_scale_prob_t;
        localscore      : fn_localscore_t;
        mutable sentexp : float;
    }

    let create 
            ?(scale_prob = uniform_scale_prob) 
            ?(localscore = uniform_localscore)
            n epsilon ruleprob ruleprior chart start
            ruleexp catexp = 
        let (insctbl, outsctbl) = 
            Inout.compute n epsilon ruleprob chart start in
        (* let ruleexp = create_pcfg n epsilon
        and catexp = ScoreTbl.create n epsilon in *)
        {
            epsilon    = epsilon;
            chart      = chart;
            ruleprob   = ruleprob;
            ruleprior  = ruleprior;
            insctbl    = insctbl;
            outsctbl   = outsctbl;
            ruleexp    = ruleexp;
            catexp     = catexp;
            start      = start;
            endpos     = List.length (Chart.get_sentence chart) - 1;
            scale_prob = scale_prob;
            localscore = localscore;
            sentexp    = 0.0;
        }

    let to_string inout =
        sprintf "<Rule Expectation>\n%s\n\n<Category Expectation>\n%s\n\n<Rule Prob>\n%s"
            (DerivGenModel.to_string inout.ruleexp)
            (ScoreTbl.to_string inout.catexp)
            (DerivGenModel.to_string inout.ruleprob)

    let get_sentexp inout =
        inout.sentexp

    let calc_lexexp inout syncat span headpos word =
        let inst = ScoreInst.create syncat word 
        and ioinst = Inout.ScoreInst.create span syncat word in
        let deriv = conv_word_to_deriv syncat word in
        (* printf ">> Calculating the expectation of %s\n\n" 
            (Deriv.to_string deriv); *)
        let prior = DerivGenModel.get_smooth_prob inout.ruleprior deriv in
        let ruleprob =
            DerivGenModel.get_smooth_prob inout.ruleprob deriv in
        let outsc = IOScoreTbl.get_smooth_prob inout.outsctbl ioinst in
        (* printf "Prior = %e\nRule prob = %e\nOutside score = %e\n"
            prior ruleprob outsc; *)
        let subtotal = prior +. (ruleprob *. outsc) in
        (* printf "Subtotal = %e\n" subtotal;
        printf "RuleExp before %e -> " (DerivGenModel.get_prob inout.ruleexp deriv); *)
        DerivGenModel.acc_prob inout.ruleexp deriv subtotal;
        (* printf "after %e\n" (DerivGenModel.get_prob inout.ruleexp deriv);
        printf "CatExp before %e -> " (ScoreTbl.get_prob inout.catexp inst); *)
        ScoreTbl.acc_prob inout.catexp inst subtotal (* ;
        printf "after %e\n\n" (ScoreTbl.get_prob inout.catexp inst) *)

    let calc_dtrsexp inout syncat span headpos word depdir dtrargs =
        let inst = ScoreInst.create syncat word 
        and ioinst = IOScoreInst.create span syncat word in
        let deriv = conv_dtrargs_to_deriv syncat word depdir dtrargs in
        (* printf ">> Calculating the expectation of %s\n\n" 
            (Deriv.to_string deriv); *)
        let prior = DerivGenModel.get_smooth_prob inout.ruleprior deriv in
        let ruleprob =
            DerivGenModel.get_smooth_prob inout.ruleprob deriv in
        let outsc = IOScoreTbl.get_smooth_prob inout.outsctbl ioinst in
        (* printf "Prior = %e\nRule prob = %e\nOutside score = %e\n"
            prior ruleprob outsc; *)
        let subtotal = prior +. List.fold_left
            (fun v (syncat', span', headpos', word') ->
                let ioinst' = IOScoreInst.create span' syncat' word' in
                v *. IOScoreTbl.get_smooth_prob inout.insctbl ioinst'
                    *. inout.localscore headpos headpos')
            (ruleprob *. outsc) dtrargs in
        (* printf "Subtotal = %e\n" subtotal;
        printf "RuleExp before %e -> " (DerivGenModel.get_prob inout.ruleexp deriv); *)
        DerivGenModel.acc_prob inout.ruleexp deriv subtotal;
        (* printf "after %e\n" (DerivGenModel.get_prob inout.ruleexp deriv);
        printf "CatExp before %e -> " (ScoreTbl.get_prob inout.catexp inst); *)
        ScoreTbl.acc_prob inout.catexp inst subtotal (* ;
        printf "after %e\n\n" (ScoreTbl.get_prob inout.catexp inst) *)

    let compute_e_step inout =
        Chart.iter_derivs inout.chart
            Chart.TopDown
            (calc_lexexp inout) (calc_dtrsexp inout)
            (0, inout.endpos)

    let compute_sentexp_lex inout syncat span headpos word =
        if Syncat.equal_ign_attrs syncat inout.start then
            let ioinst = IOScoreInst.create span syncat word in
            let subtotal = 
                IOScoreTbl.get_smooth_prob inout.insctbl ioinst in
            inout.sentexp <- inout.sentexp +. subtotal
        else ()

    let compute_sentexp_dtrs inout 
            syncat span headpos word depdir dtrargs =
        if Syncat.equal_ign_attrs syncat inout.start then
            let ioinst = IOScoreInst.create span syncat word in
            let subtotal = 
                IOScoreTbl.get_smooth_prob inout.insctbl ioinst in
            inout.sentexp <- inout.sentexp +. subtotal
        else ()

    let compute_sentexp inout =
        Chart.iter_derivs_onelevel inout.chart
            (compute_sentexp_lex inout) (compute_sentexp_dtrs inout)
            (0, inout.endpos)

    let compute 
            ?(scale_prob = uniform_scale_prob) 
            ?(localscore = uniform_localscore) 
            n epsilon ruleprob ruleprior chart start
            ruleexp catexp =
        let inout = create 
            ~scale_prob:scale_prob ~localscore:localscore
            n epsilon ruleprob ruleprior chart start
            ruleexp catexp in
        compute_e_step inout;
        (inout.ruleexp, inout.catexp)

    let compute_sentexp_from_chart 
            ?(scale_prob = uniform_scale_prob) 
            ?(localscore = uniform_localscore) 
            n epsilon ruleprob ruleprior chart start
            ruleexp catexp =
        let inout = create 
            ~scale_prob:scale_prob ~localscore:localscore
            n epsilon ruleprob ruleprior chart start
            ruleexp catexp in
        compute_sentexp inout;
        get_sentexp inout

end

module MStep(ExpCalc : EXPCALC_TYPE) = struct

    type syncat_t = Syncat.t

    type word_t = string

    type scoreinst_t = ScoreInst.t

    type scoretbl_t = ScoreTbl.t

    type genmodel_t = DerivGenModel.t

    type deriv_t = Deriv.t

    type fn_scale_prob_t = float -> deriv_t -> float

    type fn_localscore_t = int -> int -> float

    type t = {
        epsilon         : float;
        ruleprob        : genmodel_t;
        ruleprior       : genmodel_t;
        ruleexp         : genmodel_t;
        catexp          : scoretbl_t;
        start           : syncat_t;
        scale_prob      : fn_scale_prob_t;
        localscore      : fn_localscore_t;
        fwdbwd          : Hmm.fwdbwd_t;
        mutable sentexp : float;
    }

    let create 
            ?(scale_prob = uniform_scale_prob) 
            ?(localscore = uniform_localscore)
            n epsilon ruleprob ruleprior start
            ruleexp catexp fwdbwd = 
        (* let ruleexp = create_pcfg n epsilon
        and catexp = ScoreTbl.create n epsilon in *)
        {
            epsilon    = epsilon;
            ruleprob   = ruleprob;
            ruleprior  = ruleprior;
            ruleexp    = ruleexp;
            catexp     = catexp;
            start      = start;
            scale_prob = scale_prob;
            localscore = localscore;
            fwdbwd     = fwdbwd;
            sentexp    = 0.0;
        }

    let calc_ruleprob inout deriv oldruleprob =
        Scale.inc_and_display ();
        (* printf ">> Calculating the rule probability of %s\n\n"
            (Deriv.to_string deriv); *)
        let prems = Deriv.get_prems deriv in
        let syncat = Deriv.get_syncat prems
        and head = Deriv.get_head prems in
        let inst = ScoreInst.create syncat head in
        let ruleexp = DerivGenModel.get_smooth_prob inout.ruleexp deriv
        and catexp = ScoreTbl.get_smooth_prob inout.catexp inst in
        (* printf "Rule expectation = %e\nCategory expectation = %e\n"
            ruleexp catexp; *)
        let ruleprob' = ref (ExpCalc.calc_prob ruleexp catexp) in
        if Utils.is_postreg () then
            let consq = Deriv.get_consq deriv in
            if Deriv.is_consq_lex consq then begin
                let word = Deriv.get_word consq in
                let fbinst = Hmm.EmisInst.create syncat word in
                let emisprob = Hmm.get_emisprob inout.fwdbwd in
                let mult = Hmm.EmisProb.get_prob emisprob fbinst in
                let newprob = !ruleprob' *. mult in
                ruleprob' := newprob;
                Hmm.EmisProb.set_prob emisprob fbinst newprob
            end else ()
        else ();
        let ruleprob = inout.scale_prob !ruleprob' deriv in
        (* printf "Rule probability = %e\n\n" ruleprob; *)
        DerivGenModel.set_prob inout.ruleprob deriv ruleprob

    let compute_m_step inout =
        if Utils.is_postreg () then
            Hmm.estimate inout.fwdbwd
        else ();
        Scale.add 
            (* (Scale.Number (Chart.get_derivcnt inout.chart)) *)
            (Scale.WindMill 1500)
            "M-Step  " "";
        DerivGenModel.calc_normfacts inout.ruleexp;
        DerivGenModel.iter inout.ruleexp (calc_ruleprob inout);
        DerivGenModel.calc_probftrs inout.ruleprob;
        if Utils.is_use_backoff () then
            DerivGenModel.calc_backoffs inout.ruleprob
        else ();
        Scale.discard ()

    let compute 
            ?(scale_prob = uniform_scale_prob) 
            ?(localscore = uniform_localscore) 
            n epsilon ruleprob ruleprior start
            ruleexp catexp fwdbwd =
        let inout = create 
            ~scale_prob:scale_prob ~localscore:localscore
            n epsilon ruleprob ruleprior start 
            ruleexp catexp fwdbwd in
        compute_m_step inout

end

module EStepProblem (ExpCalc : EXPCALC_TYPE) = struct

    module EStep = EStep (ExpCalc)

    let identifier = "EStep"

    type init = unit

    type client_data = unit

    type collection = {
        min       : int;
        max       : int;
        epsilon   : float;
        ruleprob  : DerivGenModel.t;
        ruleprior : DerivGenModel.t;
        start     : Syncat.t;
        get_chart : int -> Chart.t;
        fwdbwd    : Hmm.fwdbwd_t;
        postreg   : bool;
    }

    type genmodellist_t = (Deriv.t * float) list

    type enum = {
        cur   : int;
        limit : int;
        ep    : float;
        rpb   : genmodellist_t;
        rpr   : genmodellist_t;
        sc    : Syncat.t;
        gc    : int -> Chart.t;
        fb    : Hmm.fwdbwd_t;
        pr    : bool;
    }

    type input = int * int * float * genmodellist_t * genmodellist_t * Chart.t * Syncat.t * Hmm.fwdbwdserl_t * bool

    type key = DerivTblInst of Deriv.t
             | ScoreTblInst of ScoreInst.t
             | FBInitInst of Hmm.InitInst.t
             | FBTransInst of Hmm.TransInst.t
             | FBEmisInst of Hmm.EmisInst.t
             | FBCatInst of Hmm.CatInst.t
             | FBTotalCatExp

    type value = float

    type output = DerivGenModel.t * ScoreTbl.t * Hmm.fwdbwd_t

    let serialize_genmodel gm =
        let result = ref [] in
        DerivGenModel.iter gm
            (fun inst _ ->
                result := (inst, DerivGenModel.get_prob gm inst) :: !result);
        !result

    let enum_start range = { 
        cur = range.min; 
        limit = range.max; 
        ep = range.epsilon;
        rpb = serialize_genmodel range.ruleprob;
        rpr = serialize_genmodel range.ruleprior;
        sc = range.start;
        gc = range.get_chart;
        fb = range.fwdbwd;
        pr = range.postreg;
    }

    let create_genmodel gml epsilon =
        let gm = create_pcfg 1000 epsilon in
        List.iter
            (fun (inst, prob) ->
                DerivGenModel.acc_prob gm inst prob)
            gml;
        DerivGenModel.calc_probftrs gm;
        DerivGenModel.calc_normfacts gm;
        gm

    let enum_step e =
        if e.cur <= e.limit then begin
            Scale.inc_and_display ();
            let chart = e.gc e.cur in
            let input = (
                e.cur, e.limit, e.ep, 
                e.rpb, e.rpr, 
                chart, e.sc, Hmm.serialize e.fb, e.pr
            ) in
            Some (input, { 
                cur = e.cur + 1; 
                limit = e.limit; 
                ep = e.ep;
                rpr = e.rpr;
                rpb = e.rpb;
                sc = e.sc;
                gc = e.gc;
                fb = e.fb;
                pr = e.pr;
            })
        end else None

    let init_client () = ()

    let compare_keys = Pervasives.compare

    let map () (
        idx, noentries, epsilon, ruleproblst, rulepriorlst, chart, start, fwdbwdserl, is_postreg
    ) =
        (* Printf.printf "map!\n"; flush stdout; *)
        let ruleprob = create_genmodel ruleproblst epsilon
        and ruleprior = create_genmodel rulepriorlst epsilon in
        let ruleexp' = 
            create_pcfg noentries epsilon
        and catexp' = 
            ScoreTbl.create noentries epsilon in
        let fwdbwd = Hmm.deserialize fwdbwdserl in
        if is_postreg then
            Hmm.train_onestep fwdbwd chart
        else ();
        (* Printf.printf "learn chart!\n"; flush stdout; *)
        ignore(
            EStep.compute
                noentries epsilon 
                ruleprob ruleprior
                chart start
                ruleexp' catexp'
        );
        (* Printf.printf "learn chart done!\n"; flush stdout; *)
        let result = ref [] in
        DerivGenModel.iter ruleexp'
            (fun inst prob ->
                result := (DerivTblInst inst, prob) :: !result);
        ScoreTbl.iter catexp'
            (fun inst prob ->
                result := (ScoreTblInst inst, prob) :: !result);
        Hmm.InitProb.iter (Hmm.get_initexp fwdbwd)
            (fun inst prob ->
                result := (FBInitInst inst, prob) :: !result);
        Hmm.TransProb.iter (Hmm.get_transexp fwdbwd)
            (fun inst prob ->
                result := (FBTransInst inst, prob) :: !result);
        Hmm.EmisProb.iter (Hmm.get_emisexp fwdbwd)
            (fun inst prob ->
                result := (FBEmisInst inst, prob) :: !result);
        Hmm.CatProb.iter (Hmm.get_catexp fwdbwd)
            (fun inst prob ->
                result := (FBCatInst inst, prob) :: !result);
        result := (FBTotalCatExp, !(Hmm.get_totalcatexp fwdbwd)) :: !result;
        (* Printf.printf "map done!\n"; flush stdout; *)
        !result

    let combine v1 v2 =
        v1 +. v2

    let reduce k prob (rp, ct, fb) =
        (* Printf.printf "reduce!\n"; flush stdout; *)
        begin
            match k with
                DerivTblInst inst ->
                    DerivGenModel.acc_prob rp inst prob
              | ScoreTblInst inst ->
                    ScoreTbl.acc_prob ct inst prob
              | FBInitInst inst ->
                    Hmm.InitProb.acc_prob (Hmm.get_initexp fb) inst prob
              | FBTransInst inst ->
                    Hmm.TransProb.acc_prob (Hmm.get_transexp fb) inst prob
              | FBEmisInst inst ->
                    Hmm.EmisProb.acc_prob (Hmm.get_emisexp fb) inst prob
              | FBCatInst inst ->
                    Hmm.CatProb.acc_prob (Hmm.get_catexp fb) inst prob
              | FBTotalCatExp ->
                    let totalcatexp = Hmm.get_totalcatexp fb in
                    totalcatexp := prob +. !totalcatexp
        end;
        (* Printf.printf "reduce done!\n"; flush stdout; *)
        (rp, ct, fb)

end

module EStepMapRed (ExpCalc : EXPCALC_TYPE) = Mapred.Make(EStepProblem(ExpCalc))

module InOutAlgo (ExpCalc : EXPCALC_TYPE) = struct

    type syncat_t = Syncat.t

    type word_t = string

    type chart_t = Chart.t

    type genmodel_t = DerivGenModel.t

    type scoreinst_t = ScoreInst.t

    type scoretbl_t = ScoreTbl.t

    module EStep = EStep (ExpCalc)
    
    module EStepProblem = EStepProblem(ExpCalc)

    module EStepMapRed = EStepMapRed(ExpCalc)
    
    module MStep = MStep (ExpCalc)

    type fn_get_chart_t = int -> chart_t

    type deriv_t = Deriv.t

    type prems_t = Deriv.prems_t
    
    type fn_scale_prob_t = float -> deriv_t -> float

    type fn_localscore_t = int -> int -> float

    type server_t = EStepMapRed.server_t

    let server conf i = 
        EStepMapRed.server conf i

    type t = {
        noentries   : int;
        diffthr     : float;
        rndthr      : int;
        epsilon     : float;
        ruleprob    : genmodel_t;
        ruleprior   : genmodel_t;
        start       : syncat_t;
        get_chart   : fn_get_chart_t;
        scale_prob  : fn_scale_prob_t;
        localscore  : fn_localscore_t;
        nocharts    : int;
        sentexps    : float array;
        fwdbwd      : Hmm.fwdbwd_t;
        mutable server : server_t option;
        mutable rnd : int;
    }

    let create 
            ?(scale_prob = uniform_scale_prob) 
            ?(localscore = uniform_localscore) 
            n diffthr rndthr 
            epsilon ruleprob ruleprior start get_chart nocharts fwdbwd =
        {
            noentries  = n;
            diffthr    = diffthr;
            rndthr     = rndthr;
            epsilon    = epsilon;
            ruleprob   = ruleprob;
            ruleprior  = ruleprior;
            start      = start;
            get_chart  = get_chart;
            scale_prob = scale_prob;
            localscore = localscore;
            nocharts   = nocharts;
            sentexps   = Array.make nocharts 0.0;
            fwdbwd     = fwdbwd;
            server     = None;
            rnd        = 0;
        }

    let set_server inout svr =
        inout.server <- svr

    let get_round inout =
        inout.rnd

    let reset_round inout =
        inout.rnd <- 0

    let compute_onernd_client inout =
        EStepMapRed.client (Utils.client_conf_estep ())

    let compute_onernd_nothr inout =
        Scale.add (Scale.Number inout.nocharts) "charts " " ";
        let oldruleprob = DerivGenModel.create 100 inout.epsilon [] in
        DerivGenModel.iter inout.ruleprob
            (fun inst prob ->
                DerivGenModel.acc_prob oldruleprob inst prob);
        let ruleexp = create_pcfg inout.noentries inout.epsilon
        and catexp = ScoreTbl.create inout.noentries inout.epsilon in
        for ind = 0 to inout.nocharts - 1 do
            let chart = inout.get_chart ind in
            Scale.inc_and_display ();
            ignore (
                EStep.compute
                    inout.noentries inout.epsilon 
                    inout.ruleprob inout.ruleprior
                    chart inout.start
                    ruleexp catexp
            )
        done;
        Scale.discard ();
        MStep.compute
            inout.noentries inout.epsilon 
            inout.ruleprob inout.ruleprior
            inout.start
            ruleexp catexp inout.fwdbwd;
        DerivGenModel.dkl inout.ruleprob oldruleprob 

    let compute_onernd_server inout =
        match inout.server with
            Some svr -> begin
                Scale.add (Scale.Number inout.nocharts) "charts " " ";
                let oldruleprob = DerivGenModel.create 100 inout.epsilon [] in
                DerivGenModel.iter inout.ruleprob
                    (fun inst prob ->
                        DerivGenModel.acc_prob oldruleprob inst prob);
                let ruleexp = create_pcfg inout.noentries inout.epsilon
                and catexp = ScoreTbl.create inout.noentries inout.epsilon in
                let chartcoll = {
                    EStepProblem.min = 0;
                    EStepProblem.max = inout.nocharts - 1;
                    EStepProblem.epsilon = inout.epsilon;
                    EStepProblem.ruleprob = inout.ruleprob;
                    EStepProblem.ruleprior = inout.ruleprior;
                    EStepProblem.start = inout.start;
                    EStepProblem.get_chart = inout.get_chart;
                    EStepProblem.fwdbwd = inout.fwdbwd;
                    EStepProblem.postreg = Utils.is_postreg ();
                } in
                ignore ( EStepMapRed.compute svr chartcoll (ruleexp, catexp, inout.fwdbwd) );
                Scale.discard ();
                MStep.compute
                    inout.noentries inout.epsilon 
                    inout.ruleprob inout.ruleprior
                    inout.start
                    ruleexp catexp inout.fwdbwd;
                DerivGenModel.dkl inout.ruleprob oldruleprob
            end
          | None -> compute_onernd_nothr inout

    let compute_onernd ?(run_as_server=false) inout =
        if run_as_server then
            compute_onernd_server inout
        else
            compute_onernd_nothr inout

    let em_term_cond inout absdkl olddkl =
        let diff = ref (absdkl -. olddkl) in
        if !diff < 0.0 then diff := -1.0 *. !diff
        else ();
        absdkl <= inout.diffthr 
            || (absdkl <= 1e-10 && !diff < 1e-10) 
            || inout.rnd >= inout.rndthr

    let compute_inout ?(run_as_server=false) ?(run_as_client=false) inout =
        if run_as_server then begin
            Scale.clearline ();
            Scale.print ">> Training with Inside/Outside Algorithm\n";
            Scale.redisplay ();
            Scale.add Scale.NumberNoMax "EM (iter " ") > ";
            let absdkl = ref infinity 
            and olddkl = ref infinity in
            while not (em_term_cond inout !absdkl !olddkl) do
                olddkl := !absdkl;
                Scale.inc_and_display ();
                let starttime = Unix.time () in
                let dkl = compute_onernd inout ~run_as_server:run_as_server in
                let endtime = Unix.time () in
                if dkl < 0.0 then absdkl := -1.0 *. dkl
                else absdkl := dkl;
                inout.rnd <- inout.rnd + 1;
                Scale.clearline ();
                let runtime = (endtime -. starttime) /. 60.0 in
                begin
                    match !goldmodel with
                        None ->
                            Scale.print (sprintf 
                                "     Iteration %d : D_KL(old | new) = %e, runtime = %.2f mins\n" 
                                inout.rnd dkl runtime)
                      | Some gm -> begin
                            let golddkl = DerivGenModel.dkl gm inout.ruleprob in 
                            Scale.print (sprintf 
                                "     Iteration %d: D_KL(old | new) = %e, D_KL(gold | new) = %e, runtime = %.2f mins\n" 
                                inout.rnd dkl golddkl runtime)
                        end;
                end;
                Scale.redisplay ()
            done;
            Scale.discard ()
        end else
            compute_onernd_client inout

    let compute 
            ?(svr = None)
            ?(scale_prob = uniform_scale_prob) 
            ?(localscore = uniform_localscore) 
            ?(run_as_server=false) ?(run_as_client=false)
            n diffthr rndthr 
            epsilon ruleprob ruleprior start get_chart nocharts fwdbwd =
        let inout = create 
            ~scale_prob:scale_prob ~localscore:localscore
            n diffthr rndthr 
            epsilon ruleprob ruleprior start get_chart nocharts fwdbwd in
        set_server inout svr;
        compute_inout inout ~run_as_server:run_as_server ~run_as_client:run_as_client

end

module HashedInt = struct

    type t = int

    let hash = Hashtbl.hash

    let equal = Pervasives.(=)

end

let leapfrogthr = ref 16

module BabystepInOutAlgo (ExpCalc : EXPCALC_TYPE) = struct

    type syncat_t = Syncat.t

    type word_t = string

    type chart_t = Chart.t

    type genmodel_t = DerivGenModel.t

    module InOutAlgo = InOutAlgo(ExpCalc)

    type inoutalgo_t = InOutAlgo.t

    module StepTbl = Hashtbl.Make(HashedInt)

    type steptbl_t = (int list) StepTbl.t

    type fn_get_chart_t = int -> chart_t

    type deriv_t = Deriv.t

    type prems_t = Deriv.prems_t
    
    type fn_scale_prob_t = float -> deriv_t -> float

    type fn_localscore_t = int -> int -> float

    module EStepMapRed = EStepMapRed(ExpCalc)

    type server_t = EStepMapRed.server_t

    let server conf i =
        EStepMapRed.server conf i

    type t = {
        noentries     : int;
        diffthr       : float;
        rndthr        : int;
        epsilon       : float;
        ruleprob      : genmodel_t;
        ruleprior     : genmodel_t;
        start         : syncat_t;
        get_chart     : fn_get_chart_t;
        scale_prob    : fn_scale_prob_t;
        localscore    : fn_localscore_t;
        nocharts      : int;
        sentexps      : float array;
        steptbl       : steptbl_t;
        fwdbwd        : Hmm.fwdbwd_t;
        mutable steps : int list ref;
        mutable inout : inoutalgo_t option;
        mutable server: server_t option;
        mutable rnd   : int;
    }

    let create 
        ?(scale_prob = uniform_scale_prob) 
        ?(localscore = uniform_localscore) 
        n diffthr rndthr 
        epsilon ruleprob ruleprior start get_chart nocharts fwdbwd =
    {
        noentries   = n;
        diffthr     = diffthr;
        rndthr      = rndthr;
        epsilon     = epsilon;
        ruleprob    = ruleprob;
        ruleprior   = ruleprior;
        start       = start;
        get_chart   = get_chart;
        scale_prob  = scale_prob;
        localscore  = localscore;
        nocharts    = nocharts;
        sentexps    = Array.make nocharts 0.0;
        steptbl     = StepTbl.create 100;
        fwdbwd      = fwdbwd;
        steps       = ref [];
        server      = None;
        inout       = None;
        rnd         = 0;
    }

    let set_server inout svr =
        inout.server <- svr

    let get_round babystep =
        babystep.rnd

    let reset_round babystep =
        babystep.rnd <- 0

    let assg_step babystep sentlen =
        if sentlen <= !leapfrogthr then sentlen
        else 1 + int_of_float (5.0 *. ceil (float_of_int (sentlen - 1) /. 5.0))

    let build_steptbl babystep =
        Scale.add Scale.StaticInfix "Building steps ..." "";
        Scale.display ();
        for i = 0 to babystep.nocharts - 1 do begin
            let chart = babystep.get_chart i in
            let sentlen = assg_step babystep (List.length (Chart.get_sentence chart)) in
            if not (StepTbl.mem babystep.steptbl sentlen) then
                StepTbl.replace babystep.steptbl sentlen []
            else ();
            let entry = StepTbl.find babystep.steptbl sentlen in
            StepTbl.replace babystep.steptbl sentlen (entry @ [i]);
            if not (List.mem sentlen !(babystep.steps)) then
                babystep.steps := (sentlen :: !(babystep.steps))
            else ()
        end done;
        babystep.steps := List.sort Pervasives.compare !(babystep.steps);
        (*
        for i = 0 to (List.length !(babystep.steps) - 2) do
            let sentlen = List.nth !(babystep.steps) i in
            let nextsentlen = List.nth !(babystep.steps) (i + 1) in
            let entry = StepTbl.find babystep.steptbl sentlen
            and nextentry = StepTbl.find babystep.steptbl nextsentlen in
            StepTbl.replace babystep.steptbl nextsentlen (entry @ nextentry)
        done;
        *)
        Scale.discard ()

    let get_chart babystep sentlen i =
        let entry = StepTbl.find babystep.steptbl sentlen in
        let sentid = List.nth entry i in
        babystep.get_chart sentid

    let count_stats babystep sentlen =
        let entry = StepTbl.find babystep.steptbl sentlen in
        let nocharts = List.length entry in
        let get_chart' = get_chart babystep sentlen in
        let derivcnt = ref 0
        and treecnt = ref 0 in
        for i = 0 to nocharts - 1 do
            let chart = get_chart' i in
            derivcnt := !derivcnt + Chart.get_derivcnt chart;
            treecnt := !treecnt + Chart.get_treecnt chart
        done;
        (nocharts, !derivcnt, !treecnt)

    let compute_babystep ?(run_as_server=false) babystep =
        for i = 0 to (List.length !(babystep.steps) - 1) do
            let sentlen = List.nth !(babystep.steps) i in
            let (nocharts, derivcnt, treecnt) = count_stats babystep sentlen in
            Scale.clearline ();
            if sentlen <= !leapfrogthr then
                Scale.print (
                    Printf.sprintf 
                        ">> Running Babystep (sentlen=%d, charts=%d, derivs=%d, trees=%d)\n" 
                        (sentlen - 1) nocharts derivcnt treecnt)
            else
                Scale.print (
                    Printf.sprintf 
                        ">> Running Leapfrog (sentlen=%d, charts=%d, derivs=%d, trees=%d)\n" 
                        (sentlen - 1) nocharts derivcnt treecnt);
            Scale.redisplay ();
            if sentlen <= !leapfrogthr then
                Scale.add Scale.StaticInfix 
                    (Printf.sprintf "Babystep (sentlen=%d" (sentlen - 1))
                    ") > "
            else
                Scale.add Scale.StaticInfix 
                    (Printf.sprintf "Leapfrog (sentlen=%d" (sentlen - 1))
                    ") > ";
            Scale.display ();
            let entry = StepTbl.find babystep.steptbl sentlen in
            let get_chart' = get_chart babystep sentlen 
            and nocharts = List.length entry in
            InOutAlgo.compute
                ~svr:babystep.server 
                ~scale_prob:babystep.scale_prob ~localscore:babystep.localscore
                babystep.noentries babystep.diffthr babystep.rndthr 
                babystep.epsilon babystep.ruleprob babystep.ruleprior 
                babystep.start get_chart' nocharts babystep.fwdbwd
                ~run_as_server:run_as_server ~run_as_client:false;
            Scale.discard ()
        done

    let compute 
            ?(svr = None)
            ?(scale_prob = uniform_scale_prob) 
            ?(localscore = uniform_localscore) 
            ?(run_as_server=false) ?(run_as_client=false)
            n diffthr rndthr 
            epsilon ruleprob ruleprior start get_chart nocharts fwdbwd =
        if not run_as_client then begin
            let babystep = create 
                ~scale_prob:scale_prob ~localscore:localscore
                n diffthr rndthr 
                epsilon ruleprob ruleprior start get_chart nocharts fwdbwd in
            set_server babystep svr;
            build_steptbl babystep;
            compute_babystep babystep ~run_as_server:run_as_server
        end else
            InOutAlgo.compute
                ~svr:svr
                ~scale_prob:scale_prob ~localscore:localscore
                ~run_as_server:run_as_server ~run_as_client:run_as_client
                n diffthr rndthr
                epsilon ruleprob ruleprior start get_chart nocharts fwdbwd

end

module StandardInOutExpCalc = struct

    let calc_prob ruleexp catexp =
        if ruleexp > 0.0 && catexp > 0.0 then 
            ruleexp /. catexp
        else 0.0

end

module VBInOutExpCalc = struct

    let calc_prob ruleexp catexp =
        if ruleexp > 0.0 && catexp > 0.0 then 
            exp (digamma ruleexp -. digamma catexp)
        else 0.0

end

module StandardInOutAlgo = InOutAlgo(StandardInOutExpCalc)

module VBInOutAlgo = InOutAlgo(VBInOutExpCalc)

module StandardBabystepInOutAlgo = BabystepInOutAlgo(StandardInOutExpCalc)

module VBBabystepInOutAlgo = BabystepInOutAlgo(VBInOutExpCalc)

module StandardMapRed = EStepMapRed(StandardInOutExpCalc)

module VBMapRed = EStepMapRed(VBInOutExpCalc)
