(** Inside-Outside Algorithm *)

(***********************************************************
*                                                          *
*      FUNGI - Fast UNsupervised Grammar Induction         *
*                                                          *
*   Originator: Prachya Boonkwan [1], [2]                  *
*               [1] School of Informatics,                 *
*                   University of Edinburgh, UK            *
*               [2] National Electronics and Computer      *
*                   Technology Center (NECTEC), Thailand   *
*                                                          *
*   Copyright (C) 2011 Prachya Boonkwan                    *
*   <p.boonkwan@sms.ed.ac.uk, kaamanita@hotmail.com>       *
*                                                          *
*   This library is free software; you can redistribute    *
*   it and/or modify it under the terms of the GNU         *
*   Lesser General Public License as published by the      *
*   Free Software Foundation; either version 2.1 of the    *
*   License, or (at your option) any later version.        *
*                                                          *
*   This library is distributed in the hope that it will   *
*   be useful, but WITHOUT ANY WARRANTY; without even      *
*   the implied warranty of MERCHANTABILITY or FITNESS     *
*   FOR A PARTICULAR PURPOSE.  See the GNU Lesser          *
*   General Public License for more details.               *
*                                                          *
*   You should have received a copy of the GNU Lesser      *
*   General Public License along with this library; if     *
*   not, write to the Free Software Foundation, Inc.,      *
*   59 Temple Place, Suite 330, Boston, MA  02111-1307     *
*   USA.                                                   *
*                                                          *
***********************************************************)

module type EXPCALC_TYPE = sig
    val calc_prob : float -> float -> float
end

val set_goldmodel : Deriv.DerivGenModel.t -> unit

module InOutAlgo :
functor (ExpCalc : EXPCALC_TYPE) ->
sig
    type syncat_t = Cg.Syncat.t
    type word_t = string
    type chart_t = Chart.Chart.t
    type genmodel_t = Deriv.DerivGenModel.t
    type fn_get_chart_t = int -> chart_t
    type deriv_t = Deriv.Deriv.t
    type fn_scale_prob_t = float -> deriv_t -> float
    type fn_localscore_t = int -> int -> float
    type server_t
    val server : JoinHelper.configuration -> unit -> server_t
    type t
    val create :
        ?scale_prob:fn_scale_prob_t -> ?localscore:fn_localscore_t ->
        int -> float -> int -> float ->
        genmodel_t -> genmodel_t -> syncat_t ->
        fn_get_chart_t -> int -> Hmm.fwdbwd_t -> t
    val set_server : t -> server_t option -> unit
    val get_round : t -> int
    val reset_round : t -> unit
    (* val compute_onernd : t -> float *)
    val compute_inout : ?run_as_server:bool -> ?run_as_client:bool -> t -> unit
    val compute :
        ?svr:server_t option -> 
        ?scale_prob:fn_scale_prob_t -> ?localscore:fn_localscore_t ->
        ?run_as_server:bool -> ?run_as_client:bool ->
        int -> float -> int -> float ->
        genmodel_t -> genmodel_t -> syncat_t ->
        fn_get_chart_t -> int -> Hmm.fwdbwd_t -> unit
end

module BabystepInOutAlgo :
functor (ExpCalc : EXPCALC_TYPE) ->
sig
    type syncat_t = Cg.Syncat.t
    type word_t = string
    type chart_t = Chart.Chart.t
    type genmodel_t = Deriv.DerivGenModel.t
    type fn_get_chart_t = int -> chart_t
    type deriv_t = Deriv.Deriv.t
    type fn_scale_prob_t = float -> deriv_t -> float
    type fn_localscore_t = int -> int -> float
    type server_t
    val server : JoinHelper.configuration -> unit -> server_t
    type t
    val create :
        ?scale_prob:fn_scale_prob_t -> ?localscore:fn_localscore_t ->
        int -> float -> int -> float ->
        genmodel_t -> genmodel_t -> syncat_t ->
        fn_get_chart_t -> int -> Hmm.fwdbwd_t -> t
    val set_server : t -> server_t option -> unit
    val get_round : t -> int
    val reset_round : t -> unit
    val compute_babystep : ?run_as_server:bool -> t -> unit
    val compute :
        ?svr:server_t option -> 
        ?scale_prob:fn_scale_prob_t -> ?localscore:fn_localscore_t ->
        ?run_as_server:bool -> ?run_as_client:bool -> 
        int -> float -> int -> float ->
        genmodel_t -> genmodel_t -> syncat_t ->
        fn_get_chart_t -> int -> Hmm.fwdbwd_t -> unit
end

(** Standard Inside-Outside Algorithm *)
module StandardInOutAlgo :
sig
    type syncat_t = Cg.Syncat.t
    type word_t = string
    type chart_t = Chart.Chart.t
    type genmodel_t = Deriv.DerivGenModel.t
    type fn_get_chart_t = int -> chart_t
    type fn_scale_prob_t = float -> Deriv.Deriv.t -> float
    type fn_localscore_t = int -> int -> float
    type server_t
    val server : JoinHelper.configuration -> unit -> server_t
    type t
    val create :
        ?scale_prob:fn_scale_prob_t ->
        ?localscore:fn_localscore_t ->
        int -> float -> int -> float ->
        genmodel_t -> genmodel_t -> 
        syncat_t -> fn_get_chart_t -> int -> Hmm.fwdbwd_t -> t
    val set_server : t -> server_t option -> unit
    val get_round : t -> int
    val reset_round : t -> unit
    val compute_inout : ?run_as_server:bool -> ?run_as_client:bool -> t -> unit
    val compute :
        ?svr:server_t option -> 
        ?scale_prob:fn_scale_prob_t ->
        ?localscore:fn_localscore_t ->
        ?run_as_server:bool -> ?run_as_client:bool -> 
        int -> float -> int -> float ->
        genmodel_t -> genmodel_t -> 
        syncat_t -> fn_get_chart_t -> int -> Hmm.fwdbwd_t -> unit
end

(** Variational-Bayesian Inside-Outside Algorithm *)
module VBInOutAlgo :
sig
    type syncat_t = Cg.Syncat.t
    type word_t = string
    type chart_t = Chart.Chart.t
    type genmodel_t = Deriv.DerivGenModel.t
    type fn_get_chart_t = int -> chart_t
    type fn_scale_prob_t = float -> Deriv.Deriv.t -> float
    type fn_localscore_t = int -> int -> float
    type server_t
    val server : JoinHelper.configuration -> unit -> server_t
    type t
    val create :
        ?scale_prob:fn_scale_prob_t ->
        ?localscore:fn_localscore_t ->
        int -> float -> int -> float ->
        genmodel_t -> genmodel_t -> 
        syncat_t -> fn_get_chart_t -> int -> Hmm.fwdbwd_t -> t
    val set_server : t -> server_t option -> unit
    val get_round : t -> int
    val reset_round : t -> unit
    val compute_inout : ?run_as_server:bool -> ?run_as_client:bool -> t -> unit
    val compute :
        ?svr:server_t option -> 
        ?scale_prob:fn_scale_prob_t ->
        ?localscore:fn_localscore_t ->
        ?run_as_server:bool -> ?run_as_client:bool -> 
        int -> float -> int -> float ->
        genmodel_t -> genmodel_t -> 
        syncat_t -> fn_get_chart_t -> int -> Hmm.fwdbwd_t -> unit
end

val leapfrogthr : int ref

(** Standard Babystep Inside-Outside Algorithm *)
module StandardBabystepInOutAlgo :
sig
    type syncat_t = Cg.Syncat.t
    type word_t = string
    type chart_t = Chart.Chart.t
    type genmodel_t = Deriv.DerivGenModel.t
    type fn_get_chart_t = int -> chart_t
    type fn_scale_prob_t = float -> Deriv.Deriv.t -> float
    type fn_localscore_t = int -> int -> float
    type server_t
    val server : JoinHelper.configuration -> unit -> server_t
    type t
    val create :
        ?scale_prob:fn_scale_prob_t ->
        ?localscore:fn_localscore_t ->
        int -> float -> int -> float ->
        genmodel_t -> genmodel_t -> 
        syncat_t -> fn_get_chart_t -> int -> Hmm.fwdbwd_t -> t
    val set_server : t -> server_t option -> unit
    val get_round : t -> int
    val reset_round : t -> unit
    val compute_babystep : ?run_as_server:bool -> t -> unit
    val compute :
        ?svr:server_t option -> 
        ?scale_prob:fn_scale_prob_t ->
        ?localscore:fn_localscore_t ->
        ?run_as_server:bool -> ?run_as_client:bool -> 
        int -> float -> int -> float ->
        genmodel_t -> genmodel_t -> 
        syncat_t -> fn_get_chart_t -> int -> Hmm.fwdbwd_t -> unit
end

(** Variational-Bayesian Babystep Inside-Outside Algorithm *)
module VBBabystepInOutAlgo :
sig
    type syncat_t = Cg.Syncat.t
    type word_t = string
    type chart_t = Chart.Chart.t
    type genmodel_t = Deriv.DerivGenModel.t
    type fn_get_chart_t = int -> chart_t
    type fn_scale_prob_t = float -> Deriv.Deriv.t -> float
    type fn_localscore_t = int -> int -> float
    type server_t
    val server : JoinHelper.configuration -> unit -> server_t
    type t
    val create :
        ?scale_prob:fn_scale_prob_t ->
        ?localscore:fn_localscore_t ->
        int -> float -> int -> float ->
        genmodel_t -> genmodel_t -> 
        syncat_t -> fn_get_chart_t -> int -> Hmm.fwdbwd_t -> t
    val set_server : t -> server_t option -> unit
    val get_round : t -> int
    val reset_round : t -> unit
    val compute_babystep : ?run_as_server:bool -> t -> unit
    val compute :
        ?svr:server_t option -> 
        ?scale_prob:fn_scale_prob_t ->
        ?localscore:fn_localscore_t ->
        ?run_as_server:bool -> ?run_as_client:bool -> 
        int -> float -> int -> float ->
        genmodel_t -> genmodel_t -> 
        syncat_t -> fn_get_chart_t -> int -> Hmm.fwdbwd_t -> unit
end

module StandardMapRed :
sig
    type init = unit
    type input
    type output
    type collection
    type server_t
    val client : JoinHelper.configuration -> unit
    val server : JoinHelper.configuration -> init -> server_t
    val compute : server_t -> collection -> output -> output
end

module VBMapRed :
sig
    type init = unit
    type input
    type output
    type collection
    type server_t
    val client : JoinHelper.configuration -> unit
    val server : JoinHelper.configuration -> init -> server_t
    val compute : server_t -> collection -> output -> output
end
