(***********************************************************
*                                                          *
*      FUNGI - Fast UNsupervised Grammar Induction         *
*                                                          *
*   Originator: Prachya Boonkwan [1], [2]                  *
*               [1] School of Informatics,                 *
*                   University of Edinburgh, UK            *
*               [2] National Electronics and Computer      *
*                   Technology Center (NECTEC), Thailand   *
*                                                          *
*   Copyright (C) 2011 Prachya Boonkwan                    *
*   <p.boonkwan@sms.ed.ac.uk, kaamanita@hotmail.com>       *
*                                                          *
*   This library is free software; you can redistribute    *
*   it and/or modify it under the terms of the GNU         *
*   Lesser General Public License as published by the      *
*   Free Software Foundation; either version 2.1 of the    *
*   License, or (at your option) any later version.        *
*                                                          *
*   This library is distributed in the hope that it will   *
*   be useful, but WITHOUT ANY WARRANTY; without even      *
*   the implied warranty of MERCHANTABILITY or FITNESS     *
*   FOR A PARTICULAR PURPOSE.  See the GNU Lesser          *
*   General Public License for more details.               *
*                                                          *
*   You should have received a copy of the GNU Lesser      *
*   General Public License along with this library; if     *
*   not, write to the Free Software Foundation, Inc.,      *
*   59 Temple Place, Suite 330, Boston, MA  02111-1307     *
*   USA.                                                   *
*                                                          *
***********************************************************)

open Cg
open Chart
open Prob
open Deriv
open Utils
open Printf

type syncat_t = Syncat.t
type word_t = string
type sent_t = word_t list
type time_t = int
type chart_t = Chart.t

module TransInst = struct

    type t = {
        src  : syncat_t;
        dest : syncat_t;
    }

    let create s d = {
        src  = s;
        dest = d;
    }

    let get_src inst = inst.src

    let get_dest inst = inst.dest

    let equal i1 i2 =
        Syncat.equal i1.src i2.src && Syncat.equal i1.dest i2.dest

    let hash inst =
        Syncat.hash inst.src + Syncat.hash inst.dest

    let to_string inst =
        sprintf "%s -> %s" 
            (Syncat.to_string inst.src) 
            (Syncat.to_string inst.dest)

end

module EmisInst = struct

    type t = {
        syncat : syncat_t;
        word   : word_t;
    }

    let create cat word = {
        syncat = cat;
        word   = word;
    }

    let get_syncat inst = inst.syncat

    let get_word inst = inst.word

    let equal e1 e2 =
        Syncat.equal e1.syncat e2.syncat && e1.word = e2.word

    let hash inst =
        Syncat.hash inst.syncat + Hashtbl.hash inst.word

    let to_string inst =
        sprintf "%s -> %s"
            (Syncat.to_string inst.syncat) inst.word

end

module InitInst = struct

    type t = syncat_t

    let create s = s

    let get_syncat s = s

    let equal = Syncat.equal

    let hash = Syncat.hash

    let to_string = Syncat.to_string

end

module CatInst = InitInst

module StatCatInst = struct

    type t = {
        syncat : syncat_t;
        time   : time_t;
    }

    let create s q = {
        syncat = s;
        time   = q;
    }

    let get_syncat inst = inst.syncat

    let get_time inst = inst.time

    let equal i1 i2 =
        Syncat.equal i1.syncat i2.syncat && i1.time = i2.time

    let hash inst =
        Syncat.hash inst.syncat + Hashtbl.hash inst.time

    let to_string inst =
        sprintf "%s @ t=%d"
            (Syncat.to_string inst.syncat) inst.time

end

module StatTransInst = struct

    type t = {
        src  : syncat_t;
        dest : syncat_t;
        time : time_t;
    }

    let create s d q = {
        src  = s;
        dest = d;
        time = q;
    }

    let get_src inst = inst.src

    let get_dest inst = inst.dest

    let get_time inst = inst.time

    let equal i1 i2 =
        Syncat.equal i1.src i2.src 
            && Syncat.equal i1.dest i2.dest
            && i1.time = i2.time

    let hash inst =
        Syncat.hash inst.src + Syncat.hash inst.dest + Hashtbl.hash inst.time

    let to_string inst =
        sprintf "%s -> %s @ t=%d" 
            (Syncat.to_string inst.src) 
            (Syncat.to_string inst.dest)
            inst.time

end

module StatInst = struct

    type t = int

    let create q = q

    let equal = (=)

    let hash = Hashtbl.hash

    let to_string = string_of_int

end

module InitProb = struct 
    include ProbTbl(InitInst)

    type list_t = (inst_t * float) list

    let serialize probtbl =
        let result = ref [] in
        iter probtbl
            (fun inst prob ->
                result := (inst, prob) :: !result);
        !result
end

module TransProb = struct
    include ProbTbl(TransInst)

    type list_t = (inst_t * float) list

    let serialize probtbl =
        let result = ref [] in
        iter probtbl
            (fun inst prob ->
                result := (inst, prob) :: !result);
        !result
end

module EmisProb = struct
    include ProbTbl(EmisInst)

    type list_t = (inst_t * float) list

    let serialize probtbl =
        let result = ref [] in
        iter probtbl
            (fun inst prob ->
                result := (inst, prob) :: !result);
        !result
end

module CatProb = struct 
    include ProbTbl(CatInst)

    type list_t = (inst_t * float) list

    let serialize probtbl =
        let result = ref [] in
        iter probtbl
            (fun inst prob ->
                result := (inst, prob) :: !result);
        !result
end

module StatCatProb = ProbTbl(StatCatInst)

module StatTransProb = ProbTbl(StatTransInst)

module StatProb = ProbTbl(StatInst)

module VisitedInst = struct

    type t =
        Trans of syncat_t * syncat_t
      | Emis of syncat_t * word_t
      | Init of syncat_t

    let create inst = inst

    let create_trans src dest = Trans (src, dest)

    let create_emis syncat word = Emis (syncat, word)

    let create_init syncat = Init syncat

    let equal = (=)

    let hash = Hashtbl.hash

    let to_string inst =
        match inst with
            Trans (src, dest) -> 
                sprintf "trans: %s -> %s"
                    (Syncat.to_string src) (Syncat.to_string dest)
          | Emis (syncat, word) ->
                sprintf "emis: %s -> %s"
                    (Syncat.to_string syncat) word
          | Init syncat ->
                sprintf "init: %s" (Syncat.to_string syncat)

end

module VisitedProb = struct
    include ProbTbl(VisitedInst)

    type list_t = (inst_t * float) list

    let serialize probtbl =
        let result = ref [] in
        iter probtbl
            (fun inst prob ->
                result := (inst, prob) :: !result);
        !result
end

module FwdBwdOneStep = struct

    type chart_t = Chart.t

    type transprob_t = TransProb.t

    type visitedprob_t = VisitedProb.t

    type initprob_t = InitProb.t

    type emisprob_t = EmisProb.t

    type catprob_t = CatProb.t

    type statcatprob_t = StatCatProb.t

    type stattransprob_t = StatTransProb.t

    type statprob_t = StatProb.t

    type t = {
        epsilon      : float;
        chart        : chart_t;
        sent         : sent_t;
        sentlen      : int;
        visited      : visitedprob_t;
        initprob     : initprob_t;
        transprob    : transprob_t;
        emisprob     : emisprob_t;
        initexp      : initprob_t;
        transexp     : transprob_t;
        emisexp      : emisprob_t;
        catexp       : catprob_t;
        totalcatexp  : float ref;
    }

    let create 
        epsilon chart 
        visited initprob transprob emisprob 
        initexp transexp emisexp catexp totalcatexp
    = {
        epsilon = epsilon;
        chart = chart;
        sent = Chart.get_sentence chart;
        sentlen = List.length (Chart.get_sentence chart);
        visited = visited;
        initprob = initprob;
        transprob = transprob;
        emisprob = emisprob;
        initexp = initexp;
        transexp = transexp;
        emisexp = emisexp;
        catexp = catexp;
        totalcatexp = totalcatexp;
    }

    let calc_fwdcnt ?(init=false) algo =
        let fwdcnt = StatCatProb.create 100 algo.epsilon in
        let s1 = algo.sentlen - 2 in
        let s2 = algo.sentlen - 1 in
        for i = 0 to s1 do
            let span = (i, i) in
            Chart.iter_derivs_onelevel algo.chart
                (fun syncat _ _ word ->
                    let inst = StatCatInst.create syncat i in
                    if i = 0 then begin
                        let prob =
                            if not init then
                                InitProb.get_prob algo.initprob syncat
                                *. EmisProb.get_prob algo.emisprob (EmisInst.create syncat word) 
                            else 1.0 in
                        StatCatProb.acc_prob fwdcnt inst prob;
                        VisitedProb.acc_prob algo.visited (VisitedInst.create_init syncat) 0.0
                    end else ();
                    VisitedProb.acc_prob algo.visited (VisitedInst.create_emis syncat word) 0.0;
                    let prob = StatCatProb.get_prob fwdcnt inst in
                    let j = i + 1 in
                    let span' = (j, j) in
                    Chart.iter_derivs_onelevel algo.chart
                        (fun syncat' _ _ word' ->
                            let inst' = StatCatInst.create syncat' j in
                            if j = s2 then
                                VisitedProb.acc_prob algo.visited (VisitedInst.create_emis syncat' word') 0.0
                            else ();
                            let prob' =
                                if not init then
                                    EmisProb.get_prob algo.emisprob (EmisInst.create syncat' word')
                                    *. prob 
                                    *. TransProb.get_prob algo.transprob (TransInst.create syncat syncat')
                                else 1.0 in
                            StatCatProb.acc_prob fwdcnt inst' prob';
                            VisitedProb.acc_prob algo.visited (VisitedInst.create_trans syncat syncat') 0.0)
                        (fun _ _ _ _ _ _ -> ())
                        span')
                (fun _ _ _ _ _ _ -> ())
                span
        done;
        fwdcnt

    let calc_bwdcnt ?(init=false) algo =
        let bwdcnt = StatCatProb.create 100 algo.epsilon in
        for j = algo.sentlen downto 1 do
            let span' = (j, j) in
            Chart.iter_derivs_onelevel algo.chart
                (fun syncat' _ _ word' ->
                    let inst' = StatCatInst.create syncat' j in
                    if j = algo.sentlen then
                        let prob' = 1.0 in
                        StatCatProb.acc_prob bwdcnt inst' prob'
                    else ();
                    let prob' = StatCatProb.get_prob bwdcnt inst' in
                    let i = j - 1 in
                    let span = (i, i) in
                    Chart.iter_derivs_onelevel algo.chart
                        (fun syncat _ _ word ->
                            let inst = StatCatInst.create syncat i in
                            let prob =
                                if not init then
                                    EmisProb.get_prob algo.emisprob (EmisInst.create syncat' word')
                                    *. TransProb.get_prob algo.transprob (TransInst.create syncat syncat')
                                    *. prob'
                                else 1.0 in
                            StatCatProb.acc_prob bwdcnt inst prob)
                        (fun _ _ _ _ _ _ -> ())
                        span)
                (fun _ _ _ _ _ _ -> ())
                span'
        done;
        bwdcnt

    let calc_statcatexp algo fwdcnt bwdcnt =
        let statcatexp = StatCatProb.create 100 algo.epsilon in
        let numexp = StatCatProb.create 100 algo.epsilon
        and denomexp = StatProb.create 100 algo.epsilon in
        StatCatProb.iter fwdcnt
            (fun inst fwd ->
                let bwd = StatCatProb.get_prob bwdcnt inst in
                let score = fwd *. bwd in
                StatCatProb.acc_prob numexp inst score;
                let inst' = StatInst.create (StatCatInst.get_time inst) in
                StatProb.acc_prob denomexp inst' score);
        StatCatProb.iter numexp
            (fun inst score ->
                let inst' = StatInst.create (StatCatInst.get_time inst) in
                let score' = StatProb.get_prob denomexp inst' in
                StatCatProb.acc_prob statcatexp inst (score /. score'));
        statcatexp

    let calc_stattransexp algo fwdcnt bwdcnt =
        let stattransexp = StatTransProb.create 100 algo.epsilon in
        let numexp = StatTransProb.create 100 algo.epsilon
        and denomexp = StatProb.create 100 algo.epsilon in
        let s1 = algo.sentlen - 1 in
        for i = 0 to s1 do
            let span = (i, i) in
            Chart.iter_derivs_onelevel algo.chart
                (fun syncat _ _ word ->
                    let inst = StatCatInst.create syncat i in
                    let j = i + 1 in
                    let span' = (j, j) in
                    Chart.iter_derivs_onelevel algo.chart
                        (fun syncat' _ _ word' ->
                            let inst' = StatCatInst.create syncat' j in
                            let prob = StatCatProb.get_prob fwdcnt inst
                                *. StatCatProb.get_prob bwdcnt inst'
                                *. TransProb.get_prob algo.transprob (TransInst.create syncat syncat')
                                *. EmisProb.get_prob algo.emisprob (EmisInst.create syncat' word') in
                            StatTransProb.acc_prob numexp 
                                (StatTransInst.create syncat syncat' i) prob;
                            StatProb.acc_prob denomexp i prob)
                        (fun _ _ _ _ _ _ -> ())
                        span')
                (fun _ _ _ _ _ _ -> ())
                span
        done;
        StatTransProb.iter numexp
            (fun inst score ->
                let inst' = StatInst.create (StatTransInst.get_time inst) in
                let score' = StatProb.get_prob denomexp inst' in
                StatTransProb.acc_prob stattransexp inst (score /. score'));
        stattransexp

    let calc_exps algo stattransexp statcatexp =
        Chart.iter_derivs_onelevel algo.chart
            (fun syncat _ _ _ ->
                let inst = InitInst.create syncat in
                let ip = StatCatProb.get_prob statcatexp (StatCatInst.create syncat 0) in
                InitProb.acc_prob algo.initexp inst ip)
            (fun _ _ _ _ _ _ -> ())
            (0, 0);
        StatTransProb.iter stattransexp
            (fun inst score ->
                let src = StatTransInst.get_src inst
                and dest = StatTransInst.get_dest inst in
                let inst' = TransInst.create src dest in
                TransProb.acc_prob algo.transexp inst' score);
        StatCatProb.iter statcatexp
            (fun inst score ->
                let syncat = StatCatInst.get_syncat inst in
                let time = StatCatInst.get_time inst in
                let word = List.nth algo.sent time in
                let inst' = EmisInst.create syncat word in
                EmisProb.acc_prob algo.emisexp inst' score;
                CatProb.acc_prob algo.catexp syncat score;
                algo.totalcatexp := score +. !(algo.totalcatexp))

    let run ?(init=false) algo =
        let fwdcnt = calc_fwdcnt algo ~init:init
        and bwdcnt = calc_bwdcnt algo ~init:init in
        let statcatexp = calc_statcatexp algo fwdcnt bwdcnt
        and stattransexp = calc_stattransexp algo fwdcnt bwdcnt in
        calc_exps algo stattransexp statcatexp

    let compute 
            epsilon chart 
            visited initprob transprob emisprob 
            initexp transexp emisexp catexp totalcatexp =
        let algo = create epsilon chart
            visited initprob transprob emisprob
            initexp transexp emisexp catexp totalcatexp in
        run algo ~init:false

    let init
            epsilon chart 
            visited initprob transprob emisprob 
            initexp transexp emisexp catexp totalcatexp =
        let algo = create epsilon chart
            visited initprob transprob emisprob
            initexp transexp emisexp catexp totalcatexp in
        run algo ~init:true

end

type fwdbwd_t = {
    epsilon   : float;
    visited   : VisitedProb.t;
    initprob  : InitProb.t;
    transprob : TransProb.t;
    emisprob  : EmisProb.t;
    initexp   : InitProb.t;
    transexp  : TransProb.t;
    emisexp   : EmisProb.t;
    catexp    : CatProb.t;
    totalcatexp : float ref;
}

let create epsilon = {
    epsilon = epsilon;
    visited = VisitedProb.create 100 epsilon;
    initprob = InitProb.create 100 epsilon;
    transprob = TransProb.create 100 epsilon;
    emisprob = EmisProb.create 100 epsilon;
    initexp = InitProb.create 100 epsilon;
    transexp = TransProb.create 100 epsilon;
    emisexp = EmisProb.create 100 epsilon;
    catexp = CatProb.create 100 epsilon;
    totalcatexp = ref 0.0;
}

let to_string algo =
    Printf.sprintf "Epsilon:\n%e\n\nVisitedProb:\n%s\n\nInitProb:\n%s\n\nTransProb:\n%s\n\nEmisProb:\n%s\n\nInitExp:\n%s\n\nTransExp:\n%s\n\nEmisExp:\n%s\n\nCatExp:\n%s\n\nTotalCatExp = %e"
        algo.epsilon (VisitedProb.to_string algo.visited)
        (InitProb.to_string algo.initprob) (TransProb.to_string algo.transprob) (EmisProb.to_string algo.emisprob)
        (InitProb.to_string algo.initexp) (TransProb.to_string algo.transexp) (EmisProb.to_string algo.emisexp)
        (CatProb.to_string algo.catexp) !(algo.totalcatexp)

let print algo =
    Printf.printf "InitProb:\n%s\n\nTransProb:\n%s\n\nEmisProb:\n%s"
        (InitProb.to_string algo.initprob) 
        (TransProb.to_string algo.transprob) 
        (EmisProb.to_string algo.emisprob)

type fwdbwdserl_t = 
    float *
    VisitedProb.list_t * InitProb.list_t * TransProb.list_t * EmisProb.list_t *
    InitProb.list_t * TransProb.list_t * EmisProb.list_t * CatProb.list_t * 
    float

let serialize algo = (
    algo.epsilon, VisitedProb.serialize algo.visited,
    InitProb.serialize algo.initprob, TransProb.serialize algo.transprob, EmisProb.serialize algo.emisprob,
    InitProb.serialize algo.initexp, TransProb.serialize algo.transexp, EmisProb.serialize algo.emisexp,
    CatProb.serialize algo.catexp,
    !(algo.totalcatexp)
)

let deserialize serl =
    let (epsilon, vs, ips, tps, eps, ies, tes, ees, ces, tce) = serl in
    let visited = VisitedProb.create 100 epsilon in
    let initprob = InitProb.create 100 epsilon in
    let transprob = TransProb.create 100 epsilon in
    let emisprob = EmisProb.create 100 epsilon in
    let initexp = InitProb.create 100 epsilon in
    let transexp = TransProb.create 100 epsilon in
    let emisexp = EmisProb.create 100 epsilon in
    let catexp = CatProb.create 100 epsilon in
    List.iter (fun (i, p) -> VisitedProb.acc_prob visited i p) vs;
    List.iter (fun (i, p) -> InitProb.acc_prob initprob i p) ips;
    List.iter (fun (i, p) -> TransProb.acc_prob transprob i p) tps;
    List.iter (fun (i, p) -> EmisProb.acc_prob emisprob i p) eps;
    List.iter (fun (i, p) -> InitProb.acc_prob initexp i p) ies;
    List.iter (fun (i, p) -> TransProb.acc_prob transexp i p) tes;
    List.iter (fun (i, p) -> EmisProb.acc_prob emisexp i p) ees;
    List.iter (fun (i, p) -> CatProb.acc_prob catexp i p) ces;
    {
        epsilon = epsilon;
        visited = visited;
        initprob = initprob;
        transprob = transprob;
        emisprob = emisprob;
        initexp = initexp;
        transexp = transexp;
        emisexp = emisexp;
        catexp = catexp;
        totalcatexp = ref tce;
    }

let merge algo serl =
    let (_, vs, _, _, _, ies, tes, ees, ces, tce) = serl in
    List.iter (fun (i, p) -> VisitedProb.acc_prob algo.visited i p) vs;
    List.iter (fun (i, p) -> InitProb.acc_prob algo.initexp i p) ies;
    List.iter (fun (i, p) -> TransProb.acc_prob algo.transexp i p) tes;
    List.iter (fun (i, p) -> EmisProb.acc_prob algo.emisexp i p) ees;
    List.iter (fun (i, p) -> CatProb.acc_prob algo.catexp i p) ces;
    algo.totalcatexp := tce +. !(algo.totalcatexp)

let get_epsilon algo = algo.epsilon

let get_visited algo = algo.visited

let get_initprob algo = algo.initprob

let get_transprob algo = algo.transprob

let get_emisprob algo = algo.emisprob

let get_initexp algo = algo.initexp

let get_transexp algo = algo.transexp

let get_emisexp algo = algo.emisexp

let get_catexp algo = algo.catexp

let get_totalcatexp algo = algo.totalcatexp

let init_onestep algo chart =
    FwdBwdOneStep.init
        algo.epsilon chart
        algo.visited algo.initprob algo.transprob algo.emisprob
        algo.initexp algo.transexp algo.emisexp algo.catexp algo.totalcatexp

let train_onestep algo chart =
    FwdBwdOneStep.compute
        algo.epsilon chart
        algo.visited algo.initprob algo.transprob algo.emisprob
        algo.initexp algo.transexp algo.emisexp algo.catexp algo.totalcatexp

let estimate algo =
    VisitedProb.iter algo.visited
        (fun inst _ ->
            match inst with
                VisitedInst.Trans (src, dest) ->
                    let inst = TransInst.create src dest
                    and inst' = CatInst.create src in
                    let score = TransProb.get_prob algo.transexp inst
                    and score' = CatProb.get_prob algo.catexp inst' in
                    TransProb.set_prob algo.transprob inst (score /. score')
              | VisitedInst.Emis (syncat, word) ->
                    let inst = EmisInst.create syncat word
                    and inst' = CatInst.create syncat in
                    let score = EmisProb.get_prob algo.emisexp inst
                    and score' = CatProb.get_prob algo.catexp inst' in
                    EmisProb.set_prob algo.emisprob inst (score /. score')
              | VisitedInst.Init syncat ->  
                    let inst = InitInst.create syncat in
                    let score = InitProb.get_prob algo.initexp inst
                    and score' = !(algo.totalcatexp) in
                    InitProb.set_prob algo.initprob inst (score /. score'));
    VisitedProb.clear algo.visited;
    InitProb.clear algo.initexp;
    TransProb.clear algo.transexp;
    EmisProb.clear algo.emisexp;
    CatProb.clear algo.catexp;
    algo.totalcatexp := 0.0

let from_file fname epsilon =
    let algo = create epsilon in
    let (initprob, transprob, emisprob) = unmarshal_from_file fname in
    InitProb.iter initprob
        (InitProb.set_prob algo.initprob);
    TransProb.iter transprob
        (TransProb.set_prob algo.transprob);
    EmisProb.iter emisprob
        (EmisProb.set_prob algo.emisprob);
    algo

let to_file algo fname =
    let data = (algo.initprob, algo.transprob, algo.emisprob) in
    marshal_to_file fname data
