(** Inside-Outside Score Calculation *)

(***********************************************************
*                                                          *
*      FUNGI - Fast UNsupervised Grammar Induction         *
*                                                          *
*   Originator: Prachya Boonkwan [1], [2]                  *
*               [1] School of Informatics,                 *
*                   University of Edinburgh, UK            *
*               [2] National Electronics and Computer      *
*                   Technology Center (NECTEC), Thailand   *
*                                                          *
*   Copyright (C) 2011 Prachya Boonkwan                    *
*   <p.boonkwan@sms.ed.ac.uk, kaamanita@hotmail.com>       *
*                                                          *
*   This library is free software; you can redistribute    *
*   it and/or modify it under the terms of the GNU         *
*   Lesser General Public License as published by the      *
*   Free Software Foundation; either version 2.1 of the    *
*   License, or (at your option) any later version.        *
*                                                          *
*   This library is distributed in the hope that it will   *
*   be useful, but WITHOUT ANY WARRANTY; without even      *
*   the implied warranty of MERCHANTABILITY or FITNESS     *
*   FOR A PARTICULAR PURPOSE.  See the GNU Lesser          *
*   General Public License for more details.               *
*                                                          *
*   You should have received a copy of the GNU Lesser      *
*   General Public License along with this library; if     *
*   not, write to the Free Software Foundation, Inc.,      *
*   59 Temple Place, Suite 330, Boston, MA  02111-1307     *
*   USA.                                                   *
*                                                          *
***********************************************************)

(** Score Instance *)
module ScoreInst :
sig

    (** Position span *)
    type span_t = int * int

    (** Syntactic category *)
    type syncat_t = Cg.Syncat.t
    
    (** Lexical item *)
    type word_t = string
    
    (** Score instance *)
    type t
    
    (** Create a score instance with a span, a syntactic 
        category, and a head word. *)
    val create : span_t -> syncat_t -> word_t -> t
    
    (** Get the span from the score instance. *)
    val get_span : t -> span_t
    
    (** Get the syntactic category from the score instance. *)
    val get_syncat : t -> syncat_t
    
    (** Get the head word from the score instance. *)
    val get_head : t -> word_t
    
    (** Check if two score instances equal. *)
    val equal : t -> t -> bool
    
    (** Calculate the hash value of the score instance. *)
    val hash : t -> int

end

(** Score Table *)
module ScoreTbl :
sig
    
    (** Score instance *)
    type inst_t = ScoreInst.t
    
    (** Score table *)
    type t
    
    (** Create a probability table with the size and the default value. *)
    val create : int -> float -> t
    
    (** Get the default value. *)
    val get_epsilon : t -> float
    
    (** Check if an instance is a member of the table. *)
    val mem : t -> inst_t -> bool
    
    (** Get the probability of the instance. *)
    val get_prob : t -> inst_t -> float
    
    (** Get the smoothened probability of the instance. *)
    val get_smooth_prob : t -> inst_t -> float
    
    (** Set the probability of the instance. *)
    val set_prob : t -> inst_t -> float -> unit
    
    (** Accumulate the probability of the instance. *)
    val acc_prob : t -> inst_t -> float -> unit

end

(** Compute the inside and outside score tables from the size, 
    the default value, the generative model of derivations, 
    the chart, and the start syntactic category. Please note that
    all scores are calculated based on smoothened probabilities. *)
val compute :
    int -> float -> Deriv.DerivGenModel.t -> Chart.Chart.t -> 
    Cg.Syncat.t -> ScoreTbl.t * ScoreTbl.t

val compute_insc :
    int -> float -> Deriv.DerivGenModel.t -> Chart.Chart.t -> 
    Cg.Syncat.t -> ScoreTbl.t
