(***********************************************************
*                                                          *
*      FUNGI - Fast UNsupervised Grammar Induction         *
*                                                          *
*   Originator: Prachya Boonkwan [1], [2]                  *
*               [1] School of Informatics,                 *
*                   University of Edinburgh, UK            *
*               [2] National Electronics and Computer      *
*                   Technology Center (NECTEC), Thailand   *
*                                                          *
*   Copyright (C) 2011 Prachya Boonkwan                    *
*   <p.boonkwan@sms.ed.ac.uk, kaamanita@hotmail.com>       *
*                                                          *
*   This library is free software; you can redistribute    *
*   it and/or modify it under the terms of the GNU         *
*   Lesser General Public License as published by the      *
*   Free Software Foundation; either version 2.1 of the    *
*   License, or (at your option) any later version.        *
*                                                          *
*   This library is distributed in the hope that it will   *
*   be useful, but WITHOUT ANY WARRANTY; without even      *
*   the implied warranty of MERCHANTABILITY or FITNESS     *
*   FOR A PARTICULAR PURPOSE.  See the GNU Lesser          *
*   General Public License for more details.               *
*                                                          *
*   You should have received a copy of the GNU Lesser      *
*   General Public License along with this library; if     *
*   not, write to the Free Software Foundation, Inc.,      *
*   59 Temple Place, Suite 330, Boston, MA  02111-1307     *
*   USA.                                                   *
*                                                          *
***********************************************************)

open Cg
open Macro
open Dict
open Utils
open Printf

module LongCatGen = struct

    (* Meaning of list option
           1. Some [x1; x2; x3; ...] => Generate syntactic categories 
              with respect to the constraints.
           2. None => Do not generate any syntactic categories.
           3. Some [] => Generate every possible syntactic category.
     *)

    type macroexp_t = MacroExpander.t

    type dict_t = Dict.t

    (* fund_ord_1_01: Word order of Subj+Verb+Obj. 
       S = subject, V = verb, and O = object. *)
    type fund_ord_1_01_t = SOV | SVO | VSO | VOS | OVS | OSV
    type fund_ord_1_01_ans_t = fund_ord_1_01_t list option
    let fund_ord_1_01_all = [ SOV; SVO; VSO; VOS; OVS; OSV ]

    (* fund_ord_1_02: How to distinguish agent from patient in 
       subject-verb-object sentence. *)
    type agtpat_distinction_t = AgtPat_with_infl | AgtPat_with_prefix
                              | AgtPat_with_suffix | AgtPat_with_infix
                              | AgtPat_with_circumfix | AgtPat_nochange
    let agtpat_distinction_all = [ AgtPat_with_infl; AgtPat_with_prefix; AgtPat_with_suffix; AgtPat_with_infix; AgtPat_with_circumfix; AgtPat_nochange ]
    type fund_ord_1_02_ans_t = {
        agt : agtpat_distinction_t;
        pat : agtpat_distinction_t;
    }

    (* fund_ord_2_01: Type of ditransitivity. *)
    type fund_ord_2_01_t = VT_and_PP | DTV
    type fund_ord_2_01_ans_t = fund_ord_2_01_t list option
    let fund_ord_2_01_all = [ VT_and_PP; DTV ]

    (* fund_ord_2_02: Word order of ditransitive verb.
       S = subject, V = verb, O = direct object, 
       and I = indirect object. *)
    type fund_ord_2_02_t = SVOI | SVIO | SIVO | ISVO
                         | SOVI | SOIV | SIOV | ISOV
                         | VSOI | VSIO | VISO | IVSO
                         | VOSI | VOIS | VIOS | IVOS
                         | OVSI | OVIS | OIVS | IOVS
                         | OSVI | OSIV | OISV | IOSV
    type fund_ord_2_02_ans_t = fund_ord_2_02_t list option
    let fund_ord_2_02_all = [ SVOI; SVIO; SIVO; ISVO; SOVI; SOIV; SIOV; ISOV; VSOI; VSIO; VISO; IVSO; VOSI; VOIS; VIOS; IVOS; OVSI; OVIS; OIVS; IOVS; OSVI; OSIV; OISV; IOSV ]

    (* fund_ord_2_03: Word order of transitive verb and adpositional 
       phrase. 
       V = verb, O = object, and P = adpositional phrase. *)
    type fund_ord_2_03_t = SVOP | SVPO | SPVO | PSVO
                         | SOVP | SOPV | SPOV | PSOV
                         | VSOP | VSPO | VPSO | PVSO
                         | VOSP | VOPS | VPOS | PVOS
                         | OVSP | OVPS | OPVS | POVS
                         | OSVP | OSPV | OPSV | POSV
    type fund_ord_2_03_ans_t = fund_ord_2_03_t list option
    let fund_ord_2_03_all = [ SVOP; SVPO; SPVO; PSVO; SOVP; SOPV; SPOV; PSOV; VSOP; VSPO; VPSO; PVSO; VOSP; VOPS; VPOS; PVOS; OVSP; OVPS; OPVS; POVS; OSVP; OSPV; OPSV; POSV ]

    (* fund_ord_2_04: Transformation from TV to DTV. *)
    type fund_ord_2_04_t = Dtv_with_infl | Dtv_without_infl
    type fund_ord_2_04_ans_t = fund_ord_2_04_t list option
    let fund_ord_2_04_all = [ Dtv_with_infl; Dtv_without_infl ]

    (* fund_ord_3_01: Adpositional marker. *)
    type fund_ord_3_01_t = Adpos_word | Adpos_infl
    type fund_ord_3_01_ans_t = fund_ord_3_01_t list option

    (* fund_ord_3_02: Word order of adposition. 
       N = core noun, P = adposition marker, and O = adpositional 
       object. *)
    type fund_ord_3_02_t = NPO | NOP | PON | OPN
    type fund_ord_3_02_ans_t = fund_ord_3_02_t list option

    (* fund_ord_3_03: Word order of inflectional adposition. *)
    type fund_ord_3_03_t = NAP | APN
    type fund_ord_3_03_ans_t = fund_ord_3_03_t list option

    (* fund_ord_3_04: Word order of VP and adposition. 
       S = subject, V = verb, O = object, and P = adpositional 
       phrase. *)
    type fund_ord_3_04_t = fund_ord_2_03_t
    type fund_ord_3_04_ans_t = fund_ord_3_04_t list option

    (* fund_ord_4_01: Possessive marker. *)
    type fund_ord_4_01_t = Poss_word | Poss_infl
    type fund_ord_4_01_ans_t = fund_ord_4_01_t list option

    (* fund_ord_4_02: Word order of possessive word. 
       For "book of girls", N1 = book and N2 = girls. *)
    type fund_ord_4_02_t = N1PossN2 | N1N2Poss | PossN2N1 | N2PossN1
    type fund_ord_4_02_ans_t = fund_ord_4_02_t list option

    (* fund_ord_4_03: Word order of possessive inflection. 
       For "book of girls", N1 = book and Poss = of girls. *)
    type fund_ord_4_03_t = N1Poss | PossN1
    type fund_ord_4_03_ans_t = fund_ord_4_03_t list option

    (* fund_ord_5_01: How adjectives behave in the language? *)
    type fund_ord_5_01_t = Adj_as_nmod | Adj_with_cl | Adj_as_verb
    type fund_ord_5_01_ans_t = fund_ord_5_01_t list option

    (* fund_ord_5_02: Word order of adjective as a noun modifier. *)
    type fund_ord_5_02_t = AdjN | NAdj
    type fund_ord_5_02_ans_t = fund_ord_5_02_t list option

    (* fund_ord_5_03: Word order of adjective with classifier. *)
    type fund_ord_5_03_t = NClAdj | NAdjCl | ClNAdj 
                         | ClAdjN | AdjNCl | AdjClN
    type fund_ord_5_03_ans_t = fund_ord_5_03_t list option

    (* fund_ord_6_01: Demonstrative pronouns. *)
    type fund_ord_6_01_t = Dempro_as_word | Dempro_with_cl 
                         | Dem_infl | DemPro_as_circumfix
    type fund_ord_6_01_ans_t = fund_ord_6_01_t list option

    (* fund_ord_6_02: Word order of demonstrative pronoun. *)
    type fund_ord_6_02_t = DemN | NDem
    type fund_ord_6_02_ans_t = fund_ord_6_02_t list option

    (* fund_ord_6_03: Word order of demonstrative pronoun with classifier. *)
    type fund_ord_6_03_t = DemClN | NDemCl | ClDemN | NClDem
    type fund_ord_6_03_ans_t = fund_ord_6_03_t list option

    (* fund_ord_7_01: Cardinal numbers. *)
    type fund_ord_7_01_t = Num_as_adj | Num_with_cl | Num_mixed
    type fund_ord_7_01_ans_t = fund_ord_7_01_t list option

    (* fund_ord_7_02: Cardinal numbers as adjective. *)
    type fund_ord_7_02_t = NumN | NNum
    type fund_ord_7_02_ans_t = fund_ord_7_02_t list option

    (* fund_ord_7_03: Cardinal numbers with classifier. *)
    type fund_ord_7_03_t = NumClN | NumNCl | NNumCl | NClNum | ClNumN | ClNNum
    type fund_ord_7_03_ans_t = fund_ord_7_03_t list option

    (* fund_ord_8_01: Relative pronouns. *)
    type fund_ord_8_01_t = Relpro_as_word | Relcl_infl | Relpro_notexists
                         | Relcl_as_aggluedunit | Relcl_extraposed | Relpro_by_wordrep
                         | Relpro_with_cl
    type fund_ord_8_01_ans_t = fund_ord_8_01_t list option
    
    (* fund_ord_8_02: Word order of relative pronoun. *)
    type fund_ord_8_02_t = NRelVp | RelVpN | NVpRel | VpRelN
    type fund_ord_8_02_ans_t = fund_ord_8_02_t list option

    (* fund_ord_8_03: Word order of inflected relative clause. *)
    type fund_ord_8_03_t = NRelinfl | RelinflN
    type fund_ord_8_03_ans_t = fund_ord_8_03_t list option

    (* fund_ord_8_04: Word order of agglutinated relative clause. *)
    type fund_ord_8_04_t = AggSVO | AggSOV | AggOSV | AggOVS | AggVSO | AggVOS
    type fund_ord_8_04_ans_t = fund_ord_8_04_t list option

    (* fund_ord_8_05: Relative clause extraposition. *)
    type fund_ord_8_05_t = SentRelc | RelcSent
    type fund_ord_8_05_ans_t = fund_ord_8_05_t list option

    (* fund_ord_8_06: Relative pronoun by repeating the core noun. *)
    type fund_ord_8_06_t = NRelcRepn | NRepnRelc | RelcRepnN | RepnRelcN
    type fund_ord_8_06_ans_t = fund_ord_8_06_t list option

    (* fund_ord_8_07: Relative pronoun with classifier. *)
    type fund_ord_8_07_t = NRelcCl | NClRelc | ClNRelc | ClRelcN | RelcNCl | RelcClN
    type fund_ord_8_07_ans_t = fund_ord_8_07_t list option

    (* fund_ord_9_01: The quantifier of adjective. *)
    type fund_ord_9_01_t = VeryAdj | AdjVery | AdjAdj | Adj_with_infl
    type fund_ord_9_01_ans_t = fund_ord_9_01_t list option

    (* fund_ord_10_01: Tense, aspect, and modality. *)
    type tam_t = TAM_word | TAM_infl | TAM_infl_word | TAM_nochange
    type fund_ord_10_01_ans_t = {
        tam_present_simple   : tam_t;
        tam_present_prog     : tam_t;
        tam_past_simple      : tam_t;
        tam_future_simple    : tam_t;
        tam_present_perfect  : tam_t;
        tam_ability          : tam_t;
        tam_possibility      : tam_t;
        tam_subjunctive      : tam_t;
        tam_past_subjunctive : tam_t;
        tam_necessity        : tam_t;
        tam_past_possibility : tam_t;
    }

    (* fund_ord_10_02: Word order of TAM markers. *)
    type tam_word_order_t = TamSVO | TamSOV | TamVSO | TamVOS | TamOSV | TamOVS
                          | STamVO | STamOV | VTamSO | VTamOS | OTamSV | OTamVS
                          | SVTamO | SOTamV | VSTamO | VOTamS | OSTamV | OVTamS
                          | SVOTam | SOVTam | VSOTam | VOSTam | OSVTam | OVSTam
    type fund_ord_10_02_ans_t = tam_word_order_t list option

    (* simp_1_01: How to express a nominative pronoun. *)
    type simp_1_01_t = Nompro_as_word | Nompro_as_infl 
                     | Nompro_as_affix | Nompro_as_indecl
    type simp_1_01_ans_t = simp_1_01_t list option

    (* simp_1_02: How to express an accusative pronoun. *)
    type simp_1_02_t = Accpro_as_word | Accpro_as_infl
                     | Accpro_as_affix | Accpro_as_indecl
    type simp_1_02_ans_t = simp_1_02_t list option

    (* simp_1_03: The orientation of nominative pronoun affix. *)
    type simp_1_03_t = NomproV | VNompro
    type simp_1_03_ans_t = simp_1_03_t list option

    (* simp_1_04: The orientation of accusative pronoun affix. *)
    type simp_1_04_t = AccproV | VAccpro
    type simp_1_04_ans_t = simp_1_04_t list option

    (* simp_1_05: The orientation of nominative and accusative pronoun affixes. *)
    type simp_1_05_t = NomVAcc | NomAccV | VNomAcc | VAccNom | AccNomV | AccVNom
    type simp_1_05_ans_t = simp_1_05_t list option

    (* simp_1_06: The omission of nominative pronoun. *)
    type simp_1_06_t = Nompro_obligatory | Nompro_optional | Nompro_disappear
    type simp_1_06_ans_t = simp_1_06_t list option

    (* simp_1_07: The omission of accusative pronoun. *)
    type simp_1_07_t = Accpro_obligatory | Accpro_optional | Accpro_disappear
    type simp_1_07_ans_t = simp_1_07_t list option

    (* simp_1_08: The use of pro-forms when nominative and accusative pronouns are both used. *)
    type use_of_pro_t = Pro_obligatory | Pro_optional | Pro_disappear
    type simp_1_08_ans_t = {
        nompro_use : use_of_pro_t;
        accpro_use : use_of_pro_t;
    }

    (* simp_2_01: The use of reflexive pronouns 'oneself'/'oneselves'. *)
    type simp_2_01_t = Reflpro_as_word | Reflpro_as_infl 
                     | Reflpro_as_circumfix | Reflpro_as_perspro
    type simp_2_01_ans_t = simp_2_01_t list option

    (* simp_2_02: Word order of reflexive pronouns. *)
    type simp_2_02_t = ReflSVAdv | ReflSAdvV | ReflVSAdv | ReflVAdvS
                     | ReflAdvSV | ReflAdvVS | SReflVAdv | SReflAdvV
                     | VReflSAdv | VReflAdvS | AdvReflSV | AdvReflVS
                     | SVReflAdv | SAdvReflV | VSReflAdv | VAdvReflS
                     | AdvSReflV | AdvVReflS | SVAdvRefl | SAdvVRefl
                     | VSAdvRefl | VAdvSRefl | AdvSVRefl | AdvVSRefl
    type simp_2_02_ans_t = simp_2_02_t list option

    (* simp_2_03: The use of reciprocal pronoun 'each other'. *)
    type simp_2_03_t = Recipro_as_word | Recipro_as_infl 
                     | Recipro_as_circumfix | Recipro_nonexists
    type simp_2_03_ans_t = simp_2_03_t list option

    (* simp_2_04: Word order of reciprocal pronoun. *)
    type simp_2_04_t = ReciSV | ReciVS | SVReci | SReciV | VSReci | VReciS
    type simp_2_04_ans_t = simp_2_04_t list option

    (* simp_2_05: Is reciprocal pronoun distinguished from reflexive pronoun. *)
    type simp_2_05_t = bool
    type simp_2_05_ans_t = simp_2_05_t list option

    (* simp_2_06: The use of reflexive adverb 'oneself' (in 'he does it himself'). *)
    type simp_2_06_t = Refladv_as_word | Refladv_as_infl 
                     | Refladv_as_circumfix | Refladv_notexists
    type simp_2_06_ans_t = simp_2_06_t list option

    (* simp_2_07: Word order of reflexive adverb. *)
    type simp_2_07_t = RefladvSVO | RefladvSOV | RefladvVSO | RefladvVOS
                     | RefladvOSV | RefladvOVS | SRefladvVO | SRefladvOV
                     | VRefladvSO | VRefladvOS | ORefladvSV | ORefladvVS
                     | SVRefladvO | SORefladvV | VSRefladvO | VORefladvS
                     | OSRefladvV | OVRefladvS | SVORefladv | SOVRefladv
                     | VSORefladv | VOSRefladv | OSVRefladv | OVSRefladv
    type simp_2_07_ans_t = simp_2_07_t list option

    (* simp_3_01: Passive construction. *)
    type simp_3_01_t = Passive_as_infl | Passive_with_marker 
                     | Passive_by_diff_wordorder | Passive_notexists
                     | Passive_with_oblique_marker
    type simp_3_01_ans_t = simp_3_01_t list option

    (* simp_3_02: Word order of passive construction with marker. *)
    type simp_3_02_t = PassSVObl | PassSOblV | PassVSObl | PassVOblS
                     | PassOblSV | PassOblVS | SPassVObl | SPassOblV
                     | VPassSObl | VPassOblS | OblPassSV | OblPassVS
                     | SVPassObl | SOblPassV | VSPassObl | VOblPassS
                     | OblSPassV | OblVPassS | SVOblPass | SOblVPass
                     | VSOblPass | VOblSPass | OblSVPass | OblVSPass
    type simp_3_02_ans_t = simp_3_02_t list option

    (* simp_3_03: Passive inflection. *)
    type simp_3_03_t = Passive_as_prefix | Passive_as_suffix
                     | Passive_as_infix | Passive_as_circumfix
    type simp_3_03_ans_t = simp_3_03_t list option

    (* simp_3_04: Passive without marker: no inflection. *)
    (* simp_3_05: Passive without marker: prefix. *)
    (* simp_3_06: Passive without marker: suffix. *)
    (* simp_3_07: Passive without marker: infix. *)
    (* simp_3_08: Passive without marker: circumfix. *)
    type passive_as_infl_order_t = Passive_SVObl | Passive_SOblV 
                                 | Passive_VSObl | Passive_VOblS 
                                 | Passive_OblSV | Passive_OblVS
    type simp_3_04_ans_t = passive_as_infl_order_t list option
    type simp_3_05_ans_t = passive_as_infl_order_t list option
    type simp_3_06_ans_t = passive_as_infl_order_t list option
    type simp_3_07_ans_t = passive_as_infl_order_t list option
    type simp_3_08_ans_t = passive_as_infl_order_t list option

    (* simp_3_09: Optionality of oblique in passive construction. *)
    type simp_3_09_t = bool
    type simp_3_09_ans_t = simp_3_09_t option

    (* simp_3_10: How to distinguish agent from patient? *)
    type simp_3_10_t = Passive_AgtPat_infl | Passive_AgtPat_marker
                     | Passive_AgtPat_infl_and_marker | Passive_AgtPat_nochange
    type simp_3_10_ans_t = {
        agt_marker : simp_3_10_t;
        pat_marker : simp_3_10_t;
    }

    (* simp_3_11: Patient marker. *)
    (* simp_3_12: Agent marker. *)
    type agtpat_marker_order_t = AgtPat_prefix | AgtPat_suffix 
                               | AgtPat_infix | AgtPat_circumfix
    type simp_3_11_ans_t = agtpat_marker_order_t list option
    type simp_3_12_ans_t = agtpat_marker_order_t list option

    (* simp_4_01: Infinitive in object control. *)
    type simp_4_01_t = ObjCtrl_infmarker_obligatory | ObjCtrl_infmarker_optional
                     | ObjCtrl_verbinfl | ObjCtrl_no_infmarker
    type simp_4_01_ans_t = simp_4_01_t list option

    (* simp_4_02: Word order of the infinitive. *)
    type simp_4_02_t = ObjCtrl_InfVO | ObjCtrl_InfOV 
                     | ObjCtrl_VOInf | ObjCtrl_VInfO
                     | ObjCtrl_OInfV | ObjCtrl_OVInf
    type simp_4_02_ans_t = simp_4_02_t list option

    (* simp_4_03: Word order of object control structure. *)
    type simp_4_03_t = SV1O1V2O2 | SV1V2O2O1 | SV1V2O1O2
                     | SO1V1O2V2 | SO1O2V2V1 | SO1O2V1V2
                     | V1SO1V2O2 | V1SV2O2O1 | V1V2O2SO1
                     | V1SV2O1O2 | V1V2SO1O2 | O1SV1O2V2
                     | O1O2V2SV1 | O1O2SV2V1 | O2O1V2SV1
                     | O2O1SV2V1 | O1V1O2V2S | O1O2V2V1S
                     | O1O2V1V2S | V1O1V2O2S | V1V2O2O1S
                     | V1V2O1O2S
    type simp_4_03_ans_t = simp_4_03_t list option

    (* simp_5_01: Negation of intransitive verb. *)
    type simp_5_01_t = NegVi_with_auxv | NegVi_with_negmarker 
                     | NegVi_with_infl | NegVi_by_possession
    type simp_5_01_ans_t = simp_5_01_t list option

    (* simp_5_02: Negation with auxiliary verb. *)
    type simp_5_02_t = AuxSNegV | AuxSVNeg | AuxNegSV | AuxNegVS
                     | AuxVSNeg | AuxVNegS | SAuxNegV | SAuxVNeg
                     | NegAuxSV | NegAuxVS | VAuxSNeg | VAuxNegS
                     | SNegAuxV | SVAuxNeg | NegSAuxV | NegVAuxS
                     | VSAuxNeg | VNegAuxS | SNegVAux | SVNegAux
                     | NegSVAux | NegVSAux | VSNegAux | VNegSAux
    type simp_5_02_ans_t = simp_5_02_t list option

    (* simp_5_03_t: Negation without auxiliary verb. *)
    type simp_5_03_t = SVNeg | SNegV | NegSV | NegVS | VSNeg | VNegS
    type simp_5_03_ans_t = simp_5_03_t list option

    (* simp_5_04: Negation of transitive verb. *)
    type simp_5_04_t = NegVt_with_auxv | NegVt_with_negmarker 
                     | NegVt_with_infl | NegVt_by_possession
    type simp_5_04_ans_t = simp_5_04_t list option

    (* simp_5_05: Negation with auxiliary verb. *)
    type simp_5_05_t = AuxSNegVO | AuxSVNegO | AuxNegSVO | AuxNegVSO
                     | AuxVSNegO | AuxVNegSO | SAuxNegVO | SAuxVNegO
                     | NegAuxSVO | NegAuxVSO | VAuxSNegO | VAuxNegSO
                     | SNegAuxVO | SVAuxNegO | NegSAuxVO | NegVAuxSO
                     | VSAuxNegO | VNegAuxSO | SNegVAuxO | SVNegAuxO
                     | NegSVAuxO | NegVSAuxO | VSNegAuxO | VNegSAuxO
                     | AuxSNegOV | AuxSVONeg | AuxNegSOV | AuxNegVOS
                     | AuxVSONeg | AuxVNegOS | SAuxNegOV | SAuxVONeg
                     | NegAuxSOV | NegAuxVOS | VAuxSONeg | VAuxNegOS
                     | SNegAuxOV | SVAuxONeg | NegSAuxOV | NegVAuxOS
                     | VSAuxONeg | VNegAuxOS | SNegVOAux | SVNegOAux
                     | NegSVOAux | NegVSOAux | VSNegOAux | VNegSOAux
                     | AuxSONegV | AuxSOVNeg | AuxNegOSV | AuxNegOVS
                     | AuxVOSNeg | AuxVONegS | SAuxONegV | SAuxOVNeg
                     | NegAuxOSV | NegAuxOVS | VAuxOSNeg | VAuxONegS
                     | SNegOAuxV | SVOAuxNeg | NegSOAuxV | NegVOAuxS
                     | VSOAuxNeg | VNegOAuxS | SNegOVAux | SVONegAux
                     | NegSOVAux | NegVOSAux | VSONegAux | VNegOSAux
                     | AuxOSNegV | AuxOSVNeg | AuxONegSV | AuxONegVS
                     | AuxOVSNeg | AuxOVNegS | SOAuxNegV | SOAuxVNeg
                     | NegOAuxSV | NegOAuxVS | VOAuxSNeg | VOAuxNegS
                     | SONegAuxV | SOVAuxNeg | NegOSAuxV | NegOVAuxS
                     | VOSAuxNeg | VONegAuxS | SONegVAux | SOVNegAux
                     | NegOSVAux | NegOVSAux | VOSNegAux | VONegSAux
                     | OAuxSNegV | OAuxSVNeg | OAuxNegSV | OAuxNegVS
                     | OAuxVSNeg | OAuxVNegS | OSAuxNegV | OSAuxVNeg
                     | ONegAuxSV | ONegAuxVS | OVAuxSNeg | OVAuxNegS
                     | OSNegAuxV | OSVAuxNeg | ONegSAuxV | ONegVAuxS
                     | OVSAuxNeg | OVNegAuxS | OSNegVAux | OSVNegAux
                     | ONegSVAux | ONegVSAux | OVSNegAux | OVNegSAux
    type simp_5_05_ans_t = simp_5_05_t list option

    (* simp_6_01: How to describe possession. *)
    type simp_6_01_t = Poss_as_v | Poss_as_of | Poss_as_pp 
                     | Poss_as_beneficative | Poss_as_bene_and_exist
                     | Poss_as_coactor | Poss_as_adj
    type simp_6_01_ans_t = simp_6_01_t list option

    (* simp_6_02: How to describe existence. *)
    type simp_6_02_t = Existence_as_exist | Existence_as_have 
                     | Existence_as_be
    type simp_6_02_ans_t = simp_6_02_t list option

    (* simp_6_03: The usage of copula and adjectives. *)
    type simp_7_01_t = AdjSent_with_obligatory_copula
                     | AdjSent_with_optional_copula
                     | AdjSent_as_v
                     | AdjSent_because_no_copula
    type simp_7_01_ans_t = simp_7_01_t list option

    (* simp_7_02: The usage of copula, adjectives, and negation. *)
    type simp_7_02_t = NegAdjSent_with_obligatory_copula
                     | NegAdjSent_with_optional_copula
                     | NegAdjSent_as_v
                     | NegAdjSent_because_no_copula
                     | NegAdjSent_as_v_with_obligatory_aux
    type simp_7_02_ans_t = simp_7_02_t list option

    (* simp_7_03: The usage of copula and nouns. *)
    type simp_7_03_t = NSent_with_obligatory_copula
                     | NSent_with_optional_copula
                     | NSent_as_v
                     | NSent_because_no_copula
    type simp_7_03_ans_t = simp_7_03_t list option

    (* simp_7_04: The usage of copula and prepositional phrase. *)
    type simp_7_04_t = PpSent_with_obligatory_copula
                     | PpSent_with_optional_copula
                     | PpSent_as_v
                     | PpSent_because_no_copula
    type simp_7_04_ans_t = simp_7_04_t list option

    (* simp_8_01: Comparative adjective. *)
    type simp_8_01_t = CompAdj_as_infl | CompAdj_with_more 
                     | CompAdj_with_than | CompAdj_with_oblique
                     | CompAdj_by_contrastive_utterance
    type simp_8_01_ans_t = simp_8_01_t list option

    (* simp_8_02: Word order of comparative adjective with than. *)
    (* simp_8_03: Word order of comparative adjective with than and prepositional oblique. *)
    type compadj_order_t = OblThan | ThanObl
    type simp_8_02_ans_t = compadj_order_t list option
    type simp_8_03_ans_t = compadj_order_t list option

    (* simp_9_01: The usage of subject control verbs. *)
    type simp_9_01_t = SubjCtrl_with_infmarker | SubjCtrl_subordcls_with_marker
                     | SubjCtrl_serialv | SubjCtrl_subordcls_without_marker
    type simp_9_01_ans_t = simp_9_01_t list option

    (* simp_9_02: Word order of subject control verb. *)
    type simp_9_02_t = SV1V2O | SV1OV2 | SOV2V1 | SOV1V2
                     | V1SV2O | V1V2OS | V1V2SO | OV2SV1
                     | OSV2V1 | V1OV2S | OV2V1S | OV1V2S
    type simp_9_02_ans_t = simp_9_02_t list option

    (* simp_9_03: The usage of object control verbs. *)
    type simp_9_03_t = ObjCtrl_with_infmarker | ObjCtrl_subordcls_with_marker
                     | ObjCtrl_subordcls_without_marker
    type simp_9_03_ans_t = simp_9_03_t list option

    (* simp_9_04: Word order of object control verb. *)
    type simp_9_04_t = simp_4_03_t
    type simp_9_04_ans_t = simp_9_04_t list option

    (* simp_10_01: How to form a serial verb construction. *)
    type simp_10_01_t = SerialV_with_connective | SerialV_with_infl
                      | SerialV_with_repeating_pron | SerialV_by_simple_concat
    type simp_10_01_ans_t = simp_10_01_t list option

    (* simp_10_02: Word order of serial verb construction. *)
    type simp_10_02_t = simp_4_03_t
    type simp_10_02_ans_t = simp_10_02_t list option

    (* simp_10_03: Word order of the infinitive in serial verb construction. *)
    type simp_10_03_t = SerialV_InfVO | SerialV_InfOV 
                      | SerialV_VOInf | SerialV_VInfO
                      | SerialV_OInfV | SerialV_OVInf
    type simp_10_03_ans_t = simp_10_03_t list option

    (* simp_10_04: The usage of subordinate clause. *)
    type simp_10_04_t = SubordCls_with_subordconj | SubordCls_with_prep
                      | SubordCls_with_simple_concat | SubordCls_with_infl
    type simp_10_04_ans_t = simp_10_04_t list option

    (* simp_10_05: Word order of subordinate clause. *)
    type simp_10_05_t = SubordCls_ConnSubordSent | SubordCls_ConnSentSubord
                      | SubordCls_SubordConnSent | SubordCls_SubordSentConn
                      | SubordCls_SentConnSubord | SubordCls_SentSubordConn
    type simp_10_05_ans_t = simp_10_05_t list option

    (* simp_10_06: The usage of resultative clause. *)
    type simp_10_06_t = ResultCls_with_subordconj | ResultCls_with_prep
                      | ResultCls_with_simple_concat | ResultCls_with_infl
    type simp_10_06_ans_t = simp_10_06_t list option

    (* simp_10_07: Word order of resultative clause. *)
    type simp_10_07_t = ResultCls_ConnResultSent | ResultCls_ConnSentResult
                      | ResultCls_ResultConnSent | ResultCls_ResultSentConn
                      | ResultCls_SentConnResult | ResultCls_SentResultConn
    type simp_10_07_ans_t = simp_10_07_t list option

    (* simp_10_08: The usage of reported speech. *)
    type simp_10_08_t = RepSpeech_with_subordconj | RepSpeech_with_prep
                      | RepSpeech_with_simple_concat | RepSpeech_with_infl
    type simp_10_08_ans_t = simp_10_08_t list option

    (* simp_10_09: Word order of reported speech with marker. *)
    type simp_10_09_t = RepSpeech_SubordSpeechSent | RepSpeech_SubordSentSpeech
                      | RepSpeech_SpeechSubordSent | RepSpeech_SpeechSentSubord
                      | RepSpeech_SentSubordSpeech | RepSpeech_SentSpeechSubord
    type simp_10_09_ans_t = simp_10_09_t list option

    (* simp_10_10: Word order of reported speech without marker. *)
    type simp_10_10_t = RepSpeech_SVSpeech | RepSpeech_SSpeechV | RepSpeech_SpeechSV
                      | RepSpeech_VSSpeech | RepSpeech_VSpeechS | RepSpeech_SpeechVS
    type simp_10_10_ans_t = simp_10_10_t list option

    (* ques_1_01: The usage of polar question. *)
    type ques_1_01_t = PolarQ_add_marker | PolarQ_drop_affirm_marker 
                     | PolarQ_by_infl | PolarQ_change_pitch 
                     | PolarQ_change_order_and_add_marker
                     | PolarQ_change_order | PolarQ_do_or_not 
                     | PolarQ_nochange
    type ques_1_01_ans_t = ques_1_01_t list option

    (* ques_1_02: Word order of polar question with marker. *)
    type ques_1_02_t = QSVAdv | QSAdvV | QVSAdv | QVAdvS
                     | SQVAdv | SQAdvV | VQSAdv | VQAdvS
                     | SVAdvQ | SAdvVQ | VSAdvQ | VAdvSQ
    type ques_1_02_ans_t = ques_1_02_t list option

    (* ques_1_03: Word order of polar question with marker -- for the case of Ancient Greek. *)
    type ques_1_03_t = PolarQ_normal 
                     | PolarQ_marker_always_follow_first_word
    type ques_1_03_ans_t = ques_1_03_t list option

    (* ques_1_04: Position of affirmative marker. *)
    type ques_1_04_t = PolarQ_drop_affirm_infront | PolarQ_drop_affirm_before_v
                     | PolarQ_drop_affirm_after_v | PolarQ_drop_affirm_at_end
    type ques_1_04_ans_t = ques_1_04_t list option

    (* ques_1_05: Word order of polar question with marker. *)
    type ques_1_05_t = ques_1_02_t
    type ques_1_05_ans_t = ques_1_05_t list option

    (* ques_1_06: Word order of polar question without marker. *)
    type ques_1_06_t = PolarQ_SVAdv | PolarQ_SAdvV 
                     | PolarQ_VSAdv | PolarQ_VAdvS 
                     | PolarQ_AdvSV | PolarQ_AdvVS
    type ques_1_06_ans_t = ques_1_06_t list option

    (* ques_2_01: How to express Wh-question: what. *)
    (* ques_2_02: How to express Wh-question: who. *)
    (* ques_2_03: How to express Wh-question: how many. *)
    type wh_question_order_t = WHQ_SVO | WHQ_SOV | WHQ_VSO 
                             | WHQ_VOS | WHQ_OVS | WHQ_OSV
    type wh_question_ans_t = {
        wh_infront      : bool;
        same_word_order : bool;
        use_qmarker     : bool;
        drop_affirm     : bool;
        changed_orders  : wh_question_order_t list option;
    }
    type ques_2_01 = wh_question_ans_t
    type ques_2_02 = wh_question_ans_t
    type ques_2_03 = wh_question_ans_t

    (* ques_2_04: Word order of Wh-question with auxiliary verb and question marker. *)
    type ques_2_04_t = WHQ_QSVO | WHQ_QSOV | WHQ_QVSO | WHQ_QVOS
                     | WHQ_QOSV | WHQ_QOVS | WHQ_SQVO | WHQ_SQOV
                     | WHQ_VQSO | WHQ_VQOS | WHQ_OQSV | WHQ_OQVS
                     | WHQ_SVQO | WHQ_SOQV | WHQ_VSQO | WHQ_VOQS
                     | WHQ_OSQV | WHQ_OVQS | WHQ_SVOQ | WHQ_SOVQ
                     | WHQ_VSOQ | WHQ_VOSQ | WHQ_OSVQ | WHQ_OVSQ
    type ques_2_04_ans_t = ques_2_04_t list option

    (************************************************************)

    let s = Syncat.create_atom "s"

    let np = Syncat.create_atom "np"
    
    let pp = Syncat.create_atom "pp"

    let conj = Syncat.conn_cat

    let num = Syncat.create_atom "num"

    let cl = Syncat.create_atom "cl"

    let vp = Syncat.create_atom "vp"

    let sbar = Syncat.create_atom "sbar"

    let neg = Syncat.create_atom "neg"

    let vi fund_ord_1_01 omit_subj =
        if not omit_subj then
            match fund_ord_1_01 with
                SOV | SVO | OSV -> 
                    Syncat.create_bwd s np Syncat.Right
              | VSO | VOS | OVS -> 
                    Syncat.create_fwd s np Syncat.Left
        else s

    let vt fund_ord_1_01 omit_subj omit_obj =
        match fund_ord_1_01 with
            SOV -> (* right branching preferred *) 
                let vi' = vi fund_ord_1_01 omit_subj in
                if not omit_obj then Syncat.create_bwd vi' np Syncat.Right
                else vi'
          | SVO -> (* right branching preferred *) 
                let vi' = vi fund_ord_1_01 omit_subj in
                if not omit_obj then Syncat.create_fwd vi' np Syncat.Left
                else vi'
          | VSO -> (* left branching preferred *) 
                let vo = 
                    if not omit_obj then Syncat.create_fwd s np Syncat.Left 
                    else s in
                if not omit_subj then Syncat.create_fwd vo np Syncat.Left
                else vo
          | VOS -> (* left branching preferred *)
                let vi' = vi fund_ord_1_01 omit_subj in
                if not omit_obj then Syncat.create_fwd vi' np Syncat.Left
                else vi'
          | OVS -> (* left branching preferred *) 
                let vi' = vi fund_ord_1_01 omit_subj in
                if not omit_obj then Syncat.create_bwd vi' np Syncat.Right
                else vi'
          | OSV -> (* right branching preferred *) 
                let ov =
                    if not omit_obj then Syncat.create_bwd s np Syncat.Right
                    else s in
                if not omit_subj then Syncat.create_bwd ov np Syncat.Right
                else ov

    let vd fund_ord_2_02 omit_subj omit_obj omit_iobj =
        match fund_ord_2_02 with
            SVOI -> (* s \> np /< np /< np: right branching preferred *)
                let s_voi = if not omit_subj then Syncat.create_bwd s np Syncat.Right
                            else s in
                let vo_i = if not omit_iobj then Syncat.create_fwd s_voi np Syncat.Left
                           else s_voi in
                let v_o = if not omit_obj then Syncat.create_fwd vo_i np Syncat.Left
                          else vo_i in
                v_o
          | SVIO -> (* s \> np /< np /< np: right branching preferred *)
                let s_vio = if not omit_subj then Syncat.create_bwd s np Syncat.Right
                            else s in
                let vi_o = if not omit_obj then Syncat.create_fwd s_vio np Syncat.Left
                           else s_vio in
                let v_i = if not omit_iobj then Syncat.create_fwd vi_o np Syncat.Left
                          else vi_o in
                v_i
          | SIVO -> (* s \> np \> np /< np: right branching preferred *)
                let s_ivo = if not omit_subj then Syncat.create_bwd s np Syncat.Right
                            else s in
                let i_vo = if not omit_iobj then Syncat.create_bwd s_ivo np Syncat.Right
                           else s_ivo in
                let v_o = if omit_obj then Syncat.create_fwd i_vo np Syncat.Left
                          else i_vo in
                v_o
          | ISVO -> (* s \> np \> np /< np: right branching preferred *)
                let i_svo = if not omit_iobj then Syncat.create_bwd s np Syncat.Right
                            else s in
                let s_vo = if not omit_subj then Syncat.create_bwd i_svo np Syncat.Right
                           else i_svo in
                let v_o = if not omit_obj then Syncat.create_fwd s_vo np Syncat.Left
                          else s_vo in
                v_o
          | SOVI -> (* s \> np \> np /< np: right branching preferred *)
                let s_ovi = if not omit_subj then Syncat.create_bwd s np Syncat.Right
                            else s in
                let o_vi = if not omit_obj then Syncat.create_bwd s_ovi np Syncat.Right
                           else s_ovi in
                let v_i = if not omit_iobj then Syncat.create_fwd o_vi np Syncat.Left
                          else o_vi in
                v_i
          | SOIV -> (* s \> np \> np \> np: right branching preferred *)
                let s_oiv = if not omit_subj then Syncat.create_bwd s np Syncat.Right
                            else s in
                let o_iv = if not omit_obj then Syncat.create_bwd s_oiv np Syncat.Right
                           else s_oiv in
                let i_v = if not omit_iobj then Syncat.create_bwd o_iv np Syncat.Right
                          else o_iv in
                i_v
          | SIOV -> (* s \> np \> np \> np: right branching preferred *)
                let s_iov = if not omit_subj then Syncat.create_bwd s np Syncat.Right
                           else s in
                let i_ov = if not omit_iobj then Syncat.create_bwd s_iov np Syncat.Right
                           else s_iov in
                let o_v = if not omit_obj then Syncat.create_bwd i_ov np Syncat.Right
                          else i_ov in
                o_v
          | ISOV -> (* s \> np \> np \> np: right branching preferred *)
                let i_sov = if not omit_iobj then Syncat.create_bwd s np Syncat.Right
                            else s in
                let s_ov = if not omit_subj then Syncat.create_bwd i_sov np Syncat.Right
                           else i_sov in
                let o_v = if not omit_iobj then Syncat.create_bwd s_ov np Syncat.Right
                          else s_ov in
                o_v
          | VSOI -> (* s /< np /< np /< np: left branching preferred *)
                let vso_i = if not omit_iobj then Syncat.create_fwd s np Syncat.Left
                            else s in
                let vs_o = if not omit_obj then Syncat.create_fwd vso_i np Syncat.Left
                           else vso_i in
                let v_s = if not omit_iobj then Syncat.create_fwd vs_o np Syncat.Left
                          else vs_o in
                v_s
          | VSIO -> (* s /< np /< np /< np: left branching preferred *)
                let vsi_o = if not omit_obj then Syncat.create_fwd s np Syncat.Left
                            else s in
                let vs_i = if not omit_iobj then Syncat.create_fwd vsi_o np Syncat.Left
                           else vsi_o in
                let v_s = if not omit_obj then Syncat.create_fwd vs_i np Syncat.Left
                          else vs_i in
                v_s
          | VISO -> (* s /< np /< np /< np: left branching preferred *)
                let vis_o = if not omit_obj then Syncat.create_fwd s np Syncat.Left
                            else s in
                let vi_s = if not omit_subj then Syncat.create_fwd vis_o np Syncat.Left
                          else vis_o in
                let v_i = if not omit_iobj then Syncat.create_fwd vi_s np Syncat.Left
                          else vi_s in
                v_i
          | IVSO -> (* s /< np /< np \> np: left branching preferred *)
                let ivs_o = if not omit_iobj then Syncat.create_fwd s np Syncat.Left
                            else s in
                let iv_s = if not omit_obj then Syncat.create_fwd ivs_o np Syncat.Left
                           else ivs_o in
                let i_v = if not omit_subj then Syncat.create_bwd iv_s np Syncat.Right
                          else iv_s in
                i_v
          | VOSI -> (* s /< np /< np /< np: left branching preferred *)
                let vos_i = if not omit_iobj then Syncat.create_fwd s np Syncat.Left
                            else s in
                let vo_s = if not omit_subj then Syncat.create_fwd vos_i np Syncat.Left
                           else vos_i in
                let v_o = if not omit_obj then Syncat.create_fwd vo_s np Syncat.Left
                          else vo_s in
                v_o
          | VOIS -> (* s /< np /< np /< np: left branching preferred *)
                let voi_s = if not omit_subj then Syncat.create_fwd s np Syncat.Left
                            else s in
                let vo_i = if not omit_iobj then Syncat.create_fwd voi_s np Syncat.Left
                           else voi_s in
                let v_o = if not omit_obj then Syncat.create_fwd vo_i np Syncat.Left
                          else vo_i in
                v_o
          | VIOS -> (* s /< np /< np /< np: left branching preferred *)
                let vio_s = if not omit_subj then Syncat.create_fwd s np Syncat.Left
                            else s in
                let vi_o = if not omit_obj then Syncat.create_fwd vio_s np Syncat.Left
                           else vio_s in
                let v_i = if not omit_iobj then Syncat.create_fwd vi_o np Syncat.Left
                          else vi_o in
                v_i
          | IVOS -> (* s /< np /< np \> np: left branching preferred *)
                let ivo_s = if not omit_subj then Syncat.create_fwd s np Syncat.Left
                            else s in
                let iv_o = if not omit_obj then Syncat.create_fwd ivo_s np Syncat.Left
                           else ivo_s in
                let i_v = if not omit_iobj then Syncat.create_bwd iv_o np Syncat.Right
                          else iv_o in
                i_v
          | OVSI -> (* s /> np /> np \> np: left branching preferred *)
                let ovs_i = if not omit_iobj then Syncat.create_fwd s np Syncat.Left
                            else s in
                let ov_s = if not omit_subj then Syncat.create_fwd ovs_i np Syncat.Left
                           else ovs_i in
                let o_v = if not omit_obj then Syncat.create_bwd ov_s np Syncat.Right
                          else ov_s in
                o_v
          | OVIS -> (* s /> np /> np \> np: left branching preferred *)
                let ovi_s = if not omit_subj then Syncat.create_fwd s np Syncat.Left
                            else s in
                let ov_i = if not omit_iobj then Syncat.create_fwd ovi_s np Syncat.Left
                           else ovi_s in
                let o_v = if not omit_obj then Syncat.create_bwd ov_i np Syncat.Right
                          else ov_i in
                o_v
          | OIVS -> (* s /< np \> np \> np: left branching preferred *)
                let oiv_s = if not omit_subj then Syncat.create_fwd s np Syncat.Left
                            else s in
                let o_iv = if not omit_obj then Syncat.create_bwd oiv_s np Syncat.Right
                           else oiv_s in
                let i_v = if not omit_iobj then Syncat.create_bwd o_iv np Syncat.Right
                          else o_iv in
                i_v
          | IOVS -> (* s /< np \> np \> np: left branching preferred *)
                let iov_s = if not omit_subj then Syncat.create_fwd s np Syncat.Left
                            else s in
                let i_ov = if not omit_iobj then Syncat.create_bwd iov_s np Syncat.Right
                           else iov_s in
                let o_v = if not omit_obj then Syncat.create_bwd i_ov np Syncat.Right
                          else i_ov in
                o_v
          | OSVI -> (* s \> np \> np /< np: right branching preferred *)
                let osv_i = if not omit_iobj then Syncat.create_bwd s np Syncat.Right
                            else s in
                let o_sv = if not omit_obj then Syncat.create_bwd osv_i np Syncat.Right
                           else osv_i in
                let s_v = if not omit_subj then Syncat.create_fwd o_sv np Syncat.Left
                          else o_sv in
                s_v
          | OSIV -> (* s \> np \> np \> np: right branching preferred *)
                let o_siv = if not omit_obj then Syncat.create_bwd s np Syncat.Right
                            else s in
                let s_iv = if not omit_subj then Syncat.create_bwd o_siv np Syncat.Right
                           else o_siv in
                let i_v = if not omit_iobj then Syncat.create_bwd s_iv np Syncat.Right
                          else s_iv in
                i_v
          | OISV -> (* s \> np \> np \> np: right branching preferred *)
                let o_isv = if not omit_obj then Syncat.create_bwd s np Syncat.Right
                            else s in
                let i_sv = if not omit_iobj then Syncat.create_bwd o_isv np Syncat.Right
                           else o_isv in
                let s_v = if not omit_subj then Syncat.create_bwd i_sv np Syncat.Right
                          else i_sv in
                s_v
          | IOSV -> (* s \> np \> np \> np: right branching preferred *)
                let i_osv = if not omit_iobj then Syncat.create_bwd s np Syncat.Right
                            else s in
                let o_sv = if not omit_obj then Syncat.create_bwd i_osv np Syncat.Right
                           else i_osv in
                let s_v = if not omit_subj then Syncat.create_bwd o_sv np Syncat.Right
                          else o_sv in
                s_v

    let vd_as_vtpp fund_ord_2_03 omit_subj omit_obj omit_pp =
        match fund_ord_2_03 with
            SVOP -> (* s \> np /< pp /< np: right branching preferred *)
                let s_vop = if not omit_subj then Syncat.create_bwd s np Syncat.Right
                            else s in
                let vo_p = if not omit_pp then Syncat.create_fwd s_vop pp Syncat.Left
                           else s_vop in
                let v_o = if not omit_obj then Syncat.create_fwd vo_p np Syncat.Left
                          else vo_p in
                v_o
          | SVPO -> (* s \> np /< np /< pp: right branching preferred *)
                let s_vpo = if not omit_subj then Syncat.create_bwd s np Syncat.Right
                            else s in
                let vp_o = if not omit_obj then Syncat.create_fwd s_vpo np Syncat.Left
                           else s_vpo in
                let v_p = if not omit_pp then Syncat.create_fwd vp_o pp Syncat.Left
                          else vp_o in
                v_p
          | SPVO -> (* s \> np \> pp /< np: right branching preferred *)
                let s_pvo = if not omit_subj then Syncat.create_bwd s np Syncat.Right
                            else s in
                let p_vo = if not omit_pp then Syncat.create_bwd s_pvo pp Syncat.Right
                           else s_pvo in
                let v_o = if omit_obj then Syncat.create_fwd p_vo np Syncat.Left
                          else p_vo in
                v_o
          | PSVO -> (* s \> pp \> np /< np: right branching preferred *)
                let p_svo = if not omit_pp then Syncat.create_bwd s pp Syncat.Right
                            else s in
                let s_vo = if not omit_subj then Syncat.create_bwd p_svo np Syncat.Right
                           else p_svo in
                let v_o = if not omit_obj then Syncat.create_fwd s_vo np Syncat.Left
                          else s_vo in
                v_o
          | SOVP -> (* s \> np \> np /< pp: right branching preferred *)
                let s_ovp = if not omit_subj then Syncat.create_bwd s np Syncat.Right
                            else s in
                let o_vp = if not omit_obj then Syncat.create_bwd s_ovp np Syncat.Right
                           else s_ovp in
                let v_p = if not omit_pp then Syncat.create_fwd o_vp pp Syncat.Left
                          else o_vp in
                v_p
          | SOPV -> (* s \> np \> np \> pp: right branching preferred *)
                let s_opv = if not omit_subj then Syncat.create_bwd s np Syncat.Right
                            else s in
                let o_pv = if not omit_obj then Syncat.create_bwd s_opv np Syncat.Right
                           else s_opv in
                let p_v = if not omit_pp then Syncat.create_bwd o_pv pp Syncat.Right
                          else o_pv in
                p_v
          | SPOV -> (* s \> np \> pp \> np: right branching preferred *)
                let s_pov = if not omit_subj then Syncat.create_bwd s np Syncat.Right
                           else s in
                let p_ov = if not omit_pp then Syncat.create_bwd s_pov pp Syncat.Right
                           else s_pov in
                let o_v = if not omit_obj then Syncat.create_bwd p_ov np Syncat.Right
                          else p_ov in
                o_v
          | PSOV -> (* s \> pp \> np \> np: right branching preferred *)
                let p_sov = if not omit_pp then Syncat.create_bwd s pp Syncat.Right
                            else s in
                let s_ov = if not omit_subj then Syncat.create_bwd p_sov np Syncat.Right
                           else p_sov in
                let o_v = if not omit_pp then Syncat.create_bwd s_ov np Syncat.Right
                          else s_ov in
                o_v
          | VSOP -> (* s /< pp /< np /< np: left branching preferred *)
                let vso_p = if not omit_pp then Syncat.create_fwd s pp Syncat.Left
                            else s in
                let vs_o = if not omit_obj then Syncat.create_fwd vso_p np Syncat.Left
                           else vso_p in
                let v_s = if not omit_pp then Syncat.create_fwd vs_o np Syncat.Left
                          else vs_o in
                v_s
          | VSPO -> (* s /< np /< pp /< np: left branching preferred *)
                let vsp_o = if not omit_obj then Syncat.create_fwd s np Syncat.Left
                            else s in
                let vs_p = if not omit_pp then Syncat.create_fwd vsp_o pp Syncat.Left
                           else vsp_o in
                let v_s = if not omit_obj then Syncat.create_fwd vs_p np Syncat.Left
                          else vs_p in
                v_s
          | VPSO -> (* s /< np /< np /< pp: left branching preferred *)
                let vps_o = if not omit_obj then Syncat.create_fwd s np Syncat.Left
                            else s in
                let vp_s = if not omit_subj then Syncat.create_fwd vps_o np Syncat.Left
                          else vps_o in
                let v_p = if not omit_pp then Syncat.create_fwd vp_s pp Syncat.Left
                          else vp_s in
                v_p
          | PVSO -> (* s /< np /< np \> pp: left branching preferred *)
                let pvs_o = if not omit_pp then Syncat.create_fwd s pp Syncat.Left
                            else s in
                let pv_s = if not omit_obj then Syncat.create_fwd pvs_o np Syncat.Left
                           else pvs_o in
                let p_v = if not omit_subj then Syncat.create_bwd pv_s pp Syncat.Right
                          else pv_s in
                p_v
          | VOSP -> (* s /< pp /< np /< np: left branching preferred *)
                let vos_p = if not omit_pp then Syncat.create_fwd s pp Syncat.Left
                            else s in
                let vo_s = if not omit_subj then Syncat.create_fwd vos_p np Syncat.Left
                           else vos_p in
                let v_o = if not omit_obj then Syncat.create_fwd vo_s np Syncat.Left
                          else vo_s in
                v_o
          | VOPS -> (* s /< np /< pp /< np: left branching preferred *)
                let vop_s = if not omit_subj then Syncat.create_fwd s np Syncat.Left
                            else s in
                let vo_p = if not omit_pp then Syncat.create_fwd vop_s pp Syncat.Left
                           else vop_s in
                let v_o = if not omit_obj then Syncat.create_fwd vo_p np Syncat.Left
                          else vo_p in
                v_o
          | VPOS -> (* s /< np /< np /< pp: left branching preferred *)
                let vpo_s = if not omit_subj then Syncat.create_fwd s np Syncat.Left
                            else s in
                let vp_o = if not omit_obj then Syncat.create_fwd vpo_s np Syncat.Left
                           else vpo_s in
                let v_p = if not omit_pp then Syncat.create_fwd vp_o pp Syncat.Left
                          else vp_o in
                v_p
          | PVOS -> (* s /< np /< np \> pp: left branching preferred *)
                let pvo_s = if not omit_subj then Syncat.create_fwd s np Syncat.Left
                            else s in
                let pv_o = if not omit_obj then Syncat.create_fwd pvo_s np Syncat.Left
                           else pvo_s in
                let p_v = if not omit_pp then Syncat.create_bwd pv_o pp Syncat.Right
                          else pv_o in
                p_v
          | OVSP -> (* s /> pp /> np \> np: left branching preferred *)
                let ovs_p = if not omit_pp then Syncat.create_fwd s pp Syncat.Left
                            else s in
                let ov_s = if not omit_subj then Syncat.create_fwd ovs_p np Syncat.Left
                           else ovs_p in
                let o_v = if not omit_obj then Syncat.create_bwd ov_s np Syncat.Right
                          else ov_s in
                o_v
          | OVPS -> (* s /> np /> pp \> np: left branching preferred *)
                let ovp_s = if not omit_subj then Syncat.create_fwd s np Syncat.Left
                            else s in
                let ov_p = if not omit_pp then Syncat.create_fwd ovp_s pp Syncat.Left
                           else ovp_s in
                let o_v = if not omit_obj then Syncat.create_bwd ov_p np Syncat.Right
                          else ov_p in
                o_v
          | OPVS -> (* s /< np \> np \> pp: left branching preferred *)
                let opv_s = if not omit_subj then Syncat.create_fwd s np Syncat.Left
                            else s in
                let o_pv = if not omit_obj then Syncat.create_bwd opv_s np Syncat.Right
                           else opv_s in
                let p_v = if not omit_pp then Syncat.create_bwd o_pv pp Syncat.Right
                          else o_pv in
                p_v
          | POVS -> (* s /< np \> pp \> np: left branching preferred *)
                let pov_s = if not omit_subj then Syncat.create_fwd s np Syncat.Left
                            else s in
                let p_ov = if not omit_pp then Syncat.create_bwd pov_s pp Syncat.Right
                           else pov_s in
                let o_v = if not omit_obj then Syncat.create_bwd p_ov np Syncat.Right
                          else p_ov in
                o_v
          | OSVP -> (* s \> np \> np /< pp: right branching preferred *)
                let osv_p = if not omit_pp then Syncat.create_bwd s pp Syncat.Right
                            else s in
                let o_sv = if not omit_obj then Syncat.create_bwd osv_p np Syncat.Right
                           else osv_p in
                let s_v = if not omit_subj then Syncat.create_fwd o_sv np Syncat.Left
                          else o_sv in
                s_v
          | OSPV -> (* s \> np \> np \> pp: right branching preferred *)
                let o_spv = if not omit_obj then Syncat.create_bwd s np Syncat.Right
                            else s in
                let s_pv = if not omit_subj then Syncat.create_bwd o_spv np Syncat.Right
                           else o_spv in
                let p_v = if not omit_pp then Syncat.create_bwd s_pv pp Syncat.Right
                          else s_pv in
                p_v
          | OPSV -> (* s \> np \> pp \> np: right branching preferred *)
                let o_psv = if not omit_obj then Syncat.create_bwd s np Syncat.Right
                            else s in
                let p_sv = if not omit_pp then Syncat.create_bwd o_psv pp Syncat.Right
                           else o_psv in
                let s_v = if not omit_subj then Syncat.create_bwd p_sv np Syncat.Right
                          else p_sv in
                s_v
          | POSV -> (* s \> pp \> np \> np: right branching preferred *)
                let p_osv = if not omit_pp then Syncat.create_bwd s pp Syncat.Right
                            else s in
                let o_sv = if not omit_obj then Syncat.create_bwd p_osv np Syncat.Right
                           else p_osv in
                let s_v = if not omit_subj then Syncat.create_bwd o_sv np Syncat.Right
                          else o_sv in
                s_v

    let adpos_for_np_genuine fund_ord_3_02 =
        match fund_ord_3_02 with
            NPO ->
                Syncat.create_fwd (Syncat.create_bwd np np Syncat.Left) np Syncat.Left
          | NOP ->
                Syncat.create_bwd (Syncat.create_bwd np np Syncat.Left) np Syncat.Right
          | PON ->
                Syncat.create_fwd (Syncat.create_fwd np np Syncat.Right) np Syncat.Left
          | OPN ->
                Syncat.create_bwd (Syncat.create_fwd np np Syncat.Right) np Syncat.Right

    let adpos_for_np_as_infl fund_ord_3_03 =
        match fund_ord_3_03 with
            NAP ->
                Syncat.create_bwd np np Syncat.Left
          | APN -> 
                Syncat.create_fwd np np Syncat.Right

    let pp_attach_for_vi fund_ord_3_04 omit_subj =
        match fund_ord_3_04 with
            SVOP ->
                Syncat.create_bwd (vi SVO omit_subj) (vi SVO omit_subj) Syncat.Left
          | SVPO ->
                Syncat.create_bwd (vi SVO omit_subj) (vi SVO omit_subj) Syncat.Left
          | SPVO ->
                Syncat.create_fwd (vi SVO omit_subj) (vi SVO omit_subj) Syncat.Right
          | PSVO ->
                Syncat.create_fwd s s Syncat.Right
          | SOVP ->
                Syncat.create_bwd (vi SOV omit_subj) (vi SOV omit_subj) Syncat.Left
          | SOPV ->
                Syncat.create_fwd (vi SOV omit_subj) (vi SOV omit_subj) Syncat.Right
          | SPOV ->
                Syncat.create_fwd (vi SOV omit_subj) (vi SOV omit_subj) Syncat.Right
          | PSOV ->
                Syncat.create_fwd s s Syncat.Right
          | VSOP ->
                Syncat.create_bwd s s Syncat.Left
          | VSPO ->
                let vs = Syncat.create_fwd s np Syncat.Left in
                Syncat.create_bwd vs vs Syncat.Left
          | VPSO ->
                Syncat.create_bwd (vi VSO omit_subj) (vi VSO omit_subj) Syncat.Left
          | PVSO ->
                let vs = Syncat.create_fwd s np Syncat.Left in
                Syncat.create_fwd vs vs Syncat.Right
          | VOSP ->
                Syncat.create_bwd s s Syncat.Left
          | VOPS ->
                Syncat.create_bwd (vi VOS omit_subj) (vi VOS omit_subj) Syncat.Left
          | VPOS ->
                Syncat.create_bwd (vi VOS omit_subj) (vi VOS omit_subj) Syncat.Left
          | PVOS ->
                Syncat.create_fwd (vi VOS omit_subj) (vi VOS omit_subj) Syncat.Right
          | OVSP ->
                Syncat.create_bwd s s Syncat.Left
          | OVPS ->
                Syncat.create_bwd (vi OVS omit_subj) (vi OVS omit_subj) Syncat.Left
          | OPVS ->
                Syncat.create_fwd (vi OVS omit_subj) (vi OVS omit_subj) Syncat.Right
          | POVS ->
                Syncat.create_fwd (vi OVS omit_subj) (vi OVS omit_subj) Syncat.Right
          | OSVP ->
                Syncat.create_bwd s s Syncat.Left
          | OSPV ->
                Syncat.create_fwd (vi OSV omit_subj) (vi OSV omit_subj) Syncat.Right
          | OPSV ->
                let sv = Syncat.create_bwd s np Syncat.Right in
                Syncat.create_fwd sv sv Syncat.Right
          | POSV ->
                Syncat.create_fwd s s Syncat.Right

    let pp_attach_for_vt fund_ord_3_04 omit_subj omit_obj =
        match fund_ord_3_04 with
            SVOP ->
                Syncat.create_bwd (vi SVO omit_subj) (vi SVO omit_subj) Syncat.Left
          | SVPO ->
                Syncat.create_bwd (vt SVO omit_subj omit_obj) (vt SVO omit_subj omit_obj) Syncat.Left
          | SPVO ->
                Syncat.create_fwd (vi SVO omit_subj) (vi SVO omit_subj) Syncat.Right
          | PSVO ->
                Syncat.create_fwd s s Syncat.Right
          | SOVP ->
                Syncat.create_bwd (vi SOV omit_subj) (vi SOV omit_subj) Syncat.Left
          | SOPV ->
                Syncat.create_fwd (vt SOV omit_subj omit_obj) (vt SOV omit_subj omit_obj) Syncat.Right
          | SPOV ->
                Syncat.create_fwd (vi SOV omit_subj) (vi SOV omit_subj) Syncat.Right
          | PSOV ->
                Syncat.create_fwd s s Syncat.Right
          | VSOP ->
                Syncat.create_bwd s s Syncat.Left
          | VSPO ->
                let vs = Syncat.create_fwd s np Syncat.Left in
                Syncat.create_bwd vs vs Syncat.Left
          | VPSO ->
                Syncat.create_bwd (vt VSO omit_subj omit_obj) (vt VSO omit_subj omit_obj) Syncat.Left
          | PVSO ->
                let vs = Syncat.create_fwd s np Syncat.Left in
                Syncat.create_fwd vs vs Syncat.Right
          | VOSP ->
                Syncat.create_bwd s s Syncat.Left
          | VOPS ->
                Syncat.create_bwd (vi VOS omit_subj) (vi VOS omit_subj) Syncat.Left
          | VPOS ->
                Syncat.create_bwd (vt VOS omit_subj omit_obj) (vt VOS omit_subj omit_obj) Syncat.Left
          | PVOS ->
                Syncat.create_fwd (vi VOS omit_subj) (vi VOS omit_subj) Syncat.Right
          | OVSP ->
                Syncat.create_bwd s s Syncat.Left
          | OVPS ->
                Syncat.create_bwd (vi OVS omit_subj) (vi OVS omit_subj) Syncat.Left
          | OPVS ->
                Syncat.create_fwd (vt OVS omit_subj omit_obj) (vt OVS omit_subj omit_obj) Syncat.Right
          | POVS ->
                Syncat.create_fwd (vi OVS omit_subj) (vi OVS omit_subj) Syncat.Right
          | OSVP ->
                Syncat.create_bwd s s Syncat.Left
          | OSPV ->
                Syncat.create_fwd (vt OSV omit_subj omit_obj) (vt OSV omit_subj omit_obj) Syncat.Right
          | OPSV ->
                let sv = Syncat.create_bwd s np Syncat.Right in
                Syncat.create_fwd sv sv Syncat.Right
          | POSV ->
                Syncat.create_fwd s s Syncat.Right

    let adpos_for_vt_genuine fund_ord_2_03 fund_ord_3_02 omit_subj omit_obj =
        let pp = pp_attach_for_vt fund_ord_2_03 omit_subj omit_obj in
        match fund_ord_3_02 with
            NPO ->
                Syncat.create_fwd pp np Syncat.Left
          | NOP ->
                Syncat.create_bwd pp np Syncat.Right
          | PON ->
                Syncat.create_fwd pp np Syncat.Left
          | OPN ->
                Syncat.create_bwd pp np Syncat.Right

    let adpos_for_vt_as_infl fund_ord_2_03 omit_subj omit_obj =
        pp_attach_for_vt fund_ord_2_03 omit_subj omit_obj

    let adpos_for_vi_genuine fund_ord_2_03 fund_ord_3_02 omit_subj =
        let pp = pp_attach_for_vi fund_ord_2_03 omit_subj in
        match fund_ord_3_02 with
            NPO ->
                Syncat.create_fwd pp np Syncat.Left
          | NOP ->
                Syncat.create_bwd pp np Syncat.Right
          | PON ->
                Syncat.create_fwd pp np Syncat.Left
          | OPN ->
                Syncat.create_bwd pp np Syncat.Right

    let adpos_for_vi_as_infl fund_ord_2_03 omit_subj =
        pp_attach_for_vi fund_ord_2_03 omit_subj

    let poss_genuine fund_ord_4_02 =
        match fund_ord_4_02 with
            N1PossN2 -> 
                Syncat.create_fwd (Syncat.create_bwd np np Syncat.Left) np Syncat.Left
          | N1N2Poss -> 
                Syncat.create_bwd (Syncat.create_bwd np np Syncat.Left) np Syncat.Right
          | PossN2N1 ->
                Syncat.create_fwd (Syncat.create_fwd np np Syncat.Right) np Syncat.Left
          | N2PossN1 ->
                Syncat.create_bwd (Syncat.create_fwd np np Syncat.Right) np Syncat.Right

    let poss_as_infl fund_ord_4_03 =
        match fund_ord_4_03 with
            N1Poss -> Syncat.create_bwd np np Syncat.Left
          | PossN1 -> Syncat.create_fwd np np Syncat.Right

    let adj_genuine fund_ord_5_02 =
        match fund_ord_5_02 with
            AdjN -> Syncat.create_fwd np np Syncat.Right
          | NAdj -> Syncat.create_bwd np np Syncat.Left

    let adj_with_cl fund_ord_5_03 =
        match fund_ord_5_03 with
            NClAdj ->
                Syncat.create_bwd (Syncat.create_bwd np np Syncat.Left) cl Syncat.Right
          | NAdjCl ->
                Syncat.create_fwd (Syncat.create_bwd np np Syncat.Left) cl Syncat.Left
          | ClNAdj ->
                Syncat.create_bwd (Syncat.create_bwd np cl Syncat.Right) np Syncat.Left
          | ClAdjN ->
                Syncat.create_bwd (Syncat.create_fwd np np Syncat.Right) cl Syncat.Right
          | AdjNCl ->
                Syncat.create_fwd (Syncat.create_fwd np cl Syncat.Left) np Syncat.Right
          | AdjClN ->
                Syncat.create_fwd (Syncat.create_fwd np np Syncat.Right) cl Syncat.Left

    let dem_genuine fund_ord_6_02 =
        match fund_ord_6_02 with
            DemN -> Syncat.create_fwd np np Syncat.Right
          | NDem -> Syncat.create_bwd np np Syncat.Left

    let dem_with_cl fund_ord_6_03 =
        match fund_ord_6_03 with
            NClDem ->
                Syncat.create_bwd (Syncat.create_bwd np np Syncat.Left) cl Syncat.Right
          | NDemCl ->
                Syncat.create_fwd (Syncat.create_bwd np np Syncat.Left) cl Syncat.Left
          | ClDemN ->
                Syncat.create_bwd (Syncat.create_fwd np np Syncat.Right) cl Syncat.Right
          | DemClN ->
                Syncat.create_fwd (Syncat.create_fwd np np Syncat.Right) cl Syncat.Left

    let numadj_genuine fund_ord_7_02 =
        match fund_ord_7_02 with
            NumN -> Syncat.create_fwd np np Syncat.Right
          | NNum -> Syncat.create_bwd np np Syncat.Left

    let numadj_with_cl fund_ord_7_03 =
        match fund_ord_7_03 with
            NClNum ->
                Syncat.create_bwd (Syncat.create_bwd np np Syncat.Left) cl Syncat.Right
          | NNumCl ->
                Syncat.create_fwd (Syncat.create_bwd np np Syncat.Left) cl Syncat.Left
          | ClNNum ->
                Syncat.create_bwd (Syncat.create_bwd np cl Syncat.Right) np Syncat.Left
          | ClNumN ->
                Syncat.create_bwd (Syncat.create_fwd np np Syncat.Right) cl Syncat.Right
          | NumNCl ->
                Syncat.create_fwd (Syncat.create_fwd np cl Syncat.Left) np Syncat.Right
          | NumClN ->
                Syncat.create_fwd (Syncat.create_fwd np np Syncat.Right) cl Syncat.Left

    let relpro_for_vi fund_ord_8_02 fund_ord_1_01 omit_np omit_subj =
        let vp = vi fund_ord_1_01 omit_subj in
        match fund_ord_8_02 with
            NRelVp ->
                let npmod = 
                    if not omit_np then Syncat.create_bwd np np Syncat.Left
                    else np in
                Syncat.create_fwd npmod vp Syncat.Left
          | RelVpN ->
                let npmod =
                    if not omit_np then Syncat.create_fwd np np Syncat.Right
                    else np in
                Syncat.create_fwd npmod vp Syncat.Left
          | NVpRel ->
                let npmod = 
                    if not omit_np then Syncat.create_bwd np np Syncat.Left
                    else np in
                Syncat.create_bwd npmod vp Syncat.Right
          | VpRelN ->
                let npmod =
                    if not omit_np then Syncat.create_fwd np np Syncat.Right
                    else np in
                Syncat.create_bwd npmod vp Syncat.Left

    let relpro_for_vt fund_ord_8_02 fund_ord_1_01 omit_np omit_subj omit_obj =
        let vp = vt fund_ord_1_01 omit_subj omit_obj in
        match fund_ord_8_02 with
            NRelVp ->
                let npmod =
                    if not omit_np then Syncat.create_bwd np np Syncat.Left
                    else np in
                Syncat.create_fwd npmod vp Syncat.Left
          | RelVpN ->
                let npmod =
                    if not omit_np then Syncat.create_fwd np np Syncat.Right
                    else np in
                Syncat.create_fwd npmod vp Syncat.Left
          | NVpRel ->
                let npmod =
                    if not omit_np then Syncat.create_bwd np np Syncat.Left
                    else np in
                Syncat.create_bwd npmod vp Syncat.Right
          | VpRelN ->
                let npmod =
                    if not omit_np then Syncat.create_fwd np np Syncat.Right
                    else np in
                Syncat.create_bwd npmod vp Syncat.Left

    let relpro_for_vd fund_ord_8_02 fund_ord_2_02 omit_np omit_subj omit_obj omit_iobj =
        let vp = vd fund_ord_2_02 omit_subj omit_obj omit_iobj in
        match fund_ord_8_02 with
            NRelVp ->
                let npmod =
                    if not omit_np then (Syncat.create_bwd np np Syncat.Left)
                    else np in
                Syncat.create_fwd npmod vp Syncat.Left
          | RelVpN ->
                let npmod =
                    if not omit_np then (Syncat.create_fwd np np Syncat.Right)
                    else np in
                Syncat.create_fwd npmod vp Syncat.Left
          | NVpRel ->
                let npmod =
                    if not omit_np then (Syncat.create_bwd np np Syncat.Left)
                    else np in
                Syncat.create_bwd npmod vp Syncat.Right
          | VpRelN ->
                let npmod =
                    if not omit_np then (Syncat.create_fwd np np Syncat.Right)
                    else np in
                Syncat.create_bwd npmod vp Syncat.Left

    let relinfl fund_ord_8_03 omit_np =
        match fund_ord_8_03 with
            NRelinfl -> 
                if not omit_np then Syncat.create_bwd np np Syncat.Left
                else np
          | RelinflN -> 
                if not omit_np then Syncat.create_fwd np np Syncat.Right
                else np

    let aggrel_for_vi fund_ord_8_04 omit_np =
        match fund_ord_8_04 with
            AggSVO | AggSOV | AggOSV ->
                if not omit_np then Syncat.create_bwd np np Syncat.Left
                else np
          | AggOVS | AggVSO | AggVOS ->
                if not omit_np then Syncat.create_fwd np np Syncat.Right
                else np

    let aggrel_for_vt fund_ord_8_04 omit_np omit_obj =
        match fund_ord_8_04 with
            AggSOV -> (* right branching preferred *)
                let aggvi' = aggrel_for_vi fund_ord_8_04 omit_np in
                if not omit_obj then Syncat.create_bwd aggvi' np Syncat.Right
                else aggvi'
          | AggSVO -> (* right branching preferred *)
                let aggvi' = aggrel_for_vi fund_ord_8_04 omit_np in
                if not omit_obj then Syncat.create_fwd aggvi' np Syncat.Left
                else aggvi'
          | AggVSO -> (* left branching preferred *)
                let aggvo =
                    if not omit_obj then Syncat.create_fwd np np Syncat.Left
                    else np in
                if not omit_np then Syncat.create_fwd aggvo np Syncat.Left
                else aggvo
          | AggVOS -> (* left branching preferred *)
                let aggvi' = aggrel_for_vi fund_ord_8_04 omit_np in
                if not omit_obj then Syncat.create_fwd aggvi' np Syncat.Left
                else aggvi'
          | AggOVS -> (* left branching preferred *)
                let aggvi' = aggrel_for_vi fund_ord_8_04 omit_np in
                if not omit_obj then Syncat.create_bwd aggvi' np Syncat.Right
                else aggvi'
          | AggOSV -> (* right branching preferred *)
                let aggov =
                    if not omit_obj then Syncat.create_bwd np np Syncat.Right
                    else np in
                if not omit_np then Syncat.create_bwd aggov np Syncat.Right
                else aggov

    let relc_extrapos fund_ord_8_05 omit_np =
        match fund_ord_8_05 with
            SentRelc ->
                if not omit_np then Syncat.create_bwd s s Syncat.Left
                else np
          | RelcSent ->
                if not omit_np then Syncat.create_fwd s s Syncat.Right
                else np

    let relpro_of_vi_for_n_dup fund_ord_8_06 fund_ord_1_01 omit_np omit_subj =
        let fund_ord_8_02 =
            match fund_ord_8_06 with
                NRelcRepn -> NVpRel
              | NRepnRelc -> NRelVp
              | RelcRepnN -> VpRelN
              | RepnRelcN -> RelVpN in
        relpro_for_vi fund_ord_8_02 fund_ord_1_01 omit_np omit_subj

    let relpro_of_vt_for_n_dup fund_ord_8_06 fund_ord_1_01 omit_np omit_subj omit_obj =
        let fund_ord_8_02 =
            match fund_ord_8_06 with
                NRelcRepn -> NVpRel
              | NRepnRelc -> NRelVp
              | RelcRepnN -> VpRelN
              | RepnRelcN -> RelVpN in
        relpro_for_vt fund_ord_8_02 fund_ord_1_01 omit_np omit_subj omit_obj

    let relpro_of_vd_for_n_dup fund_ord_8_06 fund_ord_1_01 omit_np omit_subj omit_obj omit_iobj =
        let fund_ord_8_02 =
            match fund_ord_8_06 with
                NRelcRepn -> NVpRel
              | NRepnRelc -> NRelVp
              | RelcRepnN -> VpRelN
              | RepnRelcN -> RelVpN in
        relpro_for_vd fund_ord_8_02 fund_ord_1_01 omit_np omit_subj omit_obj omit_iobj

    let cl_for_relc fund_ord_8_07 fund_ord_8_02 omit_np =
        let relc =
            match fund_ord_8_02 with
                NRelVp -> Syncat.create_bwd np np Syncat.Left
              | RelVpN -> Syncat.create_fwd np np Syncat.Right
              | NVpRel -> Syncat.create_bwd np np Syncat.Left
              | VpRelN -> Syncat.create_fwd np np Syncat.Right in
        match fund_ord_8_07 with
            NRelcCl ->
                if not omit_np then Syncat.create_bwd relc relc Syncat.Left
                else Syncat.create_bwd np relc Syncat.Left
          | NClRelc ->
                if not omit_np then Syncat.create_fwd relc relc Syncat.Right
                else Syncat.create_fwd np relc Syncat.Right
          | ClNRelc ->
                if not omit_np then Syncat.create_fwd np np Syncat.Right
                else Syncat.create_fwd np relc Syncat.Right
          | ClRelcN ->
                if not omit_np then Syncat.create_fwd relc relc Syncat.Right
                else Syncat.create_fwd np relc Syncat.Right
          | RelcNCl ->
                if not omit_np then Syncat.create_bwd np np Syncat.Left
                else Syncat.create_bwd np relc Syncat.Left
          | RelcClN ->
                if not omit_np then Syncat.create_bwd relc relc Syncat.Left
                else Syncat.create_bwd np relc Syncat.Left

    let quant_for_adj fund_ord_9_01 fund_ord_5_02 =
        let adj = adj_genuine fund_ord_5_02 in
        match fund_ord_9_01 with
            VeryAdj -> Syncat.create_fwd adj adj Syncat.Right
          | AdjVery -> Syncat.create_bwd adj adj Syncat.Left
          | AdjAdj -> Syncat.create_fwd adj adj Syncat.Right
          | Adj_with_infl -> adj

    let tam tam_word_order omit_subj omit_obj =
        match tam_word_order with
            TamSVO | TamSOV | TamVSO | TamVOS | TamOSV | TamOVS -> 
                Syncat.create_fwd s s Syncat.Right
          | STamVO | STamOV ->
                let vp = 
                    if not omit_subj then Syncat.create_bwd s np Syncat.Left 
                    else s in
                Syncat.create_fwd vp vp Syncat.Right
          | VOTamS | OVTamS ->
                let vp = 
                    if not omit_subj then Syncat.create_fwd s np Syncat.Left 
                    else s in
                Syncat.create_bwd vp vp Syncat.Left
          | VTamSO | VTamOS | SVTamO ->
                let vp =
                    if not omit_subj then Syncat.create_fwd s np Syncat.Left
                    else s in
                Syncat.create_bwd vp vp Syncat.Left
          | SOTamV | OTamVS | OSTamV ->
                let vp =
                    if not omit_subj then Syncat.create_bwd s np Syncat.Right
                    else s in
                Syncat.create_fwd vp vp Syncat.Right
          | OTamSV ->
                let sv = 
                    if not omit_obj then Syncat.create_bwd s np Syncat.Right 
                    else s in
                Syncat.create_fwd sv sv Syncat.Right
          | VSTamO ->
                let vs = 
                    if not omit_obj then Syncat.create_fwd s np Syncat.Right 
                    else s in
                Syncat.create_bwd vs vs Syncat.Left
          | SVOTam | SOVTam | VSOTam | VOSTam | OSVTam | OVSTam ->
                Syncat.create_bwd s s Syncat.Left

    let nompropref_for_vi simp_1_03 fund_ord_1_01 omit_subj =
        let vi' = vi fund_ord_1_01 omit_subj in
        match simp_1_03 with
            NomproV -> Syncat.create_fwd vi' vi' Syncat.Right
          | VNompro -> Syncat.create_bwd vi' vi' Syncat.Left

    let nompropref_for_vt simp_1_03 fund_ord_1_01 omit_subj omit_obj =
        let vt' = vt fund_ord_1_01 omit_subj omit_obj in
        match simp_1_03 with
            NomproV -> Syncat.create_fwd vt' vt' Syncat.Right
          | VNompro -> Syncat.create_bwd vt' vt' Syncat.Left

    let nompropref_for_vd simp_1_03 fund_ord_2_02 omit_subj omit_obj omit_iobj =
        let vd' = vd fund_ord_2_02 omit_subj omit_obj omit_iobj in
        match simp_1_03 with
            NomproV -> Syncat.create_fwd vd' vd' Syncat.Right
          | VNompro -> Syncat.create_bwd vd' vd' Syncat.Left

    let accpropref_for_vi simp_1_04 fund_ord_1_01 omit_subj =
        let vi' = vi fund_ord_1_01 omit_subj in
        match simp_1_04 with
            AccproV -> Syncat.create_fwd vi' vi' Syncat.Right
          | VAccpro -> Syncat.create_bwd vi' vi' Syncat.Left

    let accpropref_for_vt simp_1_04 fund_ord_1_01 omit_subj omit_obj =
        let vt' = vt fund_ord_1_01 omit_subj omit_obj in
        match simp_1_04 with
            AccproV -> Syncat.create_fwd vt' vt' Syncat.Right
          | VAccpro -> Syncat.create_bwd vt' vt' Syncat.Left

    let accpropref_for_vd simp_1_04 fund_ord_2_02 omit_subj omit_obj omit_iobj =
        let vd' = vd fund_ord_2_02 omit_subj omit_obj omit_iobj in
        match simp_1_04 with
            AccproV -> Syncat.create_fwd vd' vd' Syncat.Right
          | VAccpro -> Syncat.create_bwd vd' vd' Syncat.Left

    let reflpro simp_2_02 fund_ord_1_01 omit_subj =
        let vp' = vi fund_ord_1_01 omit_subj
        and vt' = vt fund_ord_1_01 omit_subj false in
        match simp_2_02 with
            ReflSVAdv | ReflSAdvV | ReflVSAdv | ReflVAdvS | ReflAdvSV | ReflAdvVS ->
                Syncat.create_fwd s s Syncat.Right
          | SReflVAdv | SReflAdvV | AdvReflSV | AdvReflVS ->
                Syncat.create_fwd vp' vt' Syncat.Right
          | VReflSAdv | VReflAdvS | SVReflAdv | AdvVReflS ->
                Syncat.create_bwd vp' vt' Syncat.Left
          | SAdvReflV | AdvSReflV ->
                Syncat.create_fwd vp' vt' Syncat.Right
          | VSReflAdv ->
                Syncat.create_bwd vp' vt' Syncat.Left
          | VAdvReflS ->
                Syncat.create_fwd vp' vt' Syncat.Right
          | SVAdvRefl | SAdvVRefl | VSAdvRefl | VAdvSRefl | AdvSVRefl | AdvVSRefl ->
                Syncat.create_bwd s s Syncat.Left

    let recipro simp_2_04 fund_ord_1_01 omit_subj =
        let vp' = vi fund_ord_1_01 omit_subj
        and vt' = vt fund_ord_1_01 omit_subj false in
        match simp_2_04 with
            ReciSV -> Syncat.create_fwd s s Syncat.Right
          | VSReci -> Syncat.create_bwd s s Syncat.Left
          | ReciVS | SReciV -> Syncat.create_fwd vp' vt' Syncat.Right
          | SVReci | VReciS -> Syncat.create_bwd vp' vt' Syncat.Right

    let refladv_for_vi simp_2_07 omit_subj =
        match simp_2_07 with
            SVORefladv ->
                Syncat.create_bwd (vi SVO omit_subj) (vi SVO omit_subj) Syncat.Left
          | SVRefladvO ->
                Syncat.create_bwd (vi SVO omit_subj) (vi SVO omit_subj) Syncat.Left
          | SRefladvVO ->
                Syncat.create_fwd (vi SVO omit_subj) (vi SVO omit_subj) Syncat.Right
          | RefladvSVO ->
                Syncat.create_fwd s s Syncat.Right
          | SOVRefladv ->
                Syncat.create_bwd (vi SOV omit_subj) (vi SOV omit_subj) Syncat.Left
          | SORefladvV ->
                Syncat.create_fwd (vi SOV omit_subj) (vi SOV omit_subj) Syncat.Right
          | SRefladvOV ->
                Syncat.create_fwd (vi SOV omit_subj) (vi SOV omit_subj) Syncat.Right
          | RefladvSOV ->
                Syncat.create_fwd s s Syncat.Right
          | VSORefladv ->
                Syncat.create_bwd s s Syncat.Left
          | VSRefladvO ->
                let vs = Syncat.create_fwd s np Syncat.Left in
                Syncat.create_bwd vs vs Syncat.Left
          | VRefladvSO ->
                Syncat.create_bwd (vi VSO omit_subj) (vi VSO omit_subj) Syncat.Left
          | RefladvVSO ->
                let vs = Syncat.create_fwd s np Syncat.Left in
                Syncat.create_fwd vs vs Syncat.Right
          | VOSRefladv ->
                Syncat.create_bwd s s Syncat.Left
          | VORefladvS ->
                Syncat.create_bwd (vi VOS omit_subj) (vi VOS omit_subj) Syncat.Left
          | VRefladvOS ->
                Syncat.create_bwd (vi VOS omit_subj) (vi VOS omit_subj) Syncat.Left
          | RefladvVOS ->
                Syncat.create_fwd (vi VOS omit_subj) (vi VOS omit_subj) Syncat.Right
          | OVSRefladv ->
                Syncat.create_bwd s s Syncat.Left
          | OVRefladvS ->
                Syncat.create_bwd (vi OVS omit_subj) (vi OVS omit_subj) Syncat.Left
          | ORefladvVS ->
                Syncat.create_fwd (vi OVS omit_subj) (vi OVS omit_subj) Syncat.Right
          | RefladvOVS ->
                Syncat.create_fwd (vi OVS omit_subj) (vi OVS omit_subj) Syncat.Right
          | OSVRefladv ->
                Syncat.create_bwd s s Syncat.Left
          | OSRefladvV ->
                Syncat.create_fwd (vi OSV omit_subj) (vi OSV omit_subj) Syncat.Right
          | ORefladvSV ->
                let sv = Syncat.create_bwd s np Syncat.Right in
                Syncat.create_fwd sv sv Syncat.Right
          | RefladvOSV ->
                Syncat.create_fwd s s Syncat.Right

    let refladv_for_vt simp_2_07 omit_subj omit_obj =
        match simp_2_07 with
            SVORefladv ->
                Syncat.create_bwd (vi SVO omit_subj) (vi SVO omit_subj) Syncat.Left
          | SVRefladvO ->
                Syncat.create_bwd (vt SVO omit_subj omit_obj) (vt SVO omit_subj omit_obj) Syncat.Left
          | SRefladvVO ->
                Syncat.create_fwd (vi SVO omit_subj) (vi SVO omit_subj) Syncat.Right
          | RefladvSVO ->
                Syncat.create_fwd s s Syncat.Right
          | SOVRefladv ->
                Syncat.create_bwd (vi SOV omit_subj) (vi SOV omit_subj) Syncat.Left
          | SORefladvV ->
                Syncat.create_fwd (vt SOV omit_subj omit_obj) (vt SOV omit_subj omit_obj) Syncat.Right
          | SRefladvOV ->
                Syncat.create_fwd (vi SOV omit_subj) (vi SOV omit_subj) Syncat.Right
          | RefladvSOV ->
                Syncat.create_fwd s s Syncat.Right
          | VSORefladv ->
                Syncat.create_bwd s s Syncat.Left
          | VSRefladvO ->
                let vs = Syncat.create_fwd s np Syncat.Left in
                Syncat.create_bwd vs vs Syncat.Left
          | VRefladvSO ->
                Syncat.create_bwd (vt VSO omit_subj omit_obj) (vt VSO omit_subj omit_obj) Syncat.Left
          | RefladvVSO ->
                let vs = Syncat.create_fwd s np Syncat.Left in
                Syncat.create_fwd vs vs Syncat.Right
          | VOSRefladv ->
                Syncat.create_bwd s s Syncat.Left
          | VORefladvS ->
                Syncat.create_bwd (vi VOS omit_subj) (vi VOS omit_subj) Syncat.Left
          | VRefladvOS ->
                Syncat.create_bwd (vt VOS omit_subj omit_obj) (vt VOS omit_subj omit_obj) Syncat.Left
          | RefladvVOS ->
                Syncat.create_fwd (vi VOS omit_subj) (vi VOS omit_subj) Syncat.Right
          | OVSRefladv ->
                Syncat.create_bwd s s Syncat.Left
          | OVRefladvS ->
                Syncat.create_bwd (vi OVS omit_subj) (vi OVS omit_subj) Syncat.Left
          | ORefladvVS ->
                Syncat.create_fwd (vt OVS omit_subj omit_obj) (vt OVS omit_subj omit_obj) Syncat.Right
          | RefladvOVS ->
                Syncat.create_fwd (vi OVS omit_subj) (vi OVS omit_subj) Syncat.Right
          | OSVRefladv ->
                Syncat.create_bwd s s Syncat.Left
          | OSRefladvV ->
                Syncat.create_fwd (vt OSV omit_subj omit_obj) (vt OSV omit_subj omit_obj) Syncat.Right
          | ORefladvSV ->
                let sv = Syncat.create_bwd s np Syncat.Right in
                Syncat.create_fwd sv sv Syncat.Right
          | RefladvOSV ->
                Syncat.create_fwd s s Syncat.Right

    let passive_marker simp_3_02 fund_ord_1_01 omit_subj omit_obl =
        let vt' = vt fund_ord_1_01 omit_subj false in
        match simp_3_02 with
            PassSVObl -> (* s/pp/vt/np *)
                let passsv_obl =
                    if not omit_obl then Syncat.create_fwd s pp Syncat.Left
                    else s in
                let passs_v = Syncat.create_fwd passsv_obl vt' Syncat.Right in
                if not omit_subj then Syncat.create_fwd passs_v np Syncat.Left
                else passs_v
          | PassSOblV -> (* s/vt/pp/np *)
                let passsobl_v = Syncat.create_fwd s vt' Syncat.Right in
                let passs_obl = 
                    if not omit_obl then Syncat.create_fwd passsobl_v pp Syncat.Left
                    else passsobl_v in
                if not omit_subj then Syncat.create_fwd passs_obl np Syncat.Left
                else passs_obl
          | PassVSObl -> (* s/pp/np/vt *)
                let passvs_obl =
                    if not omit_obl then Syncat.create_fwd s pp Syncat.Left
                    else s in
                let passv_s =
                    if not omit_subj then Syncat.create_fwd passvs_obl np Syncat.Left
                    else passvs_obl in
                Syncat.create_fwd passv_s vt' Syncat.Right
          | PassVOblS -> (* s/np/pp/vt *)
                let passvobl_s =
                    if not omit_subj then Syncat.create_fwd s np Syncat.Left
                    else s in
                let passv_obl =
                    if not omit_obl then Syncat.create_fwd passvobl_s pp Syncat.Left
                    else passvobl_s in
                Syncat.create_fwd passv_obl vt' Syncat.Right
          | PassOblSV -> (* s/vt/np/pp *)
                let passobls_v = Syncat.create_fwd s vt' Syncat.Right in
                let passobl_s =
                    if not omit_subj then Syncat.create_fwd passobls_v np Syncat.Left
                    else passobls_v in
                if not omit_obl then Syncat.create_fwd passobl_s pp Syncat.Left
                else passobl_s
          | PassOblVS -> (* s/np/vt/pp *)
                let passoblv_s =
                    if not omit_subj then Syncat.create_fwd s np Syncat.Left
                    else s in
                let passobl_v = Syncat.create_fwd passoblv_s vt' Syncat.Right in
                if not omit_obl then Syncat.create_fwd passobl_v pp Syncat.Left
                else passobl_v
          | SPassVObl -> (* s\np/pp/vt *)
                let s_passvobl =
                    if not omit_subj then Syncat.create_bwd s np Syncat.Right
                    else s in
                let passv_obl =
                    if not omit_obl then Syncat.create_fwd s_passvobl pp Syncat.Left
                    else s_passvobl in
                Syncat.create_fwd passv_obl vt' Syncat.Right
          | SPassOblV -> (* s\np/vt/pp *)
                let s_passoblv =
                    if not omit_subj then Syncat.create_bwd s np Syncat.Right
                    else s in
                let passobl_v = Syncat.create_fwd s_passoblv vt' Syncat.Right in
                if not omit_obl then Syncat.create_fwd passobl_v pp Syncat.Left
                else passobl_v
          | VPassSObl -> (* s/pp/np\vt *)
                let vpasss_obl =
                    if not omit_obl then Syncat.create_fwd s pp Syncat.Left
                    else s in
                let vpass_s =
                    if not omit_subj then Syncat.create_fwd vpasss_obl np Syncat.Left
                    else vpasss_obl in
                Syncat.create_bwd vpass_s vt' Syncat.Right
          | VPassOblS -> (* s/np/pp\vt *)
                let vpassobl_s =
                    if not omit_subj then Syncat.create_fwd s np Syncat.Left
                    else s in
                let vpass_obl =
                    if not omit_obl then Syncat.create_fwd vpassobl_s pp Syncat.Left
                    else vpassobl_s in
                Syncat.create_bwd vpass_obl vt' Syncat.Left
          | OblPassSV -> (* s\pp/vt/np *)
                let obl_passsv =
                    if not omit_obl then Syncat.create_bwd s pp Syncat.Right
                    else s in
                let passs_v = Syncat.create_fwd obl_passsv vt' Syncat.Right in
                if not omit_subj then Syncat.create_fwd passs_v np Syncat.Left
                else passs_v
          | OblPassVS -> (* s\pp/np/vt *)
                let obl_passvs =
                    if not omit_obl then Syncat.create_bwd s pp Syncat.Right
                    else s in
                let passv_s =
                    if not omit_subj then Syncat.create_fwd obl_passvs np Syncat.Left
                    else obl_passvs in
                Syncat.create_fwd passv_s vt' Syncat.Right
          | SVPassObl -> (* s/pp\np\vt *)
                let svpass_obl =
                    if not omit_obl then Syncat.create_fwd s pp Syncat.Left
                    else s in
                let s_vpass =
                    if not omit_subj then Syncat.create_bwd svpass_obl np Syncat.Right
                    else svpass_obl in
                Syncat.create_fwd s_vpass vt' Syncat.Right
          | SOblPassV -> (* s\np\pp/vt *)
                let s_oblpassv =
                    if not omit_subj then Syncat.create_bwd s np Syncat.Right
                    else s in
                let obl_passv =
                    if not omit_obl then Syncat.create_bwd s_oblpassv pp Syncat.Right
                    else s_oblpassv in
                Syncat.create_fwd obl_passv vt' Syncat.Right
          | VSPassObl -> (* s/pp\vt\np *)
                let vspass_obl =
                    if not omit_obl then Syncat.create_fwd s pp Syncat.Right
                    else s in
                let v_spass = Syncat.create_bwd vspass_obl vt' Syncat.Left in
                if not omit_subj then Syncat.create_bwd v_spass np Syncat.Right
                else v_spass
          | VOblPassS -> (* s/np\vt\pp *)
                let voblpass_s =
                    if not omit_subj then Syncat.create_fwd s np Syncat.Left
                    else s in
                let v_oblpass = Syncat.create_bwd voblpass_s vt' Syncat.Left in
                if not omit_obl then Syncat.create_bwd v_oblpass pp Syncat.Left
                else v_oblpass
          | OblSPassV -> (* s\pp\np/vt *)
                let obl_spassv =
                    if not omit_obl then Syncat.create_bwd s pp Syncat.Right
                    else s in
                let s_passv =
                    if not omit_subj then Syncat.create_bwd obl_spassv np Syncat.Right
                    else obl_spassv in
                Syncat.create_fwd s_passv vt' Syncat.Right
          | OblVPassS -> (* s/np\pp\vt *)
                let oblvpass_s =
                    if not omit_subj then Syncat.create_fwd s np Syncat.Left
                    else s in
                let obl_vpass =
                    if not omit_obl then Syncat.create_bwd oblvpass_s pp Syncat.Right
                    else oblvpass_s in
                Syncat.create_bwd obl_vpass vt' Syncat.Left
          | SVOblPass -> (* s\np\vt\pp *)
                let s_voblpass =
                    if not omit_subj then Syncat.create_bwd s np Syncat.Right
                    else s in
                let v_oblpass = Syncat.create_bwd s_voblpass vt' Syncat.Left in
                if not omit_obl then Syncat.create_bwd v_oblpass pp Syncat.Right
                else v_oblpass
          | SOblVPass -> (* s\np\pp\vt *)
                let s_oblvpass =
                    if not omit_subj then Syncat.create_bwd s np Syncat.Right
                    else s in
                let obl_vpass =
                    if not omit_obl then Syncat.create_bwd s_oblvpass pp Syncat.Right
                    else s_oblvpass in
                Syncat.create_bwd obl_vpass vt' Syncat.Left
          | VSOblPass -> (* s\vt\np\pp *)
                let v_soblpass = Syncat.create_bwd s vt' Syncat.Left in
                let s_oblpass =
                    if not omit_subj then Syncat.create_bwd v_soblpass np Syncat.Right 
                    else v_soblpass in
                if not omit_obl then Syncat.create_bwd s_oblpass pp Syncat.Right
                else s_oblpass
          | VOblSPass -> (* s\vt\pp\np *)
                let v_oblspass = Syncat.create_bwd s vt' Syncat.Left in
                let obl_spass =
                    if not omit_obl then Syncat.create_bwd v_oblspass pp Syncat.Right
                    else v_oblspass in
                if not omit_subj then Syncat.create_bwd obl_spass np Syncat.Right
                else obl_spass
          | OblSVPass -> (* s\pp\np\vt *)
                let obl_svpass =
                    if not omit_obl then Syncat.create_bwd s pp Syncat.Right
                    else s in
                let s_vpass =
                    if not omit_subj then Syncat.create_bwd obl_svpass np Syncat.Right
                    else obl_svpass in
                Syncat.create_fwd s_vpass vt' Syncat.Left
          | OblVSPass -> (* s\pp\vt\np *)
                let obl_vspass =
                    if not omit_obl then Syncat.create_bwd s pp Syncat.Right
                    else s in
                let v_spass = Syncat.create_bwd obl_vspass vt' Syncat.Left in
                if not omit_subj then Syncat.create_bwd v_spass np Syncat.Right
                else v_spass

    let passivized_verb passive_as_infl_order omit_subj omit_obl =
        match passive_as_infl_order with
            Passive_SOblV -> (* right branching preferred *) 
                let vi' = vi SOV omit_subj in
                if not omit_obl then Syncat.create_bwd vi' pp Syncat.Right
                else vi'
          | Passive_SVObl -> (* right branching preferred *) 
                let vi' = vi SVO omit_subj in
                if not omit_obl then Syncat.create_fwd vi' pp Syncat.Left
                else vi'
          | Passive_VSObl -> (* left branching preferred *) 
                let vo = 
                    if not omit_obl then Syncat.create_fwd s pp Syncat.Left 
                    else s in
                if not omit_subj then Syncat.create_fwd vo np Syncat.Left
                else vo
          | Passive_VOblS -> (* left branching preferred *)
                let vi' = vi VOS omit_subj in
                if not omit_obl then Syncat.create_fwd vi' pp Syncat.Left
                else vi'
          | Passive_OblVS -> (* left branching preferred *) 
                let vi' = vi OVS omit_subj in
                if not omit_obl then Syncat.create_bwd vi' pp Syncat.Right
                else vi'
          | Passive_OblSV -> (* right branching preferred *) 
                let ov =
                    if not omit_obl then Syncat.create_bwd s pp Syncat.Right
                    else s in
                if not omit_subj then Syncat.create_bwd ov np Syncat.Right
                else ov

    let infmarker_for_vi simp_4_02 fund_ord_1_01 =
        match simp_4_02 with
            ObjCtrl_InfVO | ObjCtrl_InfOV | ObjCtrl_OInfV ->
                Syncat.create_fwd vp (vi fund_ord_1_01 false) Syncat.Left
          | ObjCtrl_VOInf | ObjCtrl_OVInf | ObjCtrl_VInfO->
                Syncat.create_bwd vp (vi fund_ord_1_01 false) Syncat.Right

    let infmarker_for_vt simp_4_02 fund_ord_1_01 =
        match simp_4_02 with
            ObjCtrl_InfVO | ObjCtrl_InfOV ->
                let vi' = vi fund_ord_1_01 false in
                Syncat.create_fwd vp vi' Syncat.Left
          | ObjCtrl_VOInf | ObjCtrl_OVInf ->
                let vi' = vi fund_ord_1_01 false in
                Syncat.create_bwd vp vi' Syncat.Right
          | ObjCtrl_VInfO ->
                let vt' = vt fund_ord_1_01 false false in
                let v_info = Syncat.create_bwd vp vt' Syncat.Right in
                Syncat.create_fwd v_info np Syncat.Left
          | ObjCtrl_OInfV ->
                let vt' = vt fund_ord_1_01 false false in
                let oinf_v = Syncat.create_fwd vp vt' Syncat.Left in
                Syncat.create_bwd oinf_v np Syncat.Right

    let complementizer is_on_left =
        if is_on_left then Syncat.create_fwd sbar s Syncat.Left
        else Syncat.create_bwd sbar s Syncat.Right

    let vt_objctrl simp_4_03 omit_subj omit_obj compmark =
        let comp =
            if not compmark then
                match simp_4_03 with
                    SV1O1V2O2 | SV1V2O2O1 | SV1V2O1O2 ->
                        vi SVO false
                  | SO1V1O2V2 | SO1O2V2V1 | SO1O2V1V2 ->
                        vi SOV false
                  | V1SO1V2O2 | V1SV2O2O1 | V1V2O2SO1 | V1SV2O1O2 | V1V2SO1O2 ->
                        vi VSO false
                  | O1SV1O2V2 | O1O2V2SV1 | O1O2SV2V1 | O2O1SV2V1 ->
                        vi OSV false
                  | O2O1V2SV1 | O1V1O2V2S | O1O2V2V1S | O1O2V1V2S ->
                        vi OVS false
                  | V1O1V2O2S | V1V2O2O1S | V1V2O1O2S ->
                        vi VOS false
            else sbar in
        match simp_4_03 with
            SV1O1V2O2 -> (* s\np0/(s\np1)/np1 *)
                let s_vocomp =
                    if not omit_subj then Syncat.create_bwd s np Syncat.Right
                    else s in
                let svo_comp = Syncat.create_fwd s_vocomp comp Syncat.Left in
                if not omit_obj then Syncat.create_fwd svo_comp np Syncat.Left
                else svo_comp
          | SV1V2O2O1 -> (* s\np0/np1/(s\np1) *)
                let s_vcompo =
                    if not omit_subj then Syncat.create_bwd s np Syncat.Right
                    else s in
                let svcomp_o =
                    if not omit_obj then Syncat.create_fwd s_vcompo np Syncat.Left
                    else s_vcompo in
                Syncat.create_fwd svcomp_o comp Syncat.Left
          | SV1V2O1O2 -> (* s\np0/np1/(s\np1) *)
                let s_vcompo =
                    if not omit_subj then Syncat.create_bwd s np Syncat.Right
                    else s in
                let svcomp_o =
                    if not omit_obj then Syncat.create_fwd s_vcompo np Syncat.Left
                    else s_vcompo in
                Syncat.create_fwd svcomp_o comp Syncat.Left
          | SO1V1O2V2 -> (* s\np0\np1/(s\np1) *)
                let s_ovcomp =
                    if not omit_subj then Syncat.create_bwd s np Syncat.Right
                    else s in
                let so_vcomp =
                    if not omit_obj then Syncat.create_bwd s_ovcomp np Syncat.Right
                    else s_ovcomp in
                Syncat.create_fwd so_vcomp comp Syncat.Left
          | SO1O2V2V1 -> (* s\np0\np1\(s\np1) *)
                let s_ocompv =
                    if not omit_subj then Syncat.create_bwd s np Syncat.Right
                    else s in
                let so_compv =
                    if not omit_obj then Syncat.create_bwd s_ocompv np Syncat.Right
                    else s_ocompv in
                Syncat.create_bwd so_compv comp Syncat.Right
          | SO1O2V1V2 -> (* s\np0\np1/(s\np1) *)
                let s_ovcomp =
                    if not omit_subj then Syncat.create_bwd s np Syncat.Right
                    else s in
                let so_vcomp =
                    if not omit_obj then Syncat.create_bwd s_ovcomp np Syncat.Right
                    else s_ovcomp in
                Syncat.create_fwd so_vcomp comp Syncat.Left
          | V1SO1V2O2 -> (* s/(s/np1)/np1/np0 *)
                let vso_comp = Syncat.create_fwd s comp Syncat.Left in
                let vs_ocomp =
                    if not omit_obj then Syncat.create_fwd vso_comp np Syncat.Left
                    else vso_comp in
                if not omit_subj then Syncat.create_fwd vs_ocomp np Syncat.Left
                else vs_ocomp
          | V1SV2O2O1 -> (* s/np1/(s/np1)/np0 *)
                let vscomp_o =
                    if not omit_obj then Syncat.create_fwd s np Syncat.Left
                    else s in
                let vs_compo = Syncat.create_fwd vscomp_o comp Syncat.Left in
                if not omit_subj then Syncat.create_fwd vs_compo np Syncat.Left
                else vs_compo
          | V1V2O2SO1 -> (* s/np1/np0/(s/np1) *)
                let vcomps_o =
                    if not omit_obj then Syncat.create_fwd s np Syncat.Left
                    else s in
                let vcomp_so =
                    if not omit_subj then Syncat.create_fwd vcomps_o np Syncat.Left
                    else vcomps_o in
                Syncat.create_fwd vcomp_so comp Syncat.Left
          | V1SV2O1O2 -> (* artificial case: s/np0/np1/(s/np1) *)
                let vscomp_o =
                    if not omit_obj then Syncat.create_fwd s np Syncat.Left
                    else s in
                let vs_compo = Syncat.create_fwd vscomp_o comp Syncat.Left in
                if not omit_subj then Syncat.create_fwd vs_compo np Syncat.Left
                else vs_compo
          | V1V2SO1O2 -> (* artificial case: s/np0/np1/(s/np1) *)
                let vcomps_o =
                    if not omit_obj then Syncat.create_fwd s np Syncat.Left
                    else s in
                let vcomp_so =
                    if not omit_subj then Syncat.create_fwd vcomps_o np Syncat.Left
                    else vcomps_o in
                Syncat.create_fwd vcomp_so comp Syncat.Left
          | O1SV1O2V2 -> (* s\np1\np0/(s\np1) *)
                let o_svcomp =
                    if not omit_obj then Syncat.create_bwd s np Syncat.Right
                    else s in
                let os_vcomp =
                    if not omit_subj then Syncat.create_bwd o_svcomp np Syncat.Right
                    else o_svcomp in
                Syncat.create_fwd os_vcomp comp Syncat.Left
          | O1O2V2SV1 -> (* s\np1\(s\np1)\np0 *)
                let o_compsv =
                    if not omit_obj then Syncat.create_bwd s np Syncat.Right
                    else s in
                let ocomp_sv = Syncat.create_bwd o_compsv comp Syncat.Right in
                if not omit_subj then Syncat.create_bwd ocomp_sv np Syncat.Right
                else ocomp_sv
          | O1O2SV2V1 -> (* artificial case: s\np0\np1\(s\np1) *)
                let o_scompv =
                    if not omit_obj then Syncat.create_bwd s np Syncat.Right
                    else s in
                let os_compv =
                    if not omit_subj then Syncat.create_bwd o_scompv np Syncat.Right
                    else o_scompv in
                Syncat.create_bwd os_compv comp Syncat.Right
          | O2O1V2SV1 -> (* artificial case: s\np0\np1\(s\np1) *)
                let o_compsv =
                    if not omit_obj then Syncat.create_bwd s np Syncat.Right
                    else s in
                let ocomp_sv = Syncat.create_bwd o_compsv comp Syncat.Right in
                if not omit_subj then Syncat.create_bwd ocomp_sv np Syncat.Right
                else ocomp_sv
          | O2O1SV2V1 -> (* artificial case: s\np0\np1\(s\np1) *)
                let o_scompv =
                    if not omit_obj then Syncat.create_bwd s np Syncat.Right
                    else s in
                let os_compv =
                    if not omit_subj then Syncat.create_bwd o_scompv np Syncat.Right
                    else o_scompv in
                Syncat.create_bwd os_compv comp Syncat.Right
          | O1V1O2V2S -> (* s/np0\np1/(s/np1) *)
                let ovcomp_s =
                    if not omit_subj then Syncat.create_fwd s np Syncat.Left
                    else s in
                let o_vcomps =
                    if not omit_obj then Syncat.create_bwd ovcomp_s np Syncat.Right
                    else ovcomp_s in
                Syncat.create_fwd o_vcomps comp Syncat.Left
          | O1O2V2V1S -> (*  s/np0\np1\(s\np1) *)
                let ocompv_s =
                    if not omit_subj then Syncat.create_fwd s np Syncat.Left
                    else s in
                let o_compvs =
                    if not omit_obj then Syncat.create_bwd ocompv_s np Syncat.Right
                    else ocompv_s in
                 Syncat.create_bwd o_compvs comp Syncat.Right
          | O1O2V1V2S -> (* s/np0\np1/(s/np1) *)
                let ovcomp_s =
                    if not omit_subj then Syncat.create_fwd s np Syncat.Left
                    else s in
                let o_vcomps =
                    if not omit_obj then Syncat.create_bwd ovcomp_s np Syncat.Right
                    else ovcomp_s in
                Syncat.create_fwd o_vcomps comp Syncat.Left
          | V1O1V2O2S -> (* s/np0/(s/np1)/np1 *)
                let vocomp_s =
                    if not omit_subj then Syncat.create_fwd s np Syncat.Left
                    else s in
                let vo_comps = Syncat.create_fwd vocomp_s comp Syncat.Left in
                if not omit_obj then Syncat.create_fwd vo_comps np Syncat.Left
                else vo_comps
          | V1V2O2O1S -> (* s/np0/np1/(s/np1) *)
                let vcompo_s =
                    if not omit_subj then Syncat.create_fwd s np Syncat.Left
                    else s in
                let vcomp_os =
                    if not omit_obj then Syncat.create_fwd vcompo_s np Syncat.Left
                    else vcompo_s in
                Syncat.create_fwd vcomp_os comp Syncat.Left
          | V1V2O1O2S -> (* s/np0/np1/(s/np1) *)
                let vcompo_s =
                    if not omit_subj then Syncat.create_fwd s np Syncat.Left
                    else s in
                let vcomp_os =
                    if not omit_obj then Syncat.create_fwd vcompo_s np Syncat.Left
                    else vcompo_s in
                Syncat.create_fwd vcomp_os comp Syncat.Left

    let aux_for_negwithaux_for_vi simp_5_02 omit_subj omit_vp =
        if not omit_vp then
            match simp_5_02 with
                AuxSNegV | AuxSVNeg | AuxNegSV | AuxNegVS | AuxVSNeg | AuxVNegS | NegAuxSV | NegAuxVS->
                    Syncat.create_fwd s s Syncat.Right
              | SAuxNegV | SAuxVNeg | SNegAuxV | NegSAuxV ->
                    let vi' = vi SVO omit_subj in
                    Syncat.create_fwd vi' vi' Syncat.Right
              | VAuxSNeg | VAuxNegS | NegVAuxS | VNegAuxS ->
                    let vi' = vi VSO omit_subj in
                    Syncat.create_bwd vi' vi' Syncat.Left
              | SNegVAux | SVNegAux | NegSVAux | NegVSAux | VSNegAux | VNegSAux | SVAuxNeg | VSAuxNeg ->
                    Syncat.create_bwd s s Syncat.Left
        else
            match simp_5_02 with
                AuxSNegV | AuxSVNeg | AuxVSNeg | VAuxSNeg ->
                    let auxs_neg = Syncat.create_fwd s neg Syncat.Left in
                    if not omit_subj then Syncat.create_fwd auxs_neg np Syncat.Left
                    else auxs_neg
              | AuxNegSV | AuxNegVS | AuxVNegS | VAuxNegS ->
                    let auxneg_s = 
                        if not omit_subj then Syncat.create_fwd s np Syncat.Left 
                        else s in
                    Syncat.create_fwd auxneg_s neg Syncat.Left
              | SAuxNegV | SAuxVNeg | SVAuxNeg | VSAuxNeg ->
                    let s_auxneg =
                        if not omit_subj then Syncat.create_bwd s np Syncat.Right
                        else s in
                    Syncat.create_fwd s_auxneg neg Syncat.Left
              | NegAuxSV | NegAuxVS | NegVAuxS | VNegAuxS ->
                    let negaux_s =
                        if not omit_subj then Syncat.create_fwd s np Syncat.Left
                        else s in
                    Syncat.create_bwd negaux_s neg Syncat.Right
              | SNegAuxV | SNegVAux | SVNegAux | VSNegAux ->
                    let s_negaux =
                        if not omit_subj then Syncat.create_bwd s np Syncat.Right
                        else s in
                    Syncat.create_fwd s_negaux neg Syncat.Right
              | NegSAuxV | NegSVAux | NegVSAux | VNegSAux ->
                    let neg_saux = Syncat.create_bwd s neg Syncat.Right in
                    if not omit_subj then Syncat.create_bwd neg_saux np Syncat.Right
                    else neg_saux

    let neg_for_negwithaux_for_vi simp_5_02 omit_subj omit_vp =
        if not omit_vp then
            let fund_ord_1_01 =
                match simp_5_02 with
                    AuxSNegV | AuxSVNeg | AuxNegSV | SAuxNegV 
                  | SAuxVNeg | NegAuxSV | SNegAuxV | SVAuxNeg 
                  | NegSAuxV | SNegVAux | SVNegAux | NegSVAux ->
                        SVO
                  | AuxNegVS | AuxVSNeg | AuxVNegS | NegAuxVS 
                  | VAuxSNeg | VAuxNegS | NegVAuxS | VSAuxNeg 
                  | VNegAuxS | NegVSAux | VSNegAux | VNegSAux ->
                        VOS in
            let vi' = vi fund_ord_1_01 omit_subj in
            match simp_5_02 with
                AuxSNegV ->
                    Syncat.create_fwd vi' vi' Syncat.Right
              | AuxSVNeg ->
                    Syncat.create_bwd vi' vi' Syncat.Left
              | AuxNegSV ->
                    Syncat.create_fwd s s Syncat.Right
              | AuxNegVS ->
                    Syncat.create_fwd s s Syncat.Right
              | AuxVSNeg ->
                    Syncat.create_bwd s s Syncat.Left
              | AuxVNegS ->
                    Syncat.create_bwd vi' vi' Syncat.Left
              | SAuxNegV ->
                    Syncat.create_fwd vi' vi' Syncat.Right
              | SAuxVNeg ->
                    Syncat.create_bwd vi' vi' Syncat.Left
              | NegAuxSV ->
                    Syncat.create_fwd s s Syncat.Right
              | NegAuxVS ->
                    Syncat.create_fwd s s Syncat.Right
              | VAuxSNeg ->
                    Syncat.create_bwd s s Syncat.Left
              | VAuxNegS ->
                    Syncat.create_bwd vi' vi' Syncat.Left
              | SNegAuxV ->
                    Syncat.create_fwd vi' vi' Syncat.Right
              | SVAuxNeg ->
                    Syncat.create_bwd vi' vi' Syncat.Left
              | NegSAuxV ->
                    Syncat.create_fwd s s Syncat.Right
              | NegVAuxS ->
                    Syncat.create_fwd s s Syncat.Right
              | VSAuxNeg ->
                    Syncat.create_bwd s s Syncat.Left
              | VNegAuxS ->
                    Syncat.create_bwd vi' vi' Syncat.Left
              | SNegVAux ->
                    Syncat.create_fwd vi' vi' Syncat.Right
              | SVNegAux ->
                    Syncat.create_bwd vi' vi' Syncat.Left
              | NegSVAux ->
                    Syncat.create_fwd s s Syncat.Right
              | NegVSAux ->
                    Syncat.create_fwd s s Syncat.Right
              | VSNegAux ->
                    Syncat.create_bwd s s Syncat.Left
              | VNegSAux ->
                    Syncat.create_bwd vi' vi' Syncat.Left
        else neg

    let neg_for_negwithoutaux simp_5_03 omit_subj omit_vp =
        let vi' =
            match simp_5_03 with
                SVNeg | SNegV | NegSV -> vi SVO omit_subj
              | NegVS | VSNeg | VNegS -> vi VOS omit_subj in
        if not omit_vp then
            match simp_5_03 with
                SVNeg | VNegS -> Syncat.create_bwd vi' vi' Syncat.Left
              | SNegV | NegVS -> Syncat.create_fwd vi' vi' Syncat.Right
              | NegSV -> Syncat.create_fwd s s Syncat.Right
              | VSNeg -> Syncat.create_bwd s s Syncat.Left
        else
            match simp_5_03 with
                SVNeg | SNegV | VSNeg -> Syncat.create_bwd s np Syncat.Right
              | NegSV | NegVS | VNegS -> Syncat.create_fwd s np Syncat.Left

    let aux_for_negwithaux_for_vt simp_5_05 omit_subj omit_obj omit_vt =
        let fund_1_01 =
            match simp_5_05 with
                AuxSNegVO | AuxSVNegO | AuxNegSVO | AuxNegVSO 
              | SAuxNegVO | SAuxVNegO | NegAuxSVO | SNegAuxVO 
              | SVAuxNegO | NegSAuxVO | NegVAuxSO | SNegVAuxO 
              | SVNegAuxO | NegSVAuxO | SAuxVONeg | SVAuxONeg 
              | SNegVOAux | SVNegOAux | NegSVOAux | SVOAuxNeg 
              | SVONegAux -> 
                    SVO
              | AuxVSNegO | AuxVNegSO | NegAuxVSO | VAuxSNegO 
              | VAuxNegSO | VSAuxNegO | VNegAuxSO | NegVSAuxO 
              | VSNegAuxO | VNegSAuxO | AuxNegVOS | AuxVSONeg 
              | VAuxSONeg | VSAuxONeg | NegVSOAux | VSNegOAux 
              | VNegSOAux | VSOAuxNeg | VSONegAux -> 
                    VSO
              | AuxSNegOV | AuxSVONeg | AuxNegSOV | SAuxNegOV 
              | NegAuxSOV | SNegAuxOV | NegSAuxOV | AuxSONegV 
              | AuxSOVNeg | SAuxONegV | SAuxOVNeg | SNegOAuxV 
              | NegSOAuxV | SNegOVAux | NegSOVAux | SOAuxNegV 
              | SOAuxVNeg | SONegAuxV | SOVAuxNeg | SONegVAux 
              | SOVNegAux -> 
                    SOV
              | AuxVNegOS | NegAuxVOS | VAuxNegOS | NegVAuxOS 
              | VNegAuxOS | AuxVOSNeg | AuxVONegS | VAuxOSNeg 
              | VAuxONegS | NegVOAuxS | VNegOAuxS | NegVOSAux 
              | VNegOSAux | VOAuxSNeg | VOAuxNegS | VOSAuxNeg 
              | VONegAuxS  | VOSNegAux | VONegSAux -> 
                    VOS
              | AuxNegOSV | AuxNegOVS | NegAuxOSV | AuxOSNegV 
              | AuxOSVNeg | AuxONegSV | NegOAuxSV | NegOSAuxV 
              | NegOSVAux | OAuxSNegV | OAuxSVNeg | OAuxNegSV 
              | OSAuxNegV | OSAuxVNeg | ONegAuxSV | OSNegAuxV 
              | OSVAuxNeg | ONegSAuxV | ONegSVAux -> 
                    OSV
              | NegAuxOVS | AuxONegVS | AuxOVSNeg | AuxOVNegS 
              | NegOAuxVS | NegOVAuxS | NegOVSAux | OAuxNegVS 
              | OAuxVSNeg | OAuxVNegS | ONegAuxVS | OVAuxSNeg 
              | OVAuxNegS | ONegVAuxS | OVSAuxNeg | OVNegAuxS 
              | OSNegVAux | OSVNegAux | ONegVSAux | OVSNegAux 
              | OVNegSAux -> 
                    OVS in
        let vi' = vi fund_1_01 omit_subj 
        and vt' = vt fund_1_01 omit_subj omit_obj in
        match simp_5_05 with
            AuxSNegVO | AuxSVNegO | AuxNegSVO | AuxNegVSO 
          | AuxVSNegO | AuxVNegSO | AuxSNegOV | AuxSVONeg 
          | AuxNegSOV | AuxNegVOS | AuxVSONeg | AuxVNegOS 
          | NegAuxSVO | NegAuxVSO | NegAuxSOV | NegAuxVOS 
          | AuxSONegV | AuxSOVNeg | AuxNegOSV | AuxNegOVS 
          | AuxVOSNeg | AuxVONegS | AuxOSNegV | AuxOSVNeg 
          | AuxONegSV | AuxONegVS | AuxOVSNeg | AuxOVNegS 
          | NegAuxOSV | NegAuxOVS ->
                Syncat.create_fwd s s Syncat.Right
          | SAuxNegVO | SAuxVNegO | SNegAuxVO | NegSAuxVO 
          | SAuxNegOV | SAuxVONeg | SNegAuxOV | NegSAuxOV 
          | SAuxONegV | SAuxOVNeg ->
                Syncat.create_fwd vi' vi' Syncat.Right
          | NegVOAuxS | VOAuxSNeg | VOAuxNegS | NegOVAuxS 
          | VONegAuxS | OVAuxSNeg | OVAuxNegS | ONegVAuxS 
          | OVNegAuxS | VNegOAuxS ->
                Syncat.create_bwd vi' vi' Syncat.Left
          | VAuxSNegO | VAuxNegSO | SVAuxNegO | NegVAuxSO 
          | VNegAuxSO | SNegVAuxO | SVNegAuxO | NegSVAuxO 
          | VAuxSONeg | VAuxNegOS | SVAuxONeg | NegVAuxOS 
          | VNegAuxOS | VAuxOSNeg | VAuxONegS ->
                Syncat.create_bwd vt' vt' Syncat.Left
          | SNegOAuxV | NegSOAuxV | SOAuxNegV | SOAuxVNeg 
          | NegOAuxVS | SONegAuxV | NegOSAuxV | OAuxVSNeg 
          | OAuxVNegS | OSNegAuxV | ONegSAuxV | OSAuxNegV 
          | OSAuxVNeg | ONegAuxVS ->
                Syncat.create_fwd vt' vt' Syncat.Right
          | VSAuxNegO | NegVSAuxO | VSNegAuxO | VNegSAuxO 
          | VSAuxONeg ->
                Syncat.create_bwd vi' vi' Syncat.Left
          | NegOAuxSV | OAuxSNegV | OAuxSVNeg | OAuxNegSV 
          | OAuxNegVS | ONegAuxSV ->
                Syncat.create_fwd vi' vi' Syncat.Right
          | OSNegVAux | OSVNegAux | ONegSVAux | ONegVSAux 
          | OVSNegAux | OVNegSAux | SNegVOAux | SVNegOAux 
          | NegSVOAux | NegVSOAux | VSNegOAux | VNegSOAux 
          | VSOAuxNeg | SNegOVAux | SVONegAux | NegSOVAux 
          | NegVOSAux | VSONegAux | VNegOSAux | SOVAuxNeg 
          | VOSAuxNeg | SONegVAux | SOVNegAux | NegOSVAux 
          | NegOVSAux | VOSNegAux | VONegSAux | SVOAuxNeg 
          | OSVAuxNeg | OVSAuxNeg ->
                Syncat.create_bwd s s Syncat.Left

    let neg_for_negwithaux_for_vt simp_5_05 omit_subj omit_obj omit_vt =
        let fund_1_01 =
            match simp_5_05 with
                AuxSNegVO | AuxSVNegO | AuxNegSVO | AuxNegVSO 
              | SAuxNegVO | SAuxVNegO | NegAuxSVO | SNegAuxVO 
              | SVAuxNegO | NegSAuxVO | NegVAuxSO | SNegVAuxO 
              | SVNegAuxO | NegSVAuxO | SAuxVONeg | SVAuxONeg 
              | SNegVOAux | SVNegOAux | NegSVOAux | SVOAuxNeg 
              | SVONegAux -> 
                    SVO
              | AuxVSNegO | AuxVNegSO | NegAuxVSO | VAuxSNegO 
              | VAuxNegSO | VSAuxNegO | VNegAuxSO | NegVSAuxO 
              | VSNegAuxO | VNegSAuxO | AuxNegVOS | AuxVSONeg 
              | VAuxSONeg | VSAuxONeg | NegVSOAux | VSNegOAux 
              | VNegSOAux | VSOAuxNeg | VSONegAux -> 
                    VSO
              | AuxSNegOV | AuxSVONeg | AuxNegSOV | SAuxNegOV 
              | NegAuxSOV | SNegAuxOV | NegSAuxOV | AuxSONegV 
              | AuxSOVNeg | SAuxONegV | SAuxOVNeg | SNegOAuxV 
              | NegSOAuxV | SNegOVAux | NegSOVAux | SOAuxNegV 
              | SOAuxVNeg | SONegAuxV | SOVAuxNeg | SONegVAux 
              | SOVNegAux -> 
                    SOV
              | AuxVNegOS | NegAuxVOS | VAuxNegOS | NegVAuxOS 
              | VNegAuxOS | AuxVOSNeg | AuxVONegS | VAuxOSNeg 
              | VAuxONegS | NegVOAuxS | VNegOAuxS | NegVOSAux 
              | VNegOSAux | VOAuxSNeg | VOAuxNegS | VOSAuxNeg 
              | VONegAuxS  | VOSNegAux | VONegSAux -> 
                    VOS
              | AuxNegOSV | AuxNegOVS | NegAuxOSV | AuxOSNegV 
              | AuxOSVNeg | AuxONegSV | NegOAuxSV | NegOSAuxV 
              | NegOSVAux | OAuxSNegV | OAuxSVNeg | OAuxNegSV 
              | OSAuxNegV | OSAuxVNeg | ONegAuxSV | OSNegAuxV 
              | OSVAuxNeg | ONegSAuxV | ONegSVAux -> 
                    OSV
              | NegAuxOVS | AuxONegVS | AuxOVSNeg | AuxOVNegS 
              | NegOAuxVS | NegOVAuxS | NegOVSAux | OAuxNegVS 
              | OAuxVSNeg | OAuxVNegS | ONegAuxVS | OVAuxSNeg 
              | OVAuxNegS | ONegVAuxS | OVSAuxNeg | OVNegAuxS 
              | OSNegVAux | OSVNegAux | ONegVSAux | OVSNegAux 
              | OVNegSAux -> 
                    OVS in
        let vi' = vi fund_1_01 omit_subj 
        and vt' = vt fund_1_01 omit_subj omit_obj in
        match simp_5_05 with
            AuxNegSVO | AuxNegVSO | NegAuxSVO | NegAuxVSO 
          | NegSAuxVO | NegVAuxSO | NegSVAuxO | NegVSAuxO 
          | AuxNegSOV | AuxNegVOS | NegAuxSOV | NegAuxVOS 
          | NegSAuxOV | NegVAuxOS | NegSVOAux | NegVSOAux 
          | AuxNegOSV | AuxNegOVS | NegAuxOSV | NegAuxOVS 
          | NegSOAuxV | NegVOAuxS | NegSOVAux | NegVOSAux 
          | NegOAuxSV | NegOAuxVS | NegOSAuxV | NegOVAuxS 
          | NegOSVAux | NegOVSAux ->
                Syncat.create_fwd s s Syncat.Right
          | VSONegAux | VOSNegAux | AuxSVONeg | AuxVSONeg 
          | SAuxVONeg | VAuxSONeg | SVAuxONeg | VSAuxONeg 
          | AuxSOVNeg | AuxVOSNeg | SAuxOVNeg | VAuxOSNeg 
          | SVOAuxNeg | VSOAuxNeg | SVONegAux | AuxOSVNeg 
          | AuxOVSNeg | VOAuxSNeg | SOVAuxNeg | VOSAuxNeg 
          | SOVNegAux | OAuxSVNeg | OAuxVSNeg | OSAuxVNeg 
          | OVAuxSNeg | OSVAuxNeg | OVSAuxNeg | OVSNegAux 
          | SOAuxVNeg ->
                Syncat.create_bwd s s Syncat.Left
          | AuxSNegVO | SAuxNegVO | SNegAuxVO | AuxSNegOV 
          | SAuxNegOV | SNegAuxOV | SNegVOAux | SNegOAuxV 
          | SNegOVAux | AuxONegSV | AuxONegVS | OAuxNegSV 
          | OAuxNegVS | ONegAuxSV | ONegAuxVS | ONegSAuxV 
          | ONegVAuxS | ONegSVAux | ONegVSAux | SNegVAuxO ->
                Syncat.create_fwd vi' vi' Syncat.Right
          | AuxSVNegO | VONegSAux | AuxVSNegO | SAuxVNegO 
          | VAuxSNegO | VSAuxNegO | VSNegAuxO | VSNegOAux 
          | AuxVONegS | OAuxVNegS | OVAuxNegS | AuxOVNegS 
          | VOAuxNegS | VONegAuxS | OVNegAuxS | OSVNegAux 
          | OVNegSAux | VAuxONegS ->
                Syncat.create_bwd vi' vi' Syncat.Left
          | AuxSONegV | SAuxONegV | OSAuxNegV | AuxOSNegV 
          | OAuxSNegV | OSNegAuxV | SOAuxNegV | SONegAuxV 
          | SONegVAux | OSNegVAux ->
                Syncat.create_fwd vt' vt' Syncat.Right
          | VNegOSAux | AuxVNegSO | VAuxNegSO | SVAuxNegO 
          | VNegAuxOS | SVNegOAux | VNegAuxSO | SVNegAuxO 
          | VNegSAuxO | AuxVNegOS | VAuxNegOS | VNegSOAux 
          | VNegOAuxS ->
                Syncat.create_bwd vt' vt' Syncat.Left

    let than_with_obl fund_ord_3_04 omit_subj compadj_order =
        let pp_for_vi = pp_attach_for_vi fund_ord_3_04 omit_subj in
        match compadj_order with
            OblThan ->
                Syncat.create_bwd pp_for_vi np Syncat.Right
          | ThanObl ->
                Syncat.create_fwd pp_for_vi np Syncat.Left

    let vi_subjctrl simp_9_02 omit_subj compmark =
        let comp =
            if not compmark then
                match simp_9_02 with
                    SV1V2O -> vi SVO false
                  | SV1OV2 -> vi SOV false
                  | SOV2V1 -> vi SOV false
                  | SOV1V2 -> vi SOV false
                  | V1SV2O -> vi VSO false
                  | V1V2OS -> vi VOS false
                  | V1V2SO -> vi VSO false
                  | OV2SV1 -> vi OSV false
                  | OSV2V1 -> vi OSV false
                  | V1OV2S -> vi VOS false
                  | OV2V1S -> vi OVS false
                  | OV1V2S -> vi OVS false
            else sbar in
        match simp_9_02 with
            SV1V2O -> (* s\np0/(s\np0) *)
                let s_vcomp =
                    if not omit_subj then Syncat.create_bwd s np Syncat.Right
                    else s in
                Syncat.create_fwd s_vcomp comp Syncat.Left
          | SV1OV2 -> (* s\np0/(s\np0) *)
                let s_vcomp =
                    if not omit_subj then Syncat.create_bwd s np Syncat.Right
                    else s in
                Syncat.create_fwd s_vcomp comp Syncat.Left
          | SOV2V1 -> (* s\np0\(s\np0) *)
                let s_compv =
                    if not omit_subj then Syncat.create_bwd s np Syncat.Right
                    else s in
                Syncat.create_bwd s_compv comp Syncat.Right
          | SOV1V2 -> (* s\np0/(s\np0) *)
                let s_vcomp =
                    if not omit_subj then Syncat.create_bwd s np Syncat.Right
                    else s in
                Syncat.create_fwd s_vcomp comp Syncat.Right
          | V1SV2O -> (* s/(s/np0)/np0 *)
                let vs_comp = Syncat.create_fwd s comp Syncat.Left in
                if not omit_subj then
                    Syncat.create_fwd vs_comp np Syncat.Left
                else vs_comp
          | V1V2OS -> (* s/np0/(s/np0) *)
                let vcomp_s =
                    if not omit_subj then Syncat.create_fwd s np Syncat.Left
                    else s in
                Syncat.create_fwd vcomp_s comp Syncat.Left
          | V1V2SO -> (* s/np0/(s/np0) *)
                let vcomp_s =
                    if not omit_subj then Syncat.create_fwd s np Syncat.Left
                    else s in
                Syncat.create_fwd vcomp_s comp Syncat.Left
          | OV2SV1 -> (* s\(s\np0)\np0 *)
                let comps_v =
                    if not omit_subj then Syncat.create_bwd s np Syncat.Right
                    else s in
                Syncat.create_bwd comps_v comp Syncat.Right
          | OSV2V1 -> (* s\(s\np0)\np0 *)
                let comps_v =
                    if not omit_subj then Syncat.create_bwd s np Syncat.Right
                    else s in
                Syncat.create_bwd comps_v comp Syncat.Right
          | V1OV2S -> (* s/(s/np0)/np0 *)
                let v_comps = Syncat.create_fwd s comp Syncat.Left in
                if not omit_subj then Syncat.create_fwd v_comps np Syncat.Left
                else v_comps
          | OV2V1S -> (* s/np0\(s/np0) *)
                let compv_s = 
                    if not omit_subj then Syncat.create_fwd s np Syncat.Left
                    else s in
                Syncat.create_bwd compv_s comp Syncat.Right
          | OV1V2S -> (* s/np0/(s/np0) *)
                let v_scomp =
                    if not omit_subj then Syncat.create_fwd s np Syncat.Left
                    else s in
                Syncat.create_fwd v_scomp comp Syncat.Left

    let infmarker_for_vi_serial simp_10_03 fund_ord_1_01 =
        let simp_4_02 =
            match simp_10_03 with
                SerialV_InfVO -> ObjCtrl_InfVO
              | SerialV_InfOV -> ObjCtrl_InfOV
              | SerialV_VOInf -> ObjCtrl_VOInf
              | SerialV_VInfO -> ObjCtrl_VInfO
              | SerialV_OInfV -> ObjCtrl_OInfV
              | SerialV_OVInf -> ObjCtrl_OVInf in
        infmarker_for_vi simp_4_02 fund_ord_1_01

    let infmarker_for_vt_serial simp_10_03 fund_ord_1_01 =
        let simp_4_02 =
            match simp_10_03 with
                SerialV_InfVO -> ObjCtrl_InfVO
              | SerialV_InfOV -> ObjCtrl_InfOV
              | SerialV_VOInf -> ObjCtrl_VOInf
              | SerialV_VInfO -> ObjCtrl_VInfO
              | SerialV_OInfV -> ObjCtrl_OInfV
              | SerialV_OVInf -> ObjCtrl_OVInf in
        infmarker_for_vt simp_4_02 fund_ord_1_01

    let subordconj simp_10_05 =
        match simp_10_05 with
            SubordCls_ConnSubordSent ->
                let cs_m = Syncat.create_fwd s s Syncat.Right in
                Syncat.create_fwd cs_m s Syncat.Left
          | SubordCls_ConnSentSubord ->
                let cm_s = Syncat.create_fwd s s Syncat.Left in
                Syncat.create_fwd cm_s s Syncat.Right
          | SubordCls_SubordConnSent ->
                let sc_m = Syncat.create_fwd s s Syncat.Right in
                Syncat.create_bwd sc_m s Syncat.Right
          | SubordCls_SubordSentConn ->
                let sm_c = Syncat.create_bwd s s Syncat.Left in
                Syncat.create_bwd sm_c s Syncat.Right
          | SubordCls_SentConnSubord ->
                let m_cs = Syncat.create_bwd s s Syncat.Left in
                Syncat.create_fwd m_cs s Syncat.Left
          | SubordCls_SentSubordConn ->
                let ms_c = Syncat.create_bwd s s Syncat.Right in
                Syncat.create_bwd ms_c s Syncat.Left

    let resclsconj simp_10_07 =
        match simp_10_07 with
            ResultCls_ConnResultSent ->
                let cs_m = Syncat.create_fwd s s Syncat.Right in
                Syncat.create_fwd cs_m s Syncat.Left
          | ResultCls_ConnSentResult ->
                let cm_s = Syncat.create_fwd s s Syncat.Left in
                Syncat.create_fwd cm_s s Syncat.Right
          | ResultCls_ResultConnSent ->
                let sc_m = Syncat.create_fwd s s Syncat.Right in
                Syncat.create_bwd sc_m s Syncat.Right
          | ResultCls_ResultSentConn ->
                let sm_c = Syncat.create_bwd s s Syncat.Left in
                Syncat.create_bwd sm_c s Syncat.Right
          | ResultCls_SentConnResult ->
                let m_cs = Syncat.create_bwd s s Syncat.Left in
                Syncat.create_fwd m_cs s Syncat.Left
          | ResultCls_SentResultConn ->
                let ms_c = Syncat.create_bwd s s Syncat.Right in
                Syncat.create_bwd ms_c s Syncat.Left

    let repspcconj simp_10_09 =
        match simp_10_09 with
            RepSpeech_SpeechSubordSent ->
                let cs_m = Syncat.create_fwd s s Syncat.Right in
                Syncat.create_fwd cs_m s Syncat.Left
          | RepSpeech_SpeechSentSubord ->
                let cm_s = Syncat.create_fwd s s Syncat.Left in
                Syncat.create_fwd cm_s s Syncat.Right
          | RepSpeech_SubordSpeechSent ->
                let sc_m = Syncat.create_fwd s s Syncat.Right in
                Syncat.create_bwd sc_m s Syncat.Right
          | RepSpeech_SubordSentSpeech ->
                let sm_c = Syncat.create_bwd s s Syncat.Left in
                Syncat.create_bwd sm_c s Syncat.Right
          | RepSpeech_SentSpeechSubord ->
                let m_cs = Syncat.create_bwd s s Syncat.Left in
                Syncat.create_fwd m_cs s Syncat.Left
          | RepSpeech_SentSubordSpeech ->
                let ms_c = Syncat.create_bwd s s Syncat.Right in
                Syncat.create_bwd ms_c s Syncat.Left

    let v_speech simp_10_10 omit_subj omit_speech =
        match simp_10_10 with
            RepSpeech_SVSpeech ->
                let fund_ord_1_01 = SVO in
                let vi' = vi fund_ord_1_01 omit_subj in
                if not omit_speech then Syncat.create_fwd vi' s Syncat.Left
                else vi'
          | RepSpeech_SSpeechV ->
                let fund_ord_1_01 = SOV in
                let vi' = vi fund_ord_1_01 omit_subj in
                if not omit_speech then Syncat.create_bwd vi' np Syncat.Right
                else vi'
          | RepSpeech_SpeechSV ->
                let fund_ord_1_01 = OSV in
                let vi' = vi fund_ord_1_01 omit_subj in
                if not omit_speech then Syncat.create_bwd vi' s Syncat.Right
                else vi'
          | RepSpeech_VSSpeech ->
                let fund_ord_1_01 = VSO in
                let vo = 
                    if not omit_speech then Syncat.create_fwd s s Syncat.Left 
                    else s in
                if not omit_subj then Syncat.create_fwd vo np Syncat.Left
                else vo
          | RepSpeech_VSpeechS ->
                let fund_ord_1_01 = VOS in
                let vi' = vi fund_ord_1_01 omit_subj in
                if not omit_speech then Syncat.create_fwd vi' s Syncat.Left
                else vi'
          | RepSpeech_SpeechVS ->
                let fund_ord_1_01 = OVS in
                let vi' = vi fund_ord_1_01 omit_subj in
                if not omit_speech then Syncat.create_bwd vi' s Syncat.Right
                else vi'

    let qmark_for_vi ques_1_02 fund_ord_1_01 omit_subj =
        let vi' = vi fund_ord_1_01 omit_subj in
        match ques_1_02 with
            QSVAdv | QSAdvV | QVSAdv | QVAdvS ->
                Syncat.create_fwd s s Syncat.Right
          | SQVAdv | SQAdvV ->
                Syncat.create_fwd vi' vi' Syncat.Right
          | VQSAdv | VQAdvS ->
                Syncat.create_bwd vi' vi' Syncat.Left
          | SVAdvQ | SAdvVQ | VSAdvQ | VAdvSQ ->
                Syncat.create_bwd s s Syncat.Left

    let qvi wh_question_order omit_subj =
        if not omit_subj then
            match wh_question_order with
                WHQ_SOV | WHQ_SVO | WHQ_OSV -> 
                    Syncat.create_bwd s np Syncat.Right
              | WHQ_VSO | WHQ_VOS | WHQ_OVS -> 
                    Syncat.create_fwd s np Syncat.Left
        else s

    let qvt wh_question_order omit_subj omit_obj =
        match wh_question_order with
            WHQ_SOV -> (* right branching preferred *) 
                let vi' = qvi wh_question_order omit_subj in
                if not omit_obj then Syncat.create_bwd vi' np Syncat.Right
                else vi'
          | WHQ_SVO -> (* right branching preferred *) 
                let vi' = qvi wh_question_order omit_subj in
                if not omit_obj then Syncat.create_fwd vi' np Syncat.Left
                else vi'
          | WHQ_VSO -> (* left branching preferred *) 
                let vo = 
                    if not omit_obj then Syncat.create_fwd s np Syncat.Left 
                    else s in
                if not omit_subj then Syncat.create_fwd vo np Syncat.Left
                else vo
          | WHQ_VOS -> (* left branching preferred *)
                let vi' = qvi wh_question_order omit_subj in
                if not omit_obj then Syncat.create_fwd vi' np Syncat.Left
                else vi'
          | WHQ_OVS -> (* left branching preferred *) 
                let vi' = qvi wh_question_order omit_subj in
                if not omit_obj then Syncat.create_bwd vi' np Syncat.Right
                else vi'
          | WHQ_OSV -> (* right branching preferred *) 
                let ov =
                    if not omit_obj then Syncat.create_bwd s np Syncat.Right
                    else s in
                if not omit_subj then Syncat.create_bwd ov np Syncat.Right
                else ov

    let qmark_for_vt ques_2_04 =
        match ques_2_04 with
            WHQ_QSVO | WHQ_QSOV | WHQ_QVSO | WHQ_QVOS | WHQ_QOSV | WHQ_QOVS ->
                Syncat.create_fwd s s Syncat.Right
          | WHQ_SQVO | WHQ_SQOV ->
                let vi' = vi SVO false in
                Syncat.create_fwd vi' vi' Syncat.Right
          | WHQ_VOQS | WHQ_OVQS ->
                let vi' = vi OVS false in
                Syncat.create_bwd vi' vi' Syncat.Left
          | WHQ_VQSO ->
                let vt' = vt VSO false false in
                Syncat.create_bwd vt' vt' Syncat.Left
          | WHQ_VQOS ->
                let vt' = vt VOS false false in
                Syncat.create_bwd vt' vt' Syncat.Left
          | WHQ_OQVS ->
                let vt' = vt OVS false false in
                Syncat.create_fwd vt' vt' Syncat.Right
          | WHQ_OSQV ->
                let vt' = vt OSV false false in
                Syncat.create_fwd vt' vt' Syncat.Right
          | WHQ_SOQV ->
                let vt' = vt SOV false false in
                Syncat.create_fwd vt' vt' Syncat.Right
          | WHQ_OQSV ->
                let vi' = vi SOV false in
                Syncat.create_fwd vi' vi' Syncat.Right
          | WHQ_SVQO ->
                let vt' = vt SVO false false in
                Syncat.create_bwd vt' vt' Syncat.Left
          | WHQ_VSQO ->
                let vi' = vi VSO false in
                Syncat.create_bwd vi' vi' Syncat.Left
          | WHQ_SVOQ | WHQ_SOVQ | WHQ_VSOQ | WHQ_VOSQ | WHQ_OSVQ | WHQ_OVSQ ->
                Syncat.create_bwd s s Syncat.Left

    (**********************************************************)

    let interpret_ans func ans allans =
        match ans with
            Some [] -> List.iter func allans
          | Some lst -> List.iter func lst
          | None -> ()

    let interpret_omit func omit =
        if not omit then func omit
        else List.iter func [true; false]

    let rec interpret_omitopts' func omits omitopts =
        match omitopts with
            [] ->
                func omits
          | omit :: rest ->
                if not omit then 
                    interpret_omitopts' func (false :: omits) rest
                else begin
                    interpret_omitopts' func (false :: omits) rest;
                    interpret_omitopts' func (true :: omits) rest
                end

    let interpret_omitopts func omitopts =
        interpret_omitopts' func [] omitopts

    type verb_system_params_t = {
        omit_subj          : bool;
        omit_obj           : bool;
        omit_iobj          : bool;
        omit_pp            : bool;
        svo_order          : fund_ord_1_01_ans_t;
        agtpat_distinction : fund_ord_1_02_ans_t;
        ditransitivity     : fund_ord_2_01_ans_t;
        vd_order           : fund_ord_2_02_ans_t;
        vtpp_order         : fund_ord_2_03_ans_t;
        vt_to_vd_transform : fund_ord_2_04_ans_t;
    }

    let gen_verb_system_vi macroexp verb_system_params =
        interpret_ans
            (fun fund_ord_1_01 ->
                interpret_omitopts
                    (fun [omit_subj] ->
                        let vi' = vi fund_ord_1_01 omit_subj in
                        MacroExpander.add_cats macroexp "[vi]" [vi'])
                    [verb_system_params.omit_subj])
            verb_system_params.svo_order fund_ord_1_01_all

    let gen_verb_system_vt macroexp verb_system_params =
        interpret_ans
            (fun fund_ord_1_01 ->
                interpret_omitopts
                    (fun [omit_subj; omit_obj] ->
                        let vt' = vt fund_ord_1_01 omit_subj omit_obj in
                        MacroExpander.add_cats macroexp "[vt]" [vt'])
                    [verb_system_params.omit_subj; 
                     verb_system_params.omit_obj])
            verb_system_params.svo_order fund_ord_1_01_all

    let gen_verb_system macroexp verb_system_params =
        (* intransitive verb *)
        gen_verb_system_vi macroexp verb_system_params;
        (* transitive verb *)
        gen_verb_system_vt macroexp verb_system_params;
        (* ditransitive verb *)
        interpret_ans
            (fun fund_ord_2_02 ->
                interpret_omitopts
                    (fun [omit_subj; omit_obj; omit_iobj] ->
                        let vd' = vd fund_ord_2_02 omit_subj omit_obj omit_iobj in
                        MacroExpander.add_cats macroexp "[vd]" [vd'])
                    [verb_system_params.omit_subj;
                     verb_system_params.omit_obj;
                     verb_system_params.omit_iobj])
            verb_system_params.vd_order fund_ord_2_02_all

    type adpos_system_params_t = {
        adposition         : fund_ord_3_01_ans_t;
        adpos_np_order     : fund_ord_3_02_ans_t;
        adposinfl_np_order : fund_ord_3_03_ans_t;
        adpos_vp_order     : fund_ord_3_04_ans_t;
        possessive         : fund_ord_4_01_ans_t;
        poss_order         : fund_ord_4_02_ans_t;
        possinfl_order     : fund_ord_4_03_ans_t;
    }

    type mod_system_params_t = {
        adj_type : fund_ord_5_01_ans_t;
        adj_order : fund_ord_5_02_ans_t;
        adj_cl_order : fund_ord_5_03_ans_t;
        adjcop_type : simp_7_01_ans_t;
        adjnegcop_type : simp_7_02_ans_t;
        nsent_type : simp_7_03_ans_t;
        ppsent_type : simp_7_04_ans_t;
        very_order : fund_ord_9_01_ans_t;
        tam_type : fund_ord_10_01_ans_t;
        tam_order : fund_ord_10_02_ans_t;
        neg_vi_type : simp_5_01_ans_t;
        neg_viaux_order : simp_5_02_ans_t;
        neg_vi_order : simp_5_03_ans_t;
        neg_vt_type : simp_5_04_ans_t;
        neg_vtaux_order : simp_5_05_ans_t;
        compadj_type : simp_8_01_ans_t;
        compadj_order : simp_8_02_ans_t;
    }

    type pronoun_system_params_t = {
        nompro_type : simp_1_01_ans_t;
        nompro_omit : simp_1_06_ans_t;
        nompro_order : simp_1_03_ans_t;
        accpro_type : simp_1_02_ans_t;
        accpro_omit : simp_1_07_ans_t;
        accpro_order : simp_1_04_ans_t;
        nomacc_omit : simp_1_08_ans_t;
        nomacc_order : simp_1_05_ans_t;
        dempro_type : fund_ord_6_01_ans_t;
        dempro_order : fund_ord_6_02_ans_t;
        relpro_type : fund_ord_8_01_ans_t;
        relpro_order : fund_ord_8_02_ans_t;
        relproinfl_order : fund_ord_8_03_ans_t;
        relproagg_order : fund_ord_8_04_ans_t;
        relproext_order : fund_ord_8_05_ans_t;
        relprorep_order : fund_ord_8_06_ans_t;
        relprocl_order  : fund_ord_8_07_ans_t;
        reflpro_type : simp_2_01_ans_t;
        reflpro_order : simp_2_02_ans_t;
        recipro_type : simp_2_03_ans_t;
        recipro_order : simp_2_04_ans_t;
        recipro_reflpro_diff : simp_2_05_ans_t;
        refladv_type : simp_2_06_ans_t;
        refladv_order : simp_2_07_ans_t;
    }

    type number_system_params_t = {
        number_type : fund_ord_7_01_ans_t;
        num_order : fund_ord_7_02_ans_t;
        numcl_order : fund_ord_7_03_ans_t;
    }

    type marker_system_params_t = {
        passive_type : simp_3_01_ans_t;
        passive_order : simp_3_02_ans_t;
        passiveinfl_type : simp_3_03_ans_t;
        passiveinfl_order : simp_3_04_ans_t;
        passive_obl : simp_3_09_ans_t;
        passive_agtpat_diff : simp_3_10_ans_t;
        passive_agtmarker : simp_3_11_ans_t;
        passive_patmatker : simp_3_12_ans_t;
    }

    type control_system_params_t = {
        objctrl_inf_type : simp_4_01_ans_t;
        objctrl_inf_order : simp_4_02_ans_t;
        objctrl_order : simp_4_03_ans_t;
        sbjctrl_inf_type : simp_9_01_ans_t;
        sbjctrl_inf_order : simp_9_02_ans_t;
        srlv_type : simp_10_01_ans_t;
        srlv_order : simp_10_02_ans_t;
        srlv_inf_order : simp_10_03_ans_t;
    }

    type conj_system_params_t = {
        subord_type : simp_10_04_ans_t;
        subord_order : simp_10_05_ans_t;
        result_type : simp_10_06_ans_t;
        result_order : simp_10_07_ans_t;
        repspc_type : simp_10_08_ans_t;
        repspc_marker_order : simp_10_09_ans_t;
        repspc_order : simp_10_10_ans_t;
    }

    type ques_system_params_t = {
        polar_type : ques_1_01_ans_t;
        polar_order : ques_1_02_ans_t;
        polar_marker_pos : ques_1_03_ans_t;
        polar_affdrop_pos : ques_1_04_ans_t;
        polar_order_adv : ques_1_06_ans_t;
        what_ques : ques_2_01;
        who_ques : ques_2_02;
        howmany_ques : ques_2_03;
        wh_aux_ques : ques_2_04_ans_t;
    }

end
