(***********************************************************************)
(*                                                                     *)
(*                           Objective Caml                            *)
(*                                                                     *)
(*            Luc Maranget, projet Moscova, INRIA Rocquencourt         *)
(*                                                                     *)
(*  Copyright 2004 Institut National de Recherche en Informatique et   *)
(*  en Automatique.  All rights reserved.  This file is distributed    *)
(*  under the terms of the Q Public License version 1.0.               *)
(*                                                                     *)
(*  Version 1.1: Modified by Prachya Boonkwan (September, 2011).       *)
(*      - Migrated the code to JoinPool.Shared to enable task          *)
(*        interruption.                                                *)
(*                                                                     *)
(***********************************************************************)

(* $Id$ *)

module type Problem = sig
  
    val identifier : string
  
    type init
  
    type client_data

    type collection

    type enum
  
    type input
  
    type key
  
    type value
  
    type output

    val enum_start : collection -> enum

    val enum_step : enum -> (input * enum) option
  
    val init_client : init -> client_data
  
    val compare_keys : key -> key -> int
  
    val map : client_data -> input -> (key * value) list
  
    val combine : value -> value -> value
  
    val reduce : key -> value -> output -> output
  
end

module type S = sig
  
    type init
  
    type input
  
    type output

    type collection
  
    type server_t

    val client : JoinHelper.configuration -> unit

    val server :
        JoinHelper.configuration -> init -> server_t
        
    val compute : 
        server_t -> collection -> output -> output
  
end

module Make (P : Problem) : S
    with type input = P.input 
    and type output = P.output 
    and type init = P.init
    and type collection = P.collection
= struct
  
    let register_id = String.copy P.identifier
  
    type init = P.init
  
    type input = P.input
  
    type output = P.output

    type collection = P.collection
  
    module Config = struct
        let debug = false
        let nagain = 0
    end
  
    module Enumerable = struct
        type t = P.collection
        type elt = P.input
        type enum = P.enum

        let start collection =
            P.enum_start collection

        let step state =
            P.enum_step state
    end

    module Pool = JoinPool.Shared.Make(Config)(Enumerable)
  
    module M = Map.Make (struct type t = P.key let compare = P.compare_keys end)

    type server_t = ((M.key * P.value) list, P.value M.t) Pool.t
  
    let kill_pids pids =
        match pids with
            [] -> ()
          | pid :: rest ->
                Unix.kill pid Sys.sigkill

    let client cfg =
        let _, _, (
            i, 
            register : (P.init * ((P.input -> (P.key * P.value) list)) Join.chan)
        ) =
            JoinHelper.init_client_with_lookup
                ~at_fail:JoinHelper.exit_at_fail
                ~lookup:(JoinHelper.lookup_times 0 1.0)
                cfg
                register_id in
        let cd = P.init_client i in
        def work(x) = reply (P.map cd x) to work in
        spawn register(work);
        JoinHelper.wait_forever ()
  
    let server cfg i =
        let pool = Pool.create () in
        let _ =
            JoinHelper.init_server_with_register cfg register_id
                (i, pool.Pool.register) in
        pool

    let compute svr items z =
        let comb l m =
            List.fold_left
                (fun m (k, v) ->
                    let curr = 
                        try Some (M.find k m) 
                        with Not_found -> None in
                    let v' = 
                        match curr with 
                            Some x -> P.combine x v
                          | None -> v in
                    M.add k v' m)
                m l in
        let map = svr.Pool.fold items comb M.empty in
        let result : output = M.fold P.reduce map z in
        result

end
