(***********************************************************
*                                                          *
*      FUNGI - Fast UNsupervised Grammar Induction         *
*                                                          *
*   Originator: Prachya Boonkwan [1], [2]                  *
*               [1] School of Informatics,                 *
*                   University of Edinburgh, UK            *
*               [2] National Electronics and Computer      *
*                   Technology Center (NECTEC), Thailand   *
*                                                          *
*   Copyright (C) 2011 Prachya Boonkwan                    *
*   <p.boonkwan@sms.ed.ac.uk, kaamanita@hotmail.com>       *
*                                                          *
*   This library is free software; you can redistribute    *
*   it and/or modify it under the terms of the GNU         *
*   Lesser General Public License as published by the      *
*   Free Software Foundation; either version 2.1 of the    *
*   License, or (at your option) any later version.        *
*                                                          *
*   This library is distributed in the hope that it will   *
*   be useful, but WITHOUT ANY WARRANTY; without even      *
*   the implied warranty of MERCHANTABILITY or FITNESS     *
*   FOR A PARTICULAR PURPOSE.  See the GNU Lesser          *
*   General Public License for more details.               *
*                                                          *
*   You should have received a copy of the GNU Lesser      *
*   General Public License along with this library; if     *
*   not, write to the Free Software Foundation, Inc.,      *
*   59 Temple Place, Suite 330, Boston, MA  02111-1307     *
*   USA.                                                   *
*                                                          *
***********************************************************)

open Printf

open Utils

type flag_t = ReadOnly | WriteOnly | ReadWrite | Create

module type MARSHALABLE_TYPE = sig
    type t
    val marshal : t -> string
    val unmarshal : ?begpos:int -> string -> t
end

module MDB
    (DataKey : MARSHALABLE_TYPE) 
    (DataVal : MARSHALABLE_TYPE) (MetaVal : MARSHALABLE_TYPE)
= struct

    type datakey_t = DataKey.t

    type dataval_t = DataVal.t

    type metaval_t = MetaVal.t

    type key_t =
        Key_data of datakey_t
      | Key_meta of string

    type value_t =
        Value_data of dataval_t
      | Value_meta of metaval_t

    module Key = struct

        type t = key_t

        let marshal key = 
            match key with
                Key_data k -> sprintf "D%s" (DataKey.marshal k)
              | Key_meta k -> sprintf "M%s" k

        let unmarshal key =
            let len = String.length key - 1 in
            match key.[0] with
                'D' -> 
                    let content = DataKey.unmarshal ~begpos:1 key in
                    Key_data content
              | 'M' -> Key_meta (String.sub key 1 len)
              | _ -> failwith "Key.unmarshal"
                
    end

    module Value = struct

        type t = value_t

        let marshal value =
            match value with
                Value_data v -> sprintf "D%s" (DataVal.marshal v)
              | Value_meta v -> sprintf "M%s" (MetaVal.marshal v)

        let unmarshal value =
            match value.[0] with
                'D' -> 
                    let content = DataVal.unmarshal ~begpos:1 value in
                    Value_data content
              | 'M' -> 
                    let content = MetaVal.unmarshal ~begpos:1 value in
                    Value_meta content
              | _ -> failwith "Value.unmarshal"

    end

    type t = Dbm.t

    let rec conv_flags flags =
        match flags with
            [] -> []
          | ReadOnly :: rest-> Dbm.Dbm_rdonly :: (conv_flags rest)
          | WriteOnly :: rest -> Dbm.Dbm_wronly :: (conv_flags rest)
          | ReadWrite :: rest -> Dbm.Dbm_rdwr :: (conv_flags rest)
          | Create :: rest -> Dbm.Dbm_create :: (conv_flags rest)

    let opendb ?(dbperm = 0o644) fname flags =
        Dbm.opendbm fname (conv_flags flags) dbperm

    let close db =
        Dbm.close db

    let marshal_datakey datakey =
        Key.marshal (Key_data datakey)

    let marshal_metakey metakey =
        Key.marshal (Key_meta metakey)

    let unmarshal_datakey mstr =
        match Key.unmarshal mstr with
            Key_data datakey -> datakey
          | Key_meta _ -> failwith "unmarshal_datakey"

    let unmarshal_metakey mstr =
        match Key.unmarshal mstr with
            Key_meta metakey -> metakey
          | Key_data _ -> failwith "unmarshal_metakey"

    let unmarshal_dataval mstr =
        match Value.unmarshal mstr with
            Value_data dataval -> dataval
          | Value_meta _ -> 
                failwith "unmarshal_dataval"

    let unmarshal_metaval mstr =
        match Value.unmarshal mstr with
            Value_meta metaval -> metaval
          | Value_data _ -> 
                failwith "unmarshal_metaval"

    let mem_data db datakey =
        let mkey = marshal_datakey datakey in
        try
            let _ = Dbm.find db mkey in
            true
        with Not_found -> false

    let mem_meta db metakey =
        let mkey = marshal_metakey metakey in
        try
            let _ = Dbm.find db mkey in
            true
        with Not_found -> false

    let find_data db datakey =
        let mstr = Dbm.find db (marshal_datakey datakey) in
        unmarshal_dataval mstr

    let find_meta db metakey =
        let mstr = Dbm.find db (marshal_metakey metakey) in
        unmarshal_metaval mstr

    let add_data db datakey dataval =
        Dbm.add db 
            (marshal_datakey datakey) 
            (Value.marshal (Value_data dataval))

    let add_meta db metakey metaval =
        Dbm.add db 
            (marshal_metakey metakey) 
            (Value.marshal (Value_meta metaval))

    let replace_data db datakey dataval =
        Dbm.replace db 
            (marshal_datakey datakey) 
            (Value.marshal (Value_data dataval))

    let replace_meta db metakey metaval =
        Dbm.replace db 
            (marshal_metakey metakey) 
            (Value.marshal (Value_meta metaval))

    let remove_data db datakey =
        Dbm.remove db (marshal_datakey datakey)

    let remove_meta db metakey =
        Dbm.remove db (marshal_metakey metakey)

    let iter_data fn db =
        Dbm.iter 
            (fun k v -> 
                if k.[0] = 'D' then
                    let key = unmarshal_datakey k
                    and value = unmarshal_dataval v in
                    fn key value
                else ())
            db

    let iter_meta fn db =
        Dbm.iter 
            (fun k v -> 
                if k.[0] = 'M' then
                    let key = unmarshal_metakey k
                    and value = unmarshal_metaval v in
                    fn key value
                else ())
            db

end

module MarshalableInt = struct

    type t = int

    let marshal value = 
        string_of_int value

    let unmarshal ?(begpos = 0) mstr =
        let len = String.length mstr in
        int_of_string (String.sub mstr begpos (len - begpos))

end

module SeqMDB (DataVal : MARSHALABLE_TYPE) = struct

    type datakey_t = int

    type dataval_t = DataVal.t

    type metaval_t = int

    module MDB = MDB(MarshalableInt)(DataVal)(MarshalableInt)

    type t = MDB.t

    let opendb ?(dbperm = 0o644) fname flags =
        let db = MDB.opendb fname flags ~dbperm:dbperm in
        if not (MDB.mem_meta db "itemno") then
            MDB.replace_meta db "itemno" 0
        else ();
        db

    let close = MDB.close

    let mem = MDB.mem_data
    
    let mem_meta = MDB.mem_meta

    let find = MDB.find_data

    let find_meta = MDB.find_meta

    let length db =
        if mem_meta db "itemno" then
            find_meta db "itemno"
        else raise Not_found

    let replace = MDB.replace_data
    
    let replace_meta = MDB.replace_meta

    let add db data =
        let curcnt = length db in
        let newcnt = curcnt + 1 in
        replace db curcnt data;
        replace_meta db "itemno" newcnt

    let remove = MDB.remove_data

    let iter_data fn db =
        let itemno = find_meta db "itemno" in
        for datakey = 0 to itemno - 1 do
            try
                let dataval = find db datakey in
                fn datakey dataval
            with Not_found -> ()
        done

    let iter_meta fn db =
        let itemno = find_meta db "itemno" in
        for datakey = 0 to itemno - 1 do
            if mem db datakey then
                fn datakey
            else ()
        done

end

module HashedInt = struct
    type t = int
    let equal = Pervasives.(=)
    let hash = Hashtbl.hash
end

module SeqMDBWrapper (DataVal : MARSHALABLE_TYPE) = struct

    module SeqMDB = SeqMDB(DataVal)

    type datakey_t = SeqMDB.datakey_t

    type dataval_t = SeqMDB.dataval_t

    type metaval_t = SeqMDB.metaval_t

    type seqmdb_t = SeqMDB.t

    type t = {
        seqmdbs  : seqmdb_t array;
        nodbs    : int;
        noitems  : int;
        rangetbl : int array;
    }

    let init wrapper =
        let sum = ref 0 in
        for idx = 0 to wrapper.nodbs - 1 do
            wrapper.rangetbl.(idx) <- !sum;
            sum := !sum + SeqMDB.length wrapper.seqmdbs.(idx)
        done

    let create seqmdbs =
        let seqmdb_arr = Array.of_list seqmdbs in
        let nodbs = List.length seqmdbs in
        let noitems = List.fold_left
            (fun v seqmdb -> v + SeqMDB.length seqmdb) 
            0 seqmdbs in
        let wrapper = {
            seqmdbs  = seqmdb_arr;
            nodbs    = nodbs;
            noitems  = noitems;
            rangetbl = Array.make nodbs 0;
        } in
        init wrapper;
        wrapper

    let opendbs ?(dbperm = 0o644) fnames flags =
        let seqmdbs = List.map 
            (fun fname -> SeqMDB.opendb fname flags ~dbperm:dbperm)
            fnames in
        create seqmdbs

    let get_nodbs wrapper =
        wrapper.nodbs

    let get_noitems wrapper =
        wrapper.noitems

    let nth wrapper dbid =
        wrapper.seqmdbs.(dbid)

    let close wrapper =
        Array.iter (SeqMDB.close) wrapper.seqmdbs

    let calc_relidx wrapper idx =
        let i = ref 0 in
        while !i < wrapper.nodbs - 1 && idx > wrapper.rangetbl.(!i) do
            incr i
        done;
        (!i, idx - wrapper.rangetbl.(!i))

    let mem wrapper idx =
        let (dbid, relidx) = calc_relidx wrapper idx in
        SeqMDB.mem wrapper.seqmdbs.(dbid) relidx

    let mem_meta wrapper dbid metakey =
        SeqMDB.mem wrapper.seqmdbs.(dbid) metakey

    let find wrapper idx =
        let (dbid, relidx) = calc_relidx wrapper idx in
        SeqMDB.find wrapper.seqmdbs.(dbid) relidx

    let find_meta wrapper dbid metakey =
        SeqMDB.find_meta wrapper.seqmdbs.(dbid) metakey

    let length wrapper =
        wrapper.noitems

    let replace wrapper idx dataval =
        let (dbid, relidx) = calc_relidx wrapper idx in
        SeqMDB.replace wrapper.seqmdbs.(dbid) relidx dataval

    let replace_meta wrapper dbid metakey metaval =
        SeqMDB.replace_meta wrapper.seqmdbs.(dbid) metakey metaval

    let iter_data fn wrapper =
        for dbid = 0 to wrapper.nodbs - 1 do
            SeqMDB.iter_data fn wrapper.seqmdbs.(dbid)
        done

    let iter_meta fn wrapper =
        for dbid = 0 to wrapper.nodbs - 1 do
            SeqMDB.iter_meta fn wrapper.seqmdbs.(dbid)
        done

end
