(** Marshalable Database *)

(***********************************************************
*                                                          *
*      FUNGI - Fast UNsupervised Grammar Induction         *
*                                                          *
*   Originator: Prachya Boonkwan [1], [2]                  *
*               [1] School of Informatics,                 *
*                   University of Edinburgh, UK            *
*               [2] National Electronics and Computer      *
*                   Technology Center (NECTEC), Thailand   *
*                                                          *
*   Copyright (C) 2011 Prachya Boonkwan                    *
*   <p.boonkwan@sms.ed.ac.uk, kaamanita@hotmail.com>       *
*                                                          *
*   This library is free software; you can redistribute    *
*   it and/or modify it under the terms of the GNU         *
*   Lesser General Public License as published by the      *
*   Free Software Foundation; either version 2.1 of the    *
*   License, or (at your option) any later version.        *
*                                                          *
*   This library is distributed in the hope that it will   *
*   be useful, but WITHOUT ANY WARRANTY; without even      *
*   the implied warranty of MERCHANTABILITY or FITNESS     *
*   FOR A PARTICULAR PURPOSE.  See the GNU Lesser          *
*   General Public License for more details.               *
*                                                          *
*   You should have received a copy of the GNU Lesser      *
*   General Public License along with this library; if     *
*   not, write to the Free Software Foundation, Inc.,      *
*   59 Temple Place, Suite 330, Boston, MA  02111-1307     *
*   USA.                                                   *
*                                                          *
***********************************************************)

(** Database opening flag *)
type flag_t = 
    ReadOnly        (** Read only *)
  | WriteOnly       (** Write only *)
  | ReadWrite       (** Read and write *)
  | Create          (** Write if found; otherwise create a new one *)
    
(** Marshalable Type Module Signature *)
module type MARSHALABLE_TYPE =
sig
    (** Type *)
    type t
    val marshal : t -> string
    val unmarshal : ?begpos:int -> string -> t
end

(** Marshaled Database with Metadata Annotation *)
module MDB :
functor (DataKey : MARSHALABLE_TYPE) ->
functor (DataVal : MARSHALABLE_TYPE) ->
functor (MetaVal : MARSHALABLE_TYPE) ->
sig

    (** Data key *)
    type datakey_t = DataKey.t

    (** Data value *)
    type dataval_t = DataVal.t
   
    (** Metadata value *)
    type metaval_t = MetaVal.t
    
    (** Database *)
    type t
    
    (** Open the database file with the flags and the Unix-based 
        file permission (default = 0o644). *)
    val opendb : ?dbperm:int -> string -> flag_t list -> t
    
    (** Close the database. *)
    val close : t -> unit

    (** Check if a data key is in the database. *)
    val mem_data : t -> datakey_t -> bool

    (** Check if a metadata key is in the database. *)
    val mem_meta : t -> string -> bool
    
    (** Find the value given the data key. *)
    val find_data : t -> datakey_t -> dataval_t
    
    (** Find the value given the metadata key. *)
    val find_meta : t -> string -> metaval_t
    
    (** Add a new datakey-value pair to the database. *)
    val add_data : t -> datakey_t -> dataval_t -> unit
    
    (** Add a new metakey-value pair to the database. *)
    val add_meta : t -> string -> metaval_t -> unit
    
    (** Replace the datakey-value pair to the database. *)
    val replace_data : t -> datakey_t -> dataval_t -> unit
    
    (** Replace the metakey-value pair to the database. *)
    val replace_meta : t -> string -> metaval_t -> unit
    
    (** Remove the key from the database. *)
    val remove_data : t -> datakey_t -> unit
    
    (** Remove the key from the database. *)
    val remove_meta : t -> string -> unit
    
    (** Iterate over all datakey-value pairs in the database with 
        the function. *)
    val iter_data : (datakey_t -> dataval_t -> unit) -> t -> unit
    
    (** Iterate over all metakey-value pairs in the database with 
        the function. *)
    val iter_meta : (string -> metaval_t -> unit) -> t -> unit
end

module MarshalableInt :
sig
    type t = int
    val marshal : int -> string
    val unmarshal : ?begpos:int -> string -> int
end

(** Sequential Marshaled Database *)
module SeqMDB :
functor (DataVal : MARSHALABLE_TYPE) ->
sig
    
    (** Data key *)
    type datakey_t = int
    
    (** Data value *)
    type dataval_t = DataVal.t
    
    (** Metadata value *)
    type metaval_t = int
    
    (** Database *)
    type t
    
    (** Open the database file with the flags and the Unix-based 
        file permission (default = 0o644). The key to items is
        an integer starting from 0. *)
    val opendb : ?dbperm:int -> string -> flag_t list -> t
    
    (** Close the database. *)
    val close : t -> unit
    
    (** Check if a data key is in the database. *)
    val mem : t -> int -> bool
    
    (** Find the value given the data key. *)
    val find : t -> int -> dataval_t
    
    (** Find the number of items in the database. *)
    val length : t -> int
    
    (** Add a new value to the database. *)
    val add : t -> dataval_t -> unit
    
    (** Replace the value at the specified position. *)
    val replace : t -> int -> dataval_t -> unit
    
    (** Remove the value at the specified position. *)
    val remove : t -> int -> unit
    
    (** Iterate over all index-value pairs in the database with 
        the function. *)
    val iter_data : (int -> dataval_t -> unit) -> t -> unit
    
    (** Iterate over all indices in the database with 
        the function. *)
    val iter_meta : (int -> unit) -> t -> unit
end

module SeqMDBWrapper :
functor (DataVal : MARSHALABLE_TYPE) ->
sig
    type datakey_t = SeqMDB(DataVal).datakey_t
    type dataval_t = SeqMDB(DataVal).dataval_t
    type metaval_t = SeqMDB(DataVal).metaval_t
    type seqmdb_t = SeqMDB(DataVal).t
    type t
    val opendbs : ?dbperm:int -> string list -> flag_t list -> t
    val get_noitems : t -> int
    val close : t -> unit
    val calc_relidx : t -> int -> int * int
    val mem : t -> int -> bool
    val mem_meta : t -> int -> int -> bool
    val find : t -> int -> dataval_t
    val find_meta : t -> int -> string -> metaval_t
    val length : t -> int
    val replace : t -> int -> dataval_t -> unit
    val replace_meta : t -> int -> string -> metaval_t -> unit
    val iter_data : (int -> dataval_t -> unit) -> t -> unit
    val iter_meta : (int -> unit) -> t -> unit
end
