(** Probability Models for Syntactic Derivation.

    There are four different probabilities available in this module.
    {ol
        {- PCFG: P(tree) = product of P(rule) for all rules constructing the tree.}
        {- Role-emission probability: P(role | category, word), where role can be 'head,' 'dependent,' or 'connective.'}
        {- Lexicon-emission probability: P(word | category) for all preterminal nodes.}
        {- Head-emission probability: P(word | category) for all nodes.}
    }
*)

(***********************************************************
*                                                          *
*      FUNGI - Fast UNsupervised Grammar Induction         *
*                                                          *
*   Originator: Prachya Boonkwan [1], [2]                  *
*               [1] School of Informatics,                 *
*                   University of Edinburgh, UK            *
*               [2] National Electronics and Computer      *
*                   Technology Center (NECTEC), Thailand   *
*                                                          *
*   Copyright (C) 2011 Prachya Boonkwan                    *
*   <p.boonkwan@sms.ed.ac.uk, kaamanita@hotmail.com>       *
*                                                          *
*   This library is free software; you can redistribute    *
*   it and/or modify it under the terms of the GNU         *
*   Lesser General Public License as published by the      *
*   Free Software Foundation; either version 2.1 of the    *
*   License, or (at your option) any later version.        *
*                                                          *
*   This library is distributed in the hope that it will   *
*   be useful, but WITHOUT ANY WARRANTY; without even      *
*   the implied warranty of MERCHANTABILITY or FITNESS     *
*   FOR A PARTICULAR PURPOSE.  See the GNU Lesser          *
*   General Public License for more details.               *
*                                                          *
*   You should have received a copy of the GNU Lesser      *
*   General Public License along with this library; if     *
*   not, write to the Free Software Foundation, Inc.,      *
*   59 Temple Place, Suite 330, Boston, MA  02111-1307     *
*   USA.                                                   *
*                                                          *
***********************************************************)

type probflag_t = RoleEmis | JuliaEmis | LexEmis | HeadEmis

(** Create PCFG. *)
val create_pcfg : int -> float -> Deriv.DerivGenModel.t

(** Create Model 1 (PCFG). *)
val create_model1 : int -> float -> Deriv.DerivGenModel.t

(** Create Model 2 (PCFG + role-emission probability). *)
val create_model2 : int -> float -> Deriv.DerivGenModel.t

(** Create Model 3 (PCFG + lexicon-emission probability). *)
val create_model3 : int -> float -> Deriv.DerivGenModel.t

(** Create Model 4 (PCFG + role-emission probability + 
    lexicon-emission probability). *)
val create_model4 : int -> float -> Deriv.DerivGenModel.t

(** Create Model 5 (PCFG + head-emission probability). *)
val create_model5 : int -> float -> Deriv.DerivGenModel.t

(** Create Model 6 (PCFG + role-emission probability + 
    head-emission probability). *)
val create_model6 : int -> float -> Deriv.DerivGenModel.t

val conv_probflags : probflag_t list -> int -> float -> Deriv.DerivGenModel.probftr_t list

val create_model : int -> float -> probflag_t list -> Deriv.DerivGenModel.t
