(***********************************************************
*                                                          *
*      FUNGI - Fast UNsupervised Grammar Induction         *
*                                                          *
*   Originator: Prachya Boonkwan [1], [2]                  *
*               [1] School of Informatics,                 *
*                   University of Edinburgh, UK            *
*               [2] National Electronics and Computer      *
*                   Technology Center (NECTEC), Thailand   *
*                                                          *
*   Copyright (C) 2011 Prachya Boonkwan                    *
*   <p.boonkwan@sms.ed.ac.uk, kaamanita@hotmail.com>       *
*                                                          *
*   This library is free software; you can redistribute    *
*   it and/or modify it under the terms of the GNU         *
*   Lesser General Public License as published by the      *
*   Free Software Foundation; either version 2.1 of the    *
*   License, or (at your option) any later version.        *
*                                                          *
*   This library is distributed in the hope that it will   *
*   be useful, but WITHOUT ANY WARRANTY; without even      *
*   the implied warranty of MERCHANTABILITY or FITNESS     *
*   FOR A PARTICULAR PURPOSE.  See the GNU Lesser          *
*   General Public License for more details.               *
*                                                          *
*   You should have received a copy of the GNU Lesser      *
*   General Public License along with this library; if     *
*   not, write to the Free Software Foundation, Inc.,      *
*   59 Temple Place, Suite 330, Boston, MA  02111-1307     *
*   USA.                                                   *
*                                                          *
***********************************************************)

open Cg
open Dict
open Dictloader
open Chart
open Cky
open Chartdb
open Utils
open Printf

module ParserRunner = struct

    type syncat_t = Syncat.t

    type t = {
        dict     : Dict.t;
        chartdb  : ChartDb.t;
        cdbfname : string;
        fnames   : string list;
        start    : syncat_t;
        depstrs  : Dep.DepStruct.t array option;
    }

    let create dictfname cdbfname fnames start golddeps conlldeps =
        let depstrs = 
            if List.length golddeps = 0 && List.length conlldeps = 0 then
                None
            else if List.length golddeps = 0 then
                Some (Array.of_list (Conlldep.ConllDep.from_files conlldeps))
            else if List.length conlldeps = 0 then begin
                let treedb = Bank.DepBank.from_texts golddeps in
                Some (Array.of_list (List.map Tree.Tree.to_depstr treedb))
            end else None
        in
        {
            dict     = DictLoader.load dictfname false;
            chartdb  = ChartDb.create cdbfname;
            cdbfname = cdbfname;
            fnames   = fnames;
            start    = start;
            depstrs  = depstrs;
        }

    let tryparse_elt dict sent enumlevel start depstr =
        Dict.set_enumlevel dict enumlevel;
        let chart = 
            CKY.parse 
                ~transf:(Dict.get_unarydrv dict)
                ~depstr:depstr
                sent dict start in
        (chart, Chart.is_success chart)
    
    let rec tryparse dict sent enumlevels start depstr result =
        match enumlevels with
            [] -> result
          | enumlevel :: rest ->
                let (chart, success) = tryparse_elt dict sent enumlevel start depstr in
                if success then 
                    tryparse dict sent [] start depstr (Some chart, enumlevel, success)
                else 
                    tryparse dict sent rest start depstr (Some chart, enumlevel, success)
    
    let list_enumlevels enumlevel =
        match enumlevel with
            Dict.NoEnum -> 
                [Dict.NoEnum]
          | Dict.EnumOnUncertain -> 
                [Dict.NoEnum; Dict.EnumOnUncertain]
          | Dict.EnumOnAll -> 
                [Dict.NoEnum; Dict.EnumOnUncertain; Dict.EnumOnAll]
    
    let parse dict chartdb sent start depstr =
        let enumlevel = Dict.get_enumlevel dict in
        let enumlevels = list_enumlevels enumlevel in
        let (ochart, enumlevel', success) = 
            tryparse dict sent enumlevels start depstr (None, Dict.NoEnum, false) in
        Dict.set_enumlevel dict enumlevel;
        match ochart with
            Some chart -> 
            begin
                ChartDb.add chartdb chart;
                (success, enumlevel', chart)
            end
          | None -> failwith "parse"
    
    let parse_file dict chartdb start fname depstrs =
        let fhdl = open_in fname in
        let cnt = ref 0 in
        fprintf stderr "Parsing file: %s\n" fname;
        flush stderr;
        try
            while true do
                flush stdout;
                incr cnt;
                let line = input_line fhdl in
                let sent = split line in
                fprintf stderr "Line %d: [%s] ... " !cnt (String.concat " " sent);
                flush stderr;
                let starttime = Unix.time () in
                let depstr =
                    match depstrs with
                        None -> None
                      | Some depstrarr -> Some depstrarr.(!cnt - 1) in
                let (success, enumlevel, chart) = 
                    parse dict chartdb sent start depstr in
                let endtime = Unix.time () in
                let runtime = endtime -. starttime in
                if success then 
                    fprintf stderr "successful (%d derivations, %d trees, runtime = %6.4f secs).\n" 
                        (Chart.get_derivcnt chart) (Chart.get_treecnt chart)
                        runtime
                else 
                    fprintf stderr "[1munsuccessful (runtime = %6.4f secs).[0m\n" 
                        runtime;
                flush stderr
            done
        with End_of_file -> close_in fhdl

    let run psr =
        List.iter 
            (fun fname -> 
                parse_file psr.dict psr.chartdb psr.start fname psr.depstrs) 
            psr.fnames;
        ChartDb.to_file psr.chartdb psr.cdbfname

    let compute dictfname cdbfname fnames start golddeps conlldeps =
        let psr = create dictfname cdbfname fnames start golddeps conlldeps in
        run psr

end
