(***********************************************************
*                                                          *
*      FUNGI - Fast UNsupervised Grammar Induction         *
*                                                          *
*   Originator: Prachya Boonkwan [1], [2]                  *
*               [1] School of Informatics,                 *
*                   University of Edinburgh, UK            *
*               [2] National Electronics and Computer      *
*                   Technology Center (NECTEC), Thailand   *
*                                                          *
*   Copyright (C) 2011 Prachya Boonkwan                    *
*   <p.boonkwan@sms.ed.ac.uk, kaamanita@hotmail.com>       *
*                                                          *
*   This library is free software; you can redistribute    *
*   it and/or modify it under the terms of the GNU         *
*   Lesser General Public License as published by the      *
*   Free Software Foundation; either version 2.1 of the    *
*   License, or (at your option) any later version.        *
*                                                          *
*   This library is distributed in the hope that it will   *
*   be useful, but WITHOUT ANY WARRANTY; without even      *
*   the implied warranty of MERCHANTABILITY or FITNESS     *
*   FOR A PARTICULAR PURPOSE.  See the GNU Lesser          *
*   General Public License for more details.               *
*                                                          *
*   You should have received a copy of the GNU Lesser      *
*   General Public License along with this library; if     *
*   not, write to the Free Software Foundation, Inc.,      *
*   59 Temple Place, Suite 330, Boston, MA  02111-1307     *
*   USA.                                                   *
*                                                          *
***********************************************************)

module RangePool = struct

    module Config = struct
        let debug = false
        let nagain = 0
    end

    module Enumerable = struct
        type elt = int
        type t = {
            begidx : int;
            endidx : int;
        }
        type enum = {
            curidx : int;
            maxidx : int;
        }
        let create begidx endidx = {
            begidx = begidx;
            endidx = endidx;
        }
        let start range = {
            curidx = range.begidx;
            maxidx = range.endidx;
        }
        let step state =
            if state.curidx <= state.maxidx then
                let nextstate = {
                    curidx = state.curidx + 1;
                    maxidx = state.maxidx;
                } in
                Some (state.curidx, nextstate)
            else None
    end

    module Pool = JoinPool.Shared.Make(Config)(Enumerable)

    type range_t = Enumerable.t

    type ('a, 'b) pool_t = ('a, 'b) Pool.t

    type 'a mapfn_t = int -> 'a

    type ('a, 'b) reducefn_t = 'a -> 'b -> 'b

    type ('input, 'output) t = {
        pool     : ('input, 'output) pool_t;
        workerno : int;
        begidx   : int;
        endidx   : int;
        initval  : 'output;
    }

    let create workerno begidx endidx (initval : 'output) : ('input, 'output) t = {
        pool     = Pool.create ();
        workerno = workerno;
        begidx   = begidx;
        endidx   = endidx;
        initval  = initval;
    }

    let register (pool : ('a, 'b) pool_t) (mapfn : 'a mapfn_t) =
        spawn pool.Pool.register(mapfn)

    let combine (pool : ('a, 'b) pool_t) (range : range_t) (reducefn : ('a, 'b) reducefn_t) initval =
        pool.Pool.fold range reducefn initval

    let compute workerno begidx endidx (initval : 'output) 
            ~(map : 'input mapfn_t) 
            ~(reduce : ('input, 'output) reducefn_t) =
        let p = create workerno begidx endidx initval in
        for i = 1 to p.workerno do
            register p.pool map
        done;
        let range = Enumerable.create begidx endidx in
        combine p.pool range reduce p.initval

end
