(***********************************************************
*                                                          *
*      FUNGI - Fast UNsupervised Grammar Induction         *
*                                                          *
*   Originator: Prachya Boonkwan [1], [2]                  *
*               [1] School of Informatics,                 *
*                   University of Edinburgh, UK            *
*               [2] National Electronics and Computer      *
*                   Technology Center (NECTEC), Thailand   *
*                                                          *
*   Copyright (C) 2011 Prachya Boonkwan                    *
*   <p.boonkwan@sms.ed.ac.uk, kaamanita@hotmail.com>       *
*                                                          *
*   This library is free software; you can redistribute    *
*   it and/or modify it under the terms of the GNU         *
*   Lesser General Public License as published by the      *
*   Free Software Foundation; either version 2.1 of the    *
*   License, or (at your option) any later version.        *
*                                                          *
*   This library is distributed in the hope that it will   *
*   be useful, but WITHOUT ANY WARRANTY; without even      *
*   the implied warranty of MERCHANTABILITY or FITNESS     *
*   FOR A PARTICULAR PURPOSE.  See the GNU Lesser          *
*   General Public License for more details.               *
*                                                          *
*   You should have received a copy of the GNU Lesser      *
*   General Public License along with this library; if     *
*   not, write to the Free Software Foundation, Inc.,      *
*   59 Temple Place, Suite 330, Boston, MA  02111-1307     *
*   USA.                                                   *
*                                                          *
***********************************************************)

open Dep
open Conlldep
open Printf

let load_content sentlen fnames =
    let content = ConllDep.compile_many false fnames in
    let content = ConllDep.elim_longsents true sentlen content in
    printf "    Size = %d\n" (List.length content);
    flush stdout;
    content

let divide_content content =
    let size = List.length content in
    let size' =
        if size mod 10 = 0 then size
        else size + 10 - (size mod 10) in
    let divsize = size' / 10 in
    let divsents = Array.make 10 [] in
    let diventries = Array.make 10 [] in
    let id = ref 0 in
    List.iter
        (fun entries ->
            let (_, sent, _, _) = entries in
            let divid = !id / divsize in
            divsents.(divid) <- divsents.(divid) @ [sent];
            diventries.(divid) <- diventries.(divid) @ [entries];
            incr id)
        content;
    (divsents, diventries)

let write_to_files oprefix divsents diventries =
    for i = 0 to 9 do
        printf "Preparing section %d ... " i;
        flush stdout;
        let ofname_sents = sprintf "%s_sents_%d.txt" oprefix i
        and ofname_deps = sprintf "%s_deps_%d.txt" oprefix i in
        let ofhdl_sents = open_out ofname_sents
        and ofhdl_deps = open_out ofname_deps in
        List.iter
            (fun sent ->
                fprintf ofhdl_sents "%s\n" (String.concat " " sent))
            divsents.(i);
        let sentid = ref 0 in
        List.iter
            (fun entries ->
                incr sentid;
                fprintf ofhdl_deps "# Sentence %d\n" !sentid;
                fprintf ofhdl_deps "%s\n\n" (ConllDep.entries_to_conll entries))
            diventries.(i);
        close_out ofhdl_sents;
        close_out ofhdl_deps;
        printf "done\n";
        printf "    Size = %d entries\n" (List.length divsents.(i));
        printf "    Sentence file: %s\n    Dependency file: %s\n"
            ofname_sents ofname_deps;
        flush stdout
    done

let write_desc oprefix content =
    let (postags, deprels) = ConllDep.read_postags_deprels content in
    let ofname_postags = sprintf "%s_postags.txt" oprefix
    and ofname_deprels = sprintf "%s_deprels.txt" oprefix in
    let ofhdl_postags = open_out ofname_postags
    and ofhdl_deprels = open_out ofname_deprels in
    List.iter
        (fun postag ->
            fprintf ofhdl_postags "%s\n" postag)
        (List.sort Pervasives.compare postags);
    List.iter
        (fun deprel ->
            fprintf ofhdl_deprels "%s\n" deprel)
        (List.sort Pervasives.compare deprels);
    close_out ofhdl_postags;
    close_out ofhdl_deprels

let prepare_files oprefix sentlen fnames =
    let content = load_content sentlen fnames in
    let (divsents, diventries) = divide_content content in
    write_to_files oprefix divsents diventries;
    write_desc oprefix content

let main =
    Utils.print_version ();
    let fnames = ref [] in
    let oprefix = ref "output" in
    let sentlen = ref None in
    let args = [
        ("--oprefix", 
            Arg.Set_string oprefix,
            "[path_prefix]   Set the path prefix for the output (default = output)");
        ("--sentlen",
            Arg.Int (fun l -> sentlen := Some l),
            "[sentlen]   Set the length of longest sentences allowed in the corpus (default = none)")]
    and fn_anon = (fun f -> fnames := !fnames @ [f])
    and usagemsg =
        "Prepare the input files from CoNLL2006-formatted dependency files for an experiment.\n\n" 
        ^ "prepareconll [options] depfile1 depfile2 ..." in
    Arg.parse args fn_anon usagemsg;
    prepare_files !oprefix !sentlen !fnames
