(***********************************************************
*                                                          *
*      FUNGI - Fast UNsupervised Grammar Induction         *
*                                                          *
*   Originator: Prachya Boonkwan [1], [2]                  *
*               [1] School of Informatics,                 *
*                   University of Edinburgh, UK            *
*               [2] National Electronics and Computer      *
*                   Technology Center (NECTEC), Thailand   *
*                                                          *
*   Copyright (C) 2011 Prachya Boonkwan                    *
*   <p.boonkwan@sms.ed.ac.uk, kaamanita@hotmail.com>       *
*                                                          *
*   This library is free software; you can redistribute    *
*   it and/or modify it under the terms of the GNU         *
*   Lesser General Public License as published by the      *
*   Free Software Foundation; either version 2.1 of the    *
*   License, or (at your option) any later version.        *
*                                                          *
*   This library is distributed in the hope that it will   *
*   be useful, but WITHOUT ANY WARRANTY; without even      *
*   the implied warranty of MERCHANTABILITY or FITNESS     *
*   FOR A PARTICULAR PURPOSE.  See the GNU Lesser          *
*   General Public License for more details.               *
*                                                          *
*   You should have received a copy of the GNU Lesser      *
*   General Public License along with this library; if     *
*   not, write to the Free Software Foundation, Inc.,      *
*   59 Temple Place, Suite 330, Boston, MA  02111-1307     *
*   USA.                                                   *
*                                                          *
***********************************************************)

open Utils
open Printf

module type INSTANCE_TYPE = sig
    type t
    val equal : t -> t -> bool
    val hash : t -> int
    val to_string : t -> string
end

module type COND_INSTANCE_TYPE = sig
    include INSTANCE_TYPE
    type prems_t
    type consq_t
    val create : prems_t -> consq_t -> t
    val get_prems : t -> prems_t
    val get_consq : t -> consq_t
    val equal_prems : prems_t -> prems_t -> bool
    val hash_prems : prems_t -> int
    val prems_to_string : prems_t -> string
    val get_nobackoffs : unit -> int
end

module type BACKOFF_INSTANCE_TYPE = sig
    include COND_INSTANCE_TYPE
    val is_backoff : t -> bool
    val backoffs : t -> t option list
end

module ProbTbl (Inst : INSTANCE_TYPE) = struct

    type inst_t = Inst.t

    module InstTbl = Hashtbl.Make(Inst)

    type t = {
        insttbl : float InstTbl.t;
        epsilon : float;
    }

    let create n epsilon = {
        insttbl = InstTbl.create n;
        epsilon = epsilon;
    }

    let get_epsilon probtbl =
        probtbl.epsilon

    let mem probtbl inst =
        InstTbl.mem probtbl.insttbl inst

    let length probtbl =
        let cnt = ref 0 in
        InstTbl.iter 
            (fun _ _ -> incr cnt)
            probtbl.insttbl;
        !cnt

    let get_prob probtbl inst =
        if mem probtbl inst then
            InstTbl.find probtbl.insttbl inst
        else probtbl.epsilon

    let get_smooth_prob probtbl inst =
        let prob = get_prob probtbl inst in
        (prob +. probtbl.epsilon) /. (1.0 +. probtbl.epsilon)

    let set_prob probtbl inst prob =
        InstTbl.replace probtbl.insttbl inst prob

    let acc_prob probtbl inst prob =
        let oldprob = 
            if mem probtbl inst then get_prob probtbl inst
            else 0.0 in
        let newprob = oldprob +. prob in
        set_prob probtbl inst newprob

    let iter probtbl fn =
        InstTbl.iter fn probtbl.insttbl

    let to_string probtbl =
        let result = ref [] in
        InstTbl.iter
            (fun inst prob ->
                let s = sprintf "Prob(%s) = %.2e" 
                    (Inst.to_string inst) prob in
                result := !result @ [s])
            probtbl.insttbl;
        sprintf "%s\nEpsilon = %.2e" 
            (String.concat "\n" !result) probtbl.epsilon

    let clear probtbl =
        InstTbl.clear probtbl.insttbl

end

module CondProbTbl (Inst : COND_INSTANCE_TYPE) = struct

    type prems_t = Inst.prems_t
    type consq_t = Inst.consq_t
    type inst_t = Inst.t

    module HashedPrems = struct
        type t = prems_t
        let equal = Inst.equal_prems
        let hash = Inst.hash_prems
    end

    module HashedInst = struct
        type t = Inst.t
        let equal = Inst.equal
        let hash = Inst.hash
        let to_string = Inst.to_string
    end

    module PremsTbl = Hashtbl.Make(HashedPrems)

    module ProbTbl = ProbTbl(HashedInst)

    type premstbl_t = float PremsTbl.t

    type probtbl_t = ProbTbl.t

    module InstTbl = Hashtbl.Make(Inst)

    type t = {
        probtbl  : probtbl_t;
        premstbl : premstbl_t;
        epsilon  : float;
    }

    let create n epsilon = {
        probtbl  = ProbTbl.create n epsilon;
        premstbl = PremsTbl.create n;
        epsilon  = epsilon;
    }

    let get_epsilon cprobtbl =
        cprobtbl.epsilon

    let mem cprobtbl inst =
        ProbTbl.mem cprobtbl.probtbl inst

    let length cprobtbl =
        ProbTbl.length cprobtbl.probtbl

    let mem_prems cprobtbl prems =
        PremsTbl.mem cprobtbl.premstbl prems

    let get_prems_prob cprobtbl prems =
        if mem_prems cprobtbl prems then
            PremsTbl.find cprobtbl.premstbl prems
        else cprobtbl.epsilon

    let get_prob cprobtbl inst =
        let prems = Inst.get_prems inst in
        if mem_prems cprobtbl prems then
            if ProbTbl.mem cprobtbl.probtbl inst then
                let prems_prob = get_prems_prob cprobtbl prems
                and inst_prob = ProbTbl.get_prob cprobtbl.probtbl inst in
                if prems_prob <> 0.0 && inst_prob <> 0.0 then
                    inst_prob /. prems_prob
                else cprobtbl.epsilon
            else cprobtbl.epsilon
        else cprobtbl.epsilon

    let get_smooth_prob cprobtbl inst =
        let prob = get_prob cprobtbl inst in
        let p = (prob +. cprobtbl.epsilon) /. (1.0 +. cprobtbl.epsilon) in
        p

    let set_prob cprobtbl inst prob =
        ProbTbl.set_prob cprobtbl.probtbl inst prob

    let set_prems_prob cprobtbl prems prob =
        PremsTbl.replace cprobtbl.premstbl prems prob

    let acc_prob cprobtbl inst prob =
        let prems = Inst.get_prems inst in
        ProbTbl.acc_prob cprobtbl.probtbl inst prob;
        let prems_prob =
            if mem_prems cprobtbl prems then
                get_prems_prob cprobtbl prems
            else 0.0 in
        let new_prems_prob = prems_prob +. prob in
        set_prems_prob cprobtbl prems new_prems_prob

    let iter cprobtbl fn =
        ProbTbl.iter cprobtbl.probtbl fn

    let iter_prems cprobtbl fn =
        PremsTbl.iter fn cprobtbl.premstbl

    let to_string_premstbl premstbl =
        let result = ref [] in
        PremsTbl.iter
            (fun prems prob ->
                let s = sprintf "Premise_prob(%s) = %.2e" 
                    (Inst.prems_to_string prems) prob in
                result := !result @ [s])
            premstbl;
        String.concat "\n" !result

    let to_string cprobtbl =
        sprintf "%s\n%s\nEpsilon = %.2e" 
            (ProbTbl.to_string cprobtbl.probtbl) 
            (to_string_premstbl cprobtbl.premstbl)
            cprobtbl.epsilon

end

module GenModel (Inst : COND_INSTANCE_TYPE) = struct

    type inst_t = Inst.t

    type prems_t = Inst.prems_t

    module ProbTbl = ProbTbl(Inst)

    module CondProbTbl = CondProbTbl(Inst)

    type probtbl_t = ProbTbl.t

    type cprobtbl_t = CondProbTbl.t

    module InstTbl = Hashtbl.Make(Inst)

    type insttbl_t = float InstTbl.t

    module HashedPrems = struct
        type t = Inst.prems_t
        let equal = Inst.equal_prems
        let hash = Inst.hash_prems
    end

    module PremsTbl = Hashtbl.Make(HashedPrems)

    type premstbl_t = float PremsTbl.t

    type fn_acc_prob_t = inst_t -> float -> unit
    
    type fn_get_prob_t = inst_t -> float

    type fn_set_prob_t = inst_t -> float -> unit

    type probftr_t = {
        get_prob        : fn_get_prob_t;
            (** How to calculate the probability of an instance *)
        get_smooth_prob : fn_get_prob_t;
            (** How to calculate the smoothened probability of 
                an instance *)
        set_prob        : fn_set_prob_t;
            (** How to set the probability of an instance *)
        acc_prob        : fn_acc_prob_t;
            (** How to accumulate the probability of an instance *)
    }

    let create_probftr 
            get_prob get_smooth_prob set_prob acc_prob = 
    {
        get_prob        = get_prob;
        get_smooth_prob = get_smooth_prob;
        set_prob        = set_prob;
        acc_prob        = acc_prob;
    }

    type t = {
        probtbl           : probtbl_t;
        probftrs          : probftr_t list;
        n                 : int;
        epsilon           : float;
        mutable normfacts : float array;
    }

    let create n epsilon probftrs = {
        probtbl   = ProbTbl.create n epsilon;
        probftrs  = probftrs;
        n         = n;
        epsilon   = epsilon;
        normfacts = Array.make (Inst.get_nobackoffs () + 1) 0.0;
    }

    let get_epsilon genmodel = genmodel.epsilon

    let length genmodel =
        ProbTbl.length genmodel.probtbl

    let get_ftrprob genmodel inst =
        List.fold_left
            (fun prob probftr ->
                let prob' = probftr.get_prob inst in
                prob *. prob')
            1.0 genmodel.probftrs

    let get_smooth_ftrprob genmodel inst =
        List.fold_left
            (fun prob probftr ->
                let prob' = probftr.get_smooth_prob inst in
                prob *. prob')
            1.0 genmodel.probftrs

    let get_prob_noftr genmodel inst =
        ProbTbl.get_prob genmodel.probtbl inst
    
    let get_prob genmodel inst =
        let fundprob = get_prob_noftr genmodel inst in
        let p = fundprob *. get_ftrprob genmodel inst in
        p

    let mem genmodel inst =
        ProbTbl.mem genmodel.probtbl inst

    let get_smooth_prob_noftr genmodel inst =
        ProbTbl.get_smooth_prob genmodel.probtbl inst

    let get_smooth_prob genmodel inst =
        let fundprob = get_smooth_prob_noftr genmodel inst in
        fundprob *. get_smooth_ftrprob genmodel inst

    let set_prob genmodel inst prob =
        ProbTbl.set_prob genmodel.probtbl inst prob

    let acc_prob genmodel inst prob =
        ProbTbl.acc_prob genmodel.probtbl inst prob
        (* ;
        List.iter
            (fun probftr ->
                probftr.acc_prob inst prob)
            genmodel.probftrs *)

    let iter genmodel fn =
        ProbTbl.iter genmodel.probtbl fn

    let dkl genmodel1 genmodel2 =
        let result = ref 0.0 
        and divp = ref 0.0 
        and divq = ref 0.0 in
        iter
            genmodel1
            (fun inst p -> divp := !divp +. p);
        iter
            genmodel2
            (fun inst q -> divq := !divq +. q);
        if !divp = 0.0 || !divq = 0.0 then infinity
        else begin
            iter
                genmodel1
                (fun inst p ->
                    let q = get_prob genmodel2 inst in
                    let p' = p /. !divp
                    and q' = q /. !divq in
                    if (p' > 0.0) && (q' > 0.0) then begin
                        let r = q' *. (log q' -. log p') in
                        result := !result +. r
                    end else ());
            !result
        end

    let calc_probftrs genmodel =
        iter genmodel
            (fun inst _ ->
                List.iter
                    (fun probftr ->
                        probftr.set_prob inst 0.0)
                    genmodel.probftrs);
        iter genmodel
            (fun inst prob ->
                List.iter
                    (fun probftr ->
                        probftr.acc_prob inst prob)
                    genmodel.probftrs)

    let calc_backoffs _ = ()

    let calc_normfacts _ = ()

    let to_string genmodel =
        let result = ref [] in
        ProbTbl.iter genmodel.probtbl
            (fun inst prob ->
                let line = sprintf "Prob(%s) = %.2e (smoothed: %.2e)"
                    (Inst.to_string inst)
                    (get_prob genmodel inst)
                    (get_smooth_prob genmodel inst) in
                result := !result @ [line]);
        String.concat "\n" !result

    let print genmodel =
        ProbTbl.iter genmodel.probtbl
            (fun inst prob ->
                let p = get_prob genmodel inst in
                if p > 0.0 then
                    printf "Prob(%s) = %.2e (smoothed: %.2e)\n"
                        (Inst.to_string inst)
                        (get_prob genmodel inst)
                        (get_smooth_prob genmodel inst)
                else ())

    let from_file fname epsilon =
        let insttbl = unmarshal_from_file fname in
        let result = create 100 epsilon [] in
        InstTbl.iter
            (fun inst prob ->
                acc_prob result inst prob)
            insttbl;
        calc_normfacts result;
        result

    let to_file (genmodel : t) fname =
        let insttbl = InstTbl.create genmodel.n in
        iter genmodel
            (fun inst _ ->
                let prob = get_prob genmodel inst in
                InstTbl.replace insttbl inst prob);
        marshal_to_file fname insttbl

end

module GenModelBackoff (Inst : BACKOFF_INSTANCE_TYPE) = struct 

    include GenModel(Inst) 

    let to_string genmodel =
        let result = ref [] in
        ProbTbl.iter genmodel.probtbl
            (fun inst prob ->
                if not (Inst.is_backoff inst) then
                    let line = sprintf "Prob(%s) = %.2e (smoothed: %.2e)"
                        (Inst.to_string inst)
                        (get_prob genmodel inst)
                        (get_smooth_prob genmodel inst) in
                    result := !result @ [line]
                else ());
        String.concat "\n" !result

    let print genmodel =
        ProbTbl.iter genmodel.probtbl
            (fun inst prob ->
                if not (Inst.is_backoff inst) then
                    printf "Prob(%s) = %.2e (smoothed: %.2e)\n"
                        (Inst.to_string inst)
                        (get_prob genmodel inst)
                        (get_smooth_prob genmodel inst)
                else ())

    let calc_backoffs_populate genmodel premstbl insttbl newtbl =
        iter genmodel
            (fun inst prob -> 
                let boinsts = Inst.backoffs inst in
                List.iter
                    (fun inst' ->
                        match inst' with
                            Some inst'' -> begin
                                if not (InstTbl.mem insttbl inst'') then
                                    InstTbl.replace insttbl inst'' prob
                                else begin
                                    let prob' = InstTbl.find insttbl inst'' in
                                    InstTbl.replace insttbl inst'' (prob +. prob')
                                end;
                                let prems'' = Inst.get_prems inst'' in
                                if not (PremsTbl.mem premstbl prems'') then
                                    PremsTbl.replace premstbl prems'' prob
                                else begin
                                    let prob'' = PremsTbl.find premstbl prems'' in
                                    PremsTbl.replace premstbl prems'' (prob +. prob'')
                                end
                            end
                          | None -> ())
                    boinsts)

    let calc_backoffs_marginalize genmodel premstbl insttbl newtbl =
        iter genmodel
            (fun inst _ ->
                let boinsts = Inst.backoffs inst in
                List.iter
                    (fun inst' ->
                        match inst' with
                            Some inst'' -> begin
                                let prems'' = Inst.get_prems inst'' in
                                let prob' = InstTbl.find insttbl inst''
                                and prob'' = PremsTbl.find premstbl prems'' in
                                InstTbl.replace newtbl inst'' (prob' /. prob'')
                            end
                          | None -> ())
                    boinsts)

    let calc_normfacts genmodel =
        let normfacts = Array.make (Inst.get_nobackoffs () + 1) 0.0 in
        if Utils.is_use_backoff () then begin
            let total = ref 0.0 in
            iter genmodel
                (fun inst p ->
                    normfacts.(0) <- normfacts.(0) +. p;
                    total := !total +. p;
                    let boinsts = Array.of_list (Inst.backoffs inst) in
                    for i = 1 to Inst.get_nobackoffs () do
                        let inst' = boinsts.(i - 1) in
                        match inst' with
                            Some inst'' -> begin
                                let prob = get_prob genmodel inst'' in
                                normfacts.(i) <- normfacts.(i) +. prob;
                                total := !total +. prob
                            end
                          | None -> ()
                    done);
            for i = 0 to Inst.get_nobackoffs () do
                normfacts.(i) <- normfacts.(i) /. !total
            done;
            total := !total -. normfacts.(0) +. 1.0;
            normfacts.(0) <- 1.0;
            for i = 1 to Inst.get_nobackoffs () do
                normfacts.(i) <- normfacts.(i) /. !total
            done
        end else
            normfacts.(0) <- 1.0;
        genmodel.normfacts <- normfacts

    let calc_backoffs genmodel =
        if Utils.is_use_backoff () then begin
            let premstbl = PremsTbl.create genmodel.n in
            let insttbl = InstTbl.create genmodel.n in
            let newtbl = InstTbl.create genmodel.n in
            calc_backoffs_populate genmodel premstbl insttbl newtbl;
            calc_backoffs_marginalize genmodel premstbl insttbl newtbl;
            InstTbl.iter
                (fun inst prob ->
                    set_prob genmodel inst prob)
                newtbl
        end else ();
        calc_normfacts genmodel

    let get_smooth_prob_noftr genmodel inst =
        let p = get_smooth_prob_noftr genmodel inst in
        if Utils.is_use_backoff () then begin
            let result = ref (p *. genmodel.normfacts.(0)) in
            let boinsts = Array.of_list (Inst.backoffs inst) in
            for i = 1 to Inst.get_nobackoffs () do
                let inst' = boinsts.(i - 1) in
                match inst' with
                    Some inst'' -> begin
                        let p = get_smooth_prob_noftr genmodel inst'' in
                        result := !result +. p *. genmodel.normfacts.(i)
                    end
                  | None -> ()
            done;
            !result
        end else p

    let get_smooth_prob genmodel inst =
        let p = get_smooth_prob genmodel inst in
        if Utils.is_use_backoff () then begin
            let result = ref (p *. genmodel.normfacts.(0)) in
            let boinsts = Array.of_list (Inst.backoffs inst) in
            for i = 1 to Inst.get_nobackoffs () do
                let inst' = boinsts.(i - 1) in
                match inst' with
                    Some inst'' -> begin
                        let p = get_smooth_prob genmodel inst'' in
                        result := !result +. p *. genmodel.normfacts.(i)
                    end
                  | None -> ()
            done;
            !result
        end else p

end
