(** Probability Tables and Generative Model *)

(***********************************************************
*                                                          *
*      FUNGI - Fast UNsupervised Grammar Induction         *
*                                                          *
*   Originator: Prachya Boonkwan [1], [2]                  *
*               [1] School of Informatics,                 *
*                   University of Edinburgh, UK            *
*               [2] National Electronics and Computer      *
*                   Technology Center (NECTEC), Thailand   *
*                                                          *
*   Copyright (C) 2011 Prachya Boonkwan                    *
*   <p.boonkwan@sms.ed.ac.uk, kaamanita@hotmail.com>       *
*                                                          *
*   This library is free software; you can redistribute    *
*   it and/or modify it under the terms of the GNU         *
*   Lesser General Public License as published by the      *
*   Free Software Foundation; either version 2.1 of the    *
*   License, or (at your option) any later version.        *
*                                                          *
*   This library is distributed in the hope that it will   *
*   be useful, but WITHOUT ANY WARRANTY; without even      *
*   the implied warranty of MERCHANTABILITY or FITNESS     *
*   FOR A PARTICULAR PURPOSE.  See the GNU Lesser          *
*   General Public License for more details.               *
*                                                          *
*   You should have received a copy of the GNU Lesser      *
*   General Public License along with this library; if     *
*   not, write to the Free Software Foundation, Inc.,      *
*   59 Temple Place, Suite 330, Boston, MA  02111-1307     *
*   USA.                                                   *
*                                                          *
***********************************************************)

(** Instance Type Signature *)
module type INSTANCE_TYPE =
sig 
    type t 
    val equal : t -> t -> bool 
    val hash : t -> int 
    val to_string : t -> string
end

(** Conditional Instance Type Signature *)
module type COND_INSTANCE_TYPE =
sig
    include INSTANCE_TYPE

    (** Premise *)
    type prems_t

    (** Consequence *)
    type consq_t

    (** Create an instance from the premise and the consequence. *)
    val create : prems_t -> consq_t -> t

    (** Get the premise. *)
    val get_prems : t -> prems_t

    (** Get the consequence. *)
    val get_consq : t -> consq_t

    (** Check if two premises equal. *)
    val equal_prems : prems_t -> prems_t -> bool

    (** Calculate the hash value of the premise. *)
    val hash_prems : prems_t -> int

    val prems_to_string : prems_t -> string
    
    val get_nobackoffs : unit -> int
end

(** Backoff Instance Type Signature *)
module type BACKOFF_INSTANCE_TYPE = sig
    include COND_INSTANCE_TYPE
    val is_backoff : t -> bool
    val backoffs : t -> t option list
end

(** Probability Table *)
module ProbTbl :
functor (Inst : INSTANCE_TYPE) ->
sig
    (** Instance *)
    type inst_t = Inst.t

    (** Probability table *)
    type t

    (** Create a probability table with the size and the default value. *)
    val create : int -> float -> t

    (** Get the default value. *)
    val get_epsilon : t -> float

    (** Check if an instance is a member of the table. *)
    val mem : t -> inst_t -> bool

    val length : t -> int

    (** Get the probability of the instance. *)
    val get_prob : t -> inst_t -> float
    
    (** Get the smoothened probability of the instance. *)
    val get_smooth_prob : t -> inst_t -> float
    
    (** Set the probability of the instance. *)
    val set_prob : t -> inst_t -> float -> unit

    (** Accumulate the probability of the instance. *)
    val acc_prob : t -> inst_t -> float -> unit

    (** Iterate all entries with the function. *)
    val iter : t -> (inst_t -> float -> unit) -> unit

    val to_string : t -> string

    val clear : t -> unit
end

(** Conditional Probability Table *)
module CondProbTbl :
functor (Inst : COND_INSTANCE_TYPE) ->
sig
    (** Premise *)
    type prems_t = Inst.prems_t

    (** Consequence *)
    type consq_t = Inst.consq_t

    (** Instance *)
    type inst_t = Inst.t

    (** Conditional probability table *)
    type t

    (** Create a conditional probability table with the size and 
        the default value. *)
    val create : int -> float -> t

    (** Get the default value. *)
    val get_epsilon : t -> float

    (** Check if the instance is a member of the table. *)
    val mem : t -> inst_t -> bool

    val length : t -> int

    (** Get the probability of the instance. *)
    val get_prob : t -> inst_t -> float
    
    (** Get the smoothened probability of the instance. *)
    val get_smooth_prob : t -> inst_t -> float

    (** Set the probability of the instance. *)
    val set_prob : t -> inst_t -> float -> unit

    (** Accumulate the probability of the instance. *)
    val acc_prob : t -> inst_t -> float -> unit

    (** Iterate all entries with the function. *)
    val iter : t -> (inst_t -> float -> unit) -> unit
    
    (** Iterate all premises with the function. *)
    val iter_prems : t -> (prems_t -> float -> unit) -> unit

    val to_string : t -> string
end

(** Generative Model *)
module GenModel :
functor (Inst : COND_INSTANCE_TYPE) ->
sig
    (** Instance *)
    type inst_t = Inst.t

    type prems_t = Inst.prems_t

    (** Function for accumulating the probability of an instance *)
    type fn_acc_prob_t = inst_t -> float -> unit
    
    (** Function for calculating the probability (optionally 
        smoothened) of an instance *)
    type fn_get_prob_t = inst_t -> float
    
    (** Function for setting the probability of an instance *)
    type fn_set_prob_t = inst_t -> float -> unit
   
    (** Feature probability table *)
    type probftr_t
   
    (** Create a feature probability table from the 
        four functions: get_prob, get_smooth_prob, set_prob, 
        and acc_prob. *)
    val create_probftr :
        fn_get_prob_t -> fn_get_prob_t -> fn_set_prob_t -> fn_acc_prob_t 
        -> probftr_t
   
    type probtbl_t = ProbTbl(Inst).t

    (** Generative model *)
    type t
    
    (** Create a generative model with the size, the default value, 
        and the list of generative probability tables. *)
    val create : int -> float -> probftr_t list -> t

    val get_epsilon : t -> float

    val length : t -> int

    val get_ftrprob : t -> inst_t -> float

    val get_smooth_ftrprob : t -> inst_t -> float

    val get_prob_noftr : t -> inst_t -> float
    
    (** Get the probability of the instance. *)
    val get_prob : t -> inst_t -> float

    val get_smooth_prob_noftr : t -> inst_t -> float
    
    (** Get the smoothened probability of the instance. *)
    val get_smooth_prob : t -> inst_t -> float
    
    (** Set the probability of the instance. *)
    val set_prob : t -> inst_t -> float -> unit
    
    (** Accumulate the probability of the instance. *)
    val acc_prob : t -> inst_t -> float -> unit
    
    (** Iterate all entries with the function. *)
    val iter : t -> (inst_t -> float -> unit) -> unit

    val dkl : t -> t -> float

    val calc_probftrs : t -> unit
    
    val calc_backoffs : t -> unit
    
    val calc_normfacts : t -> unit

    (** Read a generative model from the file. *)
    val from_file : string -> float -> t
    
    (** Write the generative model from a file. *)
    val to_file : t -> string -> unit

    val to_string : t -> string

    val print : t -> unit
end

(** Generative Model *)
module GenModelBackoff :
functor (Inst : BACKOFF_INSTANCE_TYPE) ->
sig
    (** Instance *)
    type inst_t = Inst.t
    
    type prems_t = Inst.prems_t

    (** Function for accumulating the probability of an instance *)
    type fn_acc_prob_t = inst_t -> float -> unit
    
    (** Function for calculating the probability (optionally 
        smoothened) of an instance *)
    type fn_get_prob_t = inst_t -> float
    
    (** Function for setting the probability of an instance *)
    type fn_set_prob_t = inst_t -> float -> unit
   
    (** Feature probability table *)
    type probftr_t
   
    (** Create a feature probability table from the 
        four functions: get_prob, get_smooth_prob, set_prob, 
        and acc_prob. *)
    val create_probftr :
        fn_get_prob_t -> fn_get_prob_t -> fn_set_prob_t -> fn_acc_prob_t 
        -> probftr_t
   
    type probtbl_t = ProbTbl(Inst).t

    (** Generative model *)
    type t
    
    (** Create a generative model with the size, the default value, 
        and the list of generative probability tables. *)
    val create : int -> float -> probftr_t list -> t

    val get_epsilon : t -> float

    val length : t -> int

    val get_ftrprob : t -> inst_t -> float

    val get_smooth_ftrprob : t -> inst_t -> float

    val get_prob_noftr : t -> inst_t -> float
    
    (** Get the probability of the instance. *)
    val get_prob : t -> inst_t -> float

    val mem : t -> inst_t -> bool

    val get_smooth_prob_noftr : t -> inst_t -> float
    
    (** Get the smoothened probability of the instance. *)
    val get_smooth_prob : t -> inst_t -> float
    
    (** Set the probability of the instance. *)
    val set_prob : t -> inst_t -> float -> unit
    
    (** Accumulate the probability of the instance. *)
    val acc_prob : t -> inst_t -> float -> unit

    (** Iterate all entries with the function. *)
    val iter : t -> (inst_t -> float -> unit) -> unit

    val dkl : t -> t -> float

    (** Read a generative model from the file. *)
    val from_file : string -> float -> t
    
    (** Write the generative model from a file. *)
    val to_file : t -> string -> unit

    val to_string : t -> string

    val print : t -> unit

    val calc_probftrs : t -> unit

    val calc_backoffs : t -> unit
    
    val calc_normfacts : t -> unit
end
