(***********************************************************
*                                                          *
*      FUNGI - Fast UNsupervised Grammar Induction         *
*                                                          *
*   Originator: Prachya Boonkwan [1], [2]                  *
*               [1] School of Informatics,                 *
*                   University of Edinburgh, UK            *
*               [2] National Electronics and Computer      *
*                   Technology Center (NECTEC), Thailand   *
*                                                          *
*   Copyright (C) 2011 Prachya Boonkwan                    *
*   <p.boonkwan@sms.ed.ac.uk, kaamanita@hotmail.com>       *
*                                                          *
*   This library is free software; you can redistribute    *
*   it and/or modify it under the terms of the GNU         *
*   Lesser General Public License as published by the      *
*   Free Software Foundation; either version 2.1 of the    *
*   License, or (at your option) any later version.        *
*                                                          *
*   This library is distributed in the hope that it will   *
*   be useful, but WITHOUT ANY WARRANTY; without even      *
*   the implied warranty of MERCHANTABILITY or FITNESS     *
*   FOR A PARTICULAR PURPOSE.  See the GNU Lesser          *
*   General Public License for more details.               *
*                                                          *
*   You should have received a copy of the GNU Lesser      *
*   General Public License along with this library; if     *
*   not, write to the Free Software Foundation, Inc.,      *
*   59 Temple Place, Suite 330, Boston, MA  02111-1307     *
*   USA.                                                   *
*                                                          *
***********************************************************)

open Cg
open Chart
open Prob
open Deriv
open Hmm
open Printf

module ScoreInst = struct

    type syncat_t = Syncat.t
    type word_t = string

    type t = {
        syncat : syncat_t;
        head   : word_t;
    }

    let create syncat head = {
        syncat = syncat;
        head   = head;
    }

    let get_syncat inst =
        inst.syncat

    let get_head inst =
        inst.head

    let equal inst1 inst2 =
        Syncat.equal inst1.syncat inst2.syncat
        && inst1.head = inst2.head

    let hash inst =
        Syncat.hash inst.syncat + Hashtbl.hash inst.head

    let to_string inst =
        sprintf "%s{%s}" (Syncat.to_string inst.syncat) inst.head

end

module ScoreTbl = ProbTbl(ScoreInst)

module ProbInit = struct

    type chart_t = Chart.t

    type genmodel_t = DerivGenModel.t

    type scoretbl_t = ScoreTbl.t

    type fn_get_chart_t = int -> chart_t

    type t = {
        ruleprob         : genmodel_t;
        rulebias         : genmodel_t;
        get_chart        : fn_get_chart_t;
        nocharts         : int;
        catcnttbl        : scoretbl_t;
        fwdbwd           : Hmm.fwdbwd_t;
        mutable derivcnt : int;
        mutable treecnt  : int;
    }

    let create ruleprob rulebias get_chart nocharts fwdbwd =
        Random.self_init ();
        {
            ruleprob  = ruleprob;
            rulebias  = rulebias;
            get_chart = get_chart;
            nocharts  = nocharts;
            catcnttbl = ScoreTbl.create 1000 0.0;
            fwdbwd    = fwdbwd;
            derivcnt  = 0;
            treecnt   = 0;
        }

    let to_string probinit =
        sprintf "<Rule Prob>\n%s\n\n<Category Count>\n%s"
            (DerivGenModel.to_string probinit.ruleprob)
            (ScoreTbl.to_string probinit.catcnttbl)

    let learn_deriv_lex ruleprob catcnttbl syncat span headpos word =
        let deriv = conv_word_to_deriv syncat word in
        let syncat = Deriv.get_syncat (Deriv.get_prems deriv) in
        let inst = ScoreInst.create syncat word in
        DerivGenModel.acc_prob ruleprob deriv 1.0;
        ScoreTbl.acc_prob catcnttbl inst 1.0

    let calc_dtrargs_score headpos dtrargs =
        let sum = ref 0.0 in
        List.iter
            (fun (_, _, dtrpos, _) ->
                if dtrpos <> headpos then
                    sum := 1.0 /. (float_of_int (abs (headpos - dtrpos)) +. 2.0)
                else ())
            dtrargs;
        !sum

    let learn_deriv_dtrs ruleprob catcnttbl
            syncat span headpos word depdir dtrargs =
        let deriv = conv_dtrargs_to_deriv syncat word depdir dtrargs in
        let inst = ScoreInst.create syncat word in
        let score = calc_dtrargs_score headpos dtrargs in
        DerivGenModel.acc_prob ruleprob deriv score;
        ScoreTbl.acc_prob catcnttbl inst score

    let learn_chart ruleprob catcnttbl chart =
        Chart.iter_derivs
            chart Chart.TopDown
            (learn_deriv_lex ruleprob catcnttbl) 
            (learn_deriv_dtrs ruleprob catcnttbl)
            (0, List.length (Chart.get_sentence chart) - 1)

    module LearnProbinitProblem = struct

        let identifier = "LearnProbinit"

        type init = unit

        type client_data = unit

        type collection = {
            min      : int;
            max      : int;
            probinit : t;
            fwdbwd   : Hmm.fwdbwd_t;
            postreg  : bool;
        }

        type enum = {
            cur      : int;
            limit    : int;
            pi       : t;
            fb       : Hmm.fwdbwd_t;
            pr       : bool;
        }

        type input = int * chart_t * Hmm.fwdbwdserl_t * bool

        type key = DerivTblInst of Deriv.t
                 | ScoreTblInst of ScoreInst.t
                 | FBVisited of Hmm.VisitedInst.t
                 | FBInitInst of Hmm.InitInst.t
                 | FBTransInst of Hmm.TransInst.t
                 | FBEmisInst of Hmm.EmisInst.t
                 | FBCatInst of Hmm.CatInst.t
                 | FBTotalCatExp

        type value = float

        type output = genmodel_t * scoretbl_t * Hmm.fwdbwd_t

        let enum_start range = { 
            cur = range.min; 
            limit = range.max; 
            pi = range.probinit;
            fb = range.fwdbwd;
            pr = range.postreg;
        }

        let enum_step e =
            if e.cur <= e.limit then begin
                Scale.inc_and_display ();
                let chart = e.pi.get_chart e.cur in
                let input = (e.cur, chart, Hmm.serialize e.fb, e.pr) in
                Some (input, { 
                    cur = e.cur + 1; 
                    limit = e.limit; 
                    pi = e.pi; 
                    fb = e.fb;
                    pr = e.pr;
                })
            end else None

        let init_client () = ()

        let compare_keys = Pervasives.compare

        let map () (idx, chart, fwdbwdserl, is_postreg) =
            let ruleprob' = DerivGenModel.create 100 0.0 []
            and catcnttbl' = ScoreTbl.create 100 0.0 in
            let fwdbwd = Hmm.deserialize fwdbwdserl in
            learn_chart ruleprob' catcnttbl' chart;
            if is_postreg then
                Hmm.init_onestep fwdbwd chart
            else ();
            let result = ref [] in
            DerivGenModel.iter ruleprob'
                (fun inst prob ->
                    result := (DerivTblInst inst, prob) :: !result);
            ScoreTbl.iter catcnttbl'
                (fun inst prob ->
                    result := (ScoreTblInst inst, prob) :: !result);
            Hmm.InitProb.iter (Hmm.get_initexp fwdbwd)
                (fun inst prob ->
                    result := (FBInitInst inst, prob) :: !result);
            Hmm.TransProb.iter (Hmm.get_transexp fwdbwd)
                (fun inst prob ->
                    result := (FBTransInst inst, prob) :: !result);
            Hmm.EmisProb.iter (Hmm.get_emisexp fwdbwd)
                (fun inst prob ->
                    result := (FBEmisInst inst, prob) :: !result);
            Hmm.CatProb.iter (Hmm.get_catexp fwdbwd)
                (fun inst prob ->
                    result := (FBCatInst inst, prob) :: !result);
            Hmm.VisitedProb.iter (Hmm.get_visited fwdbwd)
                (fun inst prob ->
                    result := (FBVisited inst, prob) :: !result);
            result := (FBTotalCatExp, !(Hmm.get_totalcatexp fwdbwd)) :: !result;
            !result

        let combine v1 v2 =
            v1 +. v2

        let reduce (k : key) (prob : value) ((rp, ct, fb) : output) =
            begin
                match k with
                    DerivTblInst inst ->
                        DerivGenModel.acc_prob rp inst prob
                  | ScoreTblInst inst ->
                        ScoreTbl.acc_prob ct inst prob
                  | FBVisited inst ->
                        Hmm.VisitedProb.acc_prob (Hmm.get_visited fb) inst prob
                  | FBInitInst inst ->
                        Hmm.InitProb.acc_prob (Hmm.get_initexp fb) inst prob
                  | FBTransInst inst ->
                        Hmm.TransProb.acc_prob (Hmm.get_transexp fb) inst prob
                  | FBEmisInst inst ->
                        Hmm.EmisProb.acc_prob (Hmm.get_emisexp fb) inst prob
                  | FBCatInst inst ->
                        Hmm.CatProb.acc_prob (Hmm.get_catexp fb) inst prob
                  | FBTotalCatExp ->
                        let totalcatexp = Hmm.get_totalcatexp fb in
                        totalcatexp := prob +. !totalcatexp
            end;
            (rp, ct, fb)

    end

    module MR = Mapred.Make(LearnProbinitProblem)

    let learn_probinit_client probinit =
        MR.client (Utils.client_conf_probinit ())

    let learn_probinit_server probinit =
        Scale.add (Scale.Number probinit.nocharts) "charts " " ";
        Scale.display ();
        let chartcoll = {
            LearnProbinitProblem.min = 0;
            LearnProbinitProblem.max = probinit.nocharts - 1;
            LearnProbinitProblem.probinit = probinit;
            LearnProbinitProblem.fwdbwd = probinit.fwdbwd;
            LearnProbinitProblem.postreg = Utils.is_postreg ();
        } in
        let svr = MR.server (Utils.server_conf_probinit ()) () in
        ignore (
            MR.compute svr
                chartcoll (probinit.ruleprob, probinit.catcnttbl, probinit.fwdbwd) );
        for idx = 0 to probinit.nocharts - 1 do
            let chart = probinit.get_chart idx in
            probinit.derivcnt <- 
                probinit.derivcnt + (Chart.get_derivcnt chart);
            probinit.treecnt <-
                probinit.treecnt + (Chart.get_treecnt chart)
        done;
        Scale.discard ()
    
    let marginalize_deriv probinit deriv oldprob =
        Scale.inc_and_display ();
        let prems = Deriv.get_prems deriv in
        let word = Deriv.get_head prems in
        let syncat = Deriv.get_syncat prems in
        let inst = ScoreInst.create syncat word in
        let denom = DerivGenModel.get_prob probinit.ruleprob deriv
        and nom = ScoreTbl.get_prob probinit.catcnttbl inst in
        let prob = ref (denom /. nom) in
        if Utils.is_postreg () then
            let consq = Deriv.get_consq deriv in
            if Deriv.is_consq_lex consq then begin
                let word = Deriv.get_word consq in
                let inst' = Hmm.EmisInst.create syncat word in
                let emisprob = Hmm.get_emisprob probinit.fwdbwd in
                let newprob = !prob *. Hmm.EmisProb.get_prob emisprob inst' in
                prob := newprob;
                Hmm.EmisProb.set_prob emisprob inst' newprob
            end else ()
        else ();
        (*
        let random_mult = ref 0 in
        while !random_mult = 0 do
            random_mult := Random.int 101
        done; 
        prob := !prob *. 0.01 *. (float_of_int !random_mult)
        *)
        DerivGenModel.set_prob probinit.ruleprob deriv !prob;
        DerivGenModel.set_prob probinit.rulebias deriv !prob

    let marginalize_probinit probinit =
        Scale.add (Scale.WindMill 1500) "randomizing each parameter " "";
        (* Random.self_init (); *)
        Hmm.estimate probinit.fwdbwd;
        DerivGenModel.iter probinit.ruleprob (marginalize_deriv probinit);
        DerivGenModel.calc_probftrs probinit.ruleprob;
        if Utils.is_use_backoff () then
            DerivGenModel.calc_backoffs probinit.ruleprob
        else ();
        Scale.discard ()

    let learn ?(run_as_server=false) ?(run_as_client=false) ruleprob rulebias get_chart nocharts fwdbwd =
        let probinit = 
            create ruleprob rulebias get_chart nocharts fwdbwd in
        if run_as_server then begin
            let starttime = Unix.time () in
            fprintf stderr "Initializing the probability table ... ";
            flush stderr;
            learn_probinit_server probinit;
            (* printf "Before marginalization\n";
            DerivGenModel.print probinit.ruleprob; *)
            marginalize_probinit probinit;
            (* printf "\nAfter marginalization\n";
            DerivGenModel.print probinit.ruleprob; *)
            let endtime = Unix.time () in
            let runtime = (endtime -. starttime) /. 60.0 in
            fprintf stderr "done.\n";
            fprintf stderr "     Charts      : %d\n" nocharts;
            fprintf stderr "     Rules       : %d\n" (DerivGenModel.length ruleprob);
            fprintf stderr "     Derivaions  : %d\n" probinit.derivcnt;
            fprintf stderr "     Trees       : %d\n" probinit.treecnt;
            fprintf stderr "     Runtime     : %.2f mins\n" runtime;
            flush stderr
        end else begin
            learn_probinit_client probinit
        end

end
