(***********************************************************
*                                                          *
*      FUNGI - Fast UNsupervised Grammar Induction         *
*                                                          *
*   Originator: Prachya Boonkwan [1], [2]                  *
*               [1] School of Informatics,                 *
*                   University of Edinburgh, UK            *
*               [2] National Electronics and Computer      *
*                   Technology Center (NECTEC), Thailand   *
*                                                          *
*   Copyright (C) 2011 Prachya Boonkwan                    *
*   <p.boonkwan@sms.ed.ac.uk, kaamanita@hotmail.com>       *
*                                                          *
*   This library is free software; you can redistribute    *
*   it and/or modify it under the terms of the GNU         *
*   Lesser General Public License as published by the      *
*   Free Software Foundation; either version 2.1 of the    *
*   License, or (at your option) any later version.        *
*                                                          *
*   This library is distributed in the hope that it will   *
*   be useful, but WITHOUT ANY WARRANTY; without even      *
*   the implied warranty of MERCHANTABILITY or FITNESS     *
*   FOR A PARTICULAR PURPOSE.  See the GNU Lesser          *
*   General Public License for more details.               *
*                                                          *
*   You should have received a copy of the GNU Lesser      *
*   General Public License along with this library; if     *
*   not, write to the Free Software Foundation, Inc.,      *
*   59 Temple Place, Suite 330, Boston, MA  02111-1307     *
*   USA.                                                   *
*                                                          *
***********************************************************)

open Cg
open Dict
open Dictloader
open Models
open Samplerrunner
open Printf

let main =
    Utils.print_version ();
    let output = ref ""
    and probtbl = ref ""
    and hmm = ref ""
    and probflags = ref []
    and flagstrs = ref []
    and rndthr = ref 100
    and starttemp = ref 0
    and start = ref (Syncat.create_atom "ROOT")
    and dictfname = ref ""
    and defprior = ref 1.0
    and knownderivprior = ref 5.0
    and startderivprior = ref 100.0
    and epsilon = ref 1e-25
    and goldmodel = ref ""
    and hostname = ref "localhost"
    and port = ref 12345
    and clientpath = ref "/usr/local/bin"
    and noclients = ref 4
    and fnames = ref [] in
    let args = [
        ("--probftr", 
            Arg.Symbol (
                ["roleemis"; "depemis"; "lexemis"; "heademis"], 
                set_probflags (probflags, flagstrs)), 
            "  Enable the specified probability feature");
        ("--rndthr",
            Arg.Set_int rndthr,
            "[int]   Set the threshold number of rounds of the dual decomposition (default = 100)");
        ("--epsilon",
            Arg.Set_float epsilon,
            "[float]   Set the value of epsilon (a very small value) (default = 1e-25)");
        ("--dict",
            Arg.Set_string dictfname,
            "[somefile.txt]   Set the dictionary file name (required)");
        ("--defprior",
            Arg.Set_float defprior,
            "[float]   Set the default rule prior (default = 1.0)");
        ("--knownderivprior",
            Arg.Set_float knownderivprior,
            "[float]   Set the prior for known lexicon items (default = 5.0)");
        ("--startderivprior",
            Arg.Set_float startderivprior,
            "[float]   Set the prior for sentence formation derivations (default = 100.0)");
        ("--starttemp",
            Arg.Set_int starttemp,
            "[int]   Set the starting temperature (impose structural annealing when more than 0)");
        ("--host",
            Arg.Set_string hostname,
            "[hostname]   Set the Map/Reduce server's hostname");
        ("--port",
            Arg.Set_int port,
            "[int]   Set the Map/Reduce server's port (default = 12345)");
        ("--clientpath",
            Arg.Set_string clientpath,
            "[path/to/bin]   Set the path of client programs (default = /usr/local/bin)");
        ("--noclients",
            Arg.Set_int noclients,
            "[int]   Set the number of clients to automatically fork (default = 4)");
        ("--goldmodel",
            Arg.Set_string goldmodel,
            "[output.probdb]   Set the file name of the gold probability table");
        ("--treedb",
            Arg.Set_string output,
            "[output-treedb.tdb]   Specify the tree database for the sampled trees (required at least one of these)");
        ("--output",
            Arg.Set_string probtbl,
            "[output-model.pdb]   Specify the output probability file (required as least one of these)");
        ("--hmm",
            Arg.Set_string hmm,
            "[output-model.hmm]   Specify the output HMM file (optional - this option imposes mutual posterior regularization to the parameter estimation process)")
    ]
    and fn_anon = (fun f -> fnames := !fnames @ [f])
    and usagemsg =
        "Run the Metropolis-Hastings sampler.\n\n" 
        ^ (sprintf "Usage: %s [options] chart1 chart2 ..." Sys.argv.(0)) in
    Arg.parse args fn_anon usagemsg;

    if String.length !dictfname = 0 then
        Arg.usage args usagemsg
    else ();

    if String.length !output = 0 && String.length !probtbl = 0 then
        Arg.usage args usagemsg
    else ();

    if String.length !clientpath = 0 then
        Arg.usage args usagemsg
    else
        Utils.set_clientpath !clientpath;
    Utils.set_poolsize !noclients;

    if String.length !hmm <> 0 then
        Utils.set_postreg true
    else Utils.set_postreg false;

    if !noclients > 0 then begin
        let arglist = [
            "--host"; !hostname;
            "--port"; string_of_int !port;
            "--clientpath"; !clientpath ] in
        let args = Array.of_list arglist in
        let forked_sampler = !clientpath ^ "/fungi_samplerclient" in
        ignore (
            JoinHelper.do_forks 
                forked_sampler
                (JoinHelper.Same_arguments args) 
                !noclients )
    end else ();

    Utils.set_hostname !hostname;
    Utils.set_port_estep !port;
    
    let dict = DictLoader.load !dictfname false in

    let treedbfname = !output 
    and probdbfname = !probtbl in

    if !goldmodel <> "" then
        let genmodel = Deriv.DerivGenModel.from_file !goldmodel !epsilon in
        Sampler.set_goldmodel genmodel
    else ();

    fprintf stderr ">> Input and output\n";
    fprintf stderr "     Input charts  : %s\n" (String.concat ", " !fnames);
    fprintf stderr "     Dictionary    : %s\n" !dictfname;
    if !goldmodel <> "" then
        fprintf stderr "     Gold model    : %s\n" !goldmodel
    else ();
    if !probtbl <> "" then
        fprintf stderr "     Output PCFG   : %s\n" !probtbl
    else ();
    if !hmm <> "" then
        fprintf stderr "     Output HMM    : %s\n" !hmm
    else ();
    if !output <> "" then
        fprintf stderr "     Sampled Trees : %s\n" !output
    else ();
    fprintf stderr ">> Algorithm\n";
    fprintf stderr "     Iterations    : %d\n" !rndthr;
    fprintf stderr "     Used features : %s\n" (String.concat ", " !flagstrs);
    if Utils.is_postreg () then begin
        fprintf stderr "     Posterior reg : yes\n"
    end else 
        fprintf stderr "     Posterior reg : no\n";
    fprintf stderr "     Epsilon       : %e\n" !epsilon;
    if !starttemp > 0 then
        fprintf stderr "     Starting temp : %d\n" !starttemp
    else ();
    fprintf stderr ">> Dirichlet priors\n";
    fprintf stderr "     Default       : %e\n" !defprior;
    fprintf stderr "     Known lexicon : %e\n" !knownderivprior;
    fprintf stderr "     Sentence form : %e\n" !startderivprior;
    fprintf stderr ">> Server\n";
    fprintf stderr "     Hostname      : %s (%s)\n" !hostname (Utils.get_inet_addr !hostname);
    fprintf stderr "     Port          : %d\n" !port;
    if !noclients > 0 then
        fprintf stderr "     Clients       : %d (more client connections are still welcome :D)\n" !noclients
    else
        fprintf stderr "     Clients       : 0 (waiting for client connection)\n";
    fprintf stderr ">> If you want to run the client side on another machine, please copy and paste the following command:\n\n";
    fprintf stderr "     fungi_samplerclient --host %s --port %d --clientpath PPP --noclients XXX\n\n" !hostname !port;
    fprintf stderr ">> Then change PPP to your desired path to the client code on such machine and XXX to your desired number of concurrent processes (normally equal to the number of processing cores of the machine).\n";
    fprintf stderr ">> The specified hostname must be exactly the same as one in the above command.\n\n";

    fprintf stderr "[1m>> Sampling parses from the charts [%s].[0m\n"
        (String.concat ", " !fnames);
    flush stderr;
    let ann =
        if !starttemp > 0 then Sampler.StructAnneal !starttemp
        else Sampler.StandardSampling
    in
    SamplerRunner.compute
        !probflags !rndthr !epsilon !start
        dict !defprior !knownderivprior !startderivprior
        !fnames treedbfname probdbfname !hmm ann;
    fprintf stderr ">> Done\n";
    flush stderr
