(***********************************************************
*                                                          *
*      FUNGI - Fast UNsupervised Grammar Induction         *
*                                                          *
*   Originator: Prachya Boonkwan [1], [2]                  *
*               [1] School of Informatics,                 *
*                   University of Edinburgh, UK            *
*               [2] National Electronics and Computer      *
*                   Technology Center (NECTEC), Thailand   *
*                                                          *
*   Copyright (C) 2011 Prachya Boonkwan                    *
*   <p.boonkwan@sms.ed.ac.uk, kaamanita@hotmail.com>       *
*                                                          *
*   This library is free software; you can redistribute    *
*   it and/or modify it under the terms of the GNU         *
*   Lesser General Public License as published by the      *
*   Free Software Foundation; either version 2.1 of the    *
*   License, or (at your option) any later version.        *
*                                                          *
*   This library is distributed in the hope that it will   *
*   be useful, but WITHOUT ANY WARRANTY; without even      *
*   the implied warranty of MERCHANTABILITY or FITNESS     *
*   FOR A PARTICULAR PURPOSE.  See the GNU Lesser          *
*   General Public License for more details.               *
*                                                          *
*   You should have received a copy of the GNU Lesser      *
*   General Public License along with this library; if     *
*   not, write to the Free Software Foundation, Inc.,      *
*   59 Temple Place, Suite 330, Boston, MA  02111-1307     *
*   USA.                                                   *
*                                                          *
***********************************************************)

open Cg
open Chart
open Prob
open Deriv
open Models
open Inout
open Tree
open Utils
open Printf

module ScoreInst = struct

    type syncat_t = Syncat.t
    type word_t = string

    type t = {
        syncat : syncat_t;
        head   : word_t;
    }

    let create syncat head = {
        syncat = syncat;
        head   = head;
    }

    let from_deriv deriv =
        let prems = Deriv.get_prems deriv in
        let syncat = Deriv.get_syncat prems
        and headword = Deriv.get_head prems in
        create syncat headword

    let get_syncat inst =
        inst.syncat

    let get_head inst =
        inst.head

    let equal inst1 inst2 =
        Syncat.equal inst1.syncat inst2.syncat
        && inst1.head = inst2.head

    let hash inst =
        Syncat.hash inst.syncat + Hashtbl.hash inst.head

    let to_string inst =
        sprintf "%s{%s}" (Syncat.to_string inst.syncat) inst.head

end

module ScoreTbl = ProbTbl(ScoreInst)

module PriorTbl = ProbTbl(Syncat)

module HashedInt = struct
    type t = int
    let equal = Pervasives.(=)
    let hash = Hashtbl.hash
end

module IntTbl = Hashtbl.Make(HashedInt)

type treetbl_t = (Tree.t * float) IntTbl.t

module type EXPCALC_TYPE = sig
    val calc_prob : float -> float -> float
end

module OnestepSampler = struct

    type syncat_t = Syncat.t

    type word_t = string

    type chart_t = Chart.t

    type scoreinst_t = ScoreInst.t

    type scoretbl_t = ScoreTbl.t

    module IOScoreInst = Inout.ScoreInst

    module IOScoreTbl = Inout.ScoreTbl

    type ioscoretbl_t = IOScoreTbl.t

    type genmodel_t = DerivGenModel.t

    type deriv_t = Deriv.t

    type span_t = Chart.span_t

    type dtrargs_t = Chart.dtrargs_t

    type tree_t = Tree.t

    type t = {
        epsilon         : float;
        chart           : chart_t;
        ruleprob        : genmodel_t;
        insctbl         : ioscoretbl_t;
        ruleexp         : genmodel_t;
        catexp          : scoretbl_t;
        start           : syncat_t;
        endpos          : int;
    }

    let create n epsilon ruleprob chart start ruleexp catexp =
        let insctbl = Inout.compute_insc n epsilon ruleprob chart start in
        {
            epsilon = epsilon;
            chart = chart;
            ruleprob = ruleprob;
            insctbl = insctbl;
            ruleexp = ruleexp;
            catexp = catexp;
            start = start;
            endpos = List.length (Chart.get_sentence chart) - 1;
        }
    
    let to_string sampler =
        sprintf "<Rule Expectation>\n%s\n\n<Category Expectation>\n%s\n\n<Rule Prob>\n%s"
            (DerivGenModel.to_string sampler.ruleexp)
            (ScoreTbl.to_string sampler.catexp)
            (DerivGenModel.to_string sampler.ruleprob)

    let rec sample_tree ?(uniform=false) sampler syncat span =
        let (i, j) = span in
        if i = j then begin
            let word = List.nth (Chart.get_sentence sampler.chart) i in
            let deriv = Deriv.create_deriv_lex syncat word in
            let scoreinst = ScoreInst.create syncat word in
            DerivGenModel.acc_prob sampler.ruleexp deriv 1.0;
            ScoreTbl.acc_prob sampler.catexp scoreinst 1.0;
            let lex = Tree.create_term word span in
            let tree = Tree.create_nonterm syncat word [lex] span i Syncat.Unary in
            let insc = IOScoreTbl.get_prob 
                sampler.insctbl (IOScoreInst.create span syncat word) in
            (tree, insc)
        end else begin
            let (deriv, headpos, headword, dtrargs) = 
                sample_deriv sampler syncat span ~uniform:uniform in
            let scoreinst = ScoreInst.from_deriv deriv in
            DerivGenModel.acc_prob sampler.ruleexp deriv 1.0;
            ScoreTbl.acc_prob sampler.catexp scoreinst 1.0;
            let dtrs = ref [] in
            List.iter
                (fun (syncat', span', _, _) ->
                    let (dtr, _) = sample_tree sampler syncat' span' ~uniform:uniform in
                    dtrs := !dtrs @ [dtr])
                dtrargs;
            let depdir = Deriv.get_depdir (Deriv.get_consq deriv) in
            let insc = IOScoreTbl.get_prob 
                sampler.insctbl (IOScoreInst.create span syncat headword) in
            let tree = Tree.create_nonterm syncat headword !dtrs span headpos depdir in
            (tree, insc)
        end

    and sample_deriv ?(uniform=false) sampler syncat span =
        let cprob = ref 0.0
        and dist = ref [] in
        Chart.iter_derivs_onelevel sampler.chart
            (fun _ _ _ _ -> ())
            (fun syncat' _ headpos word depdir dtrargs ->
                if Syncat.equal syncat syncat' 
                        || (span = (0, sampler.endpos) && Syncat.equal_ign_attrs syncat syncat') then begin
                    let deriv = conv_dtrargs_to_deriv syncat' word depdir dtrargs in
                    let prob = ref 1.0 in
                    if not uniform then begin
                        prob := DerivGenModel.get_prob sampler.ruleprob deriv;
                        List.iter
                            (fun (syncat'', span'', headpos'', headword'') ->
                                let inst = IOScoreInst.create span'' syncat'' headword'' in
                                let insc = IOScoreTbl.get_prob sampler.insctbl inst in
                                prob := !prob *. insc)
                            dtrargs;
                        let inst = IOScoreInst.create span syncat word in
                        let insc = IOScoreTbl.get_prob sampler.insctbl inst in
                        prob := !prob /. insc
                    end else ();
                    cprob := !cprob +. !prob;
                    dist := !dist @ [(!cprob, deriv, headpos, word, dtrargs)]
                end else ())
            span ~backoff:false;
        let r = Random.float !cprob in
        let rec find_deriv r dist =
            match dist with
                [] -> failwith "find_deriv in sample_deriv"
              | (p, deriv, headpos, headword, dtrargs) :: tl ->
                    if r < p then (deriv, headpos, headword, dtrargs)
                    else find_deriv r tl
        in
        find_deriv r !dist

    let run ?(uniform=false) sampler =
        sample_tree sampler sampler.start (0, sampler.endpos) ~uniform:uniform

    let compute ?(uniform=false) n epsilon ruleprob chart start ruleexp catexp =
        Random.self_init ();
        let sampler = 
            create n epsilon ruleprob chart start ruleexp catexp in
        let (tree, insc) = run sampler ~uniform:uniform in
        (tree, insc, sampler.ruleexp, sampler.catexp)

    let rec count_from_tree ruleexp catexp tree =
        if Tree.is_nonterm tree then begin
            let syncat = Tree.get_syncat tree in
            let headword = Tree.get_word tree in
            let depdir = Tree.get_depdir tree in
            let dtrs = Tree.get_dtrs tree in
            match dtrs with
                [] -> failwith "count_from_tree"
              | [dtr] -> begin
                    if Tree.is_term dtr then begin
                        let deriv = Deriv.create_deriv_lex syncat headword in
                        let scoreinst = ScoreInst.create syncat headword in
                        DerivGenModel.acc_prob ruleexp deriv 1.0;
                        ScoreTbl.acc_prob catexp scoreinst 1.0
                    end else failwith "count_from_tree"
                end
              | _ -> begin
                    let dtrs' = List.map 
                        (fun tree ->
                            (Tree.get_syncat tree, Tree.get_word tree))
                        dtrs in
                    let deriv = Deriv.create_deriv_dtrs
                        syncat headword dtrs' depdir in
                    let scoreinst = ScoreInst.create syncat headword in
                    DerivGenModel.acc_prob ruleexp deriv 1.0;
                    ScoreTbl.acc_prob catexp scoreinst 1.0;
                    List.iter (count_from_tree ruleexp catexp) dtrs
                end
        end else ()

    let rec deduct_from_tree ruleexp catexp tree =
        if Tree.is_nonterm tree then begin
            let syncat = Tree.get_syncat tree in
            let headword = Tree.get_word tree in
            let depdir = Tree.get_depdir tree in
            let dtrs = Tree.get_dtrs tree in
            match dtrs with
                [] -> failwith "count_from_tree"
              | [dtr] -> begin
                    if Tree.is_term dtr then begin
                        let deriv = Deriv.create_deriv_lex syncat headword in
                        let scoreinst = ScoreInst.create syncat headword in
                        DerivGenModel.acc_prob ruleexp deriv (-1.0);
                        ScoreTbl.acc_prob catexp scoreinst (-1.0)
                    end else failwith "count_from_tree"
                end
              | _ -> begin
                    let dtrs' = List.map 
                        (fun tree ->
                            (Tree.get_syncat tree, Tree.get_word tree))
                        dtrs in
                    let deriv = Deriv.create_deriv_dtrs
                        syncat headword dtrs' depdir in
                    let scoreinst = ScoreInst.create syncat headword in
                    DerivGenModel.acc_prob ruleexp deriv (-1.0);
                    ScoreTbl.acc_prob catexp scoreinst (-1.0);
                    List.iter (count_from_tree ruleexp catexp) dtrs
                end
        end else ()

end

let add_priors ruleexp catexp priorlist =
    DerivGenModel.iter ruleexp
        (fun deriv _ ->
            let scoreinst = ScoreInst.from_deriv deriv in
            let syncat = ScoreInst.get_syncat scoreinst in
            let prior = List.assoc syncat priorlist in
            DerivGenModel.acc_prob ruleexp deriv prior;
            ScoreTbl.acc_prob catexp scoreinst prior
        )

let add_priors_fn ruleexp catexp scoreprior =
    DerivGenModel.iter ruleexp
        (fun deriv _ ->
            let scoreinst = ScoreInst.from_deriv deriv in
            let prior = scoreprior (ScoreInst.get_syncat scoreinst) in
            DerivGenModel.acc_prob ruleexp deriv prior;
            ScoreTbl.acc_prob catexp scoreinst prior)

let dec_priors ruleexp catexp priorlist =
    DerivGenModel.iter ruleexp
        (fun deriv _ ->
            let scoreinst = ScoreInst.from_deriv deriv in
            let prior = (-1.0) 
                *. List.assoc (ScoreInst.get_syncat scoreinst) priorlist in
            DerivGenModel.acc_prob ruleexp deriv prior;
            ScoreTbl.acc_prob catexp scoreinst prior)

let dec_priors_fn ruleexp catexp scoreprior =
    DerivGenModel.iter ruleexp
        (fun deriv _ ->
            let scoreinst = ScoreInst.from_deriv deriv in
            let prior = (-1.0) 
                *. scoreprior (ScoreInst.get_syncat scoreinst) in
            DerivGenModel.acc_prob ruleexp deriv prior;
            ScoreTbl.acc_prob catexp scoreinst prior)

let replace_exps ruleexp catexp oldtree newtree =
    OnestepSampler.deduct_from_tree ruleexp catexp oldtree;
    OnestepSampler.count_from_tree ruleexp catexp newtree

let complete_priors ruleexp1 catexp1 ruleexp2 catexp2 priorlist =
    DerivGenModel.iter ruleexp1
        (fun deriv _ ->
            if DerivGenModel.mem ruleexp2 deriv
                    && not (DerivGenModel.mem ruleexp1 deriv)
            then begin
                let inst = ScoreInst.from_deriv deriv in
                let prior = List.assoc (ScoreInst.get_syncat inst) priorlist in
                DerivGenModel.acc_prob ruleexp2 deriv prior;
                ScoreTbl.acc_prob catexp2 inst prior
            end else ())

let copy_exps ruleexp catexp =
    let ruleexp' = create_pcfg 1000 0.0
    and catexp' = ScoreTbl.create 1000 0.0 in
    DerivGenModel.iter ruleexp
        (fun deriv prob ->
            DerivGenModel.set_prob ruleexp' deriv prob);
    ScoreTbl.iter catexp
        (fun inst prob ->
            ScoreTbl.set_prob catexp' inst prob);
    (ruleexp', catexp')

let calc_score ruleexp catexp insc =
    let score = ref (-1.0 *. log insc) in
    DerivGenModel.iter ruleexp
        (fun inst prob ->
            score := !score +. gammaln prob);
    ScoreTbl.iter catexp
        (fun inst prob ->
            score := !score -. gammaln prob);
    !score

module OnestepSamplingProblem = struct

    let identifier = "Sampler"

    type opr_t = 
        TreeInit
      | TreeSampling

    type init = unit

    type client_data = unit

    type priorlist_t = (Syncat.t * float) list

    type collection = {
        min : int;
        max : int;
        epsilon : float;
        ruleprob : DerivGenModel.t;
        start : Syncat.t;
        get_chart : int -> Chart.t;
        ruleprior : priorlist_t;
        ruleexp : DerivGenModel.t;
        catexp : ScoreTbl.t;
        fwdbwd : Hmm.fwdbwd_t;
        treetbl : treetbl_t;
        is_postreg : bool;
    }
    
    type genmodellist_t = (Deriv.t * float) list

    type scoretbllist_t = (ScoreInst.t * float) list

    type enum = {
        cur : int;
        limit : int;
        ep : float;
        rpb : genmodellist_t;
        sc : Syncat.t;
        gc : int -> Chart.t;
        rp : priorlist_t;
        rexp : genmodellist_t;
        cexp : scoretbllist_t;
        fb : Hmm.fwdbwd_t;
        tt : treetbl_t;
        pr : bool;
    }

    type input = 
        int * int * float * genmodellist_t * Chart.t * Syncat.t 
        * priorlist_t * genmodellist_t * scoretbllist_t * Hmm.fwdbwdserl_t
        * (Tree.t * float) option * bool

    type key = 
        KeyTree of int * Tree.t
      | KeyRuleExp of Deriv.t
      | KeyCatExp of ScoreInst.t
      | KeyNonterms of int * Syncat.t list

    type value = float

    type output = bool * treetbl_t * DerivGenModel.t * ScoreTbl.t * Hmm.fwdbwd_t * (Syncat.t list ref)

    let serialize_genmodel gm =
        let result = ref [] in
        DerivGenModel.iter gm
            (fun inst _ ->
                result := (inst, DerivGenModel.get_prob gm inst) :: !result);
        !result

    let serialize_scoretbl st =
        let result = ref [] in
        ScoreTbl.iter st
            (fun inst _ ->
                result := (inst, ScoreTbl.get_prob st inst) :: !result);
        !result

    let serialize_priors rp chart =
        let lst = ref [] in
        List.iter
            (fun syncat ->
                lst := (syncat, rp syncat) :: !lst)
            (Chart.get_nonterms chart);
        !lst

    let enum_start range = 
        {
            cur = range.min;
            limit = range.max;
            ep = range.epsilon;
            rpb = serialize_genmodel range.ruleprob;
            sc = range.start;
            gc = range.get_chart;
            rp = range.ruleprior;
            rexp = serialize_genmodel range.ruleexp;
            cexp = serialize_scoretbl range.catexp;
            fb = range.fwdbwd;
            tt = range.treetbl;
            pr = range.is_postreg;
        }

    let create_genmodel gml epsilon =
        let gm = create_pcfg 1000 epsilon in
        List.iter
            (fun (inst, prob) ->
                DerivGenModel.acc_prob gm inst prob)
            gml;
        DerivGenModel.calc_probftrs gm;
        DerivGenModel.calc_normfacts gm;
        gm

    let create_scoretbl stl epsilon =
        let st = ScoreTbl.create 1000 epsilon in
        List.iter
            (fun (inst, prob) ->
                ScoreTbl.acc_prob st inst prob)
            stl;
        st

    let enum_step e =
        if e.cur <= e.limit then begin
            Scale.inc_and_display ();
            let chart = e.gc e.cur in
            let treeinst =
                if not (IntTbl.mem e.tt e.cur) then None
                else Some (IntTbl.find e.tt e.cur) in
            let input = (
                e.cur, e.limit, e.ep,
                e.rpb, chart, e.sc, 
                e.rp, e.rexp, e.cexp,
                Hmm.serialize e.fb, treeinst, e.pr
            ) in
            Some (input, {
                cur = e.cur + 1;
                limit = e.limit;
                ep = e.ep;
                rpb = e.rpb;
                sc = e.sc;
                gc = e.gc;
                rp = e.rp;
                rexp = e.rexp;
                cexp = e.cexp;
                fb = e.fb;
                tt = e.tt;
                pr = e.pr;
            })
        end else None

    let init_client () = ()

    let compare_keys = Pervasives.compare

    let map () (
        idx, noentries, epsilon, ruleproblst, chart, start, 
        priorlist, rexp, cexp, fbl, treeinst, is_postreg
    ) =
        let ruleprob = create_genmodel ruleproblst epsilon in
        let ruleexp = create_genmodel rexp epsilon
        and catexp = create_scoretbl cexp epsilon 
        and fwdbwd = Hmm.deserialize fbl in
        let result = ref [] in
        begin
            match treeinst with
                None -> begin
                    let ruleexp' = Models.create_pcfg 1000 epsilon
                    and catexp' = ScoreTbl.create 1000 epsilon in
                    let (tree', insc', _, _) = 
                        OnestepSampler.compute 
                            noentries epsilon ruleprob chart start 
                            ruleexp' catexp' ~uniform:false in
                    if is_postreg then
                        let chart' = Tree.to_chart tree' in
                        Hmm.init_onestep fwdbwd chart'
                    else ();
                    DerivGenModel.iter ruleexp'
                        (fun deriv prob ->
                            result := (KeyRuleExp deriv, prob) :: !result);
                    ScoreTbl.iter catexp'
                        (fun inst prob ->
                            result := (KeyCatExp inst, prob) :: !result);
                    result := (KeyTree (idx, tree'), neg_infinity) :: !result;
                    result := (KeyNonterms (idx, Chart.get_nonterms chart), 0.0) :: !result
                end
              | Some (tree, oldscore) -> begin
                    let (ruleexp', catexp') = copy_exps ruleexp catexp in
                    let (tree', insc', _, _) = 
                        OnestepSampler.compute 
                            noentries epsilon ruleprob chart start 
                            ruleexp' catexp' ~uniform:false in
                    add_priors ruleexp' catexp' priorlist;
                    complete_priors ruleexp catexp ruleexp' catexp' priorlist;
                    let score' = calc_score ruleexp' catexp' insc' in
                    if score' > oldscore then begin
                        result := (KeyTree (idx, tree'), score') :: !result;
                        DerivGenModel.iter ruleexp'
                            (fun deriv prob ->
                                result := (KeyRuleExp deriv, prob) :: !result);
                        ScoreTbl.iter catexp'
                            (fun inst prob ->
                                result := (KeyCatExp inst, prob) :: !result)
                    end else begin
                        result := (KeyTree (idx, tree), oldscore) :: !result
                    end
                end
        end;
        !result

    let combine v1 v2 = 
        v1 +. v2

    let reduce k prob (is_init, treetbl, ruleexp, catexp, fwdbwd, nonterms) =
        begin
            match k with
                KeyTree (idx, tree) -> begin
                    IntTbl.replace treetbl idx (tree, prob);
                    let chart = Tree.to_chart tree in
                    if is_init then Hmm.init_onestep fwdbwd chart
                    else Hmm.train_onestep fwdbwd chart
                end
              | KeyRuleExp deriv ->
                    DerivGenModel.acc_prob ruleexp deriv prob
              | KeyCatExp inst ->
                    ScoreTbl.acc_prob catexp inst prob
              | KeyNonterms (idx, nts) ->
                    nonterms := Utils.union nts !nonterms
        end;
        (is_init, treetbl, ruleexp, catexp, fwdbwd, nonterms)

end

module OnestepSamplerMapRed = Mapred.Make(OnestepSamplingProblem)

let goldmodel = ref None

let set_goldmodel genmodel =
    goldmodel := Some genmodel

type fn_scale_prob_t = int -> float -> float

let default_scale_prob _ prob = prob

let step = ref 0.0

let structanneal starttemp rndthr rnd prob =
    if !step = 0.0 then
        step := (float_of_int rndthr) /. (float_of_int starttemp)
    else ();
    exp (
        (float_of_int rnd) *. (log prob) *. !step
    )

type anneal_t = 
    StandardSampling
  | StructAnneal of int

module Sampler = struct

    type syncat_t = Syncat.t

    type word_t = string

    type chart_t = Chart.t

    type scoreinst_t = ScoreInst.t

    type scoretbl_t = ScoreTbl.t

    module IOScoreInst = Inout.ScoreInst

    module IOScoreTbl = Inout.ScoreTbl

    type ioscoretbl_t = IOScoreTbl.t

    type genmodel_t = DerivGenModel.t

    type deriv_t = Deriv.t

    type span_t = Chart.span_t

    type dtrargs_t = Chart.dtrargs_t

    type tree_t = Tree.t

    type fn_get_chart_t = int -> chart_t

    exception Server_not_found

    module Problem = OnestepSamplingProblem

    module MR = OnestepSamplerMapRed

    type server_t = MR.server_t

    let server conf i = 
        MR.server conf i

    type t = {
        noentries          : int;
        rndthr             : int;
        epsilon            : float;
        ruleprob           : genmodel_t;
        ruleprior          : syncat_t -> float;
        start              : syncat_t;
        get_chart          : fn_get_chart_t;
        nocharts           : int;
        treetbl            : treetbl_t;
        fwdbwd             : Hmm.fwdbwd_t;
        mutable scale_prob : fn_scale_prob_t;
        mutable server     : server_t option;
        mutable rnd        : int;
        mutable nonterms   : syncat_t list ref;
        mutable priorlist  : (syncat_t * float) list ref;
    }

    let create
        n rndthr epsilon ruleprob ruleprior start get_chart nocharts fwdbwd =
    {
        noentries  = n;
        rndthr     = rndthr;
        epsilon    = epsilon;
        ruleprob   = ruleprob;
        ruleprior  = ruleprior;
        start      = start;
        get_chart  = get_chart;
        nocharts   = nocharts;
        treetbl    = IntTbl.create nocharts;
        fwdbwd     = fwdbwd;
        scale_prob = default_scale_prob;
        server     = None;
        rnd        = 0;
        nonterms   = ref [];
        priorlist  = ref [];
    }

    let set_server sampler svr =
        sampler.server <- svr

    let get_round sampler =
        sampler.rnd

    let reset_round sampler =
        sampler.rnd <- 0

    let set_scale_prob sampler fn =
        sampler.scale_prob <- fn

    let compute_onernd_client () =
        MR.client (Utils.client_conf_treesampler ())

    let init_treetbl sampler =
        match sampler.server with
            Some svr -> begin
                Scale.add (Scale.Number sampler.nocharts) ">> Initializing charts " " ";
                let ruleexp = create_pcfg sampler.nocharts 0.0
                and catexp = ScoreTbl.create sampler.nocharts 0.0 in
                let chartcoll = {
                    Problem.min = 0;
                    Problem.max = sampler.nocharts - 1;
                    Problem.epsilon = sampler.epsilon;
                    Problem.ruleprob = sampler.ruleprob;
                    Problem.start = sampler.start;
                    Problem.get_chart = sampler.get_chart;
                    Problem.ruleprior = [];
                    Problem.ruleexp = ruleexp;
                    Problem.catexp = catexp;
                    Problem.fwdbwd = sampler.fwdbwd;
                    Problem.treetbl = sampler.treetbl;
                    Problem.is_postreg = Utils.is_postreg ();
                } in
                let starttime = Unix.time () in
                ignore (MR.compute svr chartcoll (
                    true,
                    sampler.treetbl, ruleexp, catexp, sampler.fwdbwd,
                    sampler.nonterms
                ));
                let endtime = Unix.time () in
                Scale.discard ();
                let runtime = (endtime -. starttime) /. 60.0 in
                Scale.print (sprintf 
                    "     Initialization: no charts = %d, runtime = %.2f mins\n" 
                    sampler.nocharts runtime);
                List.iter
                    (fun syncat ->
                        sampler.priorlist := (syncat, sampler.ruleprior syncat) :: !(sampler.priorlist))
                    !(sampler.nonterms);
                (ruleexp, catexp)
            end
          | None -> failwith "init_treetbl"

    let compute_params sampler ruleexp catexp =
        add_priors_fn ruleexp catexp sampler.ruleprior;
        if Utils.is_postreg () then
            Hmm.estimate sampler.fwdbwd
        else ();
        let oldruleprob = DerivGenModel.create 1000 sampler.epsilon [] in
        DerivGenModel.iter sampler.ruleprob
            (fun inst prob ->
                DerivGenModel.acc_prob oldruleprob inst prob);
        Scale.add (Scale.WindMill 1000) ">> Computing parameters " " ";
        DerivGenModel.iter ruleexp
            (fun deriv _ ->
                Scale.inc_and_display ();
                let prob = DerivGenModel.get_prob ruleexp deriv in
                let prems = Deriv.get_prems deriv in
                let syncat = Deriv.get_syncat prems in
                let scoreinst = ScoreInst.from_deriv deriv in
                let prob' = ScoreTbl.get_prob catexp scoreinst in
                let newprob = prob /. prob' in
                let result = ref (sampler.scale_prob sampler.rnd newprob) in
                if Utils.is_postreg () then begin
                    let consq = Deriv.get_consq deriv in
                    if Deriv.is_consq_lex consq then begin
                        let word = Deriv.get_word consq in
                        let inst = Hmm.EmisInst.create syncat word in
                        let emisprob = Hmm.get_emisprob sampler.fwdbwd in
                        let mult = Hmm.EmisProb.get_prob emisprob inst in
                        result := !result *. (sampler.scale_prob sampler.rnd mult);
                        Hmm.EmisProb.set_prob emisprob inst !result
                    end else ()
                end else ();
                DerivGenModel.set_prob sampler.ruleprob deriv !result);
        dec_priors_fn ruleexp catexp sampler.ruleprior;
        DerivGenModel.calc_probftrs sampler.ruleprob;
        let dkl = DerivGenModel.dkl oldruleprob sampler.ruleprob in
        Scale.discard ();
        dkl

    let compute_onernd_server sampler ruleexp catexp =
        match sampler.server with
            Some svr -> begin
                Scale.add (Scale.Number sampler.nocharts) "Sampling charts " " ";
                let chartcoll = {
                    Problem.min = 0;
                    Problem.max = sampler.nocharts - 1;
                    Problem.epsilon = sampler.epsilon;
                    Problem.ruleprob = sampler.ruleprob;
                    Problem.start = sampler.start;
                    Problem.get_chart = sampler.get_chart;
                    Problem.ruleprior = !(sampler.priorlist);
                    Problem.ruleexp = ruleexp;
                    Problem.catexp = catexp;
                    Problem.fwdbwd = sampler.fwdbwd;
                    Problem.treetbl = sampler.treetbl;
                    Problem.is_postreg = Utils.is_postreg ();
                } in
                let ruleexp' = create_pcfg 1000 0.0
                and catexp' = ScoreTbl.create 1000 0.0 in
                ignore (MR.compute svr chartcoll (
                     true, sampler.treetbl, ruleexp', catexp', sampler.fwdbwd,
                     sampler.nonterms
                ));
                Scale.discard ();
                if Utils.is_postreg () then begin
                    Scale.add (Scale.WindMill 1000) "Estimating HMM parameters " " ";
                    IntTbl.iter
                        (fun idx (tree, score) ->
                            Scale.inc_and_display ();
                            let chart = Tree.to_chart tree in
                            Hmm.train_onestep sampler.fwdbwd chart)
                        sampler.treetbl;
                    Scale.discard ()
                end else ();
                (ruleexp', catexp')
            end
          | None -> raise Server_not_found

    let run sampler =
        let (ruleexp, catexp) = init_treetbl sampler in
        let r = ref ruleexp
        and c = ref catexp in
        Scale.add (Scale.Number sampler.rndthr) ">> Iteration " " > ";
        for i = 1 to sampler.rndthr do
            sampler.rnd <- sampler.rnd + 1;
            Scale.inc_and_display ();
            let starttime = Unix.time () in
            let dkl = compute_params sampler !r !c in
            let (ruleexp', catexp') = compute_onernd_server sampler ruleexp catexp in
            r := ruleexp';
            c := catexp';
            let endtime = Unix.time () in
            let runtime = (endtime -. starttime) /. 60.0 in
            Scale.clearline ();
            begin
                match !goldmodel with
                    None ->
                        Scale.print (sprintf 
                            "     Iteration %d : D_KL(old | new) = %e, runtime = %.2f mins\n" 
                            sampler.rnd dkl runtime)
                  | Some gm -> begin
                        let golddkl = DerivGenModel.dkl gm sampler.ruleprob in 
                        Scale.print (sprintf 
                            "     Iteration %d: D_KL(old | new) = %e, D_KL(gold | new) = %e, runtime = %.2f mins\n" 
                            sampler.rnd dkl golddkl runtime)
                    end;
            end;
            Scale.redisplay ()
        done;
        Scale.discard ()

    let compute
            ?(svr = None) ?(scale_prob = default_scale_prob)
            n rndthr epsilon ruleprob ruleprior start get_chart nocharts fwdbwd
            : (tree_t * float) list =
        let sampler = create 
            n rndthr epsilon ruleprob ruleprior start get_chart nocharts fwdbwd in
        set_server sampler svr;
        set_scale_prob sampler scale_prob;
        run sampler;
        let result = ref [] in
        for i = sampler.nocharts - 1 downto 0 do
            let (tree, score) = IntTbl.find sampler.treetbl i in
            result := (tree, score) :: !result
        done;
        !result

end
