(***********************************************************
*                                                          *
*      FUNGI - Fast UNsupervised Grammar Induction         *
*                                                          *
*   Originator: Prachya Boonkwan [1], [2]                  *
*               [1] School of Informatics,                 *
*                   University of Edinburgh, UK            *
*               [2] National Electronics and Computer      *
*                   Technology Center (NECTEC), Thailand   *
*                                                          *
*   Copyright (C) 2011 Prachya Boonkwan                    *
*   <p.boonkwan@sms.ed.ac.uk, kaamanita@hotmail.com>       *
*                                                          *
*   This library is free software; you can redistribute    *
*   it and/or modify it under the terms of the GNU         *
*   Lesser General Public License as published by the      *
*   Free Software Foundation; either version 2.1 of the    *
*   License, or (at your option) any later version.        *
*                                                          *
*   This library is distributed in the hope that it will   *
*   be useful, but WITHOUT ANY WARRANTY; without even      *
*   the implied warranty of MERCHANTABILITY or FITNESS     *
*   FOR A PARTICULAR PURPOSE.  See the GNU Lesser          *
*   General Public License for more details.               *
*                                                          *
*   You should have received a copy of the GNU Lesser      *
*   General Public License along with this library; if     *
*   not, write to the Free Software Foundation, Inc.,      *
*   59 Temple Place, Suite 330, Boston, MA  02111-1307     *
*   USA.                                                   *
*                                                          *
***********************************************************)

open Cg
open Chart
open Dict
open Cky
open Chartdb
open Prob
open Deriv
open Models
open Inout
open Sampler
open Printf
open Tree
open Treedb
open Utils

module SamplerRunner = struct

    type syncat_t = Syncat.t

    let is_bo_consq consq =
        if not (Deriv.is_consq_dtrs consq) then false
        else
            let cwlist = Deriv.get_dtrs consq in
            let rec check cwlist' =
                match cwlist' with
                    [] -> false
                  | (c, w) :: rest ->
                        if Syncat.is_bo_cat c then true
                        else check rest in
            check cwlist

    let score_prior 
            deriv dict start defprior knownderivprior startderivprior =
        let prems = Deriv.get_prems deriv
        and consq = Deriv.get_consq deriv in
        if Syncat.is_bo_cat (Deriv.get_syncat prems) || is_bo_consq consq then
            defprior
        else if Deriv.is_consq_lex consq 
                && not (Dict.is_uncertain dict (Deriv.get_word consq)) then
            knownderivprior
        else if Syncat.equal_ign_attrs (Deriv.get_syncat prems) start then
            startderivprior
        else defprior

    let score_syncat
            syncat dict start defprior knownderivprior startderivprior =
        if Syncat.is_bo_cat syncat then defprior
        else if Syncat.equal_ign_attrs syncat start then startderivprior
        else knownderivprior
    
    let decapsulate_tree start tree =
        if Syncat.equal_ign_attrs (Tree.get_syncat tree) start then begin
            if Tree.is_nonterm tree then begin
                let dtrs = Tree.get_dtrs tree in
                let depdir = Tree.get_depdir tree in
                match depdir with
                    Syncat.Left -> List.nth dtrs 0
                  | Syncat.Right -> List.nth dtrs 1
                  | Syncat.Coor -> List.nth dtrs 0
                  | Syncat.Unary -> List.nth dtrs 0
            end else tree
        end else tree

    let compute 
            probflags
            rndthr epsilon start 
            dict defprior knownderivprior startderivprior 
            fnames ofname probfname hmmfname 
            anneal =
        let n = 1000 in
        let svr = 
            Some (Sampler.server (Utils.server_conf_treesampler ()) ()) in
        let wrapper = ChartDbWrapper.from_files fnames in
        let get_chart = ChartDbWrapper.find wrapper in
        let nocharts = ChartDbWrapper.length wrapper in
        let ruleprob = Models.create_model nocharts epsilon probflags in
        let ruleprior = (
            fun syncat -> 
                score_syncat
                    syncat dict start defprior knownderivprior startderivprior
        ) in
        let fwdbwd = Hmm.create epsilon in
        let treeprobs = 
            match anneal with
                StandardSampling ->
                    Sampler.compute
                        n rndthr epsilon ruleprob ruleprior start 
                        get_chart nocharts fwdbwd ~svr:svr 
              | StructAnneal starttemp ->
                    Sampler.compute
                        n rndthr epsilon ruleprob ruleprior start 
                        get_chart nocharts fwdbwd ~svr:svr 
                        ~scale_prob:(structanneal starttemp rndthr)
        in
        if String.length ofname > 0 then
            let treedb = TreeDb.create ofname in begin
                List.iter
                    (fun (tree, prob) ->
                        let tree' = decapsulate_tree start tree in
                        let depstr = Tree.to_depstr tree' in
                        TreeDb.add treedb (MarshTree.create tree' prob depstr))
                    treeprobs;
                TreeDb.to_file treedb ofname
        end else ();
        if String.length probfname > 0 then
            DerivGenModel.to_file ruleprob probfname
        else ();
        if Utils.is_postreg () then
            Hmm.to_file fwdbwd hmmfname
        else ()

end

let set_probflags (probflags, flagstrs) str =
    let probflag = match str with
        "roleemis" -> RoleEmis
      | "depemis" -> JuliaEmis
      | "lexemis" -> LexEmis
      | "heademis" -> HeadEmis 
      | _ -> failwith "set_probflags" in
    probflags := probflag :: !probflags;
    match str with
        "roleemis" -> flagstrs := "role emission" :: !flagstrs
      | "depemis" -> flagstrs := "dependent word emission" :: !flagstrs
      | "lexemis" -> flagstrs := "lexicon emission" :: !flagstrs
      | "heademis" -> flagstrs := "head-word emission" :: !flagstrs
      | _ -> failwith "set_probflags"
