(***********************************************************
*                                                          *
*      FUNGI - Fast UNsupervised Grammar Induction         *
*                                                          *
*   Originator: Prachya Boonkwan [1], [2]                  *
*               [1] School of Informatics,                 *
*                   University of Edinburgh, UK            *
*               [2] National Electronics and Computer      *
*                   Technology Center (NECTEC), Thailand   *
*                                                          *
*   Copyright (C) 2011 Prachya Boonkwan                    *
*   <p.boonkwan@sms.ed.ac.uk, kaamanita@hotmail.com>       *
*                                                          *
*   This library is free software; you can redistribute    *
*   it and/or modify it under the terms of the GNU         *
*   Lesser General Public License as published by the      *
*   Free Software Foundation; either version 2.1 of the    *
*   License, or (at your option) any later version.        *
*                                                          *
*   This library is distributed in the hope that it will   *
*   be useful, but WITHOUT ANY WARRANTY; without even      *
*   the implied warranty of MERCHANTABILITY or FITNESS     *
*   FOR A PARTICULAR PURPOSE.  See the GNU Lesser          *
*   General Public License for more details.               *
*                                                          *
*   You should have received a copy of the GNU Lesser      *
*   General Public License along with this library; if     *
*   not, write to the Free Software Foundation, Inc.,      *
*   59 Temple Place, Suite 330, Boston, MA  02111-1307     *
*   USA.                                                   *
*                                                          *
***********************************************************)

open Printf

type disp_t =
    Guage of int
  | Number of int
  | Percentage of int
  | NumberNoMax
  | WindMill of int
  | GuageWithWindMill of int * int
  | GuageWithNumber of int
  | GuageWithPercentage of int
  | PercentageWithWindMill of int * int
  | StaticInfix
  | MapReduce
  | MapCombineReduce
  | MapReduceWithWindMill of int

module Scale = struct

    type t = {
        mutable indstk  : int ref list;
        mutable dispstk : disp_t list;
        mutable cntstk  : int ref list;
        mutable cnt2stk : int ref list;
        mutable cnt3stk : int ref list;
        mutable oldstrs : string ref list;
        mutable prfstks : string list;
        mutable sfxstks : string list;
    }

    let guagelen = 40.0

    let create () = 
    {
        indstk  = [];
        dispstk = [];
        cntstk  = [];
        cnt2stk = [];
        cnt3stk = [];
        oldstrs = [];
        prfstks = [];
        sfxstks = [];
    }

    let bsp n =
        fprintf stderr "%s[K" (String.make n '\x08');
        flush stderr

    let length scale =
        List.length scale.indstk

    let peek_top scale =
        (
            List.hd scale.indstk, List.hd scale.dispstk, 
            List.hd scale.cntstk, List.hd scale.cnt2stk, 
            List.hd scale.cnt3stk, List.hd scale.oldstrs,
            List.hd scale.prfstks, List.hd scale.sfxstks
        )

    let pop scale =
        scale.indstk <- List.tl scale.indstk;
        scale.dispstk <- List.tl scale.dispstk;
        scale.cntstk <- List.tl scale.cntstk;
        scale.cnt2stk <- List.tl scale.cnt2stk;
        scale.cnt3stk <- List.tl scale.cnt3stk;
        scale.oldstrs <- List.tl scale.oldstrs;
        scale.prfstks <- List.tl scale.prfstks;
        scale.sfxstks <- List.tl scale.sfxstks

    let push scale disp cnt cnt2 cnt3 prefix suffix =
        scale.indstk <- ref 0 :: scale.indstk;
        scale.dispstk <- disp :: scale.dispstk;
        scale.cntstk <- ref cnt :: scale.cntstk;
        scale.cnt2stk <- ref cnt2 :: scale.cnt2stk;
        scale.cnt3stk <- ref cnt3 :: scale.cnt3stk;
        scale.oldstrs <- ref "" :: scale.oldstrs;
        scale.prfstks <- prefix :: scale.prfstks;
        scale.sfxstks <- suffix :: scale.sfxstks

    let gen_guage cnt max =
        let scl = float_of_int cnt *. guagelen /. float_of_int max in
        let bar = String.make (int_of_float scl) '=' in
        "|" ^ bar ^ ">"

    let gen_windmill cnt friction =
        let millno = (cnt / friction) mod 4 in
        String.sub "|/-\\" millno 1

    let gen_number cnt max =
        sprintf "%d/%d" cnt max

    let gen_number_nomax cnt =
        sprintf "%d" cnt

    let gen_percentage cnt max =
        let percentage = 
            100.0 *. (float_of_int cnt) /. (float_of_int max) in
        sprintf "%6.2f%%" percentage

    let gen_twocounters cnt cnt2 =
        sprintf "mapped %d/reduced %d" cnt cnt2

    let gen_threecounters cnt cnt2 cnt3 =
        sprintf "sent %d/recv %d/comp %d" cnt cnt2 cnt3

    let show_meter ?(force=false) ind disp cnt cnt2 cnt3 oldstr prefix suffix =
        let inner = match disp with
            Guage max -> gen_guage !cnt max
          | Number max -> gen_number !cnt max
          | Percentage max -> gen_percentage !cnt max
          | NumberNoMax -> gen_number_nomax !cnt
          | WindMill friction -> gen_windmill !cnt friction
          | GuageWithWindMill (max, friction) ->
                (gen_guage !cnt max) ^ " " ^ (gen_windmill !cnt friction)
          | GuageWithNumber max ->
                (gen_guage !cnt max) ^ " " ^ (gen_number !cnt max)
          | GuageWithPercentage max ->
                (gen_guage !cnt max) ^ " " ^ (gen_percentage !cnt max)
          | PercentageWithWindMill (max, friction) ->
                (gen_percentage !cnt max) ^ " "
                ^ (gen_windmill !cnt friction) 
          | StaticInfix -> "" 
          | MapReduce -> gen_twocounters !cnt !cnt2 
          | MapCombineReduce -> gen_threecounters !cnt !cnt3 !cnt2 
          | MapReduceWithWindMill (friction) ->
                (gen_twocounters !cnt !cnt2) ^ " " ^ (gen_windmill !cnt3 friction)
        in
        let str = prefix ^ inner ^ suffix in
        if str <> !oldstr || force then begin
            bsp !ind;
            let len = String.length str in
            ind := len;
            fprintf stderr "%s" str;
            flush stderr;
            oldstr := str
        end else ()

    let add scale disp prefix suffix =
        push scale disp 0 0 0 prefix suffix

    let discard scale =
        if length scale > 0 then
            let (ind, _, _, _, _, _, _, _) = peek_top scale in
            bsp !ind;
            pop scale
        else ()

    let clearline scale =
        List.iter (fun ind -> bsp !ind) scale.indstk

    let inc scale =
        if length scale > 0 then
            let (_, _, cnt, _, _, _, _, _) = peek_top scale in
            cnt := !cnt + 1
        else ()

    let inc2 scale =
        if length scale > 0 then
            let (_, _, _, cnt2, _, _, _, _) = peek_top scale in
            cnt2 := !cnt2 + 1
        else ()

    let inc3 scale =
        if length scale > 0 then
            let (_, _, _, _, cnt3, _, _, _) = peek_top scale in
            cnt3 := !cnt3 + 1
        else ()

    let display scale =
        if length scale > 0 then
            let (ind, disp, cnt, cnt2, cnt3, oldstr, prefix, suffix) = 
                peek_top scale in
            show_meter ind disp cnt cnt2 cnt3 oldstr prefix suffix
        else ()

    let inc_and_display scale =
        if length scale > 0 then
            let (_, _, cnt, _, _, _, _, _) = peek_top scale in
            cnt := !cnt + 1
        else ();
        if length scale > 0 then
            let (ind, disp, cnt, cnt2, cnt3, oldstr, prefix, suffix) = 
                peek_top scale in
            show_meter ind disp cnt cnt2 cnt3 oldstr prefix suffix
        else ()

    let inc2_and_display scale =
        if length scale > 0 then
            let (_, _, _, cnt2, _, _, _, _) = peek_top scale in
            cnt2 := !cnt2 + 1
        else ();
        if length scale > 0 then
            let (ind, disp, cnt, cnt2, cnt3, oldstr, prefix, suffix) = 
                peek_top scale in
            show_meter ind disp cnt cnt2 cnt3 oldstr prefix suffix
        else ()

    let inc3_and_display scale =
        if length scale > 0 then
            let (_, _, _, _, cnt3, _, _, _) = peek_top scale in
            cnt3 := !cnt3 + 1
        else ();
        if length scale > 0 then
            let (ind, disp, cnt, cnt2, cnt3, oldstr, prefix, suffix) = 
                peek_top scale in
            show_meter ind disp cnt cnt2 cnt3 oldstr prefix suffix
        else ()

    let redisplay scale =
        for i = List.length scale.indstk - 1 downto 0 do
            let oldstr = List.nth scale.oldstrs i in
            fprintf stderr "%s" !oldstr;
            flush stderr
        done

end

let scale = Scale.create ()

let add disp prefix suffix =
    Scale.add scale disp prefix suffix

let discard () =
    Scale.discard scale

let inc () =
    Scale.inc scale

let inc2 () =
    Scale.inc2 scale

let inc3 () =
    Scale.inc3 scale

let display () =
    Scale.display scale

let inc_and_display () =
    Scale.inc_and_display scale

let inc2_and_display () =
    Scale.inc2_and_display scale

let inc3_and_display () =
    Scale.inc3_and_display scale

let print text =
    fprintf stderr "%s" text;
    flush stderr

let clearline () =
    Scale.clearline scale

let redisplay () =
    Scale.redisplay scale
