(***********************************************************
*                                                          *
*      FUNGI - Fast UNsupervised Grammar Induction         *
*                                                          *
*   Originator: Prachya Boonkwan [1], [2]                  *
*               [1] School of Informatics,                 *
*                   University of Edinburgh, UK            *
*               [2] National Electronics and Computer      *
*                   Technology Center (NECTEC), Thailand   *
*                                                          *
*   Copyright (C) 2011 Prachya Boonkwan                    *
*   <p.boonkwan@sms.ed.ac.uk, kaamanita@hotmail.com>       *
*                                                          *
*   This library is free software; you can redistribute    *
*   it and/or modify it under the terms of the GNU         *
*   Lesser General Public License as published by the      *
*   Free Software Foundation; either version 2.1 of the    *
*   License, or (at your option) any later version.        *
*                                                          *
*   This library is distributed in the hope that it will   *
*   be useful, but WITHOUT ANY WARRANTY; without even      *
*   the implied warranty of MERCHANTABILITY or FITNESS     *
*   FOR A PARTICULAR PURPOSE.  See the GNU Lesser          *
*   General Public License for more details.               *
*                                                          *
*   You should have received a copy of the GNU Lesser      *
*   General Public License along with this library; if     *
*   not, write to the Free Software Foundation, Inc.,      *
*   59 Temple Place, Suite 330, Boston, MA  02111-1307     *
*   USA.                                                   *
*                                                          *
***********************************************************)

open Dep
open Tree
open Treedb
open Printf

let ubracket_to_string ubracket =
    let (begpos, endpos) = ubracket in
    sprintf "(%d, %d)" begpos endpos

let ubrackets_to_string ubrackets =
    sprintf "[%s]"
        (String.concat ", " 
            (List.map ubracket_to_string ubrackets))

let bracket_to_string bracket =
    let ((begpos, endpos), head) = bracket in
    sprintf "(%d, %d, %s)" begpos endpos head

let brackets_to_string brackets =
    sprintf "[%s]"
        (String.concat ", " 
            (List.map bracket_to_string brackets))

let showtreedb showtree showdep showprob showubrackets showbrackets fname =
    let treedb = (* TreeDb.opendb fname [Mdb.ReadOnly] in *)
        TreeDb.from_file fname in
    TreeDb.iter
        (fun id treemonad ->
            match treemonad with
                Some (tree, prob, depstr) -> 
                begin
                    printf "# Tree %d\n" (id + 1);
                    if showtree then
                        printf "%s\n" (Tree.pp tree)
                    else ();
                    if showdep then
                        printf "%s\n" (Tree.to_conll tree)
                    else ();
                    if showubrackets then
                        printf "<Unlabeled Brackets>\n%s\n"
                            (ubrackets_to_string (Tree.to_ubrackets tree))
                    else ();
                    if showbrackets then
                        printf "<Labeled Brackets>\n%s\n"
                            (brackets_to_string (Tree.to_brackets tree))
                    else ();
                    if showprob then
                        printf "<Probability>\nexp(%e)\n" prob
                    else ();
                    printf "\n"
                end
              | None ->
                begin
                    printf "<Tree %d>\nno parses\n\n" id
                end)
        treedb (* ;
    TreeDb.close treedb *)

let main =
    Utils.print_version ();
    let fnames = ref [] in
    let showtree = ref false
    and showdep = ref false 
    and showprob = ref false 
    and showubrackets = ref false
    and showbrackets = ref false in
    let args = [
        ("--tree", 
            Arg.Set showtree, 
            "  Also display tree structures");
        ("--dep", 
            Arg.Set showdep, 
            "  Also display dependency structures in the CoNLL2006 format");
        ("--prob",
            Arg.Set showprob, 
            "  Also display the probability of each tree");
        ("--ubrackets",
            Arg.Set showubrackets, 
            "  Also display unlabeled brackets of each tree");
        ("--brackets",
            Arg.Set showbrackets, 
            "  Also display labeled brackets of each tree")]
    and fn_anon = (fun f -> fnames := !fnames @ [f])
    and usagemsg = 
        "Display all trees in the specified tree databases.\n\n"
        ^ (sprintf "Usage: %s [options] treedb1 treedb2 ..." Sys.argv.(0)) in
    Arg.parse args fn_anon usagemsg;
    if not (!showtree || !showdep || !showprob || !showubrackets || !showbrackets) then
        showtree := true
    else ();
    List.iter (showtreedb !showtree !showdep !showprob !showubrackets !showbrackets) !fnames
