(***********************************************************
*                                                          *
*      FUNGI - Fast UNsupervised Grammar Induction         *
*                                                          *
*   Originator: Prachya Boonkwan [1], [2]                  *
*               [1] School of Informatics,                 *
*                   University of Edinburgh, UK            *
*               [2] National Electronics and Computer      *
*                   Technology Center (NECTEC), Thailand   *
*                                                          *
*   Copyright (C) 2011 Prachya Boonkwan                    *
*   <p.boonkwan@sms.ed.ac.uk, kaamanita@hotmail.com>       *
*                                                          *
*   This library is free software; you can redistribute    *
*   it and/or modify it under the terms of the GNU         *
*   Lesser General Public License as published by the      *
*   Free Software Foundation; either version 2.1 of the    *
*   License, or (at your option) any later version.        *
*                                                          *
*   This library is distributed in the hope that it will   *
*   be useful, but WITHOUT ANY WARRANTY; without even      *
*   the implied warranty of MERCHANTABILITY or FITNESS     *
*   FOR A PARTICULAR PURPOSE.  See the GNU Lesser          *
*   General Public License for more details.               *
*                                                          *
*   You should have received a copy of the GNU Lesser      *
*   General Public License along with this library; if     *
*   not, write to the Free Software Foundation, Inc.,      *
*   59 Temple Place, Suite 330, Boston, MA  02111-1307     *
*   USA.                                                   *
*                                                          *
***********************************************************)

open Cg
open Dict
open Dictloader
open Models
open Annealrunner
open Printf

let main =
    Utils.print_version ();
    let probflags = ref []
    and flagstrs = ref []
    and inoutmode = ref StandardInOut
    and inoutstr = ref "standard"
    and babystep = ref false
    and leapfrogthr = ref 15
    and annmode = ref NoAnn
    and annstr = ref "None"
    and diffthr = ref 1e-4
    and rndthr = ref 5
    and epsilon = ref 1e-25
    and start = ref (Syncat.create_atom "ROOT")
    and dictfname = ref ""
    and defprior = ref 1.0
    and knownderivprior = ref 5.0
    and startderivprior = ref 100.0
    and initparam = ref 0.0
    and epoch = ref 0.0
    and finalparam = ref 0.0
    and servermode = ref true
    and clientmode_probinit = ref false
    and clientmode_estep = ref false
    and clientpath = ref "/usr/local/bin"
    and hostname = ref "localhost"
    and port_probinit = ref 12345
    and port_estep = ref 12346
    and noclients = ref 4
    and backoff_flag = ref false
    and goldmodel = ref ""
    and postreg_flag = ref false
    and fnames = ref []
    and ofname = ref "" 
    and ohmm = ref "" in
    let args = [
        ("--probftr", 
            Arg.Symbol (
                ["roleemis"; "depemis"; "lexemis"; "heademis"], 
                set_probflags (probflags, flagstrs)), 
            "  Enable the specified probability feature");
        ("--inout",
            Arg.Symbol (["standard"; "vb"], set_inoutmode (inoutmode, inoutstr)),
            "  Select the Inside/Outside Algorithm (default = standard)");
        ("--babystep",
            Arg.Set babystep,
            "  Use the Babystep estimation (Spitkovsky et al, 2010) (default = false)");
        ("--leapfrogthr",
            Arg.Set_int leapfrogthr,
            "[int]   Set the leapfrog threshold (default = 15)");
        ("--annealing", 
            Arg.Symbol (["none"; "da"; "sda"; "sa"], set_annmode (annmode, annstr)), 
            "  Select the simulated annealing algorithm (default = none)");
        ("--diffthr",
            Arg.Set_float diffthr,
            "[float]   Set the termination threshold of expectation difference measured by Kullback-Leibler distance (default = 1e-4)");
        ("--rndthr",
            Arg.Set_int rndthr,
            "[int]   Set the termination threshold of iteration number (default = 5)");
        ("--epsilon",
            Arg.Set_float epsilon,
            "[float]   Set the value of epsilon (a very small value) (default = 1e-25)");
        ("--defprior",
            Arg.Set_float defprior,
            "[float]   Set the default rule prior (default = 1.0)");
        ("--knownderivprior",
            Arg.Set_float knownderivprior,
            "[float]   Set the prior for known lexicon items (default = 5.0)");
        ("--startderivprior",
            Arg.Set_float startderivprior,
            "[float]   Set the prior for sentence formation derivations (default = 100.0)");
        ("--initparam",
            Arg.Set_float initparam,
            "[float]   Set the simulated annealing's initial parameter value (required by da, sda, and sa)");
        ("--epoch",
            Arg.Set_float epoch,
            "[float]   Set the simulated annealing's epoch value (required by da, sda, and sa)");
        ("--finalparam",
            Arg.Set_float finalparam,
            "[float]   Set the simulated annealing's final parameter value (required by sa)");
        ("--host",
            Arg.Set_string hostname,
            "[hostname]   Set the Map/Reduce server's hostname");
        ("--port-probinit",
            Arg.Set_int port_probinit,
            "[int]   Set the Map/Reduce server's port");
        ("--port-estep",
            Arg.Set_int port_estep,
            "[int]   Set the Map/Reduce server's port");
        ("--clientpath",
            Arg.Set_string clientpath,
            "[path/to/bin]   Set the path of client programs (default = /usr/local/bin)");
        ("--noclients",
            Arg.Set_int noclients,
            "[int]   Set the number of clients to automatically fork (default = 4)");
        ("--backoff",
            Arg.Set backoff_flag,
            "   Use the backoff model in addition to the PCFG (default = false).");
        ("--postreg",
            Arg.Set postreg_flag,
            "   Use the mutual posterior regularization (default = false).");
        ("--dict",
            Arg.Set_string dictfname,
            "[somefile.txt]   Set the dictionary file name (required)");
        ("--goldmodel",
            Arg.Set_string goldmodel,
            "[output.probdb]   Set the file name of the gold probability table");
        ("--output",
            Arg.Set_string ofname,
            "[output.probdb]   Set the output file name for the result probability table (required)");
        ("--hmm",
            Arg.Set_string ohmm,
            "[output.probdb]   Set the output file name for the result HMM (optional)")
    ]
    and fn_anon = (fun f -> fnames := !fnames @ [f])
    and usagemsg =
        "Train the model with the specified chart databases.\n\n"
        ^ (sprintf "Usage: %s [options] chartdb1 chartdb2 ..." Sys.argv.(0)) in
    Arg.parse args fn_anon usagemsg;
    if !dictfname = "" || !ofname = "" then
        Arg.usage args usagemsg
    else ();
    if !clientpath = "" then
        Arg.usage args usagemsg
    else
        Utils.set_clientpath !clientpath;
    Utils.set_poolsize !noclients;

    if !hostname = "localhost" then begin
        try
            let h = Unix.getenv "GE_O_HOST" in
            if h <> "" then hostname := h
            else ()
        with Not_found -> ()
    end else ();
    if !goldmodel <> "" then
        let genmodel = Deriv.DerivGenModel.from_file !goldmodel !epsilon in
        Em.set_goldmodel genmodel
    else ();
    if !backoff_flag then
        Utils.use_backoff ()
    else ();
    Utils.set_postreg !postreg_flag;
    match !annmode, !initparam, !epoch, !finalparam with
        DetAnn, _, 0.0, _
      | DetAnn, 0.0, _, _
      | SkewedDetAnn, _, 0.0, _
      | SkewedDetAnn, 0.0, _, _
      | StructAnn, _, 0.0, _ -> Arg.usage args usagemsg
      | _, _, _, _ -> ();
    let annparam = conv_annparam !annmode !initparam !epoch !finalparam in
    if !clientmode_probinit || !clientmode_estep then
        servermode := false
    else
        servermode := true;
    let dict = DictLoader.load !dictfname false in
    if not !clientmode_probinit && not !clientmode_estep then begin
        fprintf stderr ">> Input and output\n";
        fprintf stderr "     Input charts  : %s\n" (String.concat ", " !fnames);
        fprintf stderr "     Dictionary    : %s\n" !dictfname;
        fprintf stderr "     Output PCFG   : %s\n" !ofname;
        if !ohmm <> "" then
            fprintf stderr "     Output HMM    : %s\n" !ohmm
        else ();
        if !goldmodel <> "" then
            fprintf stderr "     Gold model    : %s\n" !goldmodel
        else ();
        fprintf stderr ">> Algorithm\n";
        fprintf stderr "     Type of EM    : %s\n" !inoutstr;
        fprintf stderr "     Used features : %s\n" (String.concat ", " !flagstrs);
        if !babystep then begin
            fprintf stderr "     Babystep      : yes\n";
            fprintf stderr "     Leapfrog      : %d\n" !leapfrogthr
        end else
            fprintf stderr "     Babystep      : no\n";
        if !backoff_flag then begin
            fprintf stderr "     Backoff       : yes\n"
        end else 
            fprintf stderr "     Backoff       : no\n";
        if Utils.is_postreg () then begin
            fprintf stderr "     Posterior reg : yes\n"
        end else 
            fprintf stderr "     Posterior reg : no\n";
        fprintf stderr "     Min D_KL      : %e\n" !diffthr;
        fprintf stderr "     Max round no  : %d\n" !rndthr;
        fprintf stderr "     Epsilon       : %e\n" !epsilon;
        fprintf stderr ">> Dirichlet priors\n";
        fprintf stderr "     Default       : %e\n" !defprior;
        fprintf stderr "     Known lexicon : %e\n" !knownderivprior;
        fprintf stderr "     Sentence form : %e\n" !startderivprior;
        fprintf stderr ">> Annealing technique\n";
        fprintf stderr "     Technique     : %s\n" !annstr;
        if !annstr = "Deterministic Annealing" || !annstr = "Skewed Deterministic Annealing" || !annstr = "Structural Annealing" then begin
            fprintf stderr "     Initial param : %e\n" !initparam
        end else ();
        if !annstr = "Deterministic Annealing" || !annstr = "Skewed Deterministic Annealing" || !annstr = "Structural Annealing" then begin
            fprintf stderr "     Epoch         : %e\n" !epoch
        end else ();
        if !annstr = "Structural Annealing" then begin
            fprintf stderr "     Final param   : %e\n" !finalparam
        end else
            fprintf stderr "     Final param   : 1.0\n";
        fprintf stderr ">> Server\n";
        fprintf stderr "     Hostname           : %s (%s)\n" !hostname (Utils.get_inet_addr !hostname);
        fprintf stderr "     Port for prob init : %d\n" !port_probinit;
        fprintf stderr "     Port for E Step    : %d\n" !port_estep;
        if !noclients > 0 then
            fprintf stderr "     Clients            : %d (more client connections are still welcome :D)\n" !noclients
        else
            fprintf stderr "     Clients            : 0 (waiting for client connection)\n";
        fprintf stderr ">> If you want to run the client side on another machine, please copy and paste the following command:\n\n";
        fprintf stderr "     fungi_client --host %s --port-probinit %d --port-estep %d --clientpath PPP --noclients XXX\n\n" !hostname !port_probinit !port_estep;
        fprintf stderr ">> Then change PPP to your desired path to the client code on such machine and XXX to your desired number of concurrent processes (normally equal to the number of processing cores of the machine).\n";
        fprintf stderr ">> The specified hostname must be exactly the same as one in the above command.\n\n";
        fprintf stderr ">> ";
        flush stderr
    end else ();
    leapfrogthr := !leapfrogthr + 1;
    if !noclients > 0 then begin
        let arglist_probinit = [
            "--host"; !hostname;
            "--port"; string_of_int !port_probinit;
            "--clientpath"; !clientpath ]
        and arglist_estep = [
            "--host"; !hostname;
            "--port"; string_of_int !port_estep;
            "--clientpath"; !clientpath ] in
        let args_probinit = Array.of_list arglist_probinit in
        let args_estep = Array.of_list arglist_estep in
        let forked_probinit = !clientpath ^ "/fungi_probinit" 
        and forked_estep = !clientpath ^ "/fungi_estep" in
        ignore (
            JoinHelper.do_forks 
                forked_probinit 
                (JoinHelper.Same_arguments args_probinit) 
                !noclients );
        ignore (
            JoinHelper.do_forks 
                forked_estep 
                (JoinHelper.Same_arguments args_estep) 
                !noclients )
    end else ();
    let starttime = Unix.time () in
    Utils.set_hostname !hostname;
    Utils.set_port_probinit !port_probinit;
    Utils.set_port_estep !port_estep;
    AnnealRunner.compute 
        !probflags !inoutmode !babystep annparam !diffthr !rndthr !epsilon 
        !start dict !defprior !knownderivprior !startderivprior !leapfrogthr
        !fnames !ofname !ohmm !servermode !clientmode_probinit !clientmode_estep;
    let endtime = Unix.time () in
    if not !clientmode_probinit && not !clientmode_estep then begin
        fprintf stderr ">> Done\n";
        flush stderr;
        let duration = endtime -. starttime in
        fprintf stderr ">> Total runtime: %.2f mins\n" (duration /. 60.0)
    end else ();

