(***********************************************************
*                                                          *
*      FUNGI - Fast UNsupervised Grammar Induction         *
*                                                          *
*   Originator: Prachya Boonkwan [1], [2]                  *
*               [1] School of Informatics,                 *
*                   University of Edinburgh, UK            *
*               [2] National Electronics and Computer      *
*                   Technology Center (NECTEC), Thailand   *
*                                                          *
*   Copyright (C) 2011 Prachya Boonkwan                    *
*   <p.boonkwan@sms.ed.ac.uk, kaamanita@hotmail.com>       *
*                                                          *
*   This library is free software; you can redistribute    *
*   it and/or modify it under the terms of the GNU         *
*   Lesser General Public License as published by the      *
*   Free Software Foundation; either version 2.1 of the    *
*   License, or (at your option) any later version.        *
*                                                          *
*   This library is distributed in the hope that it will   *
*   be useful, but WITHOUT ANY WARRANTY; without even      *
*   the implied warranty of MERCHANTABILITY or FITNESS     *
*   FOR A PARTICULAR PURPOSE.  See the GNU Lesser          *
*   General Public License for more details.               *
*                                                          *
*   You should have received a copy of the GNU Lesser      *
*   General Public License along with this library; if     *
*   not, write to the Free Software Foundation, Inc.,      *
*   59 Temple Place, Suite 330, Boston, MA  02111-1307     *
*   USA.                                                   *
*                                                          *
***********************************************************)

open Printf

open Cg
open Dep
open Chart
open Dict
open Utils

module Tree = struct

    type syncat_t = CG.syncat_t
    type depdir_t = CG.depdir_t
    type span_t = Chart.span_t
    type word_t = Dict.word_t
    type depstruct_t = DepStruct.t

    type t =
        Term of word_t * span_t
      | Nonterm of syncat_t * word_t * t list * span_t * int * depdir_t

    let create_term word span =
        Term (word, span)

    let create_nonterm syncat word dtrs span headpos depdir =
        Nonterm (syncat, word, dtrs, span, headpos, depdir)

    let is_term tree =
        match tree with
            Term _ -> true
          | Nonterm _ -> false

    let is_nonterm tree =
        match tree with
            Term _ -> false
          | Nonterm _ -> true

    let get_word tree =
        match tree with
            Term (word, _) -> word
          | Nonterm (_, word, _, _, _, _) -> word

    let get_span tree =
        match tree with
            Term (_, span) -> span
          | Nonterm (_, _, _, span, _, _) -> span

    let get_syncat tree =
        match tree with
            Term _ -> failwith "get_syncat"
          | Nonterm (syncat, _, _, _, _, _) -> syncat

    let get_dtrs tree =
        match tree with
            Term _ -> failwith "get_dtrs"
          | Nonterm (_, _, dtrs, _, _, _) -> dtrs

    let get_headpos tree =
        match tree with
            Term (_, (headpos, _)) -> headpos
          | Nonterm (_, _, _, _, headpos, _) -> headpos

    let get_depdir tree =
        match tree with
            Term _ -> failwith "get_depdir"
          | Nonterm (_, _, _, _, _, depdir) -> depdir

    let rec to_string tree =
        match tree with
            Term (word, _) -> word
          | Nonterm (_, word, dtrs, _, _, _) ->
                sprintf "(%s %s)" word 
                    (String.concat " " (List.map to_string dtrs))

    let rec print_spaces indent =
        let result = ref "" in
        for i = 1 to indent do
            result := " " ^ !result;
        done;
        !result

    let rec pp ?(indent=0) tree =
        match tree with
            Term (word, _) -> word
          | Nonterm (_, word, dtrs, _, _, depdir) ->
                let newindent = indent + (String.length word) + 4 in
                let leading = sprintf "\n%s" (print_spaces newindent) in
                sprintf "(%s %s %s)" word
                    (Syncat.depdir_to_string (get_depdir tree))
                    (String.concat leading
                        (List.map 
                            (fun dtr -> pp dtr ~indent:newindent) dtrs))

    let rec pp_syncat ?(indent=0) tree =
        match tree with
            Term (word, _) -> word
          | Nonterm (syncat, word, dtrs, _, _, depdir) ->
                let newindent = indent 
                    + (String.length (Syncat.to_string syncat)) 
                    + (String.length word) + 5 in
                let leading = sprintf "\n%s" (print_spaces newindent) in
                sprintf "(%s:%s %s %s)" 
                    (Syncat.to_string syncat) word
                    (Syncat.depdir_to_string (get_depdir tree))
                    (String.concat leading
                        (List.map 
                            (fun dtr -> 
                                pp_syncat dtr ~indent:newindent) 
                            dtrs))

    let get_rootpos tree =
        match tree with
            Term (_, (begpos, _)) -> begpos
          | Nonterm (_, _, _, _, headpos, _) -> headpos

    let rec tree_to_sent tree =
        match tree with
            Term (word, _) -> [word]
          | Nonterm (_, _, dtrs, _, _, _) ->
                List.concat (List.map tree_to_sent dtrs)

    let rec tree_to_catseq tree =
        match tree with
            Term _ -> []
          | Nonterm (syncat, _, dtrs, _, _, _) -> begin
                match dtrs with
                    [Term (word, _)] -> [syncat]
                  | _ -> List.concat (List.map tree_to_catseq dtrs)
            end

    let rec tree_to_depstr tree depstr =
        match tree with
            Term (_, _) -> ()
          | Nonterm (_, _, dtrs, _, headpos, _) ->
            begin
                List.iter
                    (fun dtr -> 
                        let dtrpos = get_rootpos dtr in
                        DepStruct.add depstr headpos dtrpos)
                    dtrs;
                List.iter
                    (fun dtr -> tree_to_depstr dtr depstr)
                    dtrs
            end

    let to_depstr tree =
        let sent = tree_to_sent tree in
        let rootpos = get_rootpos tree in
        let depstr = DepStruct.create (List.length sent) sent rootpos in
        tree_to_depstr tree depstr;
        depstr

    let rec to_ubrackets tree : span_t list =
        match tree with
            Term (_, span) -> 
                [span]
          | Nonterm (_, _, dtrs, span, _, _) ->
                union [span] (List.flatten (List.map to_ubrackets dtrs))

    let rec to_brackets tree : (span_t * word_t) list =
        match tree with
            Term (head, span) ->
                [(span, head)]
          | Nonterm (_, head, dtrs, span, _, _) ->
                union [(span, head)] 
                    (List.flatten (List.map to_brackets dtrs))

    let rec to_catwordlist tree : (syncat_t * word_t) list =
        match tree with
            Term _ -> []
          | Nonterm (syncat, headword, dtrs, span, headpos, depdir) -> begin
                match depdir with
                    Syncat.Left | Syncat.Right | Syncat.Coor ->
                        List.concat (List.map to_catwordlist dtrs)
                  | Syncat.Unary -> [(syncat, headword)]
            end

    let rec conv_to_chart chart tree =
        match tree with
            Term _ -> 0
          | Nonterm (syncat, headword, dtrs, span, headpos, depdir) -> begin
                match dtrs with
                    [] -> failwith "conv_to_chart"
                  | [dtr] -> begin
                        match dtr with
                            Term _ ->
                                let prod = Item.create_prod_lex headword headpos in
                                let item = Chart.add_deriv_ret_item 
                                    chart span syncat headpos prod in
                                Item.get_id item
                          | Nonterm _ -> failwith "conv_to_chart"
                    end
                  | _ -> begin
                        let dtrids = List.map (conv_to_chart chart) dtrs in
                        let prod = Item.create_prod_ids dtrids headpos depdir in
                        let item = Chart.add_deriv_ret_item
                            chart span syncat headpos prod in
                        Item.get_id item
                    end
            end

    let to_chart tree : Chart.t =
        let sent = tree_to_sent tree in
        let chart = Chart.create sent in
        ignore (conv_to_chart chart tree);
        chart

    let to_conll tree =
        let sent = tree_to_sent tree in
        let catseq = tree_to_catseq tree in
        let rootpos = get_rootpos tree in
        let depstr = DepStruct.create (List.length sent) sent rootpos in
        tree_to_depstr tree depstr;
        let lines = ref [] in
        for i = 0 to List.length sent - 1 do
            let id = i + 1 in
            let form = List.nth sent i in
            let lemma = List.nth sent i in
            let cpostag = Syncat.to_string (List.nth catseq i) in
            let postag = cpostag in
            let ftrs = "_" in
            let head =
                try (DepStruct.find_head depstr i) + 1
                with Not_found -> 0 in
            let deprel = 
                if head = 0 then "ROOT" 
                else "HEAD" in
            let phead = head in
            let pdeprel = deprel in
            let line = Printf.sprintf
                "%d\t%s\t%s\t%s\t%s\t%s\t%d\t%s\t%d\t%s"
                id form lemma cpostag postag ftrs head deprel phead pdeprel
            in
            lines := !lines @ [line]
        done;
        String.concat "\n" !lines

end
