(** Syntactic Tree. *)

(***********************************************************
*                                                          *
*      FUNGI - Fast UNsupervised Grammar Induction         *
*                                                          *
*   Originator: Prachya Boonkwan [1], [2]                  *
*               [1] School of Informatics,                 *
*                   University of Edinburgh, UK            *
*               [2] National Electronics and Computer      *
*                   Technology Center (NECTEC), Thailand   *
*                                                          *
*   Copyright (C) 2011 Prachya Boonkwan                    *
*   <p.boonkwan@sms.ed.ac.uk, kaamanita@hotmail.com>       *
*                                                          *
*   This library is free software; you can redistribute    *
*   it and/or modify it under the terms of the GNU         *
*   Lesser General Public License as published by the      *
*   Free Software Foundation; either version 2.1 of the    *
*   License, or (at your option) any later version.        *
*                                                          *
*   This library is distributed in the hope that it will   *
*   be useful, but WITHOUT ANY WARRANTY; without even      *
*   the implied warranty of MERCHANTABILITY or FITNESS     *
*   FOR A PARTICULAR PURPOSE.  See the GNU Lesser          *
*   General Public License for more details.               *
*                                                          *
*   You should have received a copy of the GNU Lesser      *
*   General Public License along with this library; if     *
*   not, write to the Free Software Foundation, Inc.,      *
*   59 Temple Place, Suite 330, Boston, MA  02111-1307     *
*   USA.                                                   *
*                                                          *
***********************************************************)

(** Syntactic Tree. *)
module Tree :
sig

    (** Syntactic category. *)
    type syncat_t = Cg.CG.syncat_t
    
    (** Dependency direction. *)
    type depdir_t = Cg.CG.depdir_t
   
    (** Span. *)
    type span_t = Chart.Chart.span_t
    
    (** Lexical item. *)
    type word_t = Dict.Dict.word_t

    (** Dependency structure. *)
    type depstruct_t = Dep.DepStruct.t
    
    (** Syntactic tree. *)
    type t
    
    (** [create_atom w s] creates a terminal node from a word [w] 
        and a span [s]. *)
    val create_term : word_t -> span_t -> t
    
    (** [create_nonterm c w dtrs s h d] creates a nonterminal node from 
        a syntactic category [c], a node label [w], a list of daughter 
        nodes [dtrs], a span [s], a head word's position [h], and a
        dependency direction [d]. *)
    val create_nonterm :
        syncat_t -> word_t -> t list -> span_t -> int -> depdir_t -> t
    
    (** [is_term t] checks if a tree [t] is a terminal node. *)
    val is_term : t -> bool
    
    (** [is_nonterm t] checks if a node [t] is a nonterminal node. *)
    val is_nonterm : t -> bool
    
    (** [get_word t] get the label of the node [t]. *)
    val get_word : t -> word_t
    
    (** [get_span t] gets the span of a node [t]. *)
    val get_span : t -> span_t
    
    (** [get_syncat t] gets the syntactic category of a nonterminal 
        node [t]. If [t] is not a nonterminal node, a failure exception
        will be thrown. *)
    val get_syncat : t -> syncat_t
    
    (** [get_dtrs t] gets the daughter list of a nonterminal node [t].
        If [t] is not a nonterminal node, a failure exception will be
        thrown. *)
    val get_dtrs : t -> t list
    
    (** [get_headpos t] gets the position of the head word of a 
        nonterminal node [t]. *)
    val get_headpos : t -> int
    
    (** [get_depdir t] gets the dependency direction of a nonterminal 
        node [t]. *)
    val get_depdir : t -> depdir_t
    
    (** to_string t] represents a tree [t] with a string. *)
    val to_string : t -> string
    
    (** [pp t] pretty-prints a tree [t] into a string. *)
    val pp : ?indent:int -> t -> string
    
    (** [pp_syncat t] pretty-prints a tree [t] into a string, but only
        syntactic categories are displayed. *)
    val pp_syncat : ?indent:int -> t -> string

    (** [to_depstr t] converts a tree [t] into an equivalent dependency 
        structure. *)
    val to_depstr : t -> depstruct_t

    (** [to_ubrackets t] converts a tree [t] into a set of unlabeled
        brackets. *)
    val to_ubrackets : t -> span_t list
    
    (** [to_brackets t] converts a tree [t] into a set of labeled
        brackets. *)
    val to_brackets : t -> (span_t * word_t) list

    val to_catwordlist : t -> (syncat_t * word_t) list

    val to_chart : t -> Chart.Chart.t

    val to_conll : t -> string

end
