(***********************************************************
*                                                          *
*      FUNGI - Fast UNsupervised Grammar Induction         *
*                                                          *
*   Originator: Prachya Boonkwan [1], [2]                  *
*               [1] School of Informatics,                 *
*                   University of Edinburgh, UK            *
*               [2] National Electronics and Computer      *
*                   Technology Center (NECTEC), Thailand   *
*                                                          *
*   Copyright (C) 2011 Prachya Boonkwan                    *
*   <p.boonkwan@sms.ed.ac.uk, kaamanita@hotmail.com>       *
*                                                          *
*   This library is free software; you can redistribute    *
*   it and/or modify it under the terms of the GNU         *
*   Lesser General Public License as published by the      *
*   Free Software Foundation; either version 2.1 of the    *
*   License, or (at your option) any later version.        *
*                                                          *
*   This library is distributed in the hope that it will   *
*   be useful, but WITHOUT ANY WARRANTY; without even      *
*   the implied warranty of MERCHANTABILITY or FITNESS     *
*   FOR A PARTICULAR PURPOSE.  See the GNU Lesser          *
*   General Public License for more details.               *
*                                                          *
*   You should have received a copy of the GNU Lesser      *
*   General Public License along with this library; if     *
*   not, write to the Free Software Foundation, Inc.,      *
*   59 Temple Place, Suite 330, Boston, MA  02111-1307     *
*   USA.                                                   *
*                                                          *
***********************************************************)

let rec cartprod list1 list2 =
    match list1 with
        [] -> []
      | hd1 :: tl1 ->
            (cartprod_elt hd1 list2) @ (cartprod tl1 list2)

and cartprod_elt hd1 list2 =
    match list2 with
        [] -> []
      | hd2 :: tl2 -> 
            (hd1, hd2) :: (cartprod_elt hd1 tl2)

let rec cartmap fn list1 list2 =
    match list1 with
        [] -> []
      | hd1 :: tl1 ->
            (cartmap_elt fn hd1 list2) @ (cartmap fn tl1 list2)

and cartmap_elt fn hd1 list2 =
    match list2 with
        [] -> []
      | hd2 :: tl2 -> 
            (fn hd1 hd2) :: (cartmap_elt fn hd1 tl2)

let rec cartmap_noerr fn list1 list2 =
    match list1 with
        [] -> []
      | hd1 :: tl1 ->
            (cartmap_noerr_elt fn hd1 list2) 
                @ (cartmap_noerr fn tl1 list2)

and cartmap_noerr_elt fn hd1 list2 =
    match list2 with
        [] -> []
      | hd2 :: tl2 -> 
            try
                (fn hd1 hd2) :: (cartmap_noerr_elt fn hd1 tl2)
            with _ ->
                cartmap_noerr_elt fn hd1 tl2

let rec cartiter fn list1 list2 =
    match list1 with
        [] -> ()
      | hd1 :: tl1 ->
        begin
            cartiter_elt fn hd1 list2;
            cartiter fn tl1 list2
        end

and cartiter_elt fn hd1 list2 =
    match list2 with
        [] -> ()
      | hd2 :: tl2 -> 
        begin
            fn hd1 hd2;
            cartiter_elt fn hd1 tl2
        end

let rec cartiter_noerr fn list1 list2 =
    match list1 with
        [] -> ()
      | hd1 :: tl1 ->
        begin
            cartiter_noerr_elt fn hd1 list2;
            cartiter_noerr fn tl1 list2
        end

and cartiter_noerr_elt fn hd1 list2 =
    match list2 with
        [] -> ()
      | hd2 :: tl2 -> 
            try begin
                fn hd1 hd2;
                cartiter_noerr_elt fn hd1 tl2
            end with _ ->
                cartiter_noerr_elt fn hd1 tl2

let rec elim_dups mylist =
    elim_dups_elt mylist []

and elim_dups_elt mylist result =
    match mylist with
        [] -> result
      | elt :: rest ->
            if List.mem elt result then
                elim_dups_elt rest result
            else
                elim_dups_elt rest (elt :: result)

let rec intersect list1 list2 =
    match list1 with
        [] -> []
      | elt :: rest ->
            if List.mem elt list2 && not (List.mem elt rest) then 
                elt :: (intersect rest list2)
            else intersect rest list2

let rec diff list1 list2 =
    match list1 with
        [] -> []
      | elt :: rest ->
            if List.mem elt list2 || List.mem elt rest then 
                diff rest list2
            else elt :: (diff rest list2)

let list_to_set list1 =
    intersect list1 list1

let union list1 list2 =
    let list1' = list_to_set list1 in
    list1' @ (diff list2 list1')

let marshal value = 
    Marshal.to_string value [Marshal.Closures]

let unmarshal ?(begpos=0) mstr = 
    Marshal.from_string mstr begpos

let marshal_to_file fname value =
    let ofhdl = open_out_bin fname in
    Marshal.to_channel ofhdl value [];
    close_out ofhdl

let unmarshal_from_file fname =
    let ifhdl = open_in_bin fname in
    let result = Marshal.from_channel ifhdl in
    close_in ifhdl;
    result

let digamma x =
    assert (x > 0.0);
    let result = ref 0.0 in
    let y = ref x in
    while !y < 7.0 do
        result := !result -. 1.0 /. !y;
        y := !y +. 1.0
    done;
    y := !y -. 0.5;
    let xx = 1.0 /. !y in
    let xx2 = xx *. xx in
    let xx4 = xx2 *. xx2 in
    !result 
        +. log x
        +. xx2 /. 24.0 
        -. (7.0 /. 960.0) *. xx4
        +. (31.0 /. 8064.0) *. xx2 *. xx4
        -. (127.0 /. 30720.0) *. xx4 *. xx4

let pi = 4.0 *. atan 1.0

let gammaln x =
    let z = 1.0 /. x in
    let z2 = z *. z in
    let z3 = z *. z2 in
    let z5 = z3 *. z2 in
    let z7 = z5 *. z2 in
    let z9 = z7 *. z2 in
    let result = ref (x *. log x) in
    result := !result -. x +. 0.5 *. log (2.0 *. pi *. z);
    result := !result +. z /. 12.0;
    result := !result -. z3 /. 360.0;
    result := !result +. z5 /. 1260.0;
    result := !result -. z7 /. 1680.0;
    result := !result +. z9 /. 1188.0;
    !result

let gamma x = exp (gammaln x)

let rec except_pos list1 pos =
    except_pos_elt list1 pos 0

and except_pos_elt list1 pos idx =
    match list1 with
        [] -> []
      | hd :: rest ->
            if pos <> idx then 
                hd :: except_pos_elt rest pos (idx + 1)
            else
                except_pos_elt rest pos (idx + 1)

let rec split' text textlen idx curtok tokens =
    if textlen = idx then
        if String.length curtok = 0 then List.rev tokens
        else (List.rev tokens) @ [curtok]
    else
        let front = text.[idx] in
        if front = ' ' || front = '\t' then
            if String.length curtok = 0 then
                split' text textlen (idx + 1) "" tokens
            else
                split' text textlen (idx + 1) "" (curtok :: tokens)
        else
            let curtok' = curtok ^ (String.sub text idx 1) in
            split' text textlen (idx + 1) curtok' tokens

let split text =
    let textlen = String.length text in
    split' text textlen 0 "" []

let rec split_delim' delimlen delim text textlen idx curtok tokens =
    let textlen = String.length text in
    if idx = textlen then
        if String.length curtok = 0 then List.rev tokens
        else (List.rev tokens) @ [curtok]
    else if idx + delimlen < textlen then
        let front = String.sub text idx delimlen in
        if delim = front then
            if String.length curtok = 0 then
                split_delim' delimlen delim text textlen (idx + delimlen) "" tokens
            else
                split_delim' delimlen delim text textlen (idx + delimlen) "" (curtok :: tokens)
        else
            let curtok' = curtok ^ (String.sub text idx 1) in
            split_delim' delimlen delim text textlen (idx + 1) curtok' tokens
    else
        let curtok' = curtok ^ (String.sub text idx 1) in
        split_delim' delimlen delim text textlen (idx + 1) curtok' tokens

let split_delim delim text =
    let textlen = String.length text
    and delimlen = String.length delim in
    split_delim' delimlen delim text textlen 0 "" []

let string_of_list ?(bracket = true) fn lst =
    if bracket then
        Printf.sprintf "[%s]" (String.concat ", " (List.map fn lst))
    else
        String.concat ", " (List.map fn lst)

let ending_mark = "<eos>"

let poolsize = ref 1

let set_poolsize size = 
    poolsize := size

let get_poolsize () =
    !poolsize

let hostname = ref "localhost"

let port_probinit = ref 12345

let port_estep = ref 12346

let clientpath = ref ""

let set_hostname hn =
    hostname := hn

let set_port_probinit p =
    port_probinit := p

let set_port_estep p =
    port_estep := p

let set_clientpath path =
    let path =
        if path = "" then path
        else if path.[String.length path - 1] <> '/' then
            path ^ "/"
        else path in
    clientpath := path

let server_conf_probinit () : JoinHelper.configuration = 
    let arglist = [
        "--host"; !hostname;
        "--port"; string_of_int !port_probinit
    ] in
    let args = Array.of_list arglist in
    {
        JoinHelper.host = !hostname;
        JoinHelper.port = !port_probinit;
        JoinHelper.clients = (-1);
        JoinHelper.forked_program = !clientpath ^ "fungi_probinit";
        JoinHelper.fork_args = JoinHelper.Same_arguments args;
        JoinHelper.magic_id = "magic-number";
        JoinHelper.magic_value = "magic-value";
    }

let server_conf_estep () : JoinHelper.configuration = 
    let arglist = [
        "--host"; !hostname;
        "--port"; string_of_int !port_estep
    ] in
    let args = Array.of_list arglist in
    {
        JoinHelper.host = !hostname;
        JoinHelper.port = !port_estep;
        JoinHelper.clients = (-1);
        JoinHelper.forked_program = !clientpath ^ "fungi_estep";
        JoinHelper.fork_args = JoinHelper.Same_arguments args;
        JoinHelper.magic_id = "magic-number";
        JoinHelper.magic_value = "magic-value";
    }

let server_conf_treesampler () : JoinHelper.configuration = 
    let arglist = [
        "--host"; !hostname;
        "--port"; string_of_int !port_estep
    ] in
    let args = Array.of_list arglist in
    {
        JoinHelper.host = !hostname;
        JoinHelper.port = !port_estep;
        JoinHelper.clients = (-1);
        JoinHelper.forked_program = !clientpath ^ "fungi_samplerclient";
        JoinHelper.fork_args = JoinHelper.Same_arguments args;
        JoinHelper.magic_id = "magic-number";
        JoinHelper.magic_value = "magic-value";
    }

let get_inet_addr hostname =
    try
        let addr = Unix.gethostbyname hostname in
        Unix.string_of_inet_addr addr.Unix.h_addr_list.(0)
    with Not_found -> "*unknown-address*"

let client_conf_probinit () : JoinHelper.configuration =
    let arglist = [
        "--host"; !hostname;
        "--port"; string_of_int !port_probinit;
        "--clientpath"; !clientpath
    ] in
    let args = Array.of_list arglist in
    {
        JoinHelper.host = !hostname;
        JoinHelper.port = !port_probinit;
        JoinHelper.clients = (-1);
        JoinHelper.forked_program = !clientpath ^ "fungi_probinit";
        JoinHelper.fork_args = JoinHelper.Same_arguments args;
        JoinHelper.magic_id = "magic-number";
        JoinHelper.magic_value = "magic-value";
    }

let client_conf_estep () : JoinHelper.configuration =
    let arglist = [
        "--host"; !hostname;
        "--port"; string_of_int !port_estep;
        "--clientpath"; !clientpath;
    ] in
    let args = Array.of_list arglist in
    {
        JoinHelper.host = !hostname;
        JoinHelper.port = !port_estep;
        JoinHelper.clients = (-1);
        JoinHelper.forked_program = !clientpath ^ "fungi_estep";
        JoinHelper.fork_args = JoinHelper.Same_arguments args;
        JoinHelper.magic_id = "magic-number";
        JoinHelper.magic_value = "magic-value";
    }

let client_conf_treesampler () : JoinHelper.configuration =
    let arglist = [
        "--host"; !hostname;
        "--port"; string_of_int !port_estep;
        "--clientpath"; !clientpath;
    ] in
    let args = Array.of_list arglist in
    {
        JoinHelper.host = !hostname;
        JoinHelper.port = !port_estep;
        JoinHelper.clients = (-1);
        JoinHelper.forked_program = !clientpath ^ "fungi_samplerclient";
        JoinHelper.fork_args = JoinHelper.Same_arguments args;
        JoinHelper.magic_id = "magic-number";
        JoinHelper.magic_value = "magic-value";
    }

let backoff_flag = ref false

let use_backoff () =
    backoff_flag := true

let is_use_backoff () =
    !backoff_flag

let postreg = ref false

let set_postreg v =
    postreg := v

let is_postreg () = !postreg

let verbose_client_flag = ref false

let use_verbose_client () =
    verbose_client_flag := true

let is_verbose_client () =
    !verbose_client_flag

let version = "0.7.9"

let versionmsg =
    "\n\"FUNGI\" Fast UNsupervised Grammar Induction Software Suite\n"
    ^ "Version " ^ version  ^ " (C) 2011 by Prachya Boonkwan [1,2] and Mark Steedman [1]\n"
    ^ "    [1] School of Informatics, University of Edinburgh, United Kingdom\n"
    ^ "    [2] National Electronics and Computer Technology Center, Thailand\n\n"

let print_version () =
    Printf.fprintf stderr "%s" versionmsg;
    flush stderr

