(** Utility Functions *)

(***********************************************************
*                                                          *
*      FUNGI - Fast UNsupervised Grammar Induction         *
*                                                          *
*   Originator: Prachya Boonkwan [1], [2]                  *
*               [1] School of Informatics,                 *
*                   University of Edinburgh, UK            *
*               [2] National Electronics and Computer      *
*                   Technology Center (NECTEC), Thailand   *
*                                                          *
*   Copyright (C) 2011 Prachya Boonkwan                    *
*   <p.boonkwan@sms.ed.ac.uk, kaamanita@hotmail.com>       *
*                                                          *
*   This library is free software; you can redistribute    *
*   it and/or modify it under the terms of the GNU         *
*   Lesser General Public License as published by the      *
*   Free Software Foundation; either version 2.1 of the    *
*   License, or (at your option) any later version.        *
*                                                          *
*   This library is distributed in the hope that it will   *
*   be useful, but WITHOUT ANY WARRANTY; without even      *
*   the implied warranty of MERCHANTABILITY or FITNESS     *
*   FOR A PARTICULAR PURPOSE.  See the GNU Lesser          *
*   General Public License for more details.               *
*                                                          *
*   You should have received a copy of the GNU Lesser      *
*   General Public License along with this library; if     *
*   not, write to the Free Software Foundation, Inc.,      *
*   59 Temple Place, Suite 330, Boston, MA  02111-1307     *
*   USA.                                                   *
*                                                          *
***********************************************************)

(** [cartprod l1 l2] finds the cartesian product of two lists [l1] and
    [l2]. *)
val cartprod : 'a list -> 'b list -> ('a * 'b) list

(** [cartmap fn l1 l2] maps a function [fn] to each element of the 
    cartesian product of two lists [l1] and [l2]. *)
val cartmap : ('a -> 'b -> 'c) -> 'a list -> 'b list -> 'c list

(** [cartmap_noerr fn l1 l2] maps a function [fn] to each element of the 
    cartesian product of two lists [l1] and [l2].  This function 
    ignores all exceptions occurred during execution. *)
val cartmap_noerr : ('a -> 'b -> 'c) -> 'a list -> 'b list -> 'c list

(** [cartiter fn l1 l2] iterates a function [fn] on each element of the 
    cartesian product of two lists [l1] and [l2]. *)
val cartiter : ('a -> 'b -> 'c) -> 'a list -> 'b list -> unit

(** [cartiter_noerr fn l1 l2] iterates a function [fn] on each element of 
    the cartesian product of two lists [l1] and [l2].  This function 
    ignores all exceptions occurred during execution. *)
val cartiter_noerr : ('a -> 'b -> 'c) -> 'a list -> 'b list -> unit

(** [elim_dups lst] eliminates all duplicates in a list [lst]. *)
val elim_dups : 'a list -> 'a list

(** [intersect lst1 lst2] finds the intersection of two lists [l1]
    and [l2]. *)
val intersect : 'a list -> 'a list -> 'a list

(** [diff lst1 lst2] finds the difference of a list [lst1] from another
    list [lst2]. *)
val diff : 'a list -> 'a list -> 'a list

(** [union lst1 lst2] finds the union of two lists [lst1] and [lst2]. *)
val union : 'a list -> 'a list -> 'a list

(** [list_to_set lst] converts a list [lst] to a set. *)
val list_to_set : 'a list -> 'a list

(** [except_pos lst i] returns every element of a list [lst] but one at
    a specified position [i]. *)
val except_pos : 'a list -> int -> 'a list

(** [marshal obj] marshals a value [obj] to a string. *)
val marshal : 'a -> string

(** [unmarshal s] unmarshals a marshaled object [s]. *)
val unmarshal : ?begpos:int -> string -> 'a

(** [marshal_to_file fname obj] marshals a value [obj] into a string and
    saves it to a file [fname]. This function is NOT type-safe, so make 
    sure it is type-casted before using. *)
val marshal_to_file : string -> 'a -> unit

(** [unmarshal_from_file fname] loads a marshaled value from a file 
    [fname] and unmarshal it. This function is NOT type-safe, so make sure 
    it is type-casted before using. *)
val unmarshal_from_file : string -> 'a

val pi : float

(** [digamma v] calculates the digamma function of a value [v]. *)
val digamma : float -> float

val gammaln : float -> float

val gamma : float -> float

(** [split s] splits a string [s] into tokens by spaces. *)
val split : string -> string list

(** [split d s] splits a string [s] into tokens by a specified delimiter 
    [d] (in regular expression). *)
val split_delim : string -> string -> string list

(** [string_of_list fn lst ~bracket:b] represents a list [lst] with a 
    string, each element represented by a function [fn]. The string
    representation produced from this function can be bracketed if 
    specified. The default for this option is true. *)
val string_of_list : ?bracket:bool -> ('a -> string) -> 'a list -> string

(** The sentence ending mark for the CKY parser. *)
val ending_mark : string

val set_poolsize : int -> unit

val get_poolsize : unit -> int

val set_hostname : string -> unit

val set_port_probinit : int -> unit

val set_port_estep : int -> unit

val set_clientpath : string -> unit

val server_conf_probinit : unit -> JoinHelper.configuration

val server_conf_estep : unit -> JoinHelper.configuration

val server_conf_treesampler : unit -> JoinHelper.configuration

val client_conf_probinit : unit -> JoinHelper.configuration

val client_conf_estep : unit -> JoinHelper.configuration

val client_conf_treesampler : unit -> JoinHelper.configuration

val use_backoff : unit -> unit

val is_use_backoff : unit -> bool

val set_postreg : bool -> unit

val is_postreg : unit -> bool

val use_verbose_client : unit -> unit

val is_verbose_client : unit -> bool

val get_inet_addr : string -> string

val print_version : unit -> unit
