// view_lattice_elt.hh
//
// Mical, a regular grammar inference library
// Copyright (C) 2003 Maxime Rey
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#ifndef MICAL_VIEW_LATTICE_ELT_HH_
#define MICAL_VIEW_LATTICE_ELT_HH_

#include <vaucanson/fundamental/element.hh>
#include <vaucanson/automata/concept/automata.hh>

#include <set>
#include <vector>
#include <utility>

namespace vcsn {
  
  using namespace vcsn;
  using std::set;
  using std::vector;
  using std::map;

  struct ViewLattice
    : Structure<ViewLattice>
  {};
  
  template<typename Automaton_, typename Partition_>
  struct ImplemViewLattice
  {
    typedef Automaton_       automaton_t;
    typedef Partition_       partition_t;
  };

  template<typename T>
  struct Element<ViewLattice, T>
    :  MetaElement<Structure<ViewLattice>, T> 
  {
    typedef typename T::automaton_t                            automaton_t;
    typedef typename T::partition_t                            partition_t;
    typedef typename automaton_t::value_t                      value_t;
    typedef typename partition_t::block_t                      block_t;
    
    typedef typename automaton_traits<value_t>::label_t	       label_t;

    typedef typename vector<hstate_t>::const_iterator          state_iterator;

    typedef typename vector<hedge_t>::const_iterator           edge_iterator;
    
    typedef typename vector<hstate_t>::const_iterator	       initial_iterator;
    
    typedef typename vector<hstate_t>::const_iterator	       final_iterator;
    
    typedef Element<ViewLattice, T>           		       self_t;

    typedef typename automaton_traits<value_t>::series_t       series_t;

    typedef typename automaton_traits<value_t>::series_value_t series_value_t;

    typedef Element<series_t, series_value_t>		       series_elt_t;

    typedef typename series_t::monoid_t			       monoid_t;

    typedef typename monoid_t::alphabet_t                      alphabet_t;

    typedef typename series_elt_t::monoid_elt_t		       monoid_elt_t;

    typedef typename series_t::weights_t		       weights_t;

    typedef typename series_elt_t::weight_t		       weight_t;


    Element();
    Element(automaton_t& a, partition_t& partition);

    // fake for Vaucanson 0.1 (to be disapear)
    void emancipate();

    void create();
    /////////////////////////////////////////

    const vector<hstate_t>&
    states() const;

    const vector<hedge_t>&
    edges() const;

    const vector<hstate_t>&
    initial() const;

    const vector<hstate_t>&
    final() const;
    

    bool is_initial(hstate_t state) const;
    
    bool is_final(hstate_t state) const;

    hstate_t choose_state() const;

    bool has_state(hstate_t s) const;
    
    bool has_edge(hedge_t e) const;

    hstate_t origin_of(hedge_t e) const;

    hstate_t aim_of(hedge_t e) const;

    typename automaton_traits<value_t>::label_ret_t
    label_of(hedge_t e) const;

    template<typename Container_>
    void deltac(Container_& res, hstate_t from, delta_kind::edges k) const;

    template<typename Container_, typename L>	
    void deltac(Container_& res, 
		hstate_t from, 
		const L& query,
		delta_kind::edges k) const;

    template<typename Container_>
    void rdeltac(Container_& res, hstate_t from, delta_kind::edges k) const;

    template<typename Container_, typename L>	
    void rdeltac(Container_& res, 
		 hstate_t from, 
		 const L& query,
		 delta_kind::edges k) const;

    template<typename Container_>
    void deltac(Container_& res, hstate_t from, delta_kind::states k) const;

    template<typename Container_, typename L>	
    void deltac(Container_& res, 
		hstate_t from, 
		const L& query,
		delta_kind::states k) const;

    template<typename Container_>
    void rdeltac(Container_& res, hstate_t from, delta_kind::states k) const;

    template<typename Container_, typename L>	
    void rdeltac(Container_& res, 
		 hstate_t from, 
		 const L& query,
		 delta_kind::states k) const;



    template<typename Container_, typename L>	
    void letter_deltac(Container_& res, 
		       hstate_t from, 
		       const L& query,
		       delta_kind::edges k) const;

    template<typename Container_, typename L>	
    void letter_rdeltac(Container_& res, 
			hstate_t from, 
			const L& query,
			delta_kind::edges k) const;


    template<typename Container_, typename L>	
    void letter_deltac(Container_& res, 
		       hstate_t from, 
		       const L& query,
		       delta_kind::states k) const;
    
    template<typename Container_, typename L>	
    void letter_rdeltac(Container_& res, 
			hstate_t from, 
			const L& query,
			delta_kind::states k) const;

    void update();

    state_iterator update(hstate_t state);

    edge_iterator  update(hedge_t state);

    void update(state_iterator& lhs, state_iterator& rhs);
    
    void update(edge_iterator& lhs, edge_iterator& rhs);
    
    void update_init(initial_iterator& lhs, initial_iterator& rhs);

    void update_final(final_iterator& lhs, final_iterator& rhs);
    

    alphabet_t& alphabet() const;

    self_t&
    operator = (const self_t& other);

  private:

    hstate_t get_min_block(hstate_t state) const;
    block_t  get_block(hstate_t state) const;
    
    map<hstate_t, hstate_t> assoc_;
    
    vector<hstate_t> states_;
    vector<hedge_t>  edges_;
    vector<hstate_t> initial_;
    vector<hstate_t> final_;

    automaton_t*  auto_;
    partition_t*  partition_;
  };


} // mical

#include <mical/framework/viewer/view_lattice_elt.hxx>

#endif // MICAL_VIEW_LATTICE_ELT_HH_
