// renumerotation.hxx
//
// Mical, a regular grammar inference library
// Copyright (C) 2003 Maxime Rey
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#ifndef MICAL_RENUMEROTATION_HXX_
# define MICAL_RENUMEROTATION_HXX_

#include <mical/internal/renumerotation.hh>
#include <mical/internal/swap_states.hh>

#include <mical/walkers/breadth_search.hh>

#include <mical/tools/predefs.hh>

#include <deque>
#include <set>
#include <algorithm>

namespace mical {

  using std::deque;
  using std::set;
  using std::copy;
  using std::sort;


  // To apply to acyclic automata only
  template<class A, class T> 
  struct Renumerotation
  {
    typedef Renumerotation<A, T>          self_t;
    
    typedef InfoBreadth<A, T, self_t>     info_t;

    typedef vector<hstate_t>              delta_state_t;

    typedef Element<A, T>                 automaton_t;
    
    typedef vector<typename automaton_t::monoid_t::alphabet_t::letter_t> 
    alphabet_t;

    typedef deque<hstate_t>               queue_t;
    
    Renumerotation()
    {}

    Renumerotation(Element<A, T> a)
    {
      copy(a.series().monoid().alphabet().begin(),
	   a.series().monoid().alphabet().end(),
	   inserter(alphabet_, alphabet_.begin()));
      
      sort(alphabet_.begin(), alphabet_.end());
    }

    void operator ()(Element<A, T>& a)
    {
      auto_.create();
      auto_.series() = a.series();
      
      breadth_search(a, *this, security::responsable_t());

      a = auto_;
    }

    void prefix(info_t& info)
    {
      if (info.first_)
	{
	  state_ = auto_.add_state();
	  auto_.set_initial(state_);
	  if (info.auto_.is_final(*info.auto_.initial().begin()))
	    auto_.set_final(state_);
	}
    }

    void postfix(info_t& info)
    {}
    
    template<typename Container_>
    void 
    delta(info_t& info, Container_& res, hstate_t from)
    {
      bool first = true;
      hstate_t svg;
      for (typename alphabet_t::iterator i = alphabet_.begin();
	   i != alphabet_.end(); i++)
	{
	  vector<hstate_t> tmp;

	  info.auto_.letter_deltac(tmp, from, *i, delta_kind::states());

	  for (typename vector<hstate_t>::iterator j = tmp.begin();
	       j != tmp.end(); j++)
	    res.push_back(*j);
	  
	  for (typename vector<hstate_t>::iterator j = tmp.begin();
	       j != tmp.end(); j++)
	    {
	      hstate_t new_state = auto_.add_state();

	      if (info.auto_.is_final(*j))
		auto_.set_final(new_state);

	      queue_.push_back(new_state);
	      auto_.add_letter_edge(state_, new_state, *i);
	    }
	}
      if (!queue_.empty())
	{
	  state_ = queue_.front();
	  queue_.pop_front();
	}
    }

  private:
    
    Element<A, T> auto_;

    alphabet_t alphabet_;
    hstate_t   state_;
    queue_t    queue_;
  };


  template<typename A, typename T>
  void 
  standard_renum(Element<A, T>& a)
  {
    Renumerotation<A, T> renum(a);
    
    renum(a);
  }




  template<typename A, typename T>
  Element<A, T>
  frequency_renum_(const Element<A, T>& a, map<hstate_t, unsigned> mappy)
  {
    if (mappy.size() != a.states().size())
      return a;

    typedef typename Element<A, T>::state_iterator state_iterator;
    typedef map<hstate_t, unsigned> mappy_t;

    map<hstate_t, hstate_t> assoc;

    unsigned count = 0;
    while (!mappy.empty())
      {
	unsigned max = 0;
	mappy_t::iterator svg_iter;
	for (typename mappy_t::iterator i = mappy.begin(); i != mappy.end(); i++)
	  if (i->second > max)
	    {
	      max = i->second;
	      svg_iter = i;
	    }
	
	assoc[svg_iter->first] = count;
	mappy.erase(svg_iter);
	count++;
      }

    Element<A, T> res;
    res.create();

    res.series() = a.series();

    for (state_iterator i = a.states().begin(); i != a.states().end(); i++)
      res.add_state();
    
    for (state_iterator i = a.states().begin(); i != a.states().end(); i++)
      {
	if (a.is_final(*i))
	  res.set_final(assoc[*i]);
	if (a.is_initial(*i))
	  res.set_initial(assoc[*i]);
      }

    for (state_iterator i = a.states().begin(); i != a.states().end(); i++)
      {
	set<hedge_t> delta_edge, diff, resul;

	a.deltac(delta_edge, *i, delta_kind::edges());

	for (typename set<hedge_t>::iterator j = delta_edge.begin();
	     j != delta_edge.end(); j++)
	  add_edge(res, assoc[*i], assoc[a.aim_of(*j)], a.label_of(*j));

	a.rdeltac(diff, *i, delta_kind::edges());

	set_difference(diff.begin(), diff.end(),
		       delta_edge.begin(), delta_edge.end(),
		       inserter(resul, resul.begin()));
	
	for (typename set<hedge_t>::iterator j = resul.begin();
	     j != resul.end(); j++)
	  add_edge(res, assoc[a.origin_of(*j)], assoc[*i], a.label_of(*j));
      }
    
    return res;
  }

  template<typename A, typename T>
  void 
  frequency_renum(Element<A, T>& a, map<hstate_t, unsigned> mappy)
  {
    a = frequency_renum_(a, mappy);
  }

} // mical



#endif // MICAL_RENUMEROTATION_HXX_
