/* ------------------------------------------------------------------------------------------------------------------------
 SyntSce - Syntactical Scene Recognition
 Version: 0.4
 Author: Hemerson Pistori (pistori@ucdb.br)
 First Version: 10/01/2012
 Last Updated: 04/06/2012
 Objective:
   Apply a grammar based machine learning algorithm to the problem of scene recognition (or scene categorization). In fact, the algorithm can be applied to any kind of supervised image classification problem (simply changing the image dataset).
 Features:
 - Generate training and testing sets randomly from the images folder
 - Extract classes names from folder structure
 - Calculate some performance measures (Precision, Recall, F-Measure) for syntSce using dataSet available
 - Extract SURF features
 - Create vocabulary for BoW representation
 - Convert images to strings (using BoW vocabulary).
 - Create one parser for each grammar using linux BISON tool
 - Calculate image classification score by counting parsings errors
 - Assign each image to the class whose corresponding parse tree was created with fewer errors
 - Show the resulting classification using text and sound
 - Read all parameters from config file (so that the software does not need to be recompiled everytime we change them)
 - Create trackbars for all parameters, so that they can be changed "online"
 - Turn the sound files optional
 - Generate an ARFF file to be used with WEKA
 - Graphical interface is optional now (to facilitate running experiments several times)
 - Implement 4 different keypoint orderings: SURF, RANDOM, READING, RADIAL and READING_QUANTIZED
 - Integration with GI Toolbox Matlab 
 - Adjusted to work with OpenCV 2.4
DOING
TO-DO ASAP
 - Generate confusion matrix
 - Generate standard error
 - Integrate with Mical and try regular grammar induction
 - Integrate with Fungi
 - Enhance classification score strategy (improve error recovery)
 - Produce the same metrics used in Sun Scene Recognition benchmark in order to compare with other
 - Calculate the keypoints and features only once (it is being recalculated in each evaluation run, even when using the same dataset with different train and test partition ... turning the system extremely low for large datasets)
 TO-DO SOMEDAY
 - Save metrics results per class
 - Eliminate global vars
 - Use only C++ style
 - There are many system calls to external tools (most of them, Linux dependent) which makes the software hard to port to other platforms. These external call must be substituted by more portable code. (I did this in order to get my first results as soon as possible by reusing available software)
 - The folder name delimiter "/" is SO dependent and must be avoided.
 - Use WEKA to access other classifiers
 - Produce cross-validation results
 - Get rid of many breaks inside loops
----------------------------------------------------------------------------------------------------------------------- */

#include <opencv2/imgproc/imgproc.hpp>
#include <opencv2/highgui/highgui.hpp>
#include <opencv2/nonfree/nonfree.hpp>
#include "opencv2/features2d/features2d.hpp"
#include "opencv2/ml/ml.hpp"

#include "cv.h"
#include "cxcore.h"
#include "math.h"
#include <iostream>
#include <fstream>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sstream>
#include <time.h>
#include <sys/stat.h>
#include <pthread.h>
#include <errno.h>


#include <qt4/Qt/qapplication.h>
#include <qt4/Qt/qurl.h>
#include <qt4/Qt/qmessagebox.h>

#include <phonon/audiooutput.h>
#include <phonon/mediaobject.h>
#include <phonon/mediasource.h>
#include <phonon/videowidget.h>

using namespace cv;
using namespace std;
using namespace Phonon;










//-------------------------------------------------------------------------------------------------------------------------
//--- Global Variables

const char configFileName[] = "../data/config.xml";

// Initial position for all the windows (adjusted to a 1920x1080 screen). The values are read from the configuration file at configFileName
int woriginalX, woriginalY, wresultX, wresultY, wconfigurationX, wconfigurationY = 0;

// Types of scene recognition strategy already implemented
// TEMPLATE MATHING:
// - Extract SURF features for training images and query image
// - Calculate matching features using Fast Aproximate Nearest Neighbour (FlANN)
// - Choose the class corresponding to the training image with more matches
// SYNTSCE
// - Use Bow dictionary to convert all SURF keypoints in a strings
// - Use SEQUITUR to learg a grammar from training strings
// - Use BISON with the simpleste error recovery strategy to generate parsers
// - Parse the query string and choose the class whose corresponding grammar produced fewer errors
// WEKA
// - Generate and ARFF file with all dataset using BoW features
// - Call weka with the name of the classifier as a parameter [NOT IMPLEMENTED YET]
// - Use weka results to classify the query image [NOET IMPLEMENTED YET]


string dataSetName = "hotel_3class";
enum {TEMPLATE_MATCHING, SYNTSCE, WEKA};
int classifierType = SYNTSCE;
enum {GI_SEQUITUR, GI_KTESTABLE};
int giType = GI_SEQUITUR; // Indicate the Grammar Inference Algorithm that will be used by SyntSce


int numberOfClasses = 3; // Will be calculated automatically from the number of subfolders in the images folder
int percentTraining = 70; // Percentual of images to be used for training (1 - percentTraining/100) will be used for testing
enum {ORDER_SURF,ORDER_RANDOM,ORDER_READING,ORDER_RADIAL, ORDER_READING_QUANTIZED};
int defaultKeypointsOrder = ORDER_SURF;
int quantizeFactor = 20; // Used when ordering keypoints by ORDER_READING_QUANTIZED
float defaultSymbolRepititionThreshld = 0.03; // The higher this number the larger the strings with same symbols will be
float defaultBinValueThreshld = 0.05;
int defaultDictionarySize = 26;
int incrementDictionarySize = 3; // Add this number to the default Dictionary Size for the next execution (usefull for batch experiments)
int defaultScanningDirection = 1;
int playClassSound = true; // Indicate if a sound associated with the class should be played for each test instance
int saveIntermediateResults = false; // Save SURF descriptors graphical representation
int timeBetweenFrames = 1000; // Time to wait after the test instance is shown
int currentRun = 0; // Evaluation run
int maxRuns = 10; // Number of runs to be used during evaluation of the algorithm
int testImageIndex = 0; // test image being processed
int sizeOfDataset = 0; // This variable will contain the number training + testing samples available
int showInterface = true;

// Configure Machine Learning Related Stuff
// string dataSetName = "scene_15class";
string detectorType = "SURF";
string descriptorType = "SURF";
string matcherType = "FlannBased";
string fileWithBisonInit = "../data/bisonAux/yInit.c";
string fileWithBisonEnd = "../data/bisonAux/yEnd.c";
string fileWithConverter = "../data/bisonAux/sequiturToBison";
string fileWithErrorRecoveryAwk = "../data/bisonAux/errorRecovery.awk";
string fileWithGrammarInducer = "./sequitur";

string fileWithTrainImageNames = "../data/"+dataSetName+"/train.txt";
string fileWithTestImageNames = "../data/"+dataSetName+"/test.txt";
string fileWithClassesNames = "../data/"+dataSetName+"/classes.txt";
string fileWithAllImageNames = "../data/"+dataSetName+"/all.txt";
string dirWithImages = "../data/"+dataSetName+"/image";
string dirWithGrammars = "../data/"+dataSetName+"/grammars";
string fileWithPerformanceResults = "../data/"+dataSetName+"/results.csv";
string fileWithHistoricalResults = "../data/"+dataSetName+"/results.txt";
string arffFileName = "../data/"+dataSetName+"/dataset.arff";
string soundsDir = "../data/"+dataSetName+"/sounds";
string dirToSaveResImages = "../data/"+dataSetName+"/results";

string matlabCommand = "matlab -nodesktop -nosplash -r matlabScript";
string matlabGiToolBoxDir = "../gi/gitoolbox_2.0_matlab/";

int myInfinity=1000000;






//-------------------------------------------------------------------------------------------------------------------------
//--- Auxiliary Functions Declarations. See functions definitions for help

void adjustFileNames(void);
int getClassNumber(String trainImageName, vector<String> classesNames);
void displayInstructions();
bool fexists(const char *filename);
void createResultsFile();
bool initializeImagesAndClassesNames(vector<string>& classesNames);
void saveTrainingImage(IplImage * src, int i,int imageClass);
void saveConfiguration(void);
void readConfiguration(void);
void createAllTrackBars(void);
bool createDetectorDescriptorMatcher( const string& detectorType, const string& descriptorType, const string& matcherType,
                                      Ptr<FeatureDetector>& featureDetector,
                                      Ptr<DescriptorExtractor>& descriptorExtractor,
                                      Ptr<DescriptorMatcher>& descriptorMatcher );
void maskMatchesByTrainImgIdx( const vector<DMatch>& matches, int trainImgIdx, vector<char>& mask );
void readFilenames( const string& filename, string& dirName, vector<string>& trainFilenames );
bool readImages( const string& trainFilename,
                 vector <Mat>& trainImages, vector<string>& trainImageNames );
void detectKeypoints( const Mat& queryImage, vector<KeyPoint>& queryKeypoints,
                      const vector<Mat>& trainImages, vector<vector<KeyPoint> >& trainKeypoints,
                      Ptr<FeatureDetector>& featureDetector );
void computeDescriptors( const Mat& queryImage, vector<KeyPoint>& queryKeypoints, Mat& queryDescriptors,
                         const vector<Mat>& trainImages, vector<vector<KeyPoint> >& trainKeypoints, vector<Mat>& trainDescriptors,
                         Ptr<DescriptorExtractor>& descriptorExtractor );
void matchDescriptors( const Mat& queryDescriptors, const vector<Mat>& trainDescriptors,
                       vector<DMatch>& matches, Ptr<DescriptorMatcher>& descriptorMatcher );
void saveResultImages( const Mat& queryImage, const vector<KeyPoint>& queryKeypoints,
                       const vector<Mat>& trainImages, const vector<vector<KeyPoint> >& trainKeypoints,
                       const vector<DMatch>& matches, const vector<string>& trainImagesNames, const string& resultDir, int testIndex );
int classifyUsingTemplateMatching( const vector<DMatch>& matches, const vector<Mat>& trainImages,
                                   const vector<string>& trainImagesNames, vector<string>& classesNames, string& className );








//------------------------------------------------------------------------------------------------
// Used to convert a string to a number (float, int, ...)
template <class T>
bool from_string(T& t,
                 const std::string& s,
                 std::ios_base& (*f)(std::ios_base&))
{
    std::istringstream iss(s);
    return !(iss >> f >> t).fail();
}




















/* Class: Evaluator
 * Description: Calculate some performance metrics to evaluate the algorithm
 * Functionalities:
   - Calculate and Accumulate TP (True Positives) and FP (False Positives) value for each testing sample
   - Calculate precision, recall and F-Measure from accumulated TP and FP values
   - Print or write to disk the results
 */
class Evaluator {

public:

    vector<String> classesNames;
    vector<String> metricsNames;

    vector<float> TPs; // True Positives;
    vector<float> FPs; // False Positives;
    vector<float> Positives; // Total of testing samples for each class;

    int numberOfClasses;
    int totalOfSamples;
    int numberOfMetrics;

    String algorithmName;

    Evaluator( vector<String> names, String algorithm) {
        Initialize(names, algorithm);
    }

    Evaluator() {
    }

    void Initialize( vector<String> names, String algorithm) {
        classesNames = names;
        algorithmName = algorithm;
        numberOfClasses = classesNames.size();
        TPs.assign(numberOfClasses,0);
        FPs.assign(numberOfClasses,0);
        Positives.assign(numberOfClasses,0);
        totalOfSamples = 0;
        numberOfMetrics = 6;
        metricsNames.clear();
        metricsNames.push_back("Mean TP Rate");
        metricsNames.push_back("Mean FP Rate");
        metricsNames.push_back("Mean Precision");
        metricsNames.push_back("Mean Recall");
        metricsNames.push_back("mean F-Measure");
        metricsNames.push_back("mean Accuracy");
    }

    // Receive information about a new testing sample and increment TPs and FPs
    void newSample( int classifierClass, int groundTruthClass) {
        if (classifierClass == groundTruthClass) {
            ++TPs[classifierClass];
        } else {
            ++FPs[classifierClass];
        }
        ++Positives[groundTruthClass];
        ++totalOfSamples;
    }

    // Calculate metrics values and print
    void printResultPerClass(int classNumber, float TP, float FP, float Positives, vector<float>& means) {
        float Negatives = totalOfSamples-Positives;
        float FN = Positives - TP;
        float TN = Negatives - FP;
        float Precision = 0;
        float Acc = (TP+TN)/totalOfSamples;
        if ((TP+FP) != 0) Precision = TP/(TP+FP);
        float Recall = 0;
        if ((TP+FN) != 0) Recall = TP/(TP+FN);
        float FMeasure = 0;
        if ( (Precision+Recall) != 0 ) FMeasure = 2 * ((Precision*Recall)/(Precision+Recall));
        printf("%10s %7.2f %7.2f %7.2f %7.2f %7.2f %7.2f\n",
               classesNames[classNumber].c_str(),
               TP/Positives,
               FP/Negatives,
               Precision,
               Recall,
               FMeasure,
               Acc
              );
        means[0] += TP/Positives;
        means[1] += FP/Negatives;
        means[2] += Precision;
        means[3] += Recall;
        means[4] += FMeasure;
        means[5] += Acc;
    }

    // Print results for all classes and results means
    void printResults() {

        vector<float> means(numberOfMetrics,0);
        printf("-[ Results for %10s after %5d Samples]------------------------\n",algorithmName.c_str(),totalOfSamples);
        printf("Class      TP Rate FP Rate    Precis. Recall  F-Measure   Acc.\n");
        for ( int i = 0; i < numberOfClasses; ++i) {
            printResultPerClass(i,TPs[i],FPs[i],Positives[i],means);
        }
        printf("------------------------------------------------------------\n");
        printf("    Means:   ");
        for ( int i = 0; i < numberOfMetrics; ++i) {
            printf("%7.2f ",means[i]/numberOfClasses);
        }
        printf("\n------------------------------------------------------------\n");
    }

    // Append the results to a file in the dataset folder
    void saveResults() {


        struct tm *newtime;
        time_t ltime;

        ltime = time(&ltime);
        newtime = localtime(&ltime);

        vector<float> means(numberOfMetrics,0);
        FILE * pFile;
        pFile = fopen (fileWithHistoricalResults.c_str(),"a");

        // Calculate means;
        for ( int i = 0; i < numberOfClasses; ++i) {
            printResultPerClass(i,TPs[i],FPs[i],Positives[i],means);
        }

        fprintf(pFile,"Algorithm: %s\n",algorithmName.c_str());
        fprintf(pFile,"Date and Time: %s",asctime(newtime));
        fprintf(pFile,"Dataset Name: %s\n",dataSetName.c_str());
        fprintf(pFile,"Size of Data Set: %d\n",sizeOfDataset);
        fprintf(pFile,"Runs: %d\n",currentRun);
        fprintf(pFile,"Total of Samples: %d\n",totalOfSamples);
        fprintf(pFile,"Perc. for Training: %d\%\n",percentTraining);
        for ( int i = 0; i < numberOfMetrics; ++i) {
            fprintf(pFile,"%16s: %7.2f\n",metricsNames[i].c_str(),means[i]/numberOfClasses);
        }
        fprintf(pFile,"------------------------------------------------------------\n");

        fclose(pFile);
    }

};














//------------------------------------------------------------------------------------------------
/* Class: Grammar
 * Description: A wrapper for external tools that deals with grammar induction, parser generation and parse tree building.
 */
class Grammar {
public:
    String name; // The name of the class represented by this grammar
    vector<String> positiveStrings; // Example of strings the should be generated by this grammar

    String stringsFileName;
    String grammarFileName;

    FILE * pFile;

    String fileParserOut; // The parser writes the result, reflecting how well the string has been parsed, to this file


    Grammar(String grammarName) {
        name = grammarName;
        fileParserOut = "parserOut.txt";
    };

    void setScriptFile(FILE * ppFile) {
        pFile = ppFile;
    }

    void addPositiveString(String string) {
        positiveStrings.push_back(string);
    }

    String getGrammarName() {
        return(name);
    };

    float readParserResult() {
        // Read the result and convert to float
        float result = 0;
        ifstream file( fileParserOut.c_str() );
        if ( !file.is_open() )
            return -1;
        String str;
        getline( file, str );
        from_string<float>(result, str, std::dec);
        file.close();
        return(result);

    }

    // Create folder to receive the grammars and some intermediate files
    void createStringAndGrammarFiles() {

        mkdir(dirWithGrammars.c_str(), 0777);
        stringsFileName = dirWithGrammars + "/" + name + ".str"; // File with training strings
        grammarFileName = dirWithGrammars + "/" + name + ".grm"; // File with induced grammar in SEQUITUR format

        // Create empty files
        String command = "echo -n > " + stringsFileName;
        system(command.c_str());
        command = "echo -n > " + grammarFileName;
        system(command.c_str());
    }


    virtual void induceGrammar() {
    };

    virtual void createParser() {
        cout << "No devia ter entrado aqui ... creatParser do Pai" << endl;
    };

    virtual float parse(String string) {
        cout << "No devia ter entrado aqui ... parse do Pai" << endl;
    };

};






//------------------------------------------------------------------------------------------------
/* Class: SequiturGrammar
 * Description: A wrapper for Sequitur Grammar Inference Software
 * Functionalities:
   - The software SEQUITUR (http://sequitur.info/) is being for grammar induction.
   - Bison is being used to create parsers for the grammars induced by SEQUITUR
   - SEQUITUR output format is converted to BISON input format using another external tool (generated by the same Bison)
 */
class SequiturGrammar : public Grammar {
public:

    SequiturGrammar(String grammarName) : Grammar(grammarName) {
    };

    // SEQUITUR is called multiple times, for each string, and the resulting grammars are appended to form a new unique grammar. This function renumber the rules in order to avoid conflicts in the final grammar.
    int renumberRules(int firstNumber) {


        ifstream file( "temp2.txt" );
        if ( !file.is_open() )
            return -1;

        int largestNumber = -1;
        string command = "echo -n > temp3.txt";
        system(command.c_str());

        while ( !file.eof() )
        {
            ostringstream oss;
            ostringstream output;
            string str;
            getline( file, str );
            oss << str ;
            if ( str.empty() ) break;

            string element;
            stringstream parser(oss.str());
            while (parser>>element) {
                int number;

                int numberOfDigits = strspn(element.c_str(),"0123456789");
                if (numberOfDigits > 0) {
                    from_string<int>(number, element, std::dec);
                    output << (number+firstNumber) << " ";
                    if (number > largestNumber) largestNumber = number;
                } else {
                    output << element + " ";
                }
            }
            string command = "echo  \"" + output.str() + " \" >> temp3.txt";
            system(command.c_str());


        }
        file.close();

        return(largestNumber+firstNumber);
    }

    // Call SEQUITUR multiple times, for each string. Combine the resulting grammars using an "OR" rule: a new NON-TERMINAL at the left side and all the initial non-terminals of the previous grammars "OR"ed at the right side
    void induceGrammar() {

        vector<int> rulesIndexes; // Initial non-terminal of each grammar generated by SEQUITUR
        int lastNumber = 0;

        createStringAndGrammarFiles();

        // The initial non-terminal of the first grammar generated by SEQUITUR is always 1
        rulesIndexes.push_back(1);

        string command;

        for (int i = 0; i < positiveStrings.size(); ++i) {

            // temp.txt = just the original training string (one training string at a time)
            // temp2.txt = rules induced by sequitur for one string
            // temp3.txt = rules with the numbers changed in order not to collide

            // Add string to strings file
            command = "echo " + positiveStrings[i] + " >> " + stringsFileName;
            system(command.c_str());

            // Create file with one string to serve as the input file for SEQUITUR software
            command = "echo -n " + positiveStrings[i] + " > " + "temp.txt";
            system(command.c_str());

            // Runs SEQUITUR
            command =fileWithGrammarInducer + " <  temp.txt  >  temp2.txt";
            system(command.c_str());

            // Renumber the rules generated by SEQUITUR
            lastNumber = renumberRules(lastNumber+1);
            rulesIndexes.push_back(lastNumber+1);

            // Add the new rules to the consolidated grammar file
            command ="cat temp3.txt  >> " + grammarFileName;
            system(command.c_str());


        }

        // Create the new rule that points to each previous grammars
        ostringstream output;
        output << "echo \"0 -> 1 ";
        for (int i = 1; i < rulesIndexes.size()-1; ++i) {
            output << " | " << rulesIndexes[i];
        }



        // Insert a tag, 987654321, that indicates to the error recovery generator where the code should be put
        output << " 987654321 \" > temp4.txt";
        command = output.str();
        system(command.c_str());

        // Put the last rule in the beginning of the file (bison doesn't accept out of order rules)
        command = "cat " + grammarFileName + " >> temp4.txt";
        system(command.c_str());
        command = "cp temp4.txt " + grammarFileName;
        system(command.c_str());

    }

    // Previous to calling bison to generate the parser, this method changes the grammar in order to include a simple error recovery strategy. This strategy ignores the token that lead to a parse error and makes the parser continue from the next available token. In bison: {yyclearin;yyerrok;}
    void addErrorRecovery() {
        String inducerGrammarFileName = dirWithGrammars + "/" + name + ".grm";
        String bisonGrammarFileName = inducerGrammarFileName + ".rules";


        String command = "awk -f "+fileWithErrorRecoveryAwk+ " "+bisonGrammarFileName + " > temp";
        system(command.c_str());

        command = "cp temp "+ bisonGrammarFileName;
        system(command.c_str());


    }

    // Convert the file with the rules in Sequitur (or other inducer) to bison .y format
    void convertInducerToBison() {

        String inducerGrammarFileName = dirWithGrammars + "/" + name + ".grm";
        String bisonGrammarFileName = inducerGrammarFileName + ".rules";
        // Create .y file

        String command = fileWithConverter + " < "  + inducerGrammarFileName + " > " + bisonGrammarFileName;
        system(command.c_str());

        addErrorRecovery();

    }

    // Create a parser using BISON
    void createParser() {

        cout << "Entrou no filho ..." << endl;
        induceGrammar();

        String bisonyFileName = dirWithGrammars + "/" + name + ".y";
        String parserExecutableFileName = dirWithGrammars + "/" + name;

        String bisoncFileName = dirWithGrammars + "/" + name + ".tab.c";
        String inducerGrammarFileName = dirWithGrammars + "/" + name + ".grm";
        String bisonGrammarFileName = inducerGrammarFileName + ".rules";

        convertInducerToBison();

        // Create .y file concatenating the rules just induced with pre-available code in bisonAux folder
        String command = "cat " + fileWithBisonInit + " > " + bisonyFileName;
        command += "; cat " + bisonGrammarFileName + " >> " + bisonyFileName;
        command += "; cat " + fileWithBisonEnd + " >> " + bisonyFileName;
        system(command.c_str());

        // Create .c file from .y using BISON
        command = "bison -o" + bisoncFileName + " " + bisonyFileName;
        system(command.c_str());

        // Compile and create parser executable
        command = "gcc -o " + parserExecutableFileName + " " + bisoncFileName;
        system(command.c_str());

    };

    // Parse a string using the grammar parser
    float parse(String string) {

        String parserExecutableFileName = dirWithGrammars + "/" + name;

        // Execute the parser on the input string in temp.txt and get the result in fileParserOut
        String command = "echo -n " + string + " >  temp.txt";
        system(command.c_str());
        command = parserExecutableFileName + " < temp.txt > " + fileParserOut;
        system(command.c_str());

        return( myInfinity - readParserResult());

    };

};



//------------------------------------------------------------------------------------------------
/* Class: GIToolBoxGrammar
 * Description: A wrapper for GI Toolbox Matlab Software
 * Functionalities:
 */
class GIToolBoxGrammar : public Grammar {
public:

    int k; // Parameter used by GI K-Testable Algorithm

    GIToolBoxGrammar(String grammarName) : Grammar(grammarName) {
        k = 3;
    };



    void induceGrammar() {

        createStringAndGrammarFiles();

        for (int i = 0; i < positiveStrings.size(); ++i) {
            // Add string to strings file
            string command = "echo " + positiveStrings[i] + " >> " + stringsFileName;
            system(command.c_str());

        }

        fprintf(pFile,"dfa_%s = K2dfa('%s',%i)\n",name.c_str(),stringsFileName.c_str(),k);



    }

    void createParser() {
        induceGrammar();
    };

    // Parse a string using the grammar parser
    float parse(String string) {

        // Create the script to be executed by matlab
        fprintf(pFile,"errors = DistanceToAccept('%s',dfa_%s)\n",string.c_str(),name.c_str());
        fprintf(pFile,"results = [results,errors]\n");

        return(0);

    };

};





















class Symbol {

public:
    int x;
    int y;
    char letter;

    Symbol(int px,int py,char pletter) {
        x = px;
        y = py;
        letter = pletter;
    }
};


bool readingOrderSortPredicate(const Symbol& d1, const Symbol& d2)
{
    if ( d1.y != d2.y ) return( d1.y < d2.y );
    else return(d1.x < d2.x);
}

// Consider the points at most 10 pixels apart, vertically, are in the same line
int centralX = 0;
int centralY = 0;
bool radialOrderSortPredicate(const Symbol& d1, const Symbol& d2)
{
    float dist1 = sqrt(pow(d1.x - centralX,2)+pow(d1.y - centralY,2));
    float dist2 = sqrt(pow(d2.x - centralX,2)+pow(d2.y - centralY,2));
    return(dist1 < dist2);
}


/* Class: SyntSce
 * Description: Implements the Synthatical Scene Recognition algorithm
 * Functionalities:
   - Create the grammars and parsers for each class (learn from examples)
   - Classify testing images
 */
class SyntSce {

public:

    // These two thresholds are used only with the BOWHIST strategy ...
    float symbolRepititionThreshld; // The higher this number the larger the strings with same symbols will be
    float binValueThreshld; // The higher this number less bins will be used to form the strings
    int keypointsOrder; // The order of the keypoints used to convert the image to a string

    enum {BOWHIST, KEYPOINTS};
    int scanningDirection; // Determine how the image will be traversed to produce the string

    char hashSymbols[255]; // How table of "readable" ASCII symbols

    String name; // Name of the classifier with current parameters

    FILE * pFile; // FILE with Matlab Script


    BOWKMeansTrainer * bowTrainer;
    BOWImgDescriptorExtractor * bowDE;
    Ptr<FeatureDetector> bowFeatureDetector;
    vector<Grammar*> grammars;

    Ptr<DescriptorExtractor> descriptorExtractor;
    Ptr<DescriptorMatcher> descriptorMatcher;


    int dictionarySize;
    Mat dictionary;

    SyntSce() {
        symbolRepititionThreshld = defaultSymbolRepititionThreshld; // The higher this number the larger the strings with same symbols will be
        binValueThreshld = defaultBinValueThreshld;
        dictionarySize = defaultDictionarySize;
        scanningDirection = defaultScanningDirection;
        keypointsOrder = defaultKeypointsOrder;
        ostringstream output;
        if (scanningDirection==BOWHIST) output << "SyntSce [Dir = BOWHIST , BinValThr = " << binValueThreshld << " , RepThr = " << symbolRepititionThreshld << " , DicSize = " << dictionarySize << "]";
        else if (scanningDirection==KEYPOINTS) {
            if (keypointsOrder == ORDER_SURF) output << "SyntSce [Dir = KEYPOINTS , Ord = SURF, DicSize = " << dictionarySize << " " ;
            else if (keypointsOrder == ORDER_RANDOM) output << "SyntSce [Dir = KEYPOINTS , Ord = RANDOM, DicSize = " << dictionarySize << " ";
            else if (keypointsOrder == ORDER_READING) output << "SyntSce [Dir = KEYPOINTS , Ord = READ, DicSize = " << dictionarySize << " ";
            else if (keypointsOrder == ORDER_RADIAL) output << "SyntSce [Dir = KEYPOINTS , Ord = RADIAL, DicSize = " << dictionarySize << " ";
            else if (keypointsOrder == ORDER_READING_QUANTIZED) output << "SyntSce [Dir = KEYPOINTS , Ord = READ_QUANT " << quantizeFactor << " X " << quantizeFactor << ", DicSize = " << dictionarySize << " ";
            if ( giType == GI_KTESTABLE) output << "GI = K_TEST ]";
            else if ( giType == GI_SEQUITUR) output << "GI = SEQUITUR ]";
        }
        name = output.str();
        TermCriteria tc(CV_TERMCRIT_ITER,100,0.001);
        int retries=1;
        int flags=KMEANS_PP_CENTERS;
        bowTrainer = new BOWKMeansTrainer(dictionarySize, tc, retries, flags );
        initHashSymbols();
    }

    void initHashSymbols() {
        for (char i = 0; i < 26; ++i) {
            hashSymbols[i] = 'a' + i;
            hashSymbols[i+26] = 'A' + i;
        }
    }

    void setGrammarsNames(vector<String> classesNames) {
        for ( int i = 0; i < classesNames.size(); ++i ) {
            grammars[i]->name = classesNames[i];
        }
    }
    void createVocabulary(vector<Mat>& descriptors) {
        for ( int i = 0; i < descriptors.size(); ++i) {
            Mat descriptor = descriptors[i];
            for ( int j = 0; j < descriptor.rows; ++j) {
                bowTrainer->add(descriptor.row(j));
            }
        }
        dictionary = bowTrainer->cluster();
    };

    void setAuxiliaryAlgorithms(Ptr<DescriptorExtractor> extractor, Ptr<DescriptorMatcher> matcher,Ptr<FeatureDetector> feature) {
        bowFeatureDetector = feature;
        descriptorExtractor = extractor;
        descriptorMatcher = matcher;
        // There are two different matchers, one will be used with BoW and the other without Bow (maybe one is sufficient ... must check).
        //descriptorMatcher = DescriptorMatcher::create( matcherType );
        bowDE = new BOWImgDescriptorExtractor(extractor,matcher);
        bowDE->setVocabulary(dictionary);
        //descriptorMatcher->add( dictionary );
    }

    /* Convert the BoW Histogram to a string. Each bin generates a sequence of symbols with the size proportional to the value of the bin (counting of words represented by the bin).
    - Words with small counts or bins with small values are discarded. The size of bins to be discarded are defined a priori (binValueThreshld).
    - The numbers of symbols to be used to represent bin values are also defined a priori (symbolRepititionThreshld)
    */
    String imageToStringUsingBowHistogram(Mat image) {
        String string = "";
        int totalVocabule = 0;
        int totalBowDescriptors = 0;
        vector<KeyPoint> keypoints;
        bowFeatureDetector->detect(image,keypoints);
        Mat bowDescriptor;
        bowDE->compute(image,keypoints,bowDescriptor);
        for ( int i = 0; i < bowDescriptor.rows; ++i) {
            Mat row = bowDescriptor.row(i);
            for ( int j = 0; j < row.cols; ++j) {
                totalBowDescriptors += row.data[j];
            }
        }

        for ( int i = 0; i < bowDescriptor.rows; ++i) {
            Mat row = bowDescriptor.row(i);
            for ( int j = 0; j < row.cols; ++j) {
                totalVocabule = row.data[j];
                float percVocabule = (float)totalVocabule/(float)totalBowDescriptors;
                if ( percVocabule > binValueThreshld) {
                    totalVocabule = totalVocabule / (totalBowDescriptors*symbolRepititionThreshld);
                    for ( int k = 0; k < totalVocabule; ++k) {
                        string += hashSymbols[j];
                    }
                }
            }
        }
        return string;
    }


    void putInReadingOrder(vector<Symbol>& symbols) {
        std::sort(symbols.begin(), symbols.end(), readingOrderSortPredicate);
    }

    void putInReadingQuantizedOrder(vector<Symbol>& symbols) {

        for (int i = 0; i < symbols.size(); ++i) {
            symbols[i].x = symbols[i].x / quantizeFactor;
            symbols[i].y = symbols[i].y / quantizeFactor;
        }

        std::sort(symbols.begin(), symbols.end(), readingOrderSortPredicate);
    }

    void putInRandomOrder(vector<Symbol>& symbols) {
        random_shuffle(symbols.begin(),symbols.end());
    }


    void putInRadialOrder(vector<Symbol>& symbols) {

        centralX = 0;
        centralY = 0;

        for (int i = 0; i < symbols.size(); ++i) {
            centralX += symbols[i].x;
            centralY += symbols[i].y;
        }

        centralX = centralX / symbols.size();
        centralY = centralY / symbols.size();

        std::sort(symbols.begin(), symbols.end(), radialOrderSortPredicate);
    }

    String imageToStringUsingKeypointsOrder(Mat image) {

        String string = "";
        int totalVocabule = 0;
        int totalBowDescriptors = 0;
        vector<KeyPoint> keypoints;
        Mat descriptors;
        vector<DMatch> matches;
        vector<Symbol> symbols;

        bowFeatureDetector->detect(image,keypoints);
        descriptorExtractor->compute( image, keypoints, descriptors );

        for (int i = 0; i < descriptors.rows; ++i) {


            Mat descriptor = descriptors.row(i);
            descriptorMatcher->match( descriptor, matches );

            int vocabule = 0;
            if (matches.size()>0) {
                vocabule = matches[0].trainIdx;
                symbols.push_back(Symbol(keypoints[i].pt.x,keypoints[i].pt.y,hashSymbols[vocabule]));
            }

        }

        if (keypointsOrder == ORDER_READING) putInReadingOrder(symbols);
        else if (keypointsOrder == ORDER_RANDOM) putInRandomOrder(symbols);
        else if (keypointsOrder == ORDER_RADIAL) putInRadialOrder(symbols);
        else if (keypointsOrder == ORDER_READING_QUANTIZED) putInReadingQuantizedOrder(symbols);

        for (int i = 0; i < symbols.size(); ++i) {

            string += symbols[i].letter;
        }


        return string;
    }


    String imageToString(Mat image) {
        String string;
        if (scanningDirection == BOWHIST) string = imageToStringUsingBowHistogram(image);
        else if (scanningDirection == KEYPOINTS) string = imageToStringUsingKeypointsOrder(image);
        return string;
    }


    void createGrammars(vector<Mat>& trainImages, vector<String>& trainImagesNames, vector<String>& classesNames)
    {
        pFile = fopen ("matlabScript.m","w");
        fprintf(pFile,"addpath('%s')\n",matlabGiToolBoxDir.c_str());

        grammars.clear();
        for (int i = 0; i < classesNames.size(); ++i ) {
            if (giType == GI_SEQUITUR) grammars.push_back(new SequiturGrammar(classesNames[i]));
            else if (giType == GI_KTESTABLE) {
                grammars.push_back(new GIToolBoxGrammar(classesNames[i]));
                grammars[i]->setScriptFile(pFile);
            }
        }

        for (int i = 0; i < trainImagesNames.size(); ++i ) {
            int grammarIndex = getClassNumber(trainImagesNames[i],classesNames);
            String string = imageToString(trainImages[i]);
            grammars[grammarIndex]->addPositiveString(string);
        }

        for (int i = 0; i < classesNames.size(); ++i ) {
            grammars[i]->createParser();
        }

        fprintf(pFile,"save matlabVariables\n");
        fprintf(pFile,"exit\n");
        fclose(pFile);
        system(matlabCommand.c_str());
    }

    int calculateBestGrammar( float * confidences) {
        int best = -1;
        int bestIndex = 0;
        for ( int i = 0; i < grammars.size(); ++i) {
            if (confidences[i] > best) {
                best = confidences[i];
                bestIndex = i;
            }
        }
        return(bestIndex);

    }


    int classify(Mat testImage,  string& className, string groundTruthClass, vector<string> classesNames, Evaluator& evaluator) {

        String imageString = imageToString(testImage);
        float confidences[numberOfClasses];

        ostringstream output;
        output << "echo \"";
        output << currentRun << ";";
        output << testImageIndex << ";";
        output << imageString << ";";

        int bestGrammar;

        if (giType == GI_SEQUITUR) {

            for ( int i = 0; i < grammars.size(); ++i) {
                confidences[i] = (float)grammars[i]->parse(imageString);
                output << confidences[i] << ";";

            }
            bestGrammar = calculateBestGrammar( confidences );

        } else {

            pFile = fopen ("matlabScript.m","w");
            fprintf(pFile,"    load matlabVariables;       \n");
            fprintf(pFile,"    addpath('%s');              \n",matlabGiToolBoxDir.c_str());
            fprintf(pFile,"    results=[];                 \n");

            for ( int i = 0; i < grammars.size(); ++i) {
                grammars[i]->setScriptFile(pFile);
                grammars[i]->parse(imageString);
            }

            fprintf(pFile,"    [m,pos]=min(results);                 \n");
            fprintf(pFile,"    disp FINAL_SCORES;  \n");
            fprintf(pFile,"    results;  \n");
            fprintf(pFile,"    fileID = fopen('parserOut.txt','w');  \n");
            fprintf(pFile,"    fprintf(fileID,'\%");
            fprintf(pFile,"10f',pos);            \n");
            fprintf(pFile,"    exit                                  \n");
            fclose(pFile);
            system(matlabCommand.c_str());

            bestGrammar = grammars[0]->readParserResult() - 1;

            cout << "Best Grammar = " << bestGrammar << endl;

        }

        className = grammars[bestGrammar]->name;

        output << className << ";";
        int classIndex = getClassNumber(groundTruthClass,classesNames);
        output << classesNames[classIndex] << ";";
        output << groundTruthClass << ";";
        output << bestGrammar << ";";
        output << "\" >> " << fileWithPerformanceResults;
        String command = output.str();
        system(command.c_str());

        cout << testImageIndex << " - Classifier = " << className << " - Ground Truth  = " << classesNames[classIndex] << endl;

        evaluator.newSample(bestGrammar,classIndex);

        return(bestGrammar);

    }



};














/* Class: Weka
 * Description: Use weka to learn and classify
 * Functionalities:
 * - Generate and ARFF file with all dataset using BoW features
 * - Call weka with the name of the classifier as a parameter [NOT IMPLEMENTED YET]  *
 * - Use weka results to classify the query image [NOET IMPLEMENTED YET]
 */
class Weka {

public:

    String fileName;
    int dictionarySize;
    ofstream out; // output file

    BOWImgDescriptorExtractor * bowDE;
    Ptr<FeatureDetector> bowFeatureDetector;
    Ptr<DescriptorExtractor> descriptorExtractor;
    Ptr<DescriptorMatcher> descriptorMatcher;

    Weka() {
        fileName = arffFileName;
        dictionarySize = defaultDictionarySize;
    }
    ~Weka() {
        out.close();
    }



    void setAuxiliaryAlgorithms(Ptr<DescriptorExtractor> extractor, Ptr<DescriptorMatcher> matcher,Ptr<FeatureDetector> feature, Mat dictionary) {
        bowFeatureDetector = feature;
        descriptorExtractor = extractor;
        descriptorMatcher = matcher;
        bowDE = new BOWImgDescriptorExtractor(extractor,matcher);
        bowDE->setVocabulary(dictionary);
    }

    void insertArfffHeader(vector<String> classesNames) {

        char dicSize[6];
        sprintf(dicSize,"%02d",defaultDictionarySize);

        out.open(fileName.c_str());
        out << "% ARFF containing Bow Feature for all " << dataSetName << " images" << endl;
        out << endl;
        out << "@relation " << dataSetName+"_Dic"+dicSize << endl;
        out << endl;
        for ( int i = 0; i < dictionarySize; ++i) {
            out << "@attribute A" << i+1 << " numeric" << endl;
        }
        out << "@attribute class {";
        for ( int i = 0; i < classesNames.size()-1; ++i) {
            out << classesNames[i] << ",";
        }
        out << classesNames[classesNames.size()-1] << "}" << endl;
        out << endl;
        out << "@data" << endl;
    }


    void insertArffInstance(Mat bowDescriptor, String className) {

        double binValue;
        for ( int i = 0; i < bowDescriptor.rows; ++i) {
            Mat row = bowDescriptor.row(i);
            for ( int j = 0; j < row.cols; ++j) {
                binValue = row.data[j];
                out << binValue << ",";
            }
        }
        out << className << endl;
    }

    void insertArffInstance(Mat image, String imageFileName, vector<String> classesNames) {
        int classIndex = getClassNumber(imageFileName,classesNames);
        String className = classesNames[classIndex];
        vector<KeyPoint> keypoints;
        bowFeatureDetector->detect(image,keypoints);
        Mat bowDescriptor;
        bowDE->compute(image,keypoints,bowDescriptor);

        insertArffInstance(bowDescriptor, className);
    }

    void insertArffInstances(vector<Mat>& trainImages, vector<String>& trainImagesNames, vector<String>& classesNames) {
        insertArfffHeader(classesNames);
        for ( int i = 0; i < trainImages.size(); ++i ) {
            insertArffInstance(trainImages[i],trainImagesNames[i],classesNames);
        }
    }

    void closeArff() {
        out.close();
    }
};
























//-------------------------------------------------------------------------------------------------------------------------
//--- Main

int main( int argc, char** argv )
{
    QApplication app( argc , argv );
    readConfiguration();
    if (showInterface) displayInstructions();









//---------------------------------------------------------------------
//--- Declare objects related to the Machine Learning Module

    Ptr<FeatureDetector> featureDetector;
    Ptr<DescriptorExtractor> descriptorExtractor;
    Ptr<DescriptorMatcher> descriptorMatcher;

    SyntSce syntSce;
    Weka weka;
    Evaluator evaluator;

    Mat queryImage;
    vector<KeyPoint> queryKeypoints;
    Mat queryDescriptors;

    vector<Mat> testImages;
    vector<string> testImageNames;

    vector<Mat> trainImages;
    vector<string> trainImageNames;
    vector<string> classesNames;
    vector<vector<KeyPoint> > trainKeypoints;
    vector<Mat> trainDescriptors;

    if ( !createDetectorDescriptorMatcher( detectorType, descriptorType, matcherType, featureDetector, descriptorExtractor, descriptorMatcher ) )
    {
        cout << "Problems to create Matcher ..." << endl;
        return -1;
    }

    Mat resultImage;







//---------------------------------------------------------------------------------
//--- Learn a new model (classifier) using the training images from disk

    cout << "Separating testing and training images randomly ..." << endl;
    initializeImagesAndClassesNames(classesNames);
    cout << "Reading training images ..." << endl;
    readImages( fileWithTrainImageNames, trainImages, trainImageNames );
    cout << "Detecting keypoints..." << endl;
    featureDetector->detect( trainImages, trainKeypoints );
    cout << "Extracting features..." << endl;
    descriptorExtractor->compute( trainImages, trainKeypoints, trainDescriptors );
    if (classifierType == SYNTSCE) {
        cout << "Creating vocabulary" << endl;
        syntSce.createVocabulary(trainDescriptors);
        cout << "Seting auxiliary algorithms..." << endl;
        syntSce.setAuxiliaryAlgorithms(descriptorExtractor,descriptorMatcher,featureDetector);
        cout << "Creating grammars..." << endl;
        syntSce.createGrammars(trainImages,trainImageNames,classesNames);
        evaluator.Initialize(classesNames,syntSce.name);
    } else if (classifierType == TEMPLATE_MATCHING) {
        evaluator.Initialize(classesNames,"Template Matching");
    } else if (classifierType == WEKA) {
        cout << "Creating vocabulary" << endl;
        syntSce.createVocabulary(trainDescriptors);
        weka.setAuxiliaryAlgorithms(descriptorExtractor,descriptorMatcher,featureDetector,syntSce.dictionary);
        weka.insertArffInstances(trainImages,trainImageNames,classesNames);
    }
    cout << "Reading trainining images..." << endl;
    readImages( fileWithTestImageNames, testImages, testImageNames );
    cout << "Read to start classifying !!!" << endl;
    createResultsFile();
    sizeOfDataset = trainImages.size()+testImages.size();

//-----------------------------------------------------------------------------
//--- Create dir to save results

    if (saveIntermediateResults) {
        String command = "rm " + dirToSaveResImages + "/*";
        system(command.c_str());
        command = "mkdir " + dirToSaveResImages;
        system(command.c_str());
    }



//-----------------------------------------------------------------------------
//--- Initialize Sound Device and display instructions

    app.setApplicationName( QLatin1String("Sound Player") );
    MediaSource mediaSources[numberOfClasses];
    MediaObject * classSound = NULL;

    // Create one MediaSource for each class sound
    for (int i = 0; i < numberOfClasses; ++i ) {
        String fileName = soundsDir+"/"+classesNames[i]+".wav";
        if (! fexists(fileName.c_str())) {
            cout << "Some class sound files are missing ... sound disabled" << endl;
            playClassSound == false;
            break;
        }
        mediaSources[i] = MediaSource(fileName.c_str());
    }
    classSound = createPlayer(MusicCategory, mediaSources[0] );








//---------------------------------------------------------------------
//--- Create windows and move to saved position (configuration file - data/config.xml)

    if (showInterface) {
        namedWindow( "Original",1);  // Window for the original frames captured by the WebCam
        cvMoveWindow("Original",woriginalX,woriginalY);
        namedWindow( "Result", 1);     // Window for the tracbars with adjustable parameters
        if (wresultX || wresultY) cvMoveWindow("Result",wresultX,wresultY);
        namedWindow( "Configuration", 0);     // Window for the tracbars with adjustable parameters
        createAllTrackBars();
        cvResizeWindow("Configuration",600,160);
        if (wconfigurationX || wconfigurationY )cvMoveWindow("Configuration",wconfigurationX,wconfigurationY);
    }






//---------------------------------------------------------------------
//-- START Main loop that reads each testing image and apply scene recognition
    cout << "Starting at RUN = " << currentRun+1 << endl;

    int lastClassNumber = 0;
    int c = 0;
    testImageIndex = 0;
    while ( c != 27 )
    {

        if ( testImageIndex >= testImages.size()) {

            if (classifierType == WEKA) break; // One run is enough for weka classifier (which just create an ARFF)

            evaluator.printResults();
            ++currentRun;

            if (currentRun >= maxRuns) break;

            cout << "Starting Another RUN = " << currentRun+1 << endl;
            testImageIndex = 0;

            //---------------------------------------------------------------------------------
            //--- Learn a new model (classifier) using the training images from disk

            initializeImagesAndClassesNames(classesNames);
            readImages( fileWithTrainImageNames, trainImages, trainImageNames );
            featureDetector->detect( trainImages, trainKeypoints );
            descriptorExtractor->compute( trainImages, trainKeypoints, trainDescriptors );
            if (classifierType == SYNTSCE) {
                syntSce.createVocabulary(trainDescriptors);
                syntSce.setAuxiliaryAlgorithms(descriptorExtractor,descriptorMatcher,featureDetector);
                syntSce.createGrammars(trainImages,trainImageNames,classesNames);
            }
            readImages( fileWithTestImageNames, testImages, testImageNames );



        }

        // Get the next test image using the list of test image
        queryImage = testImages[testImageIndex];
        resultImage = queryImage.clone();

        if (showInterface) {
            imshow("Original", queryImage); // Show the webcam image
            imshow("Result", resultImage); // Show the webcam image
        }

        if ( queryImage.empty() )
        {
            cout << "Query image can not be read." << endl << ">" << endl;
            return false;
        }






//---------------------------------------------------------------------------------
//--- Apply the classifier to a new test image (read from disk)


        // Use a matcher of SURF descriptors from the queryImage (current image) to each training image and assigns the
        // the queryImage to the class of the training image that has the largest number of matches. The class is encoded
        // in the name of the training image file name

        vector<DMatch> matches;
        descriptorMatcher = DescriptorMatcher::create( matcherType );

        string className;
        int classNumber;

        if ( classifierType == TEMPLATE_MATCHING) {

            featureDetector->detect( queryImage, queryKeypoints );
            descriptorExtractor->compute( queryImage, queryKeypoints, queryDescriptors );

            matchDescriptors( queryDescriptors, trainDescriptors, matches, descriptorMatcher );

            if (matches.size()>0) {
                // Can save the images of the query and training images with the matches imposed on (very interesting to analyse the results)
                if (saveIntermediateResults) {
                    saveResultImages( queryImage, queryKeypoints, trainImages, trainKeypoints,
                                      matches, trainImageNames, dirToSaveResImages, testImageIndex);
                }

                classNumber = classifyUsingTemplateMatching(matches,trainImages,trainImageNames, classesNames, className);
                int classIndex = getClassNumber(testImageNames[testImageIndex],classesNames);
                evaluator.newSample(classNumber,classIndex);


            } else {
                cout << "No matchs" << endl;
            }

        } else if ( classifierType == SYNTSCE) {
            classNumber = syntSce.classify(queryImage, className, testImageNames[testImageIndex],classesNames,evaluator);
        } else if (classifierType == WEKA) {
            if (currentRun == 0 ) { // Only one run is enough for WEKA cause it is only creating an ARFF with all dataset
                weka.insertArffInstance(queryImage,testImageNames[testImageIndex],classesNames);
            } else { // Is this is not the first run, just break cause the ARFF has already been created at the first run
                break;
            }
        }

        ostringstream text;
        text << testImageIndex << " " << className;
        putText(resultImage, text.str(), cvPoint(10,40), FONT_HERSHEY_SIMPLEX, 1, Scalar(200, 155, 200, 0),2,8);

        if (showInterface) imshow("Result", resultImage); // Show the result Image

        // Play a sound associated with each class. The sounds are saved on data/sounds and are name class_0.wav, class_1.waw, ...
        // just change these files to have other sounds.
        if (playClassSound && classSound != NULL) {
            classSound->stop();
            if (classNumber != lastClassNumber) classSound->setCurrentSource(mediaSources[classNumber]);
            classSound->play();
        }

        lastClassNumber = classNumber;







//---------------------------------------------------------------------
//-- END Main loop that reads each testing image and apply scene recognition

        ++testImageIndex;
        c = waitKey(timeBetweenFrames);
    }

    if ( classifierType == WEKA) weka.closeArff();
    if (!(classifierType == WEKA)) evaluator.saveResults();
    saveConfiguration();
    cvDestroyAllWindows();

    return 0;

}





















// Return the index of the class from the name of the training  image
int getClassNumber(String fileImageName, vector<String> classesNames) {
    for (int i = 0; i < classesNames.size();++i) {
        int res = fileImageName.find(classesNames[i],0);
        if (res > 0) {
            return i;
        }
    }
    cout << "Problem finding the class name inside trainImageName";
}



// Test if a file exist
bool fexists(const char *filename)
{
    ifstream ifile(filename);
    return ifile;
}

// Create the log file which will receive the performance results (each time the program is run another set of results are appended)
void createResultsFile() {

    ostringstream output;
    output << "echo \"";
    output << "currentRun" << ";";
    output << "testImageIndex" << ";";
    output << "imageString" << ";";


    if (giType == GI_SEQUITUR) {

        for ( int i = 0; i < numberOfClasses; ++i) {
            output << "Conf. Class " << i << ";";

        }

    }

    output << "classifier" << ";";
    output << "groundTruthClass" << ";";
    output << "groundTruthFile" << ";";
    output << "classNumber" << ";";
    output << "\" > " << fileWithPerformanceResults;

    string command = output.str();
    system(command.c_str());


}


void displayInstructions() {
    QMessageBox msgBox;
    msgBox.setText(
        "INSTRUCTIONS\n\n- Hit <ESC> to exit (after first image appears)\n- Results will be saved to disk\n- The software must be re-started after configuration is changed\n\nFurther information: trac.gpec.ucdb.br/wiki/site_syntsce");
    msgBox.exec();
}

// Randomly select the training and testing images. Extract classes names from folder names
bool initializeImagesAndClassesNames(vector<string>& classesNames) {

    // This command will list all folders in the images folder this way getting the classes names automatically
    string command = "ls " + dirWithImages + " > " + fileWithClassesNames;
    system(command.c_str());

    string dirName;
    readFilenames( fileWithClassesNames, dirName, classesNames );
    if ( classesNames.empty() )
    {
        cout << "Could not generate classes names from directories names." << endl << ">" << endl;
        return false;
    }

    numberOfClasses = classesNames.size();

    command = "echo -n > " + fileWithTrainImageNames + "; echo -n > " + fileWithTestImageNames;
    system(command.c_str());


    for ( size_t i = 0; i < classesNames.size(); i++ )
    {
        command = "ls " + dirWithImages + "/" + classesNames[i] + " > " + "temp.txt";
        system(command.c_str());

        command = "shuf temp.txt > temp2.txt; mv temp2.txt temp.txt";
        system(command.c_str());

        vector<string> classesFiles;
        readFilenames( "temp.txt" , dirName, classesFiles );

        int trainingThreshold = classesFiles.size()*(percentTraining/100.0);

        for ( size_t j = 0; j < classesFiles.size(); j++ )
        {
            command = "echo " + dirWithImages + "/" + classesNames[i] + "/"  ;

            if ( j < trainingThreshold) {

                command += classesFiles[j] + " >> " + fileWithTrainImageNames;


            } else {

                command += classesFiles[j] + " >> " + fileWithTestImageNames;

            }
            system(command.c_str());

        }

    }
    return true;
}



//------------------------------------------------------------------------------------------------
//--- Save a training sample to the disk with the approppriate name (containing the class number)
void saveTrainingImage(IplImage * src, int i,int imageClass) {
    char indexName[20];
    char className[20];
    // TAKE CARE !!! I'm fixating here the maximum number of classes and training samples
    sprintf(indexName,"%07d",i);
    sprintf(className,"%04d",imageClass);
    string filename = dirWithImages+"train_"+className+"_"+indexName+".jpeg";
    cvSaveImage(filename.c_str(),src);
}














//------------------------------------------------------------------------------------------------
// I had to create this ugly and slow function because I could not access the elements of a Matriz using the methods available in the Internet...
void extractValuesFromMatrix(int * Result, Mat OMat)
{

    ostringstream oss;
    string element;
    oss <<  " " << OMat;
    stringstream parser(oss.str());
    int i = 0;
    while (parser>>element) {
        float f;
        if (i==0) from_string<float>(f, element.substr(1,5), std::dec);
        else from_string<float>(f, element.substr(0,5), std::dec);
        Result[i] = (int) f;
        ++i;
    }

}














//------------------------------------------------------------------------------------------------
// Create feature detector, descritor extractor and the descriptor matcher
bool createDetectorDescriptorMatcher( const string& detectorType, const string& descriptorType, const string& matcherType,
                                      Ptr<FeatureDetector>& featureDetector,
                                      Ptr<DescriptorExtractor>& descriptorExtractor,
                                      Ptr<DescriptorMatcher>& descriptorMatcher )
{
    initModule_nonfree();
    featureDetector = FeatureDetector::create( detectorType );
    descriptorExtractor = DescriptorExtractor::create( descriptorType );
    descriptorMatcher = DescriptorMatcher::create( matcherType );
    cout << ">" << endl;

    bool isCreated = !( featureDetector.empty() || descriptorExtractor.empty() || descriptorMatcher.empty() );
    if ( !isCreated )
        if( featureDetector.empty()) cout << "Can not create feature detector of given types." << endl << ">" << endl;
        if( descriptorExtractor.empty())  cout << "Can not create descriptor extractor of given types." << endl << ">" << endl;
        if( descriptorMatcher.empty())  cout << "Can not create descriptor matcher of given types." << endl << ">" << endl;

    return isCreated;
}







//------------------------------------------------------------------------------------------------
// Find the training images related to each match and construct a mask from them
void maskMatchesByTrainImgIdx( const vector<DMatch>& matches, int trainImgIdx, vector<char>& mask )
{
    mask.resize( matches.size() );
    fill( mask.begin(), mask.end(), 0 );
    for ( size_t i = 0; i < matches.size(); i++ )
    {
        if ( matches[i].imgIdx == trainImgIdx )
            mask[i] = 1;
    }
}






//------------------------------------------------------------------------------------------------
// Read the txt file containing all the image names and put the names in a vector
void readFilenames( const string& filename, string& dirName, vector<string>& trainFilenames )
{
    const char dlmtr = '/';

    trainFilenames.clear();

    ifstream file( filename.c_str() );
    if ( !file.is_open() )
        return;

    size_t pos = filename.rfind(dlmtr);
    dirName = pos == string::npos ? "" : filename.substr(0, pos) + dlmtr;
    while ( !file.eof() )
    {
        string str;
        getline( file, str );
        if ( str.empty() ) break;
        trainFilenames.push_back(str);
    }
    file.close();
}





//------------------------------------------------------------------------------------------------
// Read the training or testing images to a vector
bool readImages( const string& trainFilename,
                 vector <Mat>& trainImages, vector<string>& trainImageNames )
{
    string trainDirName;
    trainImages.clear();
    readFilenames( trainFilename, trainDirName, trainImageNames );
    if ( trainImageNames.empty() )
    {
        cout << "Train image filenames can not be read." << endl << ">" << endl;
        return false;
    }
    int readImageCount = 0;
    for ( size_t i = 0; i < trainImageNames.size(); i++ )
    {
        string filename = trainImageNames[i];
        Mat img = imread( filename, CV_LOAD_IMAGE_GRAYSCALE );
        if ( img.empty() )
            cout << "Train image " << filename << " can not be read." << endl;
        else
            readImageCount++;
        trainImages.push_back( img );
    }
    if ( !readImageCount )
    {
        cout << "All train or test images can not be read." << endl << ">" << endl;
        return false;
    }
    else
        cout << readImageCount << " images read from " << trainFilename << endl;

    return true;
}







//------------------------------------------------------------------------------------------------
// Match query and training images descriptors
void matchDescriptors( const Mat& queryDescriptors, const vector<Mat>& trainDescriptors,
                       vector<DMatch>& matches, Ptr<DescriptorMatcher>& descriptorMatcher )
{
    descriptorMatcher->add( trainDescriptors );
    descriptorMatcher->match( queryDescriptors, matches );
    CV_Assert( queryDescriptors.rows == (int)matches.size() || matches.empty() );
}







//------------------------------------------------------------------------------------------------
// Save the images showing the matches found between query and training images
void saveResultImages( const Mat& queryImage, const vector<KeyPoint>& queryKeypoints,
                       const vector<Mat>& trainImages, const vector<vector<KeyPoint> >& trainKeypoints,
                       const vector<DMatch>& matches, const vector<string>& trainImagesNames, const string& resultDir , int testIndex)
{
    Mat drawImg;
    vector<char> mask;
    for ( size_t i = 0; i < trainImages.size(); i++ )
    {
        if ( !trainImages[i].empty() )
        {
            maskMatchesByTrainImgIdx( matches, i, mask );
            drawMatches( queryImage, queryKeypoints, trainImages[i], trainKeypoints[i],
                         matches, drawImg, Scalar::all(-1), Scalar::all(-1), mask );

            mkdir(resultDir.c_str(), 0777);
            char testIndexName[20];
            char trainIndexName[20];
            sprintf(testIndexName,"%04d",testIndex);
            sprintf(trainIndexName,"%04d",i);
            string filename = resultDir + "/res_" + testIndexName + "_" + trainIndexName + ".jpg";
            if ( !imwrite( filename, drawImg ) )
                cout << "Image " << filename << " can not be saved (may be because directory " << resultDir << " does not exist)." << endl;
        }
    }
}





//------------------------------------------------------------------------------------------------
// Find the trainig image with most matchs and extract its class
int classifyUsingTemplateMatching( const vector<DMatch>& matches, const vector<Mat>& trainImages,
                                   const vector<string>& trainImagesNames, vector<string>& classesNames, string& className )
{
    int imageNumber = -1;
    int classNumber = -1;
    vector<int> counts(trainImagesNames.size(),0);

    if (matches.empty()) {
        cout << "No Matches ..." << endl;
        return(0);
    }

    for ( size_t i = 0; i < matches.size(); i++ )
    {


        if ( matches[i].imgIdx < trainImagesNames.size()) {
            ++counts[matches[i].imgIdx];
        } else {
            cout << "DEU MERDA " << endl;
            cout << "matches[i].imgIdx = " << matches[i].imgIdx << endl;;
        }
    }


    int largest = -1;
    for ( size_t i = 0; i < counts.size(); i++ )
    {
        if ( counts[i] > largest ) {
            imageNumber = i;
            largest = counts[i];
        }
    }

    string filename = trainImagesNames[imageNumber];

    cout << " i = " << imageNumber << " larg = " << largest << " filename = " << filename << endl;

    for ( size_t i = 0; i < classesNames.size(); i++ )
    {
        // cout << " class " << i << " = " << classesNames[i] << endl;

        int res = filename.find(classesNames[i],0);
        if ( res > 0 ) {

            // cout << " achou " << " res = " << res << endl;

            classNumber = i;
            className = classesNames[i];
            break;
        }
    }


    return(classNumber);


}



//------------------------------------------------------------------------------------------------
// Classifier Changed
void classifierChanged( int, void * ) {
    if (classifierType == SYNTSCE) {
        cout << "Program needs to be re-started ... exiting";
        saveConfiguration();
        exit(0);
    }
}

void adjustFileNames() {
    fileWithTrainImageNames = "../data/"+dataSetName+"/train.txt";
    fileWithTestImageNames = "../data/"+dataSetName+"/test.txt";
    fileWithClassesNames = "../data/"+dataSetName+"/classes.txt";
    fileWithAllImageNames = "../data/"+dataSetName+"/all.txt";
    dirWithImages = "../data/"+dataSetName+"/image";
    dirWithGrammars = "../data/"+dataSetName+"/grammars";
    fileWithPerformanceResults = "../data/"+dataSetName+"/results.csv";
    fileWithHistoricalResults = "../data/"+dataSetName+"/results.txt";

    char dicSize[6];
    sprintf(dicSize,"%02d",defaultDictionarySize);
    arffFileName = "../data/"+dataSetName+"/"+dataSetName+"_Dic"+dicSize+".arff";
    soundsDir = "../data/"+dataSetName+"/sounds";
    dirToSaveResImages = "../data/"+dataSetName+"/results";

    matlabCommand = "matlab -nodesktop -nosplash -r matlabScript";
    matlabGiToolBoxDir = "../gi/gitoolbox_2.0_matlab/";
}


//-------------------------------------------------------------------------------------------------

void createAllTrackBars(void) {

    createTrackbar( "Percent for Training", "Configuration", &percentTraining, 100, 0 );
    createTrackbar( "Evaluation Runs", "Configuration", &maxRuns, 100, 0);
    createTrackbar( "Classifier [0=MATCHING|1=SYNTSCE|2=WEKA]", "Configuration", &classifierType, 2, classifierChanged );
    //createTrackbar( "defaultSymbolRepititionThreshld","Configuration", &defaultSymbolRepititionThreshld);
    //createTrackbar( "defaultBinValueThreshld", "Configuration", &defaultBinValueThreshld);
    createTrackbar( "BoW Dictionary Size",  "Configuration", &defaultDictionarySize, 52, 0);
    createTrackbar( "SyntSce Keypoint Order",  "Configuration", &defaultKeypointsOrder, 4, 0);
    createTrackbar( "SyntSce Scanning Direction",  "Configuration", &defaultScanningDirection, 1, 0);
    createTrackbar( "Play Sound [0=NO|1=YES]",  "Configuration", &playClassSound, 1, 0);
    createTrackbar( "Save Matches Image [0=NO|1=YES]",  "Configuration", &saveIntermediateResults, 1,  0);
    createTrackbar( "Time to Wait Between Frames",  "Configuration", &timeBetweenFrames, 1000, 0);


}






//-------------------------------------------------------------------------------------------------
// Save configuration parameters on disk
void saveConfiguration(void) {

    CvFileStorage* fs=cvOpenFileStorage(configFileName, 0,CV_STORAGE_WRITE);

    cvWriteInt( fs, "classifierType", classifierType);
    cvWriteComment(fs,"0 = TEMPLATE_MATCHING, 1 = SYNTSCE, 2 = WEKA",1);
    cvWriteInt( fs, "giType", giType);
    cvWriteComment(fs,"0 = GI_SEQUITUR, 1 = GI_KTESTABLE",1);
    cvWriteString( fs, "dataSetName", dataSetName.c_str());
    cvWriteInt( fs, "incrementDictionarySize", incrementDictionarySize);
    cvWriteInt( fs, "defaultDictionarySize", defaultDictionarySize+incrementDictionarySize);
    cvWriteInt( fs, "percentTraining", percentTraining);
    cvWriteInt( fs, "maxRuns",maxRuns);
    cvWriteReal( fs, "defaultSymbolRepititionThreshld", defaultSymbolRepititionThreshld);
    cvWriteReal( fs, "defaultBinValueThreshld", defaultBinValueThreshld);
    cvWriteInt( fs, "defaultScanningDirection", defaultScanningDirection);
    cvWriteInt( fs, "defaultKeypointsOrder", defaultKeypointsOrder);
    cvWriteComment(fs,"0 = ORDER_SURF, 1 = ORDER_RANDOM, 2 = ORDER_READING, 3 = ORDER_RADIAL, 4 = ORDER_READING_QUANTIZED",1);
    cvWriteInt( fs, "quantizeFactor", quantizeFactor);
    cvWriteInt( fs, "playClassSound", playClassSound);
    cvWriteInt( fs, "saveIntermediateResults", saveIntermediateResults);
    cvWriteInt( fs, "timeBetweenFrames", timeBetweenFrames);
    cvWriteInt( fs, "woriginalX", woriginalX);
    cvWriteInt( fs, "woriginalY", woriginalY);
    cvWriteInt( fs, "wresultX", wresultX);
    cvWriteInt( fs, "wresultY", wresultY);
    cvWriteInt( fs, "wconfigurationX", wconfigurationX);
    cvWriteInt( fs, "wconfigurationY", wconfigurationY);
    cvWriteInt( fs, "showInterface", showInterface);
    cvReleaseFileStorage( &fs);
}




//-------------------------------------------------------------------------------------------------
// Recover configuration parameters from disk
void readConfiguration(void) {
    CvFileStorage* fs= cvOpenFileStorage(configFileName, 0, CV_STORAGE_READ);
    if (fs) {
        woriginalX = cvReadIntByName( fs, 0, "woriginalX");
        woriginalY = cvReadIntByName( fs, 0,"woriginalY");
        wresultX = cvReadIntByName( fs, 0, "wresultX");
        wresultY = cvReadIntByName( fs, 0,"wresultY");
        wconfigurationX = cvReadIntByName( fs,0, "wconfigurationX");
        wconfigurationY = cvReadIntByName( fs,0, "wconfigurationY");
        percentTraining = cvReadIntByName( fs,0, "percentTraining");
        classifierType = cvReadIntByName( fs,0, "classifierType");
        giType = cvReadIntByName( fs,0, "giType");
        defaultSymbolRepititionThreshld = cvReadRealByName( fs, 0, "defaultSymbolRepititionThreshld");
        defaultBinValueThreshld = cvReadRealByName( fs, 0, "defaultBinValueThreshld");
        incrementDictionarySize = cvReadIntByName( fs, 0, "incrementDictionarySize");
        defaultDictionarySize = cvReadIntByName( fs, 0, "defaultDictionarySize");
        defaultScanningDirection = cvReadIntByName( fs, 0, "defaultScanningDirection");
        defaultKeypointsOrder = cvReadIntByName( fs, 0, "defaultKeypointsOrder");
        quantizeFactor = cvReadIntByName( fs, 0, "quantizeFactor");
        playClassSound = cvReadIntByName( fs,0,  "playClassSound");
        saveIntermediateResults = cvReadIntByName( fs,0,  "saveIntermediateResults");
        timeBetweenFrames = cvReadIntByName( fs,0,  "timeBetweenFrames");
        maxRuns = cvReadIntByName( fs, 0, "maxRuns");
        dataSetName = cvReadStringByName( fs, 0, "dataSetName");
        showInterface = cvReadIntByName( fs, 0, "showInterface");
        adjustFileNames();
    }
    cvReleaseFileStorage( &fs);
}





